// channel.c
//

#include <pktlab.h>

#include <assert.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "channel.h"
#include "private.h"

//
// EXPORTED FUNCTION DEFINITIONS
//

// in pktlab.h
int pktlab_encode_chlist (
    const struct pktlab_channel * channels,
    void ** chlistptr, uint_fast32_t * chlistlen)
{
    uint_fast32_t totalmem;
    const struct pktlab_channel * channel;
    uint8_t * ptr;

    trace("%s(...)", __func__);

    if (chlistptr == NULL || chlistlen == NULL)
        return -1;

    totalmem = 1; // terminating null byte
    for (channel = channels; channel != NULL; channel=channel->next) {
        if (channel->channellen > PKTLAB_CHANNELID_LEN_MAX ||
            channel->channellen == 0)
            return -2; // bad channel

        if (channel->channellen >= UINT_FAST32_MAX ||
            totalmem > UINT_FAST32_MAX-channel->channellen-1)
            return -3; // wrap-around

        totalmem += channel->channellen+1;
    }

    if (totalmem > SIZE_MAX)
        return -3;

    ptr = *chlistptr = safe_malloc(totalmem);
    for (channel = channels; channel != NULL; channel=channel->next) {
        *ptr = channel->channellen; // record len
        ++ptr;

        memcpy(ptr, channel->channel, channel->channellen);
        ptr += channel->channellen;
    }

    *ptr = 0; // terminating null byte
    *chlistlen = totalmem;
    return 0;
}

int pktlab_decode_chlist (
    const void * chlistptr, uint_fast32_t chlistlen,
    struct pktlab_channel ** channels)
{
    uint_fast32_t i, channellen, totalmem;
    const uint8_t * ptr;
    struct pktlab_channel * channel;
    struct pktlab_channel * tail;
    int ret = -128;

    trace("%s(...)", __func__);

    if (chlistptr == NULL || chlistlen == 0 ||
        channels == NULL) // at least 1 byte is needed
        return -1;

    tail = *channels = NULL;
    i = 0;
    while (1) {
        assert(i < chlistlen);

        ptr = ((const uint8_t *) chlistptr)+i;
        channellen = *ptr;

        if (channellen > PKTLAB_CHANNELID_LEN_MAX) {
            ret = -2; // bad channel
            goto fail;
        }

        if (channellen == 0) {
            if (i+1 == chlistlen)
                return 0;
            ret = -2; // bad chlist (more after null byte)
            goto fail;
        }

        if (i > UINT_FAST32_MAX-channellen-1) {
            ret = -3;
            goto fail;
        }

        if (i+channellen+1 >= chlistlen) {
            ret = -2; // bad chlist
            goto fail;
        }

        totalmem = sizeof(*channel);
        if (totalmem > UINT_FAST32_MAX-channellen ||
            totalmem+channellen > SIZE_MAX) {
            ret = -3;
            goto fail;
        }
        totalmem += channellen;

        channel = safe_malloc(totalmem);
        channel->channel = (void *) (channel+1);
        channel->channellen = channellen;
        channel->next = NULL;
        memcpy(channel->channel, ptr+1, channellen);

        if (tail == NULL) {
            *channels = channel;
        } else {
            tail->next = channel;
        }
        tail = channel;

        i += channellen+1;
    }

fail:
    pktlab_free_channels(*channels);
    return ret;
}

void pktlab_free_channels (
    struct pktlab_channel * channels)
{
    struct pktlab_channel * channel;

    while (channels != NULL) {
        channel = channels->next;
        free(channels);
        channels = channel;
    }

    return;
}

// in channel.h
int find_chlistend (
    const void * ptr, uint_fast32_t len,
    const void ** chlistend, uint_fast32_t * chlistlen)
{
    uint_fast32_t i, channellen;
    const uint8_t * cursor;
    int ret = -128;

    assert(
        ptr != NULL && len != 0 &&
        chlistend != NULL && chlistlen != NULL);

    i = 0;
    while (1) {
        assert(i < len);

        cursor = ((const uint8_t *) ptr)+i;
        channellen = *cursor;

        if (channellen > PKTLAB_CHANNELID_LEN_MAX) {
            ret = -1; // bad channel
            goto fail;
        }

        if (channellen == 0) {
            *chlistend = (const void *) (cursor+1);
            *chlistlen = i+1;
            return 0;
        }

        if (i > UINT_FAST32_MAX-channellen-1) {
            ret = -2;
            goto fail;
        }

        if (i+channellen+1 >= len) {
            ret = -1; // bad chlist
            goto fail;
        }

        i += channellen+1;
    }

fail:
    return ret;
}