// hash_table.c
// Hash table C implementation.
//

#include "hash_table.h"

#include "linked_list.h"
#include "private.h"

#include <assert.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

//
// INTERNAL TYPE DEFINITION
//

struct hash_table {
    struct linked_list ** arr;
    uint_fast32_t arrsz;
    uint_fast32_t entrycnt;
    hash_t hash_func;
};

struct table_entry {
    const void * key;
    uint_fast32_t keylen;
    void * data;
};

struct free_entry_aux {
    freedata_t free_data_func;
    void * free_data_aux;
};

//
// INTERNAL FUNCTION DECLARATIONS
//

static int is_same_key (
    const void * k1, uint_fast32_t k1len,
    const void * k2, uint_fast32_t k2len);

static int find_ll (
    const void * key, uint_fast32_t keylen,
    struct linked_list ** ll, struct hash_table * table);
static int find_entry (
    const void * key, uint_fast32_t keylen,
    struct table_entry ** target, uint_fast32_t * indx,
    struct linked_list * ll);

static void free_entry(void * entry, void * aux);

//
// EXPORTED FUNCTION DEFINITIONS
//

struct hash_table * create_table(uint_fast32_t tablesz, hash_t hash_func) {
    struct hash_table * table;
    uint_fast32_t i, memlen;

    assert(tablesz > 0 && hash_func != NULL && memlen <= SIZE_MAX);

    memlen = sizeof(struct hash_table_entry *) * tablesz;
    if (tablesz != memlen/sizeof(struct hash_table_entry *) ||
        memlen > UINT_FAST32_MAX-sizeof(struct hash_table))
        return NULL;

    memlen += sizeof(struct hash_table);
    if (memlen > SIZE_MAX)
        return NULL;

    table = safe_malloc(memlen);
    memset(table, 0, memlen);

    table->arr = (struct linked_list **) (table+1);
    table->arrsz = tablesz;
    table->hash_func = hash_func;

    for (i = 0; i < tablesz; ++i) // init linked list
        table->arr[i] = create_ll();

    return table;
}

int insert_table (
    const void * key, uint_fast32_t keylen,
    void * value, struct hash_table * table)
{
    struct linked_list * ll;
    struct table_entry * entry;

    assert(key != NULL && keylen > 0 && table != NULL);

    if (find_ll(key, keylen, &ll, table) < 0)
        goto fail;

    if (find_entry(key, keylen, NULL, NULL, ll) >= 0) // found a dup key entry
        goto fail;

    // prepare table entry
    entry = safe_malloc(sizeof(*entry));
    entry->key = key;
    entry->keylen = keylen;
    entry->data = value;

    // perform the insertion
    if (push_ll(entry, ll) < 0)
        goto fail_entry;

    ++table->entrycnt;
    return 0;

fail_entry:
    free(entry);
fail:
    return -1;
}

int remove_table (
    const void * key, uint_fast32_t keylen,
    void ** value, struct hash_table * table)
{
    struct linked_list * ll;
    struct table_entry * entry;
    uint_fast32_t rmindx;

    assert(key != NULL && keylen > 0 && table != NULL);

    if (find_ll(key, keylen, &ll, table) < 0)
        return -1;

    if (find_entry(key, keylen, &entry, &rmindx, ll) < 0)
        return -1;

    if (remove_node(ll, rmindx, NULL) < 0)
        return -1; // this should not happen

    if (value != NULL)
        *value = entry->data;

    free(entry);
    --table->entrycnt;
    return 0;
}

int replace_table (
    const void * key, uint_fast32_t keylen,
    void ** oldvalue, void * newvalue,
    struct hash_table * table)
{
    struct linked_list * ll;
    struct table_entry * entry;

    assert(key != NULL && keylen > 0 && table != NULL);

    if (find_ll(key, keylen, &ll, table) < 0)
        return -1;

    if (find_entry(key, keylen, &entry, NULL, ll) < 0)
        return -1;

    if (oldvalue != NULL)
        *oldvalue = entry->data;
    entry->data = newvalue;

    return 0;
}

int lookup_table (
    const void * key, uint_fast32_t keylen,
    void ** value, struct hash_table * table)
{
    struct linked_list * ll;
    struct table_entry * entry;

    assert(key != NULL && keylen > 0 && table != NULL);

    if (find_ll(key, keylen, &ll, table) < 0)
        return -1;

    if (find_entry(key, keylen, &entry, NULL, ll) < 0)
        return -1;

    if (value != NULL)
        *value = entry->data;
    return 0;
}

void cleanup_table (
    struct hash_table * table, freedata_t free_data_func, void * free_data_aux)
{
    uint_fast32_t i;
    struct linked_list * ll;
    struct free_entry_aux aux = { .free_data_func = free_data_func, .free_data_aux = free_data_aux };

    assert(table != NULL);

    for (i = 0; i < table->arrsz; ++i) {
        ll = table->arr[i];
        assert(ll != NULL);

        cleanup_ll(ll, free_entry, &aux);
    }

    free(table);
}

uint_fast32_t size_table(const struct hash_table * table) {
    assert(table != NULL);
    return table->entrycnt;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int is_same_key (
    const void * k1, uint_fast32_t k1len,
    const void * k2, uint_fast32_t k2len)
{
    assert(k1 != NULL && k2 != NULL && k1len > 0 && k2len > 0);

    if (k1len != k2len)
        return 0;

    return (memcmp(k1, k2, k1len) == 0);
}

int find_ll (
    const void * key, uint_fast32_t keylen,
    struct linked_list ** ll, struct hash_table * table)
{
    uint_fast32_t index;

    assert(key != NULL && keylen > 0 && ll != NULL && table != NULL);

    index = table->hash_func(key, keylen) % table->arrsz;
    *ll = table->arr[index];

    assert(*ll != NULL);
    return 0;
}

int find_entry (
    const void * key, uint_fast32_t keylen,
    struct table_entry ** target, uint_fast32_t * indx,
    struct linked_list * ll)
{
    struct ll_node * node;
    struct table_entry * entry;
    uint_fast32_t cnt;

    assert(key != NULL && keylen > 0 && ll != NULL);

    if (len_ll(ll) == 0)
        return -1;

    cnt = 0;
    for (node = get_head(ll); node != NULL; node = get_next(node)) {
        entry = get_data(node);
        assert(entry != NULL);

        if (!is_same_key(key, keylen, entry->key, entry->keylen)) {
            ++cnt;
            continue;
        }

        // found a match!
        if (target != NULL)
            *target = entry;
        if (indx != NULL)
            *indx = cnt;
        return 0;
    }

    return -1; // no matching key in ll found
}

void free_entry(void * entry, void * aux) {
    struct table_entry * e = entry;
    struct free_entry_aux * ptr = aux;

    (ptr->free_data_func)(e->data, ptr->free_data_aux);
    free(entry);
    return;
}