// config.c
//

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "private.h"

#include <pktconf.h>

#define safe_str(s) ((s)?(s):"(NULL)")
#define BUF_SIZ 4096
#define OPT_SEP " \t" // space and tab are option separators

//
// INTERNAL CONSTANTS
//

enum str_search_st {
    PRE,
    IN
};

//
// INTERNAL FUNCTION DEFINITIONS
//

static int get_new_opt_ll_node (
    const char * buf, struct pktconf_opt_ll_node ** node);

static const char * strpbrk_ext (
    const char * s1, const char * s2,
    size_t * len);

//
// EXPORTED FUNCTION DEFINITIONS
//

//
// LINKED LIST LOAD AND CLEANUP FUNCTIONS
//

int pktconf_load_option (
    const char * confpath,
    struct pktconf_option_ll ** ll)
{
    FILE * fp;
    uint_fast32_t lineno;
    char buf[BUF_SIZ];
    struct pktconf_opt_ll_node * node;
    int rst, ret = -128;

    trace("%s(confpath:%s)",
        __func__, safe_str(confpath));

    if (confpath == NULL)
        return -1;

    fp = fopen(confpath, "r");
    if (fp == NULL) {
        debug("Config open failed: %s", strerror(errno));
        return -2;
    }

    *ll = safe_malloc(sizeof(**ll));
    memset(*ll, 0, sizeof(**ll));

    // from this point on, fail with goto
    lineno = 1;
    while (!feof(fp)) {
        if (fgets(buf, sizeof(buf), fp) == NULL) {
            if (ferror(fp)) {
                debug("fgets failed: %s", strerror(errno));
                ret = -2;
                goto fail;
            }
            break; // eof
        }

        // remove line end '\n' before use
        buf[strcspn(buf, "\n")] = '\0';
        rst = get_new_opt_ll_node(buf, &node);
        switch (rst) {
        case 0:
            node->lineno = lineno;
            node->next = (*ll)->head;
            (*ll)->head = node;
            ++(*ll)->len;
            break;
        case -1: // ignore line
            break;
        default:
            debug("get_new_opt_ll_node returned %d", rst);
            ret = -3;
            goto fail;
        }

        ++lineno;
    }

    fclose(fp);
    return 0;

fail:
    pktconf_cleanup_option_ll(*ll);
    fclose(fp);
    return ret;
}

void pktconf_cleanup_option_ll (
    struct pktconf_option_ll * ll)
{
    struct pktconf_opt_ll_node * next;

    trace("%s(...)", __func__);

    if (ll == NULL)
        return;

    while (ll->head != NULL) {
        next = ll->head->next;
        free(ll->head);
        ll->head = next;
    }

    free(ll);
    return;
}

//
// VALUE STRING UTILITIES
//

int pktconf_parse_sep_str (
    const char * s, uint_fast32_t len,
    char sep, uint_fast32_t max_cnt,
    struct pktconf_sep_str_ll ** ll)
{
    char seps[2] = {};
    char * scopy;
    char * ptr;
    const char * entry;
    const char * end;
    size_t entrylen, totalmem;
    struct pktconf_sep_str_entry * tail;
    struct pktconf_sep_str_entry * node;

    trace("%s(s:%s,len:%" PRIuFAST32 ",sep:%c,max_cnt:%" PRIuFAST32 ")",
        __func__, safe_str(s), len, sep, max_cnt);

    if (ll == NULL)
        return -1;

    *ll = NULL;
    if (s == NULL || len == 0 || len == SIZE_MAX ||
        sep == '\0' || max_cnt == 0)
        return -1;

    *ll = safe_malloc(sizeof(**ll));
    memset(*ll, 0, sizeof(**ll));

    scopy = pktlab_copy_str(s, len);
    entry = scopy;
    tail = NULL;
    seps[0] = sep;
    while (true) {
        // for last entry case, just treat what's left as last entry
        end = NULL;
        if ((*ll)->len < max_cnt-1)
            end = strpbrk_ext(entry, seps, &entrylen);
        if (end == NULL)
            entrylen = strlen(entry);

        if (entrylen > SIZE_MAX-sizeof(*node)-1)
            goto fail;

        totalmem = sizeof(*node)+entrylen+1;
        node = safe_malloc(totalmem);
        ptr = (char *) (node+1);
        memcpy(ptr, entry, entrylen);
        ptr[entrylen] = '\0';

        node->entry = ptr;
        node->entrylen = entrylen;
        node->next = NULL;
        if ((*ll)->head == NULL) {
            (*ll)->head = node;
            tail = node;
        } else {
            assert(tail != NULL);
            tail->next = node;
            tail = node;
        }

        ++(*ll)->len;

        if (end == NULL)
            break; // last entry
        entry = end+1;
    }

    free(scopy);
    return 0;

fail:
    free(scopy);
    pktconf_cleanup_sep_str_ll(*ll);
    return -2;
}

void pktconf_cleanup_sep_str_ll (
    struct pktconf_sep_str_ll * ll)
{
    struct pktconf_sep_str_entry * next;

    if (ll == NULL)
        return;

    while (ll->head != NULL) {
        next = ll->head->next;
        free(ll->head);
        ll->head = next;
    }

    free(ll);
    return;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int get_new_opt_ll_node (
    const char * buf, struct pktconf_opt_ll_node ** node)
{
    size_t len, totalmem, keylen, seplen;
    char * ptr;
    const char * keyend;

    assert(buf != NULL && node != NULL);

    if (buf[0] == '#' || buf[0] == '\0')
        return -1; // no node should be created

    len = strlen(buf);
    if (len > SIZE_MAX-sizeof(**node)-1)
        return -2; // cannot allocate node

    totalmem = sizeof(**node)+len+1;
    *node = safe_malloc(totalmem);
    ptr = (char *) (*node + 1);
    memcpy(ptr, buf, len+1);

    keyend = strpbrk_ext(ptr, OPT_SEP, &keylen);
    if (keyend == NULL) { // cannot find separator (space or tab) case
        (*node)->key = ptr;
        (*node)->keylen = len;
        (*node)->value = ptr+len;
        (*node)->valuelen = 0;
    } else {
        assert(len > keylen && ptr+keylen == keyend);

        (*node)->key = ptr;
        (*node)->keylen = keylen;

        seplen = strspn(keyend, OPT_SEP);
        (*node)->value = keyend+seplen;
        (*node)->valuelen = strlen((*node)->value);
        ptr[keylen] = '\0'; // key always ends in '\0'
    }

    return 0;
}

const char * strpbrk_ext (
    const char * s1, const char * s2,
    size_t * len)
{
    size_t s1len, s2len, i, j;

    assert(s1 != NULL && s2 != NULL && len != NULL);

    s1len = strlen(s1);
    s2len = strlen(s2);

    for (i = 0; i < s1len; ++i) {
        for (j = 0; j < s2len; ++j) {
            if (s2[j] == s1[i]) {
                *len = i;
                return s1+i;
            }
        }
    }

    return NULL;
}