// ctrl_openssl.c
//

// todo: check table and ll usage; check malloc freeing

#include <pktlab.h>
#include <pktctrl.h>

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "ctrl.h"
#include "hash_table.h"
#include "linked_list.h"
#include "private.h"

#include "cert_openssl.h"
#include "key_openssl.h"
#include "util_openssl.h"

#include <openssl/asn1t.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <openssl/safestack.h>
#include <openssl/ssl.h>
#include <openssl/tls1.h>
#include <openssl/x509_vfy.h>

//
// INTERNAL MACROS
//

#define HASH_TABLE_PRIME 101

//
// INTERNAL TYPE DEFINITION
//

enum store_type {
	SERVER_TYPE    = 1,
	CLIENT_TYPE    = -1,
	NOT_INIT_TYPE  = 0
};

struct snikey_entry {
	const char * snikey;
	struct pktlab_certificate ** cert_ls;
	struct pktlab_publickey ** root_ls;
	uint_fast32_t certnum;
	uint_fast32_t rootnum;
};

struct snikey_map {
	struct hash_table * map;
	struct snikey_entry * fallback; // the default one to use if no snikey is specified
};

struct pktctrl_keycert_store {
	enum store_type type;
	union {
		struct snikey_map server_map;     // for server where multiple snikey possible
		struct snikey_entry client_entry; // for client
	};
};

struct verify_aux {
	struct pktlab_publickey ** root_ls;
	uint_fast32_t rootnum;
	enum pktlab_auth_mode auth_mode;
};

struct client_hello_aux {
	SSL_CTX * await_sslctx;
	SSL_CTX * conn_sslctx;
	struct pktctrl_keycert_store * store;
	struct pktlab_privatekey * privkey;
	struct verify_aux * vaux;
	char * snikey;
};

struct ssl_aux {
	SSL_CTX * await_sslctx;
	SSL_CTX * conn_sslctx;
	SSL * sslobj;
	struct verify_aux * vaux;
	struct client_hello_aux * chaux;
};

typedef int (*ssl_cleanup_t) (SSL * sslobj);
typedef int (*ssl_setup_t) (SSL * sslobj);

//
// INTERNAL FUNCTION DECLARATIONS
//

static bool is_server_auth_mode(int auth_mode);
static bool is_client_auth_mode(int auth_mode);

static uint_fast32_t djb2(const void * ptr, uint_fast32_t len);
uint_fast32_t single_keyid_hash_func(const void * ptr, uint_fast32_t len);
uint_fast32_t double_keyid_hash_func(const void * ptr, uint_fast32_t len);

static SSL_CTX * create_sslctx(void);
static int accept_configure_sslctx_base (
	SSL_CTX * sslctx,
	struct client_hello_aux * chaux);
static void accept_configure_verify (
	SSL * sslobj,
	SSL_CTX * sslctx,
	struct verify_aux * vaux);

static int no_session_ticket (
	SSL * s, unsigned char key_name[16],
	unsigned char iv[EVP_MAX_IV_LENGTH],
	EVP_CIPHER_CTX * ctx, HMAC_CTX * hctx, int enc);

static int snikey_set_keycert(SSL * sslobj, int * al, void * aux);

static int no_verify(X509_STORE_CTX * x509_store_ctx, void * aux);  // absolutely no verify done
static int yes_verify(X509_STORE_CTX * x509_store_ctx, void * aux); // wrapper for pktlab_verify_cert_chain

static int try_setup(struct pktctrl_obj * ctrlobj, ssl_setup_t ssl_setup);

static int parse_sni_ext (
	const uint8_t * ext, size_t extlen,
	char * buf, size_t * buflen);

static int ctx_load_cert(SSL_CTX * ctx, struct snikey_entry * store);
static int ctx_load_privkey(SSL_CTX * ctx, struct pktlab_privatekey * key);

static int get_peer_certificate (
	struct pktlab_certificate *** peer_cert_ls,
	uint_fast32_t * peer_cert_num, const SSL * sslobj);

static ssize_t read_ssl (
	void * obj, void * buf, size_t buflen);
static void post_ssl_read(struct pktctrl_obj * obj, int * result);

static ssize_t write_ssl (
	void * obj, const void * buf, size_t nbyte);
static void post_ssl_write(struct pktctrl_obj * obj, int * result);

//
// INTERNAL GLOBAL VARIABLES
//

//
// EXPORTED FUNCTION DEFINITIONS
//

struct pktctrl_obj * pktctrl_create_obj(void) {
	struct pktctrl_obj * obj;
	size_t totalmem = sizeof(*obj)+sizeof(*obj->saux);

	trace("%s()", __func__);

	obj = safe_malloc(totalmem);
	memset(obj, 0, totalmem);

	obj->saux = (struct ssl_aux *) (obj+1);

	obj->saux->await_sslctx = create_sslctx();
	if (obj->saux->await_sslctx == NULL) {
		free(obj);
		return NULL;
	}

	obj->saux->conn_sslctx = create_sslctx();
	if (obj->saux->conn_sslctx == NULL) {
		free(obj);
		return NULL;
	}

	obj->type = NEITHER_TYPE;
	obj->state = ST_INIT;
	return obj;
}

struct pktctrl_keycert_store * pktctrl_create_keycert_store(void) {
	struct pktctrl_keycert_store * store;

	trace("%s()", __func__);

	store = safe_malloc(sizeof(*store));
	memset(store, 0, sizeof(*store));

	store->type = NOT_INIT_TYPE;
	return store;
}

int pktctrl_set_server_keycert (
	struct pktctrl_keycert_store * store, const char * snikey,
	struct pktlab_certificate ** cert_ls, uint_fast32_t certnum,
	struct pktlab_publickey ** root_ls, uint_fast32_t rootnum)
{
	struct hash_table * map;
	struct snikey_entry * entry;
	struct snikey_entry * oldentry;
	size_t snilen;

	trace("%s(snikey:%s,certnum:%u,rootnum:%u)", __func__,
		(snikey == NULL) ? "(NULL)" : snikey,
		(unsigned) certnum, (unsigned) rootnum);

	// cert_ls is NULL or not implies whether the call is for
	// insert/replace or remove keycert_store map entry, respectively

	if (store == NULL ||
		snikey == NULL ||
		!pktctrl_is_valid_snikey(snikey) ||
		((cert_ls == NULL) != (certnum == 0)) ||
		((root_ls == NULL) != (rootnum == 0)) ||
		(cert_ls == NULL && root_ls != NULL)) // if cert_ls is NULL, root_ls must be NULL
		return PKTCTRL_INVAL;

	switch (store->type) {
	case NOT_INIT_TYPE:
		// init store if not init yet
		store->type = SERVER_TYPE;
		store->server_map.map = create_table(HASH_TABLE_PRIME, djb2);
		break;
	case SERVER_TYPE: break;
	case CLIENT_TYPE:
	default:
		return PKTCTRL_INVAL; // bad map
	}

	snilen = strlen(snikey);
	if (snilen > SIZE_MAX-1-sizeof(*entry)) // overflow
		return PKTCTRL_UNKNOWN_FAULT;

	map = store->server_map.map;

	entry = NULL;
	if (cert_ls != NULL) { // this is an insert/replace call
		entry = safe_malloc(sizeof(*entry)+snilen+1); // the passed certs must not be freed
		entry->cert_ls = cert_ls;
		entry->certnum = certnum;
		entry->root_ls = root_ls;
		entry->rootnum = rootnum;
		entry->snikey = (char *) (entry+1);
		strncpy((char *) (entry+1), snikey, snilen+1);
	}

	if (lookup_table(snikey, snilen, NULL, map) < 0) { // snikey not in map case
		if (cert_ls != NULL &&
			insert_table(entry->snikey, snilen, entry, map) < 0) { // only do stuff if insert call
			debug("insert_table failed");
			free(entry);
			return PKTCTRL_MAP_ERR;
		}
	} else { // snikey in map case
		if (cert_ls == NULL) { // rm call
			if (remove_table(snikey, snilen, (void **) &oldentry, map) < 0) {
				debug("remove_table failed");
				return PKTCTRL_MAP_ERR;
			}
		} else { // replace call
			if (replace_table(entry->snikey, snilen, (void **) &oldentry, entry, map) < 0) {
				debug("replace_table failed");
				free(entry);
				return PKTCTRL_MAP_ERR;
			}
		}

		free(oldentry); // null is fine
	}

	return PKTCTRL_SUCCESS;
}

int pktctrl_get_server_keycert (
	struct pktctrl_keycert_store * store, const char * snikey,
	struct pktlab_certificate *** cert_ls, uint_fast32_t * certnum,
	struct pktlab_publickey *** root_ls, uint_fast32_t * rootnum)
{
	struct snikey_entry * entry;

	trace("%s(snikey:%s)", __func__, (snikey == NULL) ? "(NULL)" : snikey);

	if (store == NULL ||
		store->type != SERVER_TYPE ||
		(snikey != NULL && !pktctrl_is_valid_snikey(snikey)))
		return PKTCTRL_INVAL;

	entry = store->server_map.fallback;
	if (snikey != NULL) {
		if (lookup_table(
				snikey, strlen(snikey),
				(void **) &entry, store->server_map.map) < 0)
			entry = NULL;
	}

	if (entry == NULL)
		return PKTCTRL_KEY_NOT_FOUND;

	if (cert_ls != NULL)
		*cert_ls = entry->cert_ls;
	if (certnum != NULL)
		*certnum = entry->certnum;
	if (root_ls != NULL)
		*root_ls = entry->root_ls;
	if (rootnum != NULL)
		*rootnum = entry->rootnum;
	return PKTCTRL_SUCCESS;
}

int pktctrl_clear_server_keycert (
	struct pktctrl_keycert_store * store, const char * snikey)
{
	trace("%s(snikey:%s)", __func__, snikey);
	return pktctrl_set_server_keycert(store, snikey, NULL, 0, NULL, 0);
}

int pktctrl_set_server_keycert_default (
	struct pktctrl_keycert_store * store, const char * snikey)
{
	struct snikey_entry * entry;

	trace("%s(snikey:%s)", __func__, snikey);

	if (store == NULL ||
		store->type != SERVER_TYPE ||
		store->server_map.map == NULL ||
		(snikey != NULL && !pktctrl_is_valid_snikey(snikey)))
		return PKTCTRL_INVAL;

	if (snikey == NULL) {
		store->server_map.fallback = NULL;
		return PKTCTRL_SUCCESS;
	}

	if (lookup_table(snikey, strlen(snikey), (void **) &entry, store->server_map.map) < 0)
		return PKTCTRL_KEY_NOT_FOUND;

	store->server_map.fallback = entry;
	return PKTCTRL_SUCCESS;
}

int pktctrl_clear_server_keycert_default (
	struct pktctrl_keycert_store * store)
{
	trace("%s()", __func__);

	return pktctrl_set_server_keycert_default(store, NULL);
}

int pktctrl_set_client_keycert (
	struct pktctrl_keycert_store * store,
	struct pktlab_certificate ** cert_ls, uint_fast32_t certnum,
	struct pktlab_publickey ** root_ls, uint_fast32_t rootnum)
{
	trace("%s(certnum:%u,rootnum:%u)",
		__func__, (unsigned) certnum, (unsigned) rootnum);

	if (store == NULL ||
		((cert_ls == NULL) != (certnum == 0)) ||
		((root_ls == NULL) != (rootnum == 0)))
		return PKTCTRL_INVAL;

	switch (store->type) {
	case NOT_INIT_TYPE:
		// init store if not init yet
		store->type = CLIENT_TYPE;
		store->client_entry.snikey = NULL; // client_entry.snikey not used
		break;
	case CLIENT_TYPE: break;
	case SERVER_TYPE:
	default:
		return PKTCTRL_INVAL; // bad map
	}

	store->client_entry.cert_ls = cert_ls;
	store->client_entry.certnum = certnum;
	store->client_entry.root_ls = root_ls;
	store->client_entry.rootnum = rootnum;
	return PKTCTRL_SUCCESS;
}

int pktctrl_get_client_keycert (
	struct pktctrl_keycert_store * store,
	struct pktlab_certificate *** cert_ls, uint_fast32_t * certnum,
	struct pktlab_publickey *** root_ls, uint_fast32_t * rootnum)
{
	trace("%s()", __func__);

	if (store == NULL ||
		store->type != CLIENT_TYPE)
		return PKTCTRL_INVAL;

	if (cert_ls != NULL)
		*cert_ls = store->client_entry.cert_ls;
	if (certnum != NULL)
		*certnum = store->client_entry.certnum;
	if (root_ls != NULL)
		*root_ls = store->client_entry.root_ls;
	if (rootnum != NULL)
		*rootnum = store->client_entry.rootnum;
	return PKTCTRL_SUCCESS;
}

int pktctrl_clear_client_keycert(struct pktctrl_keycert_store * store)
{
	trace("%s()", __func__);

	return pktctrl_set_client_keycert(store, NULL, 0, NULL, 0);
}

int pktctrl_cleanup_keycert_store(struct pktctrl_keycert_store * store) {
	trace("%s()", __func__);

	if (store == NULL)
	 	return PKTCTRL_SUCCESS;

	switch(store->type) {
	case SERVER_TYPE:
		cleanup_table(store->server_map.map, wrapped_free, NULL);
		// fall through
	case CLIENT_TYPE:
	case NOT_INIT_TYPE:
	default:
		free(store);
	}

	return PKTCTRL_SUCCESS;
}

int pktctrl_accept (
	int fd, int auth_mode,
	struct pktlab_privatekey * serverkey,
	struct pktctrl_keycert_store * store,
	struct pktctrl_obj * ctrlobj, char ** client_snikey,
	struct pktlab_certificate *** peer_cert_ls,
	uint_fast32_t * peer_cert_num)
{
	struct verify_aux * vaux;
	struct client_hello_aux * chaux;
	SSL * sslobj;
	int ret;

	trace("%s(state:%d,type:%d,fd:%d,auth_mode:%d)",
		__func__, ctrlobj->state, ctrlobj->type, fd, auth_mode);

	if (ctrlobj == NULL ||
	    (ctrlobj->state != ST_INIT &&
	     ctrlobj->state != ST_HANDSHAKING)) // general bad ctrlobj check
		return PKTCTRL_INVAL;

	if (ctrlobj->state == ST_INIT &&
	    (ctrlobj->type != NEITHER_TYPE || // bad ctrlobj (already used)
	     serverkey == NULL || // server must provide both key and cert(s)
	     store == NULL ||
	     store->type != SERVER_TYPE ||
	     size_table(store->server_map.map) == 0)) // first call additional check
		return PKTCTRL_INVAL;

	if (ctrlobj->state == ST_HANDSHAKING &&
	    ctrlobj->type != SSL_TYPE) // subsequent call additional check
		return PKTCTRL_INVAL;

	if (ctrlobj->state == ST_INIT) { // first call case
		if (!is_server_auth_mode(auth_mode)) {
			debug("Bad server auth mode (got %d)", auth_mode);
			return PKTCTRL_INVAL;
		}

		vaux = safe_malloc(sizeof(*vaux));
		vaux->root_ls = NULL;
		vaux->rootnum = 0;
		vaux->auth_mode = auth_mode;

		chaux = safe_malloc(sizeof(*chaux));
		chaux->await_sslctx = ctrlobj->saux->await_sslctx;
		chaux->conn_sslctx = ctrlobj->saux->conn_sslctx;
		chaux->store = store;
		chaux->privkey = serverkey;
		chaux->vaux = vaux;
		chaux->snikey = NULL;

		ctrlobj->saux->vaux = vaux;
		ctrlobj->saux->chaux = chaux;

		if (accept_configure_sslctx_base(
				ctrlobj->saux->await_sslctx, chaux) < 0) {
			ctrlobj->state = ST_ERROR;
			return PKTCTRL_SSL_FAILURE;
		}

		if (accept_configure_sslctx_base(
				ctrlobj->saux->conn_sslctx, chaux) < 0) {
			ctrlobj->state = ST_ERROR;
			return PKTCTRL_SSL_FAILURE;
		}

		ERR_clear_error();
		sslobj = SSL_new(ctrlobj->saux->await_sslctx);
		if (sslobj == NULL) {
			error_debug_msg("SSL_new");
			ctrlobj->state = ST_ERROR;
			return PKTCTRL_SSL_FAILURE;
		}

		ERR_clear_error();
		if (SSL_set_fd(sslobj, fd) != 1) {
			error_debug_msg("SSL_set_fd");
			SSL_free(sslobj);
			ctrlobj->state = ST_ERROR;
			return PKTCTRL_SSL_FAILURE;
		}

		ctrlobj->sockfd = fd;
		ctrlobj->saux->sslobj = sslobj;
		ctrlobj->state = ST_HANDSHAKING;
		ctrlobj->type = SSL_TYPE;
	}

	ret = try_setup(ctrlobj, SSL_accept);
	switch (ret) {
	case PKTCTRL_IN_PROGRESS:
		return ret;
	case PKTCTRL_SUCCESS:
		break;
	case PKTCTRL_SSL_FAILURE:
	default:
		ctrlobj->state = ST_ERROR;
		return ret;
	}

	if (client_snikey != NULL) {
		*client_snikey = ctrlobj->saux->chaux->snikey;
	} else {
		free(ctrlobj->saux->chaux->snikey);
	}

	if (peer_cert_ls != NULL && peer_cert_num != NULL) {
		*peer_cert_ls = NULL;
		*peer_cert_num = 0;
		get_peer_certificate(peer_cert_ls, peer_cert_num, ctrlobj->saux->sslobj);
	}

	free(ctrlobj->saux->vaux);
	free(ctrlobj->saux->chaux);
	ctrlobj->saux->vaux = NULL;
	ctrlobj->saux->chaux = NULL;

	ctrlobj->reader = pktlab_create_reader(read_ssl, ctrlobj);
	if (ctrlobj->reader == NULL) {
		ctrlobj->state = ST_ERROR;
		return PKTCTRL_UNKNOWN_FAULT;
	}
	ctrlobj->readfunc = read_ssl;
	ctrlobj->readaux = ctrlobj;

	ctrlobj->writer = pktlab_create_writer(write_ssl, ctrlobj);
	if (ctrlobj->writer == NULL) {
		ctrlobj->state = ST_ERROR;
		return PKTCTRL_UNKNOWN_FAULT;
	}
	ctrlobj->writefunc = write_ssl;
	ctrlobj->writeaux = ctrlobj;

	ctrlobj->state = ST_ESTABLISHED;
	return PKTCTRL_SUCCESS;
}

int pktctrl_connect (
	int fd, int auth_mode, const char * snikey,
	struct pktlab_privatekey * clientkey,
	struct pktctrl_keycert_store * store, struct pktctrl_obj * ctrlobj,
	struct pktlab_certificate *** peer_cert_ls, uint_fast32_t * peer_cert_num)
{
	int ret;
	struct verify_aux * vaux;
	SSL * sslobj;

	trace("%s(state:%d,type:%d,fd:%d,auth_mode:%d,snikey:%s)",
		__func__, ctrlobj->state, ctrlobj->type, fd, auth_mode, snikey);

	if (ctrlobj == NULL ||
	    (ctrlobj->state != ST_INIT &&
	     ctrlobj->state != ST_HANDSHAKING)) // general bad ctrlobj check
		return PKTCTRL_INVAL;

	if (ctrlobj->state == ST_INIT &&
	    (ctrlobj->type != NEITHER_TYPE || // bad ctrlobj (already used)
	     (snikey != NULL && !pktctrl_is_valid_snikey(snikey)) || // bad sniley
	     (store != NULL && store->type != CLIENT_TYPE) || // bad keycert store
	     ((clientkey == NULL) != (store == NULL || store->client_entry.certnum == 0)))) // clientkey XOR at least one cert provided; first call additional check
		return PKTCTRL_INVAL;

	if (ctrlobj->state == ST_HANDSHAKING &&
	    ctrlobj->type != SSL_TYPE) // subsequent call additional check
		return PKTCTRL_INVAL;

	if (ctrlobj->state == ST_INIT) { // first call case
		if (!is_client_auth_mode(auth_mode)) {
			debug("Bad auth mode (got %d)", auth_mode);
			return PKTCTRL_INVAL;
		}

		// set ssl verify
		vaux = safe_malloc(sizeof(*vaux));
		ctrlobj->saux->vaux = vaux;
		vaux->root_ls = NULL;
		vaux->rootnum = 0;
		vaux->auth_mode = auth_mode;
		if (store != NULL) {
			vaux->root_ls = store->client_entry.root_ls;
			vaux->rootnum = store->client_entry.rootnum;
		}

		// currently always verify server certs
		SSL_CTX_set_verify(
			ctrlobj->saux->conn_sslctx, SSL_VERIFY_PEER, NULL);
		SSL_CTX_set_cert_verify_callback(
			ctrlobj->saux->conn_sslctx,
			auth_mode != PKTLAB_NO_VERIFY ?
				yes_verify:no_verify,
			vaux);

		// load client certificates
		if (store != NULL &&
			store->client_entry.cert_ls != NULL &&
			store->client_entry.certnum != 0 &&
			clientkey != NULL) { // only load cert & key if provided
			if (ctx_load_cert(ctrlobj->saux->conn_sslctx, &store->client_entry) < 0) {
				ctrlobj->state = ST_ERROR;
				return PKTCTRL_BAD_CERT;
			}

			if (ctx_load_privkey(ctrlobj->saux->conn_sslctx, clientkey) < 0) { // private key loaded after cert required by openssl
				ctrlobj->state = ST_ERROR;
				return PKTCTRL_BAD_CERTKEY;
			}
		}

		ERR_clear_error();
		sslobj = SSL_new(ctrlobj->saux->conn_sslctx);
		if (sslobj == NULL) {
			error_debug_msg("SSL_new");
			ctrlobj->state = ST_ERROR;
			return PKTCTRL_SSL_FAILURE;
		}

		// set snikey
		if (snikey != NULL) {
			ERR_clear_error();
			if (SSL_set_tlsext_host_name(sslobj, snikey) != 1) {
				error_debug_msg("SSL_set_tlsext_host_name");
				SSL_free(sslobj);
				ctrlobj->state = ST_ERROR;
				return PKTCTRL_SSL_FAILURE;
			}
		}

		ERR_clear_error();
		if (SSL_set_fd(sslobj, fd) != 1) {
			error_debug_msg("SSL_set_fd");
			SSL_free(sslobj);
			ctrlobj->state = ST_ERROR;
			return PKTCTRL_SSL_FAILURE;
		}

		ctrlobj->sockfd = fd;
		ctrlobj->saux->sslobj = sslobj;
		ctrlobj->state = ST_HANDSHAKING;
		ctrlobj->type = SSL_TYPE;
	}

	ret = try_setup(ctrlobj, SSL_connect);
	switch (ret) {
	case PKTCTRL_IN_PROGRESS:
		return ret;
	case PKTCTRL_SUCCESS:
		break;
	case PKTCTRL_SSL_FAILURE:
	default:
		ctrlobj->state = ST_ERROR;
		return ret;
	}

	if (peer_cert_ls != NULL && peer_cert_num != NULL) {
		*peer_cert_ls = NULL;
		*peer_cert_num = 0;
		get_peer_certificate(peer_cert_ls, peer_cert_num, ctrlobj->saux->sslobj);
	}

	free(ctrlobj->saux->vaux);
	ctrlobj->saux->vaux = NULL;

	ctrlobj->reader = pktlab_create_reader(read_ssl, ctrlobj);
	if (ctrlobj->reader == NULL) {
		ctrlobj->state = ST_ERROR;
		return PKTCTRL_UNKNOWN_FAULT;
	}
	ctrlobj->readfunc = read_ssl;
	ctrlobj->readaux = ctrlobj;

	ctrlobj->writer = pktlab_create_writer(write_ssl, ctrlobj);
	if (ctrlobj->writer == NULL) {
		ctrlobj->state = ST_ERROR;
		return PKTCTRL_UNKNOWN_FAULT;
	}
	ctrlobj->writefunc = write_ssl;
	ctrlobj->writeaux = ctrlobj;

	ctrlobj->state = ST_ESTABLISHED;
	return PKTCTRL_SUCCESS;
}

int pktctrl_close(struct pktctrl_obj * obj) {
	int ret, rst, err, fd;
	fd_set rset, wset;

	trace("%s(state:%d,type:%d,sockfd:%d)",
		__func__, obj->state, obj->type, obj->sockfd);

	if (obj == NULL)
		return PKTCTRL_SUCCESS;

	ret = PKTCTRL_SUCCESS;
	if (obj->type == SSL_TYPE &&
	    obj->state == ST_ESTABLISHED) { // perform shutdown for ssl
		fd = obj->sockfd;

		while (true) {
			ERR_clear_error();
			rst = SSL_shutdown(obj->saux->sslobj);
			if (rst >= 0)
				break;

			FD_ZERO(&rset);
			FD_ZERO(&wset);

			err = SSL_get_error(obj->saux->sslobj, rst);
			if (err == SSL_ERROR_WANT_READ) {
				FD_SET(fd, &rset);
			} else if (err == SSL_ERROR_WANT_WRITE) {
				FD_SET(fd, &wset);
			} else {
				debug("SSL_shutdown failed with %d", err); // still count as success

				if (err == SSL_ERROR_SYSCALL)
					error_debug_msg("SSL_shutdown");

				goto fail;
			}

			if (select(fd+1, &rset, &wset, NULL, NULL) < 0) {
				if (errno == EINTR)
					continue;

				debug("select failed with %d (%s)", errno, strerror(errno));
				ret = PKTCTRL_UNKNOWN_FAULT;
				goto fail;
			}
		}
	}

fail:
	if (obj->reader != NULL)
		pktlab_close_reader(obj->reader);
	if (obj->writer != NULL)
		pktlab_close_writer(obj->writer);

	if (obj->saux->sslobj != NULL)
		SSL_free(obj->saux->sslobj);

	if (obj->saux->chaux != NULL)
		free(obj->saux->chaux->snikey);

	free(obj->saux->vaux);
	free(obj->saux->chaux);

	// sockfd is NOT closed here

	SSL_CTX_free(obj->saux->await_sslctx);
	SSL_CTX_free(obj->saux->conn_sslctx);
	free(obj);
	return ret;
}

int pktctrl_gen_snikey(char ** buf) {
	uint8_t rbyte;
	int rst;
	size_t i;

	trace("%s(...)", __func__);

	if (buf == NULL)
		return PKTCTRL_INVAL;

	*buf = safe_malloc(PKTCTRL_SNIKEY_LEN_MAX+1);

	i = 0;
	while (i < PKTCTRL_SNIKEY_LEN_MAX) {
		rst = RAND_bytes(&rbyte, 1);
		if (rst != 1) {
			debug("RAND_bytes failed with %d", rst);

			if (rst == 0)
				error_debug_msg("RAND_bytes");

			free(*buf);
			return PKTCTRL_UNKNOWN_FAULT;
		}

		// only lower case and digits
		if (!isdigit(rbyte) && !islower(rbyte))
			continue;

		(*buf)[i] = rbyte;
		++i;
	}

	(*buf)[PKTCTRL_SNIKEY_LEN_MAX] = '\0';
	return PKTCTRL_SUCCESS;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

bool is_server_auth_mode(int auth_mode) {
	// PKTLAB_AUTH_AGENT can be used for both server & client
	if (auth_mode == PKTLAB_CONTROLLER_AUTH_ENDPOINT ||
		auth_mode == PKTLAB_AUTH_AGENT)
		return true;
	return false;
}

bool is_client_auth_mode(int auth_mode) {
	if (auth_mode == PKTLAB_ENDPOINT_AUTH_CONTROLLER ||
		auth_mode == PKTLAB_AUTH_AGENT)
		return true;
	return false;
}

uint_fast32_t djb2(const void * ptr, uint_fast32_t len) {
	uint8_t * head;
	uint8_t * cursor;
	uint_fast32_t hash = 5381;
	int c;

	assert(sizeof(char) == sizeof(uint8_t) && ptr != NULL && len != 0);

	head = cursor = safe_malloc(len+1);
	memcpy(head, ptr, len);
	head[len] = '\0';

	while ((c = *cursor++))
		hash = hash * 33 ^ c;

	free(head);
	return hash;
}

uint_fast32_t single_keyid_hash_func(const void * ptr, uint_fast32_t len) {
	const uint_fast32_t * hash;
	assert(len >= PKTLAB_KEYID_LEN);
	hash = ptr;
	return *hash;
}

uint_fast32_t double_keyid_hash_func(const void * ptr, uint_fast32_t len) {
	const uint_fast32_t * hash1;
	const uint_fast32_t * hash2;
	assert(len >= PKTLAB_KEYID_LEN*2);
	hash1 = ptr;
	hash2 = (const uint_fast32_t *) ((const uint8_t *) ptr+PKTLAB_KEYID_LEN);
	return *hash1 ^ *hash2;
}

SSL_CTX * create_sslctx(void) {
	SSL_CTX * sslctx;

	ERR_clear_error();
	sslctx = SSL_CTX_new(TLS_method());
	if (sslctx == NULL) {
		error_debug_msg("SSL_CTX_new");
		return NULL;
	}

	ERR_clear_error();
	if (SSL_CTX_set_min_proto_version(
			sslctx, TLS1_3_VERSION) == 0) {
		error_debug_msg("SSL_CTX_set_min_proto_version");
		SSL_CTX_free(sslctx);
		return NULL;
	}

	SSL_CTX_set_mode(sslctx, SSL_MODE_ENABLE_PARTIAL_WRITE); // mimic normal write()

	return sslctx;
}

int accept_configure_sslctx_base (
	SSL_CTX * sslctx, struct client_hello_aux * chaux)
{
	assert(sslctx != NULL && chaux != NULL);

	// get client snikey with callback
	SSL_CTX_set_client_hello_cb(sslctx, snikey_set_keycert, chaux);

	// disable session resumption
	ERR_clear_error();
	if (SSL_CTX_set_num_tickets(sslctx, 0) != 1) {
		error_debug_msg("SSL_CTX_set_num_tickets");
		return -1;
	}

	ERR_clear_error();
	if (SSL_CTX_set_tlsext_ticket_key_cb(
			sslctx, no_session_ticket) != 1) {
		error_debug_msg("SSL_CTX_set_tlsext_ticket_key_cb");
		return -1;
	}

	return 0;
}

void accept_configure_verify (
	SSL * sslobj,
	SSL_CTX * sslctx,
	struct verify_aux * vaux)
{
	int verify_flag;

	assert(sslobj != NULL && sslctx != NULL && vaux != NULL);

	// set ssl verify mode
	// For controller verify endpoint case:
	//   The base case is the server will ask the client to provide cert,
	//   but will not complain if nothing is provided
	//   (client can decide whether to provide or not).
	//   However, if the server intended to check to cert chain,
	//   not providing results in direct TLS handshake failure.
	// For verify agent case:
	//   Server will always ask the client to provide cert,
	//   and fail if no cert is provided.
	//   This is done as for both broker verify cases
	//   the remote subjkey is needed (for pub/sub verification).
	verify_flag = SSL_VERIFY_PEER|SSL_VERIFY_CLIENT_ONCE;
	if ((vaux->root_ls != NULL && vaux->rootnum != 0) ||
	    vaux->auth_mode == PKTLAB_AUTH_AGENT) {
		verify_flag |= SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
	}

	SSL_CTX_set_verify(
		sslctx, verify_flag, NULL); // set mode only
	SSL_CTX_set_cert_verify_callback(
		sslctx,
		vaux->auth_mode != PKTLAB_NO_VERIFY ?
			yes_verify:no_verify,
		vaux);

	// As it seems that for verify_mode and verify_callback in SSL_CTX,
	// they are only checked/copied when the sslobj is created.
	// We therefore set them manually in the sslobj here.
	SSL_set_verify(sslobj, verify_flag, NULL);

	return;
}

int no_session_ticket (
	SSL * s, unsigned char key_name[16],
	unsigned char iv[EVP_MAX_IV_LENGTH],
	EVP_CIPHER_CTX * ctx, HMAC_CTX * hctx, int enc)
{
	return 0;
}

int snikey_set_keycert(SSL * sslobj, int * al, void * aux) {
	struct client_hello_aux * chaux = aux;
	struct snikey_entry * entry;
	const uint8_t * sniext;
	size_t extlen;
	char snikey[PKTCTRL_SNIKEY_LEN_MAX+2] = {};
	size_t snilen;
	int ret;

	assert(
		sslobj != NULL && al != NULL &&
		chaux != NULL && chaux->store != NULL &&
		chaux->store->type == SERVER_TYPE &&
		size_table(chaux->store->server_map.map) != 0 &&
		chaux->privkey != NULL &&
		chaux->await_sslctx != NULL &&
		chaux->conn_sslctx != NULL &&
		chaux->vaux != NULL);

	trace("%s(sslobj:%p)", __func__, sslobj);

	// get entry from snikey
	ERR_clear_error();
	entry = chaux->store->server_map.fallback; // default one
	if (SSL_client_hello_get0_ext(
			sslobj, TLSEXT_TYPE_server_name, &sniext, &extlen) == 1) {

		snilen = sizeof(snikey)-1;
		ret = parse_sni_ext(sniext, extlen, snikey, &snilen);
		if (ret < 0) {
			*al = -1;
			return SSL_CLIENT_HELLO_ERROR;
		} else if (ret > 0) {
			debug("Got SNI (len:%u): %s", (unsigned) snilen, snikey);

			// Note for here if no sni match we fail the handshake
			// We can change to use default entry instead
			if (lookup_table(
					snikey, snilen, (void **) &entry,
					chaux->store->server_map.map) < 0) {
				*al = -1;
				return SSL_CLIENT_HELLO_ERROR;
			}

			chaux->snikey = pktlab_copy_str(snikey, snilen);
			assert(chaux->snikey != NULL);
		}
		// == 0 implies default
	}

	if (entry == NULL ||
		entry->cert_ls == NULL ||
		entry->certnum == 0) {
		*al = -1;
		return SSL_CLIENT_HELLO_ERROR;
	}

	// set root_ls
	chaux->vaux->root_ls = entry->root_ls;
	chaux->vaux->rootnum = entry->rootnum;

	// configure sslctx
	accept_configure_verify(sslobj, chaux->conn_sslctx, chaux->vaux);

	if (ctx_load_cert(chaux->conn_sslctx, entry) < 0) {
		*al = -1;
		return SSL_CLIENT_HELLO_ERROR;
	}

	if (ctx_load_privkey(chaux->conn_sslctx, chaux->privkey) < 0) {
		*al = -1;
		return SSL_CLIENT_HELLO_ERROR;
	}

	ERR_clear_error();
	if (SSL_set_SSL_CTX(sslobj, chaux->conn_sslctx) == NULL) {
		error_debug_msg("SSL_set_SSL_CTX");
		return SSL_CLIENT_HELLO_ERROR;
	}

	debug("TLS set cert successful");
	return SSL_CLIENT_HELLO_SUCCESS;
}

int no_verify(X509_STORE_CTX * x509_store_ctx, void * aux) {
	STACK_OF(X509) * untrusted_stack;
	STACK_OF(X509) * trusted_stack;
	int i, rst;
	X509 * x509;
	X509 * x509dup;

	assert(x509_store_ctx != NULL && aux != NULL);

	ERR_clear_error();
	untrusted_stack = X509_STORE_CTX_get0_untrusted(x509_store_ctx);
	if (untrusted_stack == NULL) {
		error_debug_msg("X509_STORE_CTX_get0_untrusted");
		return 1;
	}

	debug("Untrusted cert list len: %d", sk_X509_num(untrusted_stack));

	ERR_clear_error();
	trusted_stack = sk_X509_new_null();
	if (trusted_stack == NULL) {
		error_debug_msg("sk_X509_new_null");
		return 1;
	}

	for (i = 0; i < sk_X509_num(untrusted_stack); ++i) {
		x509 = sk_X509_value(untrusted_stack, i);

		x509dup = X509_dup(x509);
		if (x509dup == NULL)
			break; // return what we have now

		rst = sk_X509_push(trusted_stack, x509dup);
		if (rst == 0 || rst == -1)
			break; // return what we have now
	}

	X509_STORE_CTX_set0_verified_chain(x509_store_ctx, trusted_stack);
	return 1; // always accept chain
}

int yes_verify(X509_STORE_CTX * x509_store_ctx, void * aux) {
	STACK_OF(X509) * untrusted_stack;
	struct pktlab_certificate ** cert_ls;
	X509 * x509;
	X509 * x509dup;
	uint_fast32_t certnum;
	int i, verify_rst, rst, ret = 0; // default fail verify
	struct verify_aux * vaux = aux;
	struct pktlab_certificate ** verified_cert_ls;
	uint_fast32_t verified_certnum, used_root_indx, j;
	STACK_OF(X509) * trusted_stack;

	assert(x509_store_ctx != NULL && aux != NULL);

	trace("%s(authmode:%d,rootnum:%u)",
		__func__, vaux->auth_mode, (unsigned) vaux->rootnum);

	// setup verification call
	ERR_clear_error();
	untrusted_stack = X509_STORE_CTX_get0_untrusted(x509_store_ctx);
	if (untrusted_stack == NULL) {
		error_debug_msg("X509_STORE_CTX_get0_untrusted");
		return ret;
	}
	debug("Untrusted cert list len: %d", sk_X509_num(untrusted_stack));

	cert_ls = NULL;
	certnum = 0;
	if (sk_X509_num(untrusted_stack) != 0) {
		cert_ls = safe_malloc(sk_X509_num(untrusted_stack)*sizeof(*cert_ls));
		for (i = 0; i < sk_X509_num(untrusted_stack); ++i) {
			cert_ls[i] = pktlab_create_cert();
			if (cert_ls[i] == NULL)
				goto fail_cert_ls;

			++certnum;
			if (certnum == UINT_FAST32_MAX)
				goto fail_cert_ls; // overflow possible

			x509 = sk_X509_value(untrusted_stack, i);
			if (x509 == NULL)
				goto fail_cert_ls;

			cert_ls[i]->x509 = X509_dup(x509);
			if (cert_ls[i]->x509 == NULL)
				goto fail_cert_ls;
		}
	}

	// call verification function
	verify_rst = pktlab_verify_cert_chain(
		vaux->auth_mode,
		cert_ls,
		vaux->root_ls,
		&verified_cert_ls,
		certnum,
		vaux->rootnum,
		&verified_certnum,
		&used_root_indx);
	debug("Verify result: ret:%d,verified_certnum:%u/%u,used_root_indx:%u/%u",
		verify_rst,
		(unsigned) verified_certnum, (unsigned) certnum,
		(unsigned) used_root_indx, (unsigned) vaux->rootnum);
	if (verify_rst != PKTLAB_VERIFY_SUCCESS)
		goto fail_cert_ls;

	// store verified certs
	ERR_clear_error();
	trusted_stack = sk_X509_new_null();
	if (trusted_stack == NULL) {
		error_debug_msg("sk_X509_new_null");
		goto fail_verified_cert_ls;
	}

	for (j = 0; j < verified_certnum; ++j) {
		x509dup = X509_dup(verified_cert_ls[j]->x509);
		if (x509dup == NULL) {
			sk_X509_pop_free(trusted_stack, X509_free);
			goto fail_verified_cert_ls;
		}

		rst = sk_X509_push(trusted_stack, x509dup);
		if (rst == 0 || rst == -1) {
			sk_X509_pop_free(trusted_stack, X509_free);
			goto fail_verified_cert_ls;
		}
	}

	X509_STORE_CTX_set0_verified_chain(x509_store_ctx, trusted_stack);
	ret = 1;

fail_verified_cert_ls:
	free(verified_cert_ls);
fail_cert_ls:
	for (j = 0; j < certnum; ++j)
		pktlab_cleanup_cert(cert_ls[j]);
	free(cert_ls);

	return ret;
}

int try_setup(struct pktctrl_obj * ctrlobj, ssl_setup_t ssl_setup) {
	int ret, err;
	SSL * sslobj;

	assert(
		ctrlobj != NULL && ctrlobj->type == SSL_TYPE &&
		ctrlobj->state == ST_HANDSHAKING && ssl_setup != NULL);

	sslobj = ctrlobj->saux->sslobj;
	ctrlobj->hs_need_read = ctrlobj->hs_need_write = false;

	ERR_clear_error();
	ret = ssl_setup(sslobj);
	if (ret == 1)
		return PKTCTRL_SUCCESS;

	err = SSL_get_error(sslobj, ret);
	debug("SSL_get_error(%d) returned %d", ret, err);

	switch (err) {
	case SSL_ERROR_WANT_READ:
		ctrlobj->hs_need_read = true;
		return PKTCTRL_IN_PROGRESS;
	case SSL_ERROR_WANT_WRITE:
		ctrlobj->hs_need_write = true;
		return PKTCTRL_IN_PROGRESS;
	case SSL_ERROR_SYSCALL:
		debug("SSL_ERROR_SYSCALL: %s", strerror(errno));
		break;
	case SSL_ERROR_SSL:
	default:
		error_debug_msg(
			(ssl_setup == SSL_accept) ?
				"SSL_accept":"SSL_connect");
	}

	return PKTCTRL_SSL_FAILURE;
}

int parse_sni_ext (
	const uint8_t * ext, size_t extlen,
	char * buf, size_t * buflen)
{
	uint16_t fieldlen;
	const uint8_t * ptr;

	// For how to parse the extension field, refer to:
	// 1. RFC 8446 section 3 (https://datatracker.ietf.org/doc/html/rfc8446#section-3)
	// 2. RFC 8446 section 4.1.2 (https://datatracker.ietf.org/doc/html/rfc8446#section-4.1.2)
	// 3. RFC 8446 section 4.2 (https://datatracker.ietf.org/doc/html/rfc8446#section-4.2)
	// 4. RFC 6066 section 3 (https://datatracker.ietf.org/doc/html/rfc6066#section-3)
	// much better than ASN.1 imo

	assert(ext != NULL && buf != NULL && buflen != NULL && *buflen >= PKTCTRL_SNIKEY_LEN_MAX+1);

	if (extlen <= 2)
		return 0; // 0 indicates no sni specified

	ptr = ext; // pointing at start of server_name_list
	fieldlen = pktlab_get16n(ptr);
	if (fieldlen+2 != extlen) // check if extension_data len match up
		return -1; // -1 implies invalid sni extension

	ptr += 2; // pointing at name_type
	if (*ptr != 0) // 0 implies host_name
		return -1;

	ptr += 1; // pointing at host_name
	fieldlen = pktlab_get16n(ptr); // host_name len
	if (fieldlen+5 != extlen)
		return -1;
	if (fieldlen > PKTCTRL_SNIKEY_LEN_MAX || fieldlen == 0)
		return -1;

	ptr += 2; // pointing at host_name (actual name)
	strncpy(buf, (const char *) ptr, fieldlen);
	buf[fieldlen] = '\0';
	if (!pktctrl_is_valid_snikey(buf))
		return -1;

	*buflen = fieldlen;
	return 1;
}

int ctx_load_cert (
	SSL_CTX * ctx, struct snikey_entry * entry)
{
	uint_fast32_t i;
	X509 * dup;

	assert(ctx != NULL && entry != NULL && entry->certnum > 0);

	for (i = 0; i < entry->certnum; ++i) {
		if (i == 0) {
			ERR_clear_error();

			// first cert in list should be the cert corresponding to privkey
			if (SSL_CTX_use_certificate(ctx, entry->cert_ls[i]->x509) != 1) {
				error_debug_msg("SSL_CTX_use_certificate");
				return -1;
			}

			continue;
		}

		dup = X509_dup(entry->cert_ls[i]->x509);
		if (dup == NULL) {
			debug("X509_dup failed");
			return -1;
		}

		ERR_clear_error();
		if (SSL_CTX_add_extra_chain_cert(ctx, dup) != 1) {
			error_debug_msg("SSL_CTX_add_extra_chain_cert");
			X509_free(dup);
			return -1;
		}
	}

	return 0;
}

int ctx_load_privkey (
	SSL_CTX * ctx, struct pktlab_privatekey * key)
{
	assert(ctx != NULL && key != NULL && key->key != NULL);

	ERR_clear_error();
	if (SSL_CTX_use_PrivateKey(ctx, key->key) != 1) {
		error_debug_msg("SSL_CTX_use_PrivateKey");
		return -1;
	}

	return 0;
}

int get_peer_certificate (
	struct pktlab_certificate *** peer_cert_ls,
	uint_fast32_t * peer_cert_num, const SSL * sslobj)
{
	STACK_OF(X509) * x509_stack;
	uint_fast32_t totalmem;
	int cert_cnt, i;
	struct pktlab_certificate ** addrptr;
	struct pktlab_certificate * contentptr;
	X509 * x509;

	assert(peer_cert_ls != NULL && peer_cert_num != NULL && sslobj != NULL);

	ERR_clear_error();
	x509_stack = SSL_get0_verified_chain(sslobj); // do NOT free this!!
	if (x509_stack == NULL) { // fail or the peer didn't provide a cert chain
		debug("verify rst: %ld", SSL_get_verify_result(sslobj));
		debug("SSL_get0_verified_chain returned NULL");
		goto fail;
	}

	cert_cnt = sk_X509_num(x509_stack);
	if (cert_cnt < 0 || cert_cnt > UINT_FAST32_MAX)
		goto fail;

	totalmem = cert_cnt * sizeof(**peer_cert_ls);
	if (totalmem / sizeof(**peer_cert_ls) != cert_cnt || totalmem > SIZE_MAX)
		goto fail;

	addrptr = *peer_cert_ls = safe_malloc(totalmem);
	*peer_cert_num = 0;
	for (i = 0; i < cert_cnt; ++i) {
		contentptr = pktlab_create_cert();
		if (contentptr == NULL)
			goto fail_free;

		*addrptr = contentptr;
		++addrptr;
		++*peer_cert_num;

		x509 = sk_X509_value(x509_stack, i);
		contentptr->x509 = X509_dup(x509);
		if (contentptr->x509 == NULL)
			goto fail_free;
	}

	return 0;

fail_free:
	addrptr = *peer_cert_ls;
	for (i = 0; i < *peer_cert_num; ++i) {
		pktlab_cleanup_cert(*addrptr);
		++addrptr;
	}
	free(*peer_cert_ls);
fail:
	*peer_cert_ls = NULL;
	*peer_cert_num = 0;
	return -1;
}

ssize_t read_ssl (
	void * obj, void * buf, size_t buflen)
{
	int result;
	struct pktctrl_obj * ctrlobj = obj;

	trace("%s(state:%d,type:%d,sockfd:%d,buflen:%u)", __func__,
		ctrlobj->state, ctrlobj->type, ctrlobj->sockfd, (unsigned) buflen);

	if (buf == NULL ||
	    ctrlobj == NULL ||
	    ctrlobj->type != SSL_TYPE ||
	    ctrlobj->state != ST_ESTABLISHED) {
		errno = 0;
		return -1;
	}

	ERR_clear_error();
	buflen = (buflen > INT_MAX) ? INT_MAX : buflen;
	buflen = (buflen > SSIZE_MAX) ? SSIZE_MAX : buflen;
	result = SSL_read(ctrlobj->saux->sslobj, buf, buflen);
	post_ssl_read(ctrlobj, &result);
	return result;
}

void post_ssl_read(struct pktctrl_obj * obj, int * result) {
	SSL * sslobj;
	int ssl_error;

	assert(
		obj != NULL && obj->type == SSL_TYPE &&
		obj->state == ST_ESTABLISHED && result != NULL);

	if (*result > 0) { // > 0 means good result
		obj->read_need_write = false;
		return;
	}

	sslobj = obj->saux->sslobj;
	ssl_error = SSL_get_error(sslobj, *result);
	debug("SSL_read error (ret:%d,ssl error no:%d)", *result, ssl_error);

	switch (ssl_error) {
	case SSL_ERROR_WANT_READ:
		obj->read_need_write = false;
		errno = EWOULDBLOCK;
		*result = -1;
		break;
	case SSL_ERROR_WANT_WRITE:
		obj->read_need_write = true;
		errno = EWOULDBLOCK;
		*result = -1;
		break;
	case SSL_ERROR_SYSCALL:
		obj->state = ST_ERROR;

		if (errno != 0) {
			debug("SSL_read got SSL_ERROR_SYSCALL with errno: %d (%s)",
				errno, strerror(errno));
			errno = 0; // in case errno already EWOULDBLOCK or EAGAIN
			*result = -1;
			return;
		}

		// errno == 0 implies socket EOF; fall through
	case SSL_ERROR_ZERO_RETURN:
		*result = 0; // mimic EOF
		break;
	case SSL_ERROR_SSL:
		debug("SSL_read got SSL_ERROR_SSL, masking with ECONNRESET (%d)", ECONNRESET);
		error_debug_msg("SSL_read");
		// for now we mask underlying SSL error with ECONNRESET,
		// probably should change this afterwards
		errno = ECONNRESET;
		*result = -1;
		obj->state = ST_ERROR;
		break;
	default:
		debug("SSL_read got unknown ssl error: %d", ssl_error);
		error_debug_msg("SSL_read");
		errno = 0; // in case errno already EWOULDBLOCK or EAGAIN
		*result = -1;
		obj->state = ST_ERROR;
	}
}

ssize_t write_ssl (
	void * obj, const void * buf, size_t nbyte)
{
	int result;
	struct pktctrl_obj * ctrlobj = obj;

	trace("%s(state:%d,type:%d,sockfd:%d,nbyte:%u)",
		__func__, ctrlobj->state, ctrlobj->type,
		ctrlobj->sockfd, (unsigned) nbyte);

	if (buf == NULL ||
	    nbyte == 0 ||
	    ctrlobj == NULL ||
	    ctrlobj->type != SSL_TYPE ||
	    ctrlobj->state != ST_ESTABLISHED) {
		errno = 0;
		return -1;
	}

	ERR_clear_error();
	nbyte = (nbyte > INT_MAX) ? INT_MAX : nbyte;
	nbyte = (nbyte > SSIZE_MAX) ? SSIZE_MAX : nbyte;
	result = SSL_write(ctrlobj->saux->sslobj, buf, nbyte);
	post_ssl_write(ctrlobj, &result);
	return result;
}

void post_ssl_write(struct pktctrl_obj * obj, int * result) {
	SSL * sslobj;
	int ssl_error;

	assert(
		obj != NULL && obj->type == SSL_TYPE &&
		obj->state == ST_ESTABLISHED && result != NULL);

	if (*result > 0) {
		obj->write_need_read = false;
		return;
	}

	sslobj = obj->saux->sslobj;
	ssl_error = SSL_get_error(sslobj, *result);
	debug("SSL_write error (ret:%d,error no:%d)", *result, ssl_error);

	switch (ssl_error) {
	case SSL_ERROR_WANT_READ:
		obj->write_need_read = true;
		errno = EWOULDBLOCK;
		*result = -1;
		break;
	case SSL_ERROR_WANT_WRITE:
		obj->write_need_read = false;
		errno = EWOULDBLOCK;
		*result = -1;
		break;
	case SSL_ERROR_SYSCALL:
		if (errno != EPIPE) {
			debug("SSL_write got SSL_ERROR_SYSCALL with errno: %d (%s)",
				errno, strerror(errno));
			errno = 0; // in case errno already EWOULDBLOCK or EAGAIN
		}

		obj->state = ST_ERROR;
		*result = -1;
		return;
	case SSL_ERROR_ZERO_RETURN:
		errno = EPIPE;
		obj->state = ST_ERROR;
		*result = -1; // mimic broken pipe
		break;
	case SSL_ERROR_SSL:
		obj->state = ST_ERROR;
		// fall through
	default:
		debug("SSL_write got unknown ssl error: %d", ssl_error);
		error_debug_msg("SSL_write");
		errno = 0; // in case it is already EWOULDBLOCK or EAGAIN
		*result = -1;
	}
}