#include "callbackmgr.h"
#include "util/debug.h"
#include "util/gmr_chk.h"
#include <assert.h>
#include <errno.h>
#include <string.h>

#define DEFAULT_CQ_SIZE 32
#define INVOKE(_mgr, _msg, _cb)                                                \
    do {                                                                       \
        if (_cb->installed && !_cb->discard) {                                 \
            ret = ((pktif_cbfunc_t)(_cb->cb).fn)(_mgr->pktif, _msg,            \
                                                 (_cb->cb).param);             \
        }                                                                      \
    } while (0)

struct callbackmgr *callbackmgr_create(struct pktif *pktif) {
    struct callbackmgr *mgr = malloc(sizeof(*mgr));
    struct cyc_queue *cq = cyc_queue_init(DEFAULT_CQ_SIZE);
    if (!(mgr && cq)) {
        if (mgr)
            free(mgr);
        if (cq)
            cyc_queue_destroy(cq);
        return NULL;
    }

    mgr->pktif = pktif;
    mgr->msg_cb = cq;
    memset(mgr->ndata_cb, 0, sizeof(mgr->ndata_cb));
    memset(mgr->nstat_cb, 0, sizeof(mgr->nstat_cb));
    memset(mgr->ntag_cb, 0, sizeof(mgr->ntag_cb));
    memset(mgr->ndrop_cb, 0, sizeof(mgr->ndrop_cb));
    memset(mgr->data_buf, 0, sizeof(mgr->data_buf));
    return mgr;
}

int callbackmgr_teardown(struct callbackmgr *mgr) {
    if (!mgr)
        return -1;

    void *ptr;
    while ((ptr = cyc_queue_pop(mgr->msg_cb))) {
        free(ptr);
    }
    cyc_queue_destroy(mgr->msg_cb);

    for (int i = 0; i < sizeof(mgr->data_buf) / sizeof(mgr->data_buf[0]); ++i) {
        if (mgr->data_buf[i]) {
            while ((ptr = cyc_queue_pop(mgr->data_buf[i])))
                free(ptr);
            cyc_queue_destroy(mgr->data_buf[i]);
        }
    }

    free(mgr);
    return 0;
}

static inline struct general_cb *get_cbptr(struct callbackmgr *mgr,
                                           struct general_cb *cbptr) {
    switch (cbptr->type) {
    case CB_TYPE_NDATA:
        return &mgr->ndata_cb[cbptr->ndata.sktid];
    case CB_TYPE_NTAG:
        return &mgr->ntag_cb[cbptr->ntag.tag];
    case CB_TYPE_NSTAT:
        return &mgr->nstat_cb[cbptr->nstat.sktid][cbptr->nstat.tcpev];
    case CB_TYPE_NDROP:
        return &mgr->ndrop_cb[cbptr->ndrop.sktid];
    default:
        return NULL;
    }
}

int general_cb_ctl(struct callbackmgr *mgr, int op, struct general_cb *cbptr) {
    if (!mgr)
        return -1;

    if (!cbptr) {
        errno = EINVAL;
        return -1;
    }
    switch (op) {
    case PKTIF_CB_CTL_ADD: {
        if (cbptr->type == CB_TYPE_MSG) {
            cyc_queue_push(mgr->msg_cb, cbptr);
            return 0;
        } else if (cbptr->type < CB_TYPE_MAX) {
            // TODO: add default nstat handler
            struct general_cb *mgrcb_ptr = get_cbptr(mgr, cbptr);
            if (mgrcb_ptr == NULL || mgrcb_ptr->installed) {
                errno = EINVAL;
                return -1;
            }
            *mgrcb_ptr = *cbptr;
            mgrcb_ptr->installed = true;
            return 0;
        } else {
            errno = EINVAL;
            return -1;
        }
    }
    case PKTIF_CB_CTL_DEL: {
        if (cbptr->type > CB_TYPE_MSG && cbptr->type < CB_TYPE_MAX) {
            struct general_cb *mgrcb_ptr = get_cbptr(mgr, cbptr);
            // TODO: add default nstat handler
            if (mgrcb_ptr == NULL || mgrcb_ptr->installed == false) {
                errno = EINVAL;
                return -1;
            }
            *cbptr = *mgrcb_ptr;
            mgrcb_ptr->installed = false;
            return 0;
        } else {
            errno = EINVAL;
            return -1;
        }
    }
    case PKTIF_CB_CTL_MOD: {
        if (cbptr->type > CB_TYPE_MSG && cbptr->type < CB_TYPE_MAX) {
            struct general_cb *mgrcb_ptr = get_cbptr(mgr, cbptr);
            if (mgrcb_ptr == NULL || mgrcb_ptr->installed == false) {
                errno = EINVAL;
                return -1;
            }
            struct general_cb temp = *mgrcb_ptr;
            *mgrcb_ptr = *cbptr;
            *cbptr = temp;
            mgrcb_ptr->installed = true;
            return 0;
        } else {
            errno = EINVAL;
            return -1;
        }
    }
    default: {
        errno = EINVAL;
        return -1;
    }
    }
    return -1;
}

static inline bool is_notification(struct pktlab_message *msg) {
    return msg->type == PKTLAB_NTAG_MESSAGE ||
           msg->type == PKTLAB_NSTAT_MESSAGE ||
           msg->type == PKTLAB_NDATA_MESSAGE ||
           msg->type == PKTLAB_NDROP_MESSAGE;
}

static int invoke_notification_cb(struct callbackmgr *mgr,
                                  struct pktlab_message *msg) {
    if (!mgr || !msg)
        return -1;
    debug("invoke notification cb on type %d", msg->type);
    uint8_t sktid = 0;
    int ret = 0;
    struct general_cb *gcb = NULL;
    switch (msg->type) {
    case PKTLAB_NDATA_MESSAGE:
        sktid = msg->ndata.sktid;
        gcb = &mgr->ndata_cb[sktid];
        INVOKE(mgr, msg, gcb);
        // no callback registered, buffer the message
        // FIXME: empty the data queue on addcb
        if (!gcb->installed) {
            if (mgr->data_buf[sktid] == NULL)
                mgr->data_buf[sktid] = cyc_queue_init(DEFAULT_CQ_SIZE);

            assert(mgr->data_buf[sktid]);
            cyc_queue_push(mgr->data_buf[sktid], msg);
            return 0;
        }
        break;
    case PKTLAB_NSTAT_MESSAGE:
        sktid = msg->nstat.sktid;
        gcb = &mgr->nstat_cb[sktid][msg->nstat.code];
        INVOKE(mgr, msg, gcb);
        break;
    case PKTLAB_NDROP_MESSAGE:
        sktid = msg->ndrop.sktid;
        gcb = &mgr->ndrop_cb[sktid];
        INVOKE(mgr, msg, gcb);
        break;
    case PKTLAB_NTAG_MESSAGE:
        gcb = &mgr->ntag_cb[msg->ntag.tag];
        INVOKE(mgr, msg, gcb);
        break;
    default:
        warn("unknown notification type %d", msg->type);
        return -1;
    }
    free(msg);
    return ret;
}

int invoke_cb(struct callbackmgr *mgr, struct pktlab_message *msg) {
    if (!mgr || !msg)
        return -1;

    if (is_notification(msg)) {
        return invoke_notification_cb(mgr, msg);
    }
    return invoke_msg_cb(mgr, msg);
}

int invoke_msg_cb(struct callbackmgr *mgr, struct pktlab_message *msg) {
    if (!mgr || !msg)
        return -1;
    debug("invoke msg cb on type %d", msg->type);

    struct general_cb *cb = cyc_queue_pop(mgr->msg_cb);
    if (cb == NULL) {
        // FATAL ABORT
        warn("no handler for msg found");
        return -1;
    }

    if (chk_gmr(cb->msg.req_type, msg->type) == 0) {
        warn("message type grammar check failed");
    }

    int ret = 0;
    INVOKE(mgr, msg, cb);
    if (cb->discard) {
        // buffer the msg
        // or discard the msg
    }
    free(msg);
    free(cb);
    return ret;
}

int register_msg_cb(struct callbackmgr *mgr, pktif_cb_t *cb,
                    enum pktlab_message_type req_type) {
    struct general_cb *gcb = malloc(sizeof(*gcb));
    if (gcb == NULL) {
        errno = ENOMEM;
        return -1;
    }
    *gcb = (struct general_cb){
        .type = CB_TYPE_MSG,
        .installed = true,
        .cb = *cb,
        .discard = false,
        .msg.req_type = req_type,
    };
    return general_cb_ctl(mgr, PKTIF_CB_CTL_ADD, gcb);
}

int notif_cb_ctl(struct callbackmgr *mgr, int op, uint8_t type, bool discard,
                 uint8_t sktid, uint16_t tag, uint8_t tcpev, pktif_cb_t *cb) {
    if (!mgr) {
        errno = EINVAL;
        return -1;
    }
    if (!(op >= PKTIF_CB_CTL_ADD && op <= PKTIF_CB_CTL_MOD)) {
        errno = EINVAL;
        return -1;
    }
    int rv = 0;
    // construct the general callback
    struct general_cb gcb = {.type = type, .installed = true};
    if (op == PKTIF_CB_CTL_ADD || op == PKTIF_CB_CTL_MOD) {
        if (cb == NULL && !discard) {
            errno = EINVAL;
            return -1;
        }
        gcb.discard = discard;
        switch (type) {
        case CB_TYPE_NDATA:
            if (!discard) {
                gcb.cb = *cb;
                gcb.ndata.sktid = sktid;
            }
            break;
        case CB_TYPE_NTAG:
            if (!discard) {
                gcb.cb = *cb;
                gcb.ntag.tag = tag;
                if (tag == 0) {
                    // installing a callback on tag 0 will never get triggered
                    errno = EINVAL;
                    return -1;
                }
            }
            break;
        case CB_TYPE_NSTAT:
            gcb.discard = discard;
            if (!discard) {
                if (tcpev >= PKTLAB_TCPEV_MAX) {
                    errno = EINVAL;
                    error("Invalid tcpev");
                    return -1;
                }
                gcb.cb = *cb;
                gcb.nstat.sktid = sktid;
                gcb.nstat.tcpev = tcpev;
            }
            break;
        case CB_TYPE_NDROP:
            gcb.discard = discard;
            if (!discard) {
                gcb.cb = *cb;
                gcb.ndrop.sktid = sktid;
            }
            break;
        default:
            errno = EINVAL;
            return -1;
        }
    } else {
        switch (type) {
        case CB_TYPE_NDATA:
            gcb.ndata.sktid = sktid;
            break;
        case CB_TYPE_NTAG:
            if (tag == 0) {
                error("Trying to delete ntag 0 notification");
                errno = EINVAL;
                return -1;
            }
            gcb.ntag.tag = tag;
            break;
        case CB_TYPE_NSTAT:
            gcb.nstat.sktid = sktid;
            if (tcpev >= PKTLAB_TCPEV_MAX) {
                errno = EINVAL;
                error("Invalid tcpev");
                return -1;
            }
            gcb.nstat.tcpev = tcpev;
            break;
        case CB_TYPE_NDROP:
            gcb.ndrop.sktid = sktid;
            break;
        default:
            errno = EINVAL;
            return -1;
        }
    }

    // install the callback function
    rv = general_cb_ctl(mgr, op, &gcb);
    if (rv == -1)
        return -1;

    if (op == PKTIF_CB_CTL_MOD || op == PKTIF_CB_CTL_DEL) {
        if (cb) {
            *cb = gcb.cb;
        }
    }
    return rv;
}