#include "callbackmgr.h"
#include "struct_pktif.h"
#include "util/barrier.h"
#include "util/debug.h"
#include "util/util.h"
#include <errno.h>
#include <pktif.h>
#include <string.h>
#include <sys/select.h>

static void create_mread(struct pktlab_message *msg, uint32_t memaddr,
                         uint32_t bytecnt) {
    msg->type = PKTLAB_MREAD_MESSAGE;
    msg->mread.addr = memaddr;
    msg->mread.len = bytecnt;
}

static int mread_helper(struct pktif *pktif, uint32_t memaddr, uint32_t bytecnt,
                        pktif_cb_t *cb) {
    if (!pktif || !cb) {
        errno = EINVAL;
        return -1;
    }

    struct pktlab_message send_msg;
    create_mread(&send_msg, memaddr, bytecnt);

    if (try_send(pktif->ctrlobj, &send_msg) != 1)
        return -1;
    return register_msg_cb(pktif->cbmgr, cb, PKTLAB_MREAD_MESSAGE);
}

static int time_cb(struct pktif *pktif, struct pktlab_message *msg,
                   void *bar_param) {
    struct bar_param *p = bar_param;
    if (msg->type != PKTLAB_MDATA_MESSAGE ||
        msg->mdata.len != sizeof(pktlab_time_t)) {
        p->barrier = 0;
        return -1;
    }

    *(pktlab_time_t *)p->param = pktlab_get64n(msg->mdata.ptr);

    p->barrier = 0;
    return 0;
}

int pktif_mread_systime_sync(struct pktif *pktif, pktlab_time_t *time) {
    struct bar_param param = {.barrier = 1, .param = time};
    int rv;
    pktif_cb_t cb = {.fn = &time_cb, .param = &param};
    if ((rv = pktif_mread_systime_async(pktif, &cb))) {
        return rv;
    }
    SYNC_LOOP(param.barrier)
    return 0;
}

int pktif_mread_systime_async(struct pktif *pktif, pktif_cb_t *cb) {
    return mread_helper(pktif, PKTLAB_VMEMADDR_SYSPARAM_TIME,
                        sizeof(pktlab_time_t), cb);
}

int pktif_mread_sendtime_sync(struct pktif *pktif, uint16_t tidx,
                              pktlab_time_t *time) {
    struct bar_param param = {.barrier = 1, .param = time};
    int rv;
    pktif_cb_t cb = {.fn = &time_cb, .param = &param};
    if ((rv = pktif_mread_sendtime_async(pktif, tidx, &cb))) {
        return rv;
    }
    SYNC_LOOP(param.barrier)
    return 0;
}

static int copy_cb(struct pktif *pktif, struct pktlab_message *msg,
                   void *param) {
    if (msg->type != PKTLAB_MDATA_MESSAGE)
        return -1;

    struct bar_param *bar_param = param;
    memcpy(bar_param->param, msg->mdata.ptr, msg->mdata.len);
    bar_param->barrier = 0;
    return 0;
}

int pktif_mread_sendtime_async(struct pktif *pktif, uint16_t tidx,
                               pktif_cb_t *cb) {
    if (!cb)
        return 0;
    uint32_t memaddr = PKTLAB_VMEMADDR_NSENDTSTP + sizeof(pktlab_time_t) * tidx;
    return mread_helper(pktif, memaddr, sizeof(pktlab_time_t), cb);
}

int pktif_mread_sync(struct pktif *pktif, uint32_t memaddr, uint32_t bytecnt,
                     void *buffer) {
    if (!pktif || !buffer)
        return -1;

    struct bar_param param = {.barrier = 1, .param = buffer};
    pktif_cb_t cb = {.fn = &copy_cb, .param = &param};
    int rv;
    if ((rv = pktif_mread_async(pktif, memaddr, bytecnt, &cb))) {
        return rv;
    }

    SYNC_LOOP(param.barrier)

    return 0;
}

int pktif_mread_async(struct pktif *pktif, uint32_t memaddr, uint32_t bytecnt,
                      pktif_cb_t *cb) {
    return mread_helper(pktif, memaddr, bytecnt, cb);
}

int pktif_mread_ipv4_dns_addr_sync(struct pktif *pktif, int idx,
                                   struct sockaddr_in *addr) {
    if (!pktif || !addr) {
        return -1;
    }
    addr->sin_family = AF_INET;
    addr->sin_port = htons(53);
    struct bar_param param = {.barrier = 1, .param = &addr->sin_addr.s_addr};
    pktif_cb_t cb = {.fn = &copy_cb, .param = &param};
    int rv;
    if ((rv = pktif_mread_ipv4_dns_addr_async(pktif, idx, &cb))) {
        return rv;
    }
    SYNC_LOOP(param.barrier)
    return 0;
}

int pktif_mread_ipv4_dns_addr_async(struct pktif *pktif, int idx,
                                    pktif_cb_t *cb) {
    return mread_helper(pktif,
                        PKTLAB_VMEMADDR_SYSPARAM_IP4DNSLS +
                            idx * PKTLAB_VMEMADDR_SYSPARAM_IP4DNSLS_FL,
                        PKTLAB_VMEMADDR_SYSPARAM_IP4DNSLS_FL, cb);
}
