#include "callbackmgr.h"
#include "struct_pktif.h"
#include "util/barrier.h"
#include "util/debug.h"
#include "util/util.h"
#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <netdb.h>
#include <pktif.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#define PROTO_2_ADDRLEN(proto)                                                 \
    (((proto & PKTLAB_NETPROTO_MASK) == PKTLAB_IP4_PROTO)                      \
         ? PKTLAB_IPV4_WO_MSK_ADDR_LEN                                         \
         : PKTLAB_IPV6_ADDR_LEN)

#define ASSIGN_ADDR(addrptr, addrlen, portptr, portlen, addr)                  \
    do {                                                                       \
        addrptr = NULL;                                                        \
        addrlen = 0;                                                           \
        portptr = NULL;                                                        \
        portlen = 0;                                                           \
        if (addr == NULL)                                                      \
            break;                                                             \
        if (addr->sa_family == AF_INET) {                                      \
            struct sockaddr_in *addr_in = (struct sockaddr_in *)addr;          \
            addrptr = &addr_in->sin_addr;                                      \
            addrlen = sizeof(addr_in->sin_addr);                               \
            portptr = &addr_in->sin_port;                                      \
            portlen = sizeof(addr_in->sin_port);                               \
        }                                                                      \
                                                                               \
        else if (addr->sa_family == AF_INET6) {                                \
            struct sockaddr_in6 *addr_in6 = (struct sockaddr_in6 *)addr;       \
            addrptr = &addr_in6->sin6_addr;                                    \
            addrlen = sizeof(addr_in6->sin6_addr);                             \
            portptr = &addr_in6->sin6_port;                                    \
            portlen = sizeof(addr_in6->sin6_port);                             \
        }                                                                      \
    } while (0)

static inline uint8_t af_family_to_pktlab_proto(sa_family_t family) {
    return family == AF_INET ? PKTLAB_IP4_PROTO : PKTLAB_IP6_PROTO;
}

static int status_cb(struct pktif *pktif, struct pktlab_message *msg,
                     void *param) {
    struct bar_param *p = param;
    uint8_t *status = p->param;
    *status = msg->result.errid;
    p->barrier = 0;
    return msg->result.errid;
}

int pktif_ep_accept(struct sockaddr *addr) {
    // NoOp returns the fd to general ec
    return 0;
}

int pktif_ep_connect(struct sockaddr *addr) {
    // NoOp returns the fd to general ec
    return 0;
}

struct pktif *pktif_init(int sock, struct sockaddr *addr) {
    int rst;
    struct pktif *pktif = malloc(sizeof(*pktif));
    if (!pktif) {
        errno = ENOMEM;
        return NULL;
    }
    memset(pktif, 0, sizeof(*pktif));

    pktif->sock = sock;

    pktif->ctrlobj = pktctrl_create_obj();
    if (pktif->ctrlobj == NULL) {
        pktif_teardown(pktif);
        perror_exit("pktctrl_create_obj", -1);
    }

    rst = pktctrl_raw_session(sock, pktif->ctrlobj);
    if (rst != PKTCTRL_SUCCESS) {
        pktif_teardown(pktif);
        print_err_exit(-1, "pktctrl_raw_session %d", rst);
    }

    pktif->cbmgr = callbackmgr_create(pktif);
    if (!pktif->cbmgr) {
        pktif_teardown(pktif);
        perror_exit("create callbackmgr", -1);
    }
    if (addr)
        pktif->addr = *addr;
    return pktif;
}

int pktif_teardown(struct pktif *pktif) {
    if (!pktif)
        return -1;

    pktctrl_close(pktif->ctrlobj);
    callbackmgr_teardown(pktif->cbmgr);
    close(pktif->sock);
    free(pktif);
    return 0;
}

int pktif_get_fileno(struct pktif *pktif) {
    if (!pktif)
        return -1;

    return pktif->sock;
}

static void create_nopen(struct pktlab_message *msg, uint8_t sktid,
                         uint8_t proto, struct sockaddr *laddr,
                         struct sockaddr *raddr, uint32_t rbufsz) {

    msg->type = PKTLAB_NOPEN_MESSAGE;

    ASSIGN_ADDR(msg->nopen.laddrptr, msg->nopen.laddrlen, msg->nopen.lportptr,
                msg->nopen.lportlen, laddr);
    ASSIGN_ADDR(msg->nopen.raddrptr, msg->nopen.raddrlen, msg->nopen.rportptr,
                msg->nopen.rportlen, raddr);

    msg->nopen.sktid = sktid;
    if (laddr)
        msg->nopen.family = af_family_to_pktlab_proto(laddr->sa_family);
    if (raddr)
        msg->nopen.family = af_family_to_pktlab_proto(raddr->sa_family);
    msg->nopen.proto = proto;
    msg->nopen.rbufsz = rbufsz;
}

int pktif_connect_sync(struct pktif *pktif, uint8_t sktid, uint8_t proto,
                       struct sockaddr *laddr, struct sockaddr *raddr,
                       uint32_t rbufsz, pktif_cb_t *recv_cb) {
    if (!pktif || (!laddr && !raddr)) {
        errno = EINVAL;
        return -1;
    }

    uint8_t status = -1;
    struct bar_param param = {.barrier = 1, .param = &status};
    pktif_cb_t cb = {.fn = &status_cb, .param = &param};
    int rv;
    if ((rv = pktif_connect_async(pktif, sktid, proto, laddr, raddr, rbufsz,
                                  recv_cb, &cb))) {
        return rv;
    }

    SYNC_LOOP(param.barrier)

    return status;
}

int pktif_connect_async(struct pktif *pktif, uint8_t sktid, uint8_t proto,
                        struct sockaddr *laddr, struct sockaddr *raddr,
                        uint32_t rbufsz, pktif_cb_t *recv_cb,
                        pktif_cb_t *con_cb) {
    if (!pktif || (!laddr && !raddr) || !con_cb) {
        errno = EINVAL;
        return -1;
    }

    if (laddr) {
        if (laddr->sa_family != AF_INET && laddr->sa_family != AF_INET6) {
            errno = EINVAL;
            return -1;
        }
    }
    if (raddr) {
        if (raddr->sa_family != AF_INET && raddr->sa_family != AF_INET6) {
            errno = EINVAL;
            return -1;
        }
    }
    int rv = 0;
    if (recv_cb) {
        rv = notif_cb_ctl(pktif->cbmgr, PKTIF_CB_CTL_ADD, CB_TYPE_NDATA, false,
                          sktid, 0, 0, recv_cb);
        if (rv == -1)
            return rv;
    }

    struct pktlab_message send_msg;
    create_nopen(&send_msg, sktid, proto, laddr, raddr, rbufsz);

    if (try_send(pktif->ctrlobj, &send_msg) != 1) {
        if (recv_cb)
            notif_cb_ctl(pktif->cbmgr, PKTIF_CB_CTL_DEL, CB_TYPE_NDATA, false,
                         sktid, 0, 0, NULL);
        return -1;
    }
    pktif->sktinfo[sktid].proto = proto;
    pktif->sktinfo[sktid].family = send_msg.nopen.family;

    rv = register_msg_cb(pktif->cbmgr, con_cb, PKTLAB_NOPEN_MESSAGE);
    if (rv == -1) {
        if (recv_cb)
            notif_cb_ctl(pktif->cbmgr, PKTIF_CB_CTL_DEL, CB_TYPE_NDATA, false,
                         sktid, 0, 0, NULL);
        return -1;
    }
    return rv;
}

static void create_nsend(struct pktlab_message *msg, uint8_t sktid,
                         uint8_t family, uint8_t proto, pktlab_time_t time,
                         uint16_t tag, struct sockaddr *rudpaddr,
                         const void *ptr, uint32_t len) {
    msg->type = PKTLAB_NSEND_MESSAGE;
    msg->nsend.sktid = sktid;
    msg->nsend.proto = proto;
    msg->nsend.time = time;
    msg->nsend.tag = tag;
    msg->nsend.ptr = ptr;
    msg->nsend.len = len;
    msg->nsend.family = family;
    if (proto == PKTLAB_UDP_PROTO && rudpaddr) {
        ASSIGN_ADDR(msg->nsend.udp.raddrptr, msg->nsend.udp.raddrlen,
                    msg->nsend.udp.rportptr, msg->nsend.udp.rportlen, rudpaddr);
    }
}

static int time_cb(struct pktif *pktif, struct pktlab_message *msg,
                   void *bar_param) {
    struct bar_param *p = bar_param;
    assert(msg->type == PKTLAB_NTAG_MESSAGE);
    if (p->param) {
        if (msg->ntag.errid != PKTLAB_SUCCESS) {
            *(pktlab_time_t *)p->param = 0;
        } else {
            *(pktlab_time_t *)p->param = msg->ntag.time;
        }
    }

    p->barrier = 0;
    notif_cb_ctl(pktif->cbmgr, PKTIF_CB_CTL_DEL, CB_TYPE_NTAG, false, 0,
                 msg->ntag.tag, 0, NULL);
    return 0;
}

ssize_t pktif_send_sync(struct pktif *pktif, uint8_t sktid, void *buffer,
                       size_t len, uint8_t proto, pktlab_time_t send_time,
                       uint16_t tag, struct sockaddr *rudpaddr,
                       pktlab_time_t *ep_send_time) {
    uint8_t status = -1;

    struct bar_param send_param;
    send_param.barrier = 1;
    send_param.param = &status;

    struct bar_param sendtime_param;
    sendtime_param.barrier = ep_send_time ? 1 : 0;
    sendtime_param.param = ep_send_time;

    int rv;
    pktif_cb_t send_cb = {.fn = &status_cb, .param = &send_param};
    pktif_cb_t sendtime_cb = {.fn = &time_cb, .param = &sendtime_param};
    if ((rv = pktif_send_async(pktif, sktid, buffer, len, proto, send_time, tag,
                               rudpaddr, &send_cb,
                               ep_send_time ? &sendtime_cb : NULL))) {
        if (rv != len)
            return rv;
    }

    SYNC_LOOP(send_param.barrier || sendtime_param.barrier)
    // an send error occured
    if (status != PKTCTRL_SUCCESS) {
        return 0;
    }

    return len;
}

ssize_t pktif_send_async(struct pktif *pktif, uint8_t sktid, void *buffer,
                        size_t len, uint8_t proto, pktlab_time_t send_time,
                        uint16_t tag, struct sockaddr *rudpaddr,
                        pktif_cb_t *send_cb, pktif_cb_t *time_cb) {
    if (!pktif || !buffer || !send_cb) {
        errno = EINVAL;
        return -1;
    }

    struct pktlab_message send_msg;
    create_nsend(&send_msg, sktid, pktif->sktinfo[sktid].family, proto,
                 send_time, tag, rudpaddr, buffer, len);

    int rv;
    if ((rv = try_send(pktif->ctrlobj, &send_msg)) != 1)
        return -1;

    if ((rv = register_msg_cb(pktif->cbmgr, send_cb, PKTLAB_NSEND_MESSAGE)) !=
        0) {
        warn("msg sent but cb register failed");
        return rv;
    }
    if ((rv = notif_cb_ctl(pktif->cbmgr, PKTIF_CB_CTL_ADD, CB_TYPE_NTAG,
                           time_cb == NULL, 0, tag, 0, time_cb)) != 0) {
        warn("failed to register ntag callback on tag %d", tag);
        return rv;
    }
    return len;
}

ssize_t pktif_recv_sync(struct pktif *pktif, uint8_t sktid, void *buffer,
                       size_t len, struct sockaddr *raddr,
                       pktlab_time_t *recv_time) {
    struct timeval timeout = {.tv_sec = 1, .tv_usec = 0};
    if (!pktif) {
        errno = EINVAL;
        return -1;
    }

    if (!buffer) {
        errno = EFAULT;
        return -1;
    }

    if (pktif->cbmgr->ndata_cb[sktid].installed) {
        warn("recv handler exist for skt %d", sktid);
        return -1;
    }

    size_t rlen;
    while ((rlen = pktif_recv_async(pktif, sktid, buffer, len, raddr,
                                    recv_time)) == 0) {
        int set = PKTIF_SEL_READ;
        fd_set readfds;
        FD_ZERO(&readfds);
        timeout = (struct timeval){.tv_sec = 1, .tv_usec = 0};
        int nfds =
            pktif_prepare_select(pktif, set, 0, &readfds, NULL, NULL, NULL);
        nfds = select(nfds, &readfds, NULL, NULL, &timeout);
        pktif_process_select(pktif, set, &nfds, &readfds, NULL, NULL, NULL);
    }
    return rlen;
}

ssize_t pktif_recv_async(struct pktif *pktif, uint8_t sktid, void *buffer,
                        size_t len, struct sockaddr *raddr,
                        pktlab_time_t *recv_time) {
    if (!pktif) {
        errno = EINVAL;
        return -1;
    }

    if (!buffer) {
        errno = EFAULT;
        return -1;
    }

    struct callbackmgr *mgr = pktif->cbmgr;

    if (mgr->ndata_cb[sktid].installed) {
        warn("recv handler exist for skt %d", sktid);
        return -1;
    }

    struct pktlab_message *msg;
    size_t rlen = 0;
    bool first_msg = true;
    if (pktctrl_read_message(pktif->ctrlobj, NULL) == 1) {
        if (pktif_on_readable(pktif) != 0) {
            print_err_exit(-1, "pktif_on_readable failed");
        }
    }

    while ((msg = cyc_queue_peek_head(mgr->data_buf[sktid]))) {
        assert(msg->type == PKTLAB_NDATA_MESSAGE);
        // If receiving from multiple source addr, return after seeing the a
        // different addr
        // FIXME: we only return one packet at a time on a udp socket
        // no need to check for multiple source addr

        if (msg->ndata.proto == PKTLAB_UDP_PROTO) {
            if (raddr) {
                if (first_msg) {
                    first_msg = false;
                    raddr->sa_family = msg->ndata.family == PKTLAB_IP4_PROTO
                                           ? AF_INET
                                           : AF_INET6;
                    if (raddr->sa_family == AF_INET) {
                        struct sockaddr_in *addrin =
                            (struct sockaddr_in *)raddr;
                        memcpy(&addrin->sin_addr, msg->ndata.udp.raddrptr,
                               msg->ndata.udp.raddrlen);
                        memcpy(&addrin->sin_port, msg->ndata.udp.rportptr,
                               msg->ndata.udp.rportlen);
                    } else {
                        struct sockaddr_in6 *addrin6 =
                            (struct sockaddr_in6 *)raddr;
                        memcpy(&addrin6->sin6_addr, msg->ndata.udp.raddrptr,
                               msg->ndata.udp.raddrlen);
                        memcpy(&addrin6->sin6_port, msg->ndata.udp.rportptr,
                               msg->ndata.udp.rportlen);
                    }
                } else {
                    if (raddr->sa_family == AF_INET) {
                        struct sockaddr_in *addrin =
                            (struct sockaddr_in *)raddr;
                        if (memcmp(&addrin->sin_addr, msg->ndata.udp.raddrptr,
                                   msg->ndata.udp.raddrlen) ||
                            memcmp(&addrin->sin_port, msg->ndata.udp.rportptr,
                                   msg->ndata.udp.rportlen)) {
                            return rlen;
                        }
                    } else {
                        struct sockaddr_in6 *addrin6 =
                            (struct sockaddr_in6 *)raddr;
                        if (memcmp(&addrin6->sin6_addr, msg->ndata.udp.raddrptr,
                                   msg->ndata.udp.raddrlen) ||
                            memcmp(&addrin6->sin6_port, msg->ndata.udp.rportptr,
                                   msg->ndata.udp.rportlen)) {
                            return rlen;
                        }
                    }
                }
            }
        }
        // check if msg fit in buffer
        if (msg->ndata.len + rlen > len) {
            if (rlen == 0) {
                warn("insufficient buffer len for any msg");
                return -1;
            } else {
                return rlen;
            }
        }
        // copy msg content to buffer
        memcpy((int8_t *)buffer + rlen, msg->ndata.ptr, msg->ndata.len);
        rlen += msg->ndata.len;
        if (recv_time)
            *recv_time = msg->ndata.time;
        // pop msg from queue
        msg = cyc_queue_pop(mgr->data_buf[sktid]);

        // each call to recv on a raw/udp socket should only return one packet
        if (msg->ndata.proto == PKTLAB_RAW_PROTO ||
            msg->ndata.proto == PKTLAB_UDP_PROTO) {
            free(msg);
            break;
        }
        free(msg);
    }

    return rlen;
}

int pktif_prepare_select(struct pktif *pktif, int set, int nfds,
                         fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
                         struct timeval *timeout) {
    if (!pktif) {
        errno = EINVAL;
        return -1;
    }

    if (set == 0)
        return nfds;

    int set_flag = 0;
    if ((set & PKTIF_SEL_READ) && readfds) {
        FD_SET(pktif->sock, readfds);
        set_flag = 1;
    }
    if ((set & PKTIF_SEL_WRITE) && writefds) {
        FD_SET(pktif->sock, writefds);
        set_flag = 1;
    }
    if ((set & PKTIF_SEL_EXCEPT) && exceptfds) {
        FD_SET(pktif->sock, exceptfds);
        set_flag = 1;
    }
    return set_flag ? MAX(nfds, pktif->sock + 1) : nfds;
}

int pktif_process_select(struct pktif *pktif, int set, int *nfds,
                         fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
                         struct timeval *timeout) {
    if (!pktif) {
        errno = EINVAL;
        return -1;
    }
    if ((set & PKTIF_SEL_READ) && FD_ISSET(pktif->sock, readfds)) {
        (*nfds)--;
        FD_CLR(pktif->sock, readfds);
        struct pktlab_message *msg;
        do {
            util_readmsg(pktif->ctrlobj, &msg);
            debug("read in msg: %d", (msg->type));
            int ret = invoke_cb(pktif->cbmgr, msg);
            debug("invoke_cb returns %d", ret);
            (void) ret;
        } while (pktctrl_read_message(pktif->ctrlobj, NULL) == 1);
        return 0;
    }
    return -1;
}

int pktif_on_readable(struct pktif *pktif) {
    if (!pktif) {
        errno = EINVAL;
        return -1;
    }
    struct pktlab_message *msg;
    do {
        util_readmsg(pktif->ctrlobj, &msg);
        debug("read in msg: %d", msg->type);
        invoke_cb(pktif->cbmgr, msg);
    } while (pktctrl_read_message(pktif->ctrlobj, NULL) == 1);
    return 0;
}

static void create_nclose(struct pktlab_message *msg, uint8_t sktid) {
    msg->type = PKTLAB_NCLOSE_MESSAGE;
    msg->nclose.sktid = sktid;
}

int pktif_close_sync(struct pktif *pktif, uint8_t sktid) {
    if (!pktif) {
        errno = EINVAL;
        return -1;
    }

    uint8_t status = -1;
    struct bar_param param = {.barrier = 1, .param = &status};
    pktif_cb_t cb = {.fn = &status_cb, .param = &param};
    int rv;
    if ((rv = pktif_close_async(pktif, sktid, &cb))) {
        return rv;
    }

    SYNC_LOOP(param.barrier)

    return status;
}

int pktif_close_async(struct pktif *pktif, uint8_t sktid,
                      pktif_cb_t *close_cb) {
    if (!pktif) {
        errno = EINVAL;
        return -1;
    }

    struct pktlab_message send_msg;
    int rv;

    create_nclose(&send_msg, sktid);
    if ((rv = try_send(pktif->ctrlobj, &send_msg)) != 1)
        return -1;

    return register_msg_cb(pktif->cbmgr, close_cb, PKTLAB_NCLOSE_MESSAGE);
}

int pktif_ncap_sync(struct pktif *pktif, uint8_t sktid, uint8_t family,
                    uint8_t proto, pktlab_time_t endtime, const void *filter,
                    uint32_t filterlen) {
    if (!pktif || !filter || filterlen == 0) {
        errno = EINVAL;
        return -1;
    }
    uint8_t status = -1;
    struct bar_param param = {.barrier = 1, .param = &status};
    pktif_cb_t cb = {.fn = &status_cb, .param = &param};
    int rv;
    if ((rv = pktif_ncap_async(pktif, sktid, family, proto, endtime, filter,
                               filterlen, &cb))) {
        return rv;
    }
    SYNC_LOOP(param.barrier)

    return status;
}

int pktif_ncap_async(struct pktif *pktif, uint8_t sktid, uint8_t family,
                     uint8_t proto, pktlab_time_t endtime, const void *filter,
                     uint32_t filterlen, pktif_cb_t *ncap_cb) {
    if (!pktif || !filter || filterlen == 0) {
        errno = EINVAL;
        return -1;
    }
    struct pktlab_message send_msg = {.type = PKTLAB_NCAP_MESSAGE,
                                      .ncap.sktid = sktid,
                                      .ncap.family = family,
                                      .ncap.proto = proto,
                                      .ncap.endtime = endtime,
                                      .ncap.filter = filter,
                                      .ncap.filterlen = filterlen};
    int rv;
    if ((rv = try_send(pktif->ctrlobj, &send_msg)) != 1)
        return -1;

    return register_msg_cb(pktif->cbmgr, ncap_cb, PKTLAB_NCAP_MESSAGE);
}
