// prog.c
//

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <pktlab.h>

#include "linked_list.h"
#include "private.h"

//
// INTERNAL TYPE DEFINITIONS
//

typedef int (*encode_prog_t) (
    const void * prog, uint_fast32_t proglen,
    void ** buf, uint_fast32_t * buflen);

typedef int (*decode_prog_t) (
    const void * buf, uint_fast32_t buflen,
    enum pktlab_prog_type * type,
    void ** prog, uint_fast32_t * proglen);

//
// INTERNAL FUNCTION DECLARATIONS
//

static bool is_good_program_type(int type);

// internal encoders
static int encode_ebpf (
    const void * prog, uint_fast32_t proglen,
    void ** buf, uint_fast32_t * buflen);

// internal decoders
static int decode_ebpf (
    const void * buf, uint_fast32_t buflen,
    enum pktlab_prog_type * type,
    void ** prog, uint_fast32_t * proglen);

static int get_ebpf_instr (
    const char * line, struct pktlab_ebpf_instr ** instr);
static int get_ull (
    const char * start, char ** endptr,
    unsigned long long * ull);

//
// EXPORTED FUNCTION DEFINITIONS
//

int pktlab_encode_program (
    enum pktlab_prog_type type,
    const void * prog, uint_fast32_t proglen,
    void ** buf, uint_fast32_t * buflen)
{
    static const encode_prog_t encoders[] = {
        [PKTLAB_EBPF] = &encode_ebpf
    };

    trace("%s(type:%d,prog:%p,proglen:%" PRIuFAST32 ")",
        __func__, type, prog, proglen);

    if (!is_good_program_type(type) ||
        buf == NULL || buflen == NULL)
        return -1; // bad args

    if (prog == NULL && proglen != 0)
        return -2; // bad program

    assert(
        type >= 0 && type < sizeof(encoders)/sizeof(*encoders) &&
        encoders[type] != NULL);

    return (encoders[type])(prog, proglen, buf, buflen);
}

int pktlab_decode_program (
    const void * buf, uint_fast32_t buflen,
    enum pktlab_prog_type * type,
    void ** prog, uint_fast32_t * proglen)
{
    const uint8_t * ptr;
    enum pktlab_prog_type t;
    static const decode_prog_t decoders[] = {
        [PKTLAB_EBPF] = &decode_ebpf
    };

    trace("%s(buf:%p,buflen:%" PRIuFAST32 ")",
        __func__, buf, buflen);

    if (type == NULL || prog == NULL || proglen == NULL)
        return -1; // bad args

    if (buf == NULL || buflen == 0) // at least one byte needed
        return -2; // bad bytes

    ptr = buf;
    t = get8(ptr);

    if (!is_good_program_type(t))
        return -2; // bad bytes

    assert(
        t >= 0 && t < sizeof(decoders)/sizeof(*decoders) &&
        decoders[t] != NULL);
    return (decoders[t])(ptr+1, buflen-1, type, prog, proglen);
}

int pktlab_load_program (
    const char * data, uint_fast32_t datalen,
    enum pktlab_prog_type type,
    void ** prog, uint_fast32_t * proglen,
    uint_fast32_t * errlineno)
{
    char * copied_data;
    char * start;
    char * end;
    struct linked_list * ll;
    uint_fast32_t linecnt;
    struct pktlab_ebpf_instr * instr;
    struct pktlab_ebpf_instr * dstinstr;
    size_t totalmem;
    struct ll_node * node;
    int ret = -128;

    trace("%s(datalen:%" PRIuFAST32 ",type:%d)",
        __func__, datalen, type);

    if (errlineno != NULL)
        *errlineno = 0;

    if (prog == NULL || proglen == NULL ||
        !is_good_program_type(type))
        return -1; // bad arg

    assert(type == PKTLAB_EBPF); // only supporting ebpf for now

    if (data == NULL || datalen == 0) {
        *prog = NULL;
        *proglen = 0;
        return 0;
    }

    if (datalen >= SIZE_MAX)
        return -2; // overflow

    ll = create_ll();
    if (ll == NULL)
        return -3; // unknown err

    copied_data = safe_malloc(datalen+1);
    memcpy(copied_data, data, datalen);
    copied_data[datalen] = '\0';

    start = copied_data;
    linecnt = 1;
    while (1) {
        // tokenize line
        end = strchr(start, '\n');
        if (end != NULL)
            *end = '\0';

        // ignore line
        if (*start == '\0' || *start == '#' || *start == '\r') {
            if (end == NULL) // actual '\0' observed / last line
                break;

            start = end+1;
            ++linecnt;
            continue;
        }

        // process line
        if (get_ebpf_instr(start, &instr) < 0) {
            debug("Malformed ebpf instr line at line %" PRIuFAST32,
                linecnt);

            ret = -4; // bad line
            if (errlineno != NULL)
                *errlineno = linecnt;
            goto fail;
        }

        if (push_ll(instr, ll) < 0) {
            ret = -5; // unknown err
            goto fail;
        }

        if (end == NULL) // last line
            break;

        start = end+1;
        ++linecnt;
    }

    // return instructions as a blob
    totalmem = len_ll(ll) * sizeof(*instr);
    if (totalmem / sizeof(*instr) != len_ll(ll) ||
        totalmem > UINT_FAST32_MAX) {
        ret = -2; // overflow
        goto fail;
    }

    *proglen = totalmem;
    dstinstr = *prog = safe_malloc(totalmem);
    memset(*prog, 0, totalmem);

    for (node = get_head(ll);
         node != NULL; node = get_next(node)) {
        instr = get_data(node);
        memcpy(dstinstr, instr, sizeof(*dstinstr));
        ++dstinstr;
    }

    ret = 0;

fail:
    cleanup_ll(ll, wrapped_free, NULL);
    free(copied_data);
    return ret;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

bool is_good_program_type(int type) {
    switch (type) {
    case PKTLAB_EBPF:
        break;
    default:
        return false;
    }

    return true;
}

// internal encoders
int encode_ebpf (
    const void * prog, uint_fast32_t proglen,
    void ** buf, uint_fast32_t * buflen)
{
    const struct pktlab_ebpf_instr * instr;
    uint_fast32_t i, instr_cnt;
    uint8_t * ptr;
    size_t totalmem;

    assert(
        (prog != NULL || proglen == 0) &&
        buf != NULL && buflen != NULL);

    if (proglen % PKTLAB_EBPF_INSTR_LEN != 0)
        return -2; // bad prog

    // actual parsing
    instr = prog; // may be NULL, but instr_cnt will be 0 as well
    instr_cnt = proglen/PKTLAB_EBPF_INSTR_LEN;
    totalmem = instr_cnt*PKTLAB_ENC_EBPF_INSTR_LEN;
    if (totalmem/PKTLAB_ENC_EBPF_INSTR_LEN != instr_cnt ||
        totalmem >= SIZE_MAX ||
        totalmem >= UINT_FAST32_MAX)
        return -3; // overflow

    ptr = *buf = safe_malloc(totalmem+1);
    *buflen = totalmem+1;

    set8(ptr, PKTLAB_EBPF);
    ++ptr;

    for (i = 0; i < instr_cnt; ++i) {
        set8(ptr, instr[i].opcode);
        set8(ptr+1, (instr[i].dstreg << 4) | (instr[i].srcreg & 0xf));
        set16n(ptr+2, instr[i].offset);
        set32n(ptr+4, instr[i].imm);

        ptr += PKTLAB_ENC_EBPF_INSTR_LEN;
    }

    assert(ptr == ((uint8_t *) (*buf)+totalmem+1));
    return 0;
}

// internal decoders
int decode_ebpf (
    const void * buf, uint_fast32_t buflen,
    enum pktlab_prog_type * type,
    void ** prog, uint_fast32_t * proglen)
{
    const uint8_t * ptr;
    uint8_t val;
    uint_fast32_t i, instr_cnt;
    struct pktlab_ebpf_instr * instr;
    size_t totalmem;

    assert(
        buf != NULL && type != NULL &&
        prog != NULL && proglen != NULL);

    if (buflen % PKTLAB_ENC_EBPF_INSTR_LEN != 0)
        return -2; // bad bytes

    *prog = NULL;
    *proglen = 0;

    if (buflen == 0)
        return 0; // nothing allocated

    // actual parsing
    ptr = buf;
    instr_cnt = buflen/PKTLAB_ENC_EBPF_INSTR_LEN;
    totalmem = instr_cnt*PKTLAB_EBPF_INSTR_LEN;
    if (totalmem/PKTLAB_EBPF_INSTR_LEN != instr_cnt ||
        totalmem > UINT_FAST32_MAX)
        return -3; // overflow

    *type = PKTLAB_EBPF;
    instr = *prog = safe_malloc(totalmem);
    *proglen = totalmem;
    memset(*prog, 0, totalmem);

    for (i = 0; i < instr_cnt; ++i) {
        instr[i].opcode = get8(ptr);

        val = get8(ptr+1);
        instr[i].dstreg = (val >> 4);
        instr[i].srcreg = (val & 0xf);

        instr[i].offset = get16n(ptr+2);
        instr[i].imm = get32n(ptr+4);

        ptr += PKTLAB_ENC_EBPF_INSTR_LEN;
    }

    assert(ptr == ((uint8_t *) buf+buflen));
    return 0;
}

int get_ebpf_instr (
    const char * line, struct pktlab_ebpf_instr ** instr)
{
    const char * start;
    char * end;
    unsigned long long ull;
    int ret = -128;

    assert(line != NULL && instr != NULL);

    *instr = safe_malloc(sizeof(**instr));

    // parse opcode
    start = line;
    if (get_ull(start, &end, &ull) < 0 || ull > UINT8_MAX) {
        debug("Bad opcode");
        ret = -1; // bad opcode
        goto fail;
    }

    (*instr)->opcode = ull;

    // parse dstreg
    start = end;
    if (*end == '\0' || get_ull(start, &end, &ull) < 0 || ull > 0xf) {
        debug("Bad dstreg");
        ret = -2; // bad dstreg
        goto fail;
    }

    (*instr)->dstreg = ull;

    // parse srcreg
    start = end;
    if (*end == '\0' || get_ull(start, &end, &ull) < 0 || ull > 0xf) {
        debug("Bad srcreg");
        ret = -3; // bad srcreg
        goto fail;
    }

    (*instr)->srcreg = ull;

    // parse offset
    start = end;
    if (*end == '\0' || get_ull(start, &end, &ull) < 0 ||
        ull > UINT16_MAX) {
        debug("Bad offset");
        ret = -4; // bad offset
        goto fail;
    }

    (*instr)->offset = ull;

    // parse imm
    start = end;
    if (*end == '\0' || get_ull(start, &end, &ull) < 0 ||
        ull > UINT32_MAX) {
        debug("Bad imm");
        ret = -5; // bad imm
        goto fail;
    }

    (*instr)->imm = ull;

    return 0;

fail:
    free(*instr);
    return ret;
}

int get_ull (
    const char * start, char ** endptr,
    unsigned long long * ull)
{
    assert(start != NULL && endptr != NULL && ull != NULL);

    *ull = strtoull(start, endptr, 0);
    if (*ull == 0 && start == *endptr) { // more robust than EINVAL
        return -1; // conversion failed
    } else if (errno == ERANGE) {
        return -2; // value out of range
    }

    return 0;
}