// util_openssl.c
//

#include "private.h"
#include "util_openssl.h"

#include <assert.h>
#include <stdint.h>
#include <string.h>

#include <pthread.h>

#include <openssl/err.h>
#include <openssl/objects.h>

//
// INTERNAL MACROS
//

#define ERR_BUF_SZ 256

//
// INTERNAL TYPE DEFINITIONS
//

struct name_nid_map {
    const char * oid;
    int * nid;
};

//
// EXPORTED GLOBAL VARIABLE DEFINITIONS
//

bool _nid_defined;

int _nid_cert_type;
int _nid_cert_desc;
int _nid_filter_digests;
int _nid_monitor_digests;
int _nid_priority;
int _nid_del_type;

//
// INTERNAL GLOBAL VARIABLE DEFINITIONS
//

//
// INTERNAL FUNCTION DECLARATIONS
//

static int init_pktlab_nid(void);

//
// EXPORTED FUNCTION DEFINITIONS
//

int try_init_openssl(void) {
    static pthread_mutex_t init_lock = PTHREAD_MUTEX_INITIALIZER;
    int rst, ret = -128;

    // Unlike others, openssl now inits automatically
    // So for here, we will only try to init our own stuffs

    rst = pthread_mutex_lock(&init_lock);
    assert(rst == 0);

    if (init_pktlab_nid() < 0) {
        debug("init_pktlab_nid failed");
        ret = -1;
        goto fail;
    }

    ret = 0;

fail:
    rst = pthread_mutex_unlock(&init_lock);
    assert(rst == 0);

    return ret;
}

void error_debug_msg(const char * func_name) {
    long err;
    char errbuf[ERR_BUF_SZ];

    // suppress unused warning when DEBUG not defined
    (void)err;
    (void)errbuf;

    err = ERR_get_error();
    debug("%s failed: %ld (%s)",
        func_name, err,
        ERR_error_string(err, errbuf));
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int init_pktlab_nid(void) {
    const struct name_nid_map map[] = {
        { PKTLAB_EXT_CERT_TYPE,           &_nid_cert_type },
        { PKTLAB_EXT_CERT_DESCRIPTION,    &_nid_cert_desc },
        { PKTLAB_EXT_FILTER_DIGESTS,      &_nid_filter_digests },
        { PKTLAB_EXT_MONITOR_DIGESTS,     &_nid_monitor_digests },
        { PKTLAB_EXT_PRIORITY,            &_nid_priority },
        { PKTLAB_EXT_DEL_TYPE,            &_nid_del_type }
    };
    const uint_fast32_t maplen = sizeof(map)/sizeof(struct name_nid_map);
    uint_fast32_t i;
    int nid;

    if (_nid_defined)
        return 0; // already called before

    for (i = 0; i < maplen; ++i) {
        nid = OBJ_create(map[i].oid, NULL, NULL);
        if (nid == NID_undef) {
            debug("Failing OID: %s", map[i].oid);
            error_debug_msg("OBJ_create");
            return -1;
        }

        *(map[i].nid) = nid;
    }

    _nid_defined = true;
    return 0;
}