// verify.c
// Reference certificate chain verification function for PPKS
// Note that though we don't need openssl stuff here,
// we only compile this when openssl is enabled
// as the return error for the verify function is weird otherwise.
//

#include <pktlab.h>
#include <pktctrl.h>

#include <assert.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "private.h"

//
// INTERNAL CONSTANTS
//

enum phase_end_type {
    MUST_NOT_END,
    EITHER_WORKS,
    MUST_END
};

//
// INTERNAL TYPE DEFINITIONS
//

typedef int (*ppks_verify_t) (
    struct pktlab_certificate ** cert_ls,
    struct pktlab_publickey ** root_ls,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t certnum,
    uint_fast32_t rootnum,
    uint_fast32_t * verified_certnum,
    uint_fast32_t * used_root);

typedef int (*type_fields_check_t) (
    const struct pktlab_cert_info * info, void * aux);

struct check_phase {
    enum pktlab_cert_type cert_type;
    void * cert_type_aux; // for cert type specific checks
    int_fast32_t expected_cnt; // -1 means infinite
    enum phase_end_type end; // MUST_NOT_END: chain should not end in this phase
                             // EITHER_WORKS: chain could end in this phase
                             // MUST_END: chain must end in this phase
};

struct info_n_order {
    struct pktlab_cert_info * info;
    uint_fast32_t order_indx;
    bool used;
};

//
// INTERNAL FUNCTION DECLARATIONS
//

static int pubkey_check_keyid (
    struct pktlab_publickey * pubkey, const uint8_t * keyid);
static int general_cert_verify (
    const struct pktlab_cert_info * info);
static int check_signature_n_aid (
    struct pktlab_certificate * child_cert,
    struct pktlab_cert_info * child_info,
    struct pktlab_publickey ** parent_root_ls,
    uint_fast32_t parent_rootnum,
    uint_fast32_t * used_root);
static int check_chaining (
    struct pktlab_certificate * child_cert,
    struct pktlab_cert_info * child_info,
    struct pktlab_cert_info * parent_info);
static int check_pathlen (
    uint_fast32_t descendantnum,
    const struct pktlab_cert_info * info);

static int check_subcmd_fields (
    const struct pktlab_cert_info * info, void * aux);
static int check_pubcmd_fields (
    const struct pktlab_cert_info * info, void * aux);
static int check_exppriv_fields (
    const struct pktlab_cert_info * info, void * aux);
static int check_delpriv_fields (
    const struct pktlab_cert_info * info, void * aux);
static int check_agent_fields (
    const struct pktlab_cert_info * info, void * aux);
static int check_type_specific_fields (
    int cert_type, const struct pktlab_cert_info * info, void * aux);

static int find_next_unused_cert (
    const struct info_n_order * info_n_order_ls,
    uint_fast32_t certnum, const uint8_t * child_aid,
    uint_fast32_t * parent_indx);
static int check_cert (
    const struct pktlab_cert_info * info,
    int cert_type, void * cert_type_aux);
static int check_cert_chain (
    struct check_phase * check_phases,
    struct pktlab_certificate ** cert_ls,
    struct pktlab_publickey ** root_ls,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t phasecnt,
    uint_fast32_t certnum,
    uint_fast32_t rootnum,
    uint_fast32_t * verified_certnum,
    uint_fast32_t * used_root);
static void backup_verified_rst (
    const struct info_n_order * info_n_order_ls,
    uint_fast32_t totalmem, uint_fast32_t used_certnum,
    uint_fast32_t used_root_indx,
    struct info_n_order ** saved_info_n_order_ls,
    uint_fast32_t * saved_used_certnum,
    uint_fast32_t * saved_used_root_indx);
static int set_verified_rst (
    const struct info_n_order * saved_info_n_order_ls,
    struct pktlab_certificate ** cert_ls, uint_fast32_t certnum,
    uint_fast32_t saved_used_certnum, uint_fast32_t saved_used_root_indx,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t * verified_certnum, uint_fast32_t * used_root);

static int controller_verify_endpoint (
    struct pktlab_certificate ** cert_ls,
    struct pktlab_publickey ** root_ls,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t certnum,
    uint_fast32_t rootnum,
    uint_fast32_t * verified_certnum,
    uint_fast32_t * used_root);
static int endpoint_verify_controller (
    struct pktlab_certificate ** cert_ls,
    struct pktlab_publickey ** root_ls,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t certnum,
    uint_fast32_t rootnum,
    uint_fast32_t * verified_certnum,
    uint_fast32_t * used_root);
static int verify_agent (
    struct pktlab_certificate ** cert_ls,
    struct pktlab_publickey ** root_ls,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t certnum,
    uint_fast32_t rootnum,
    uint_fast32_t * verified_certnum,
    uint_fast32_t * used_root);

static uint_fast32_t safe_add (
    uint_fast32_t i, uint_fast32_t j);

static uint_fast32_t get_min_certnum (
    const struct check_phase * check_phases,
    uint_fast32_t phasecnt);
static uint_fast32_t get_max_certnum (
    const struct check_phase * check_phases,
    uint_fast32_t phasecnt);

//
// EXPORTED FUNCTION DEFINITIONS
//

int pktlab_verify_cert_chain (
    int auth_mode,
    struct pktlab_certificate ** cert_ls,
    struct pktlab_publickey ** root_ls,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t certnum,
    uint_fast32_t rootnum,
    uint_fast32_t * verified_certnum,
    uint_fast32_t * used_root)
{
    static const ppks_verify_t verifiers[] = {
        [PKTLAB_CONTROLLER_AUTH_ENDPOINT] = &controller_verify_endpoint,
        [PKTLAB_ENDPOINT_AUTH_CONTROLLER] = &endpoint_verify_controller,
        [PKTLAB_AUTH_AGENT]               = &verify_agent,
    };
    ppks_verify_t verifier;

    trace("%s(authmode:%d,certnum:%u,rootnum:%u)",
        __func__, auth_mode, (unsigned) certnum, (unsigned) rootnum);

    if (auth_mode < 0 ||
        auth_mode >= sizeof(verifiers)/sizeof(*verifiers))
        return PKTLAB_VERIFY_INVAL;

    if ((cert_ls == NULL) != (certnum == 0) ||
        (root_ls == NULL) != (rootnum == 0))
        return PKTLAB_VERIFY_INVAL;

    verifier = verifiers[auth_mode];
    if (verifier == NULL)
        return PKTLAB_VERIFY_INVAL;

    // initialize returned variables
    if (verified_cert_ls != NULL) *verified_cert_ls = NULL;
    if (verified_certnum != NULL) *verified_certnum = 0;
    if (used_root != NULL) *used_root = 0;

    return verifier(
        cert_ls, root_ls,
        verified_cert_ls,
        certnum, rootnum,
        verified_certnum,
        used_root);
}

//
// INTERNAL FUNCTION DEFINTIONS
//

int pubkey_check_keyid (
    struct pktlab_publickey * pubkey, const uint8_t * keyid)
{
    uint8_t actual_keyid[PKTLAB_KEYID_LEN];

    assert(pubkey != NULL && keyid != NULL);

    trace("%s(...)", __func__);

    if (pktlab_get_key_identifier(
            pubkey, actual_keyid, sizeof(actual_keyid)) < 0)
        return PKTLAB_VERIFY_UNKNOWN_FAULT;

    if (memcmp(actual_keyid, keyid, sizeof(actual_keyid)) != 0)
        return PKTLAB_VERIFY_BAD_CERT;

    return PKTLAB_VERIFY_SUCCESS;
}

int general_cert_verify (
    const struct pktlab_cert_info * info)
{
    int ret;
    pktlab_time_t now;
    struct pktlab_publickey * subjpubk;

    assert(info != NULL);

    trace("%s(...)", __func__);

    ret = PKTLAB_VERIFY_UNKNOWN_FAULT;

    if (info->serialno == NULL || info->serialno_len == 0) {
        ret = PKTLAB_VERIFY_BAD_CERT;
        goto fail;
    }

    now = pktlab_time_now();
    if (now < info->valid_not_before || info->valid_not_after < now) {
        ret = PKTLAB_VERIFY_EXPIRED_CERT;
        goto fail;
    }

    subjpubk = pktlab_create_publickey();
    if (pktlab_load_publickey_raw(
            info->publickey, sizeof(info->publickey), subjpubk) < 0) {
        ret = PKTLAB_VERIFY_UNKNOWN_FAULT;
        goto fail_key;
    }

    ret = pubkey_check_keyid(subjpubk, info->subject_key_id);
    if (ret != PKTLAB_VERIFY_SUCCESS)
        goto fail_key;

    ret = PKTLAB_VERIFY_SUCCESS;

fail_key:
    pktlab_cleanup_publickey(subjpubk);
fail:
    return ret;
}

int check_signature_n_aid (
    struct pktlab_certificate * child_cert,
    struct pktlab_cert_info * child_info,
    struct pktlab_publickey ** parent_root_ls,
    uint_fast32_t parent_rootnum,
    uint_fast32_t * used_root)
{
    bool good_signature = false;
    uint_fast32_t i;

    assert(
        child_cert != NULL && child_info != NULL &&
        parent_root_ls != NULL && parent_rootnum != 0);

    trace("%s(parent_rootnum:%u)",
        __func__, (unsigned) parent_rootnum);

    if (used_root != NULL)
        *used_root = 0;

    for (i = 0; i < parent_rootnum; ++i) {
        if (pubkey_check_keyid(
                parent_root_ls[i],
                child_info->authority_key_id) != PKTLAB_VERIFY_SUCCESS)
            continue;

        // got the right key matching child aid
        if (pktlab_verify_certificate_signature(
                child_cert, parent_root_ls[i]) < 0)
            return PKTLAB_VERIFY_BAD_CERT;

        if (used_root != NULL)
            *used_root = i;

        good_signature = true;
        break;
    }

    if (!good_signature) // no right key ever found
        return PKTLAB_VERIFY_NO_MATCHING_SIGN_KEY;

    return PKTLAB_VERIFY_SUCCESS;
}

int check_chaining (
    struct pktlab_certificate * child_cert,
    struct pktlab_cert_info * child_info,
    struct pktlab_cert_info * parent_info)
{
    struct pktlab_publickey * parent_pubkey;
    int ret;

    assert(
        child_cert != NULL && child_info != NULL &&
        parent_info != NULL);

    trace("%s(...)", __func__);

    parent_pubkey = pktlab_create_publickey();
    if (pktlab_load_publickey_raw(
            parent_info->publickey,
            sizeof(parent_info->publickey),
            parent_pubkey) < 0) {
        pktlab_cleanup_publickey(parent_pubkey);
        return PKTLAB_VERIFY_UNKNOWN_FAULT;
    }

    ret = check_signature_n_aid(
        child_cert, child_info, &parent_pubkey, 1, NULL);
    pktlab_cleanup_publickey(parent_pubkey);
    return ret;
}

int check_pathlen (
    uint_fast32_t descendantnum,
    const struct pktlab_cert_info * info)
{
    int_fast32_t pathlen;

    trace("%s(descendantnum:%u)",
        __func__, (unsigned) descendantnum);

    // should not be called for first cert
    assert(descendantnum > 0 && info != NULL);

    switch (info->cert_type) {
    case PKTLAB_CERT_DELPRIV:
        pathlen = info->delpriv.pathlen;
        break;
    case PKTLAB_CERT_EXPPRIV:
        pathlen = info->exppriv.pathlen;
        break;
    default:
        return PKTLAB_VERIFY_UNEXPECTED_CERT;
    }

    if (pathlen >= 0 && descendantnum-1 > pathlen)
        return PKTLAB_VERIFY_PATHLEN_EXCEEDED;

    return PKTLAB_VERIFY_SUCCESS;
}

int check_subcmd_fields (
    const struct pktlab_cert_info * info, void * aux)
{
    assert(info != NULL && aux == NULL);

    trace("%s(...)", __func__);

    (void) aux;

    if (info->cert_type != PKTLAB_CERT_SUBCMD)
        return PKTLAB_VERIFY_UNEXPECTED_CERT;

    return PKTLAB_VERIFY_SUCCESS;
}

int check_pubcmd_fields (
    const struct pktlab_cert_info * info, void * aux)
{
    assert(info != NULL && aux == NULL);

    trace("%s(...)", __func__);

    (void) aux;

    if (info->cert_type != PKTLAB_CERT_PUBCMD)
        return PKTLAB_VERIFY_UNEXPECTED_CERT;

    return PKTLAB_VERIFY_SUCCESS;
}

int check_exppriv_fields (
    const struct pktlab_cert_info * info, void * aux)
{
    assert(info != NULL && aux == NULL);

    trace("%s(...)", __func__);

    (void) aux;

    if (info->cert_type != PKTLAB_CERT_EXPPRIV)
        return PKTLAB_VERIFY_UNEXPECTED_CERT;

    if ((info->exppriv.filter_digests == NULL) !=
        (info->exppriv.ftcnt == 0))
        return PKTLAB_VERIFY_BAD_CERT;

    if ((info->exppriv.monitor_digests == NULL) !=
        (info->exppriv.mtcnt == 0))
        return PKTLAB_VERIFY_BAD_CERT;

    return PKTLAB_VERIFY_SUCCESS;
}

int check_delpriv_fields (
    const struct pktlab_cert_info * info, void * aux)
{
    uint8_t required_del_type = *((uint8_t *) aux);

    assert(info != NULL && aux != NULL);

    trace("%s(...)", __func__);

    if (info->cert_type != PKTLAB_CERT_DELPRIV)
        return PKTLAB_VERIFY_UNEXPECTED_CERT;

    if ((info->delpriv.filter_digests == NULL) !=
        (info->delpriv.ftcnt == 0))
        return PKTLAB_VERIFY_BAD_CERT;

    if ((info->delpriv.monitor_digests == NULL) !=
        (info->delpriv.mtcnt == 0))
        return PKTLAB_VERIFY_BAD_CERT;

    if ((required_del_type & info->delpriv.del_type) == 0)
        return PKTLAB_VERIFY_BAD_DEL_TYPE;

    return PKTLAB_VERIFY_SUCCESS;
}

int check_agent_fields (
    const struct pktlab_cert_info * info, void * aux)
{
    assert(info != NULL && aux == NULL);

    trace("%s(...)", __func__);

    (void) aux;

    if (info->cert_type != PKTLAB_CERT_AGENT)
        return PKTLAB_VERIFY_UNEXPECTED_CERT;

    // we can add self-signed signature check here,
    // but it doesn't seem necessary

    return PKTLAB_VERIFY_SUCCESS;
}

int check_type_specific_fields (
    int cert_type, const struct pktlab_cert_info * info, void * aux)
{
    static const type_fields_check_t check_funcs[] = {
        [PKTLAB_CERT_SUBCMD]  = &check_subcmd_fields,
        [PKTLAB_CERT_PUBCMD]  = &check_pubcmd_fields,
        [PKTLAB_CERT_EXPPRIV] = &check_exppriv_fields,
        [PKTLAB_CERT_DELPRIV] = &check_delpriv_fields,
        [PKTLAB_CERT_AGENT]   = &check_agent_fields,
    };

    trace("%s(cert_type:%d)", __func__, cert_type);

    assert(
        PKTLAB_CERT_SUBCMD <= cert_type &&
        cert_type <= PKTLAB_CERT_AGENT);

    return (check_funcs[cert_type])(info, aux);
}

int find_next_unused_cert (
    const struct info_n_order * info_n_order_ls,
    uint_fast32_t certnum, const uint8_t * child_aid,
    uint_fast32_t * parent_indx)
{
    uint_fast32_t i;

    trace("%s(certnum:%u)", __func__, (unsigned) certnum);

    assert(
        info_n_order_ls != NULL &&
        certnum != 0 && child_aid != NULL &&
        parent_indx != NULL);

    // never checks the first cert
    for (i = 1; i < certnum; ++i) {
        if (info_n_order_ls[i].used) // assuming child info used
            continue;

        if (memcmp(
                child_aid, info_n_order_ls[i].info->subject_key_id,
                PKTLAB_KEYID_LEN) != 0)
            continue;

        // found a match!
        *parent_indx = i;
        return 0;
    }

    *parent_indx = 0;
    return -1;
}

int check_cert (
    const struct pktlab_cert_info * info,
    int cert_type, void * cert_type_aux)
{
    int ret;

    trace("%s(cert_type:%d)", __func__, cert_type);

    assert(info != NULL);

    ret = general_cert_verify(info);
    if (ret != PKTLAB_VERIFY_SUCCESS)
        return ret;

    return check_type_specific_fields(
        cert_type, info, cert_type_aux);
}

int check_cert_chain (
    struct check_phase * check_phases,
    struct pktlab_certificate ** cert_ls,
    struct pktlab_publickey ** root_ls,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t phasecnt,
    uint_fast32_t certnum,
    uint_fast32_t rootnum,
    uint_fast32_t * verified_certnum,
    uint_fast32_t * used_root)
{
    uint_fast32_t totalmem, i;
    struct info_n_order * info_n_order_ls;
    struct info_n_order * saved_info_n_order_ls;
    struct pktlab_cert_info * child_info;
    bool first_cert, chain_valid = false;
    uint_fast32_t curphase, next_order_indx, used_root_indx;
    uint_fast32_t child_indx, parent_indx;
    uint_fast32_t saved_next_order_indx, saved_used_root_indx;
    int ret, rst;

    trace("%s(phasecnt:%u,certnum:%u,rootnum:%u)",
        __func__, (unsigned) phasecnt, (unsigned) certnum, (unsigned) rootnum);

    assert(
        check_phases != NULL && phasecnt != 0 &&
        (cert_ls == NULL) == (certnum == 0) &&
        (root_ls == NULL) == (rootnum == 0) &&
        (verified_cert_ls == NULL || *verified_cert_ls == NULL) &&
        (used_root == NULL || *used_root == 0) &&
        (verified_certnum == NULL || *verified_certnum == 0));

    // check cert chain min max length
    if (certnum < get_min_certnum(check_phases, phasecnt))
        return PKTLAB_VERIFY_NOT_ENOUGH_CERT;

    if (certnum > get_max_certnum(check_phases, phasecnt))
        return PKTLAB_VERIFY_TOO_MANY_CERT;

    // special case of no cert passed but match min len
    // we return success for this (though should not happen
    // in PPKS)
    if (certnum == 0)
        return set_verified_rst(
            NULL, NULL, 0, 0, 0,
            verified_cert_ls, verified_certnum,
            used_root);

    // decode all certs first
    totalmem = certnum*sizeof(*info_n_order_ls);
    if (totalmem/sizeof(*info_n_order_ls) != certnum || totalmem > SIZE_MAX)
        return PKTLAB_VERIFY_UNKNOWN_FAULT; // overflow

    // fail the verify process if any of the cert passed is bad
    // certnum from this point on stores the number of info to free
    info_n_order_ls = safe_malloc(totalmem);
    memset(info_n_order_ls, 0, totalmem);
    for (i = 0; i < certnum; ++i) {
        if (pktlab_decode_certificate(
                cert_ls[i], NULL, &info_n_order_ls[i].info)
                    != PKTLAB_DECODE_SUCCESS) {
            ret = PKTLAB_VERIFY_BAD_CERT;
            certnum = i;
            goto end;
        }
    }

    // check cert chain based on phases
    first_cert = true;
    saved_info_n_order_ls = NULL;
    used_root_indx = saved_used_root_indx = 0;
    next_order_indx = saved_next_order_indx = 0;
    child_indx = 0;
    child_info = NULL;

    for (curphase = 0; curphase < phasecnt; ++curphase) {
        // we ignore init expected_cnt == 0 phases
        // (though should not happen in PPKS)
        if (check_phases[curphase].expected_cnt == 0)
            continue;

        // check certs for current phase
        while (true) {
            // find next cert (parent) in chain
            // use head of cert_ls if first cert in chain
            parent_indx = 0;
            if (!first_cert) {
                assert(child_info != NULL);
                if (find_next_unused_cert(
                        info_n_order_ls, certnum,
                        child_info->authority_key_id,
                        &parent_indx) < 0) {
                    ret = PKTLAB_VERIFY_NO_MATCHING_SIGN_KEY;
                    goto end; // cannot find next cert
                }
            }

            debug("parent_indx: %u", (unsigned) parent_indx);
            debug("info cert_type: %d", info_n_order_ls[parent_indx].info->cert_type);

            // check validity of current cert
            rst = check_cert(
                    info_n_order_ls[parent_indx].info,
                    check_phases[curphase].cert_type,
                    check_phases[curphase].cert_type_aux);
            if (rst != PKTLAB_VERIFY_SUCCESS) {
                // for cert with non-matching cert type or delegation type,
                // may just be that the current phase has ended
                if ((rst == PKTLAB_VERIFY_UNEXPECTED_CERT ||
                     rst == PKTLAB_VERIFY_BAD_DEL_TYPE) &&
                    check_phases[curphase].expected_cnt == -1)
                    break;
                // otherwise fail
                ret = rst;
                goto end;
            }

            // check validity of chaining
            if (!first_cert) {
                // fails when too many further delegations or end-entity cert type
                rst = check_pathlen(
                    next_order_indx,
                    info_n_order_ls[parent_indx].info);
                if (rst != PKTLAB_VERIFY_SUCCESS) {
                    ret = rst;
                    goto end;
                }

                assert(child_info != NULL);
                rst = check_chaining(
                    cert_ls[child_indx], child_info,
                    info_n_order_ls[parent_indx].info);
                if (rst != PKTLAB_VERIFY_SUCCESS) {
                    ret = rst;
                    goto end;
                }
            }

            // cert valid, update used info and phase cert cnt
            info_n_order_ls[parent_indx].order_indx = next_order_indx;
            info_n_order_ls[parent_indx].used = true;
            child_info = info_n_order_ls[parent_indx].info;
            child_indx = parent_indx;
            ++next_order_indx;
            first_cert = false;

            if (check_phases[curphase].expected_cnt > 0)
                --check_phases[curphase].expected_cnt;

            // perform phase end processing
            // phase advancing behavior:
            //                   | MUST NOT END |  EITHER WORKS  |   MUST END
            // -----------------------------------------------------------------
            // expected_cnt > 0  |          (1) cont'd in cur phase
            // -----------------------------------------------------------------
            // expected_cnt == 0 | (2)          | (5) check end  | (4) check end
            //                   | cont'd to    | fail: cont'd   | fail: fail
            //                   | next phase   | to next phase  | verify
            // -----------------------------------------------------------------
            // expected_cnt < 0  | (1) cont'd   | (3) check end
            //                   | in cur phase | fail: cont'd in cur phase
            //

            // continue to be in current phase cases (1)
            if (check_phases[curphase].expected_cnt > 0 ||
                (check_phases[curphase].expected_cnt < 0 &&
                 check_phases[curphase].end == MUST_NOT_END))
                continue;

            // directly go to next phase cases (2)
            if (check_phases[curphase].expected_cnt == 0 &&
                check_phases[curphase].end == MUST_NOT_END)
                break;

            // need to check if reached cert chain end
            if (check_phases[curphase].expected_cnt <= 0 &&
                (check_phases[curphase].end == EITHER_WORKS ||
                 check_phases[curphase].end == MUST_END)) {
                // check root
                // no key passed means allow arbitrary root
                rst = PKTLAB_VERIFY_SUCCESS;
                if (root_ls != NULL && rootnum != 0) {
                    rst = check_signature_n_aid(
                        cert_ls[parent_indx],
                        info_n_order_ls[parent_indx].info,
                        root_ls, rootnum, &used_root_indx); // set used root
                    // continue on if fail or want to try match more certs in chain
                    // try match more cert behaves the same as failing root match
                }

                // check root successful means we have a valid chain
                // save results and try extend
                // (except case where no cert quota left and must end)
                if (rst == PKTLAB_VERIFY_SUCCESS) {
                    chain_valid = true;
                    free(saved_info_n_order_ls);
                    backup_verified_rst(
                        info_n_order_ls, totalmem,
                        next_order_indx, used_root_indx,
                        &saved_info_n_order_ls, &saved_next_order_indx,
                        &saved_used_root_indx);
                }

                // check root failed
                // continue in current phase cases (3)
                if (check_phases[curphase].expected_cnt < 0)
                    continue;

                assert(check_phases[curphase].expected_cnt == 0);

                // fail cases (4)
                if (check_phases[curphase].end == MUST_END) {
                    ret = rst; // rst may be success
                    goto end;
                }

                // continue to next phase cases (5)
                assert(check_phases[curphase].end == EITHER_WORKS);
                break;
            }
        }
    }

    ret = PKTLAB_VERIFY_END_OF_PHASES;

end:
    if (chain_valid) { // restore valid verify result if any
        ret = set_verified_rst(
                saved_info_n_order_ls, cert_ls, certnum,
                saved_next_order_indx, saved_used_root_indx,
                verified_cert_ls, verified_certnum, used_root);
        free(saved_info_n_order_ls);
    }

    for (i = 0; i < certnum; ++i)
        free(info_n_order_ls[i].info);
    free(info_n_order_ls);
    return ret;
}

void backup_verified_rst (
    const struct info_n_order * info_n_order_ls,
    uint_fast32_t totalmem, uint_fast32_t used_certnum,
    uint_fast32_t used_root_indx,
    struct info_n_order ** saved_info_n_order_ls,
    uint_fast32_t * saved_used_certnum,
    uint_fast32_t * saved_used_root_indx)
{
    trace("%s(used_certnum:%u,used_root_indx:%u)",
        __func__, (unsigned) used_certnum, (unsigned) used_root_indx);

    assert(
        info_n_order_ls != NULL && totalmem <= SIZE_MAX &&
        used_certnum > 0 && saved_info_n_order_ls != NULL &&
        saved_used_certnum != NULL);

    // shallow copy
    *saved_info_n_order_ls = safe_malloc(totalmem);
    memcpy(*saved_info_n_order_ls, info_n_order_ls, totalmem);
    *saved_used_certnum = used_certnum;
    *saved_used_root_indx = used_root_indx;
}

int set_verified_rst (
    const struct info_n_order * saved_info_n_order_ls,
    struct pktlab_certificate ** cert_ls, uint_fast32_t certnum,
    uint_fast32_t saved_used_certnum, uint_fast32_t saved_used_root_indx,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t * verified_certnum, uint_fast32_t * used_root)
{
    uint_fast32_t totalmem, i;

    trace("%s(saved_used_certnum:%u,saved_used_root_indx:%u)", __func__,
        (unsigned) saved_used_certnum, (unsigned) saved_used_root_indx);

    assert(
        (saved_info_n_order_ls == NULL) == (cert_ls == NULL) &&
        (cert_ls == NULL) == (certnum == 0) &&
        certnum >= saved_used_certnum);

    totalmem = saved_used_certnum*sizeof(**verified_cert_ls);
    if (totalmem/sizeof(**verified_cert_ls) != saved_used_certnum ||
        totalmem > SIZE_MAX)
        return PKTLAB_VERIFY_UNKNOWN_FAULT; // overflow

    if (verified_cert_ls != NULL) {
        *verified_cert_ls = NULL;
        if (saved_used_certnum != 0) {
            *verified_cert_ls = safe_malloc(totalmem);

            for (i = 0; i < certnum; ++i) {
                if (!saved_info_n_order_ls[i].used)
                    continue;
                (*verified_cert_ls)[saved_info_n_order_ls[i].order_indx] = cert_ls[i];
            }
        }
    }

    if (verified_certnum != NULL)
        *verified_certnum = saved_used_certnum;

    if (used_root != NULL)
        *used_root = saved_used_root_indx;

    return PKTLAB_VERIFY_SUCCESS;
}

int controller_verify_endpoint (
    struct pktlab_certificate ** cert_ls,
    struct pktlab_publickey ** root_ls,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t certnum,
    uint_fast32_t rootnum,
    uint_fast32_t * verified_certnum,
    uint_fast32_t * used_root)
{
    uint8_t reppriv_del_type = PKTLAB_DEL_TYPE_REPPRIV;
    struct check_phase check_phases[] = {
        { PKTLAB_CERT_SUBCMD, NULL, 1, EITHER_WORKS },
        { PKTLAB_CERT_DELPRIV, (void *) &reppriv_del_type, -1, MUST_END }
    };

    trace("%s(certnum:%u,rootnum:%u)",
        __func__, (unsigned) certnum, (unsigned) rootnum);

    assert(
        (cert_ls == NULL) == (certnum == 0) &&
        (root_ls == NULL) == (rootnum == 0) &&
        (verified_cert_ls == NULL || *verified_cert_ls == NULL) &&
        (used_root == NULL || *used_root == 0) &&
        (verified_certnum == NULL || *verified_certnum == 0));

    // Based on the assumption that: the pubcmd cert and exppriv cert are already checked,
    // here we only check the subcmd cert with the provided EOp public key,
    // which should appear in the exppriv cert.

    return check_cert_chain(
        check_phases, cert_ls, root_ls, verified_cert_ls,
        sizeof(check_phases)/sizeof(*check_phases),
        certnum, rootnum, verified_certnum, used_root);
}

int endpoint_verify_controller (
    struct pktlab_certificate ** cert_ls,
    struct pktlab_publickey ** root_ls,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t certnum,
    uint_fast32_t rootnum,
    uint_fast32_t * verified_certnum,
    uint_fast32_t * used_root)
{
    uint8_t exppriv_del_type = PKTLAB_DEL_TYPE_EXPPRIV;
    struct check_phase check_phases[] = {
        { PKTLAB_CERT_PUBCMD, NULL, 1, MUST_NOT_END },
        { PKTLAB_CERT_DELPRIV, (void *) &exppriv_del_type, -1, MUST_NOT_END },
        { PKTLAB_CERT_EXPPRIV, NULL, 1, MUST_END }
    };

    trace("%s(certnum:%u,rootnum:%u)",
        __func__, (unsigned) certnum, (unsigned) rootnum);

    assert(
        (cert_ls == NULL) == (certnum == 0) &&
        (root_ls == NULL) == (rootnum == 0) &&
        (verified_cert_ls == NULL || *verified_cert_ls == NULL) &&
        (used_root == NULL || *used_root == 0) &&
        (verified_certnum == NULL || *verified_certnum == 0));

    // Based on the assumption that:
    // the broker notification (notably channel) matches the local subinfo,
    // here we only check the pubcmd-exppriv cert chain
    // roots at some trusted EOp/del-EOp public key.

    return check_cert_chain(
        check_phases, cert_ls, root_ls, verified_cert_ls,
        sizeof(check_phases)/sizeof(*check_phases),
        certnum, rootnum, verified_certnum, used_root);
}

int verify_agent (
    struct pktlab_certificate ** cert_ls,
    struct pktlab_publickey ** root_ls,
    struct pktlab_certificate *** verified_cert_ls,
    uint_fast32_t certnum,
    uint_fast32_t rootnum,
    uint_fast32_t * verified_certnum,
    uint_fast32_t * used_root)
{
    struct check_phase check_phases[] = {
        { PKTLAB_CERT_AGENT, NULL, 1, MUST_END }
    };

    trace("%s(certnum:%u,rootnum:%u)",
        __func__, (unsigned) certnum, (unsigned) rootnum);

    assert(
        (cert_ls == NULL) == (certnum == 0) &&
        (root_ls == NULL) == (rootnum == 0) &&
        (verified_cert_ls == NULL || *verified_cert_ls == NULL) &&
        (used_root == NULL || *used_root == 0) &&
        (verified_certnum == NULL || *verified_certnum == 0));

    // The agent verify case should check for trusted root or self-signed.
    // Note we don't check signature for self-signed as it does no harm.
    // We can certainly add signature check in check_agent_fields though,
    // just need to pass in aux on whether the cert is expected
    // to be self-signed (based on whether root_ls is supplied).

    return check_cert_chain(
        check_phases, cert_ls, root_ls, verified_cert_ls,
        sizeof(check_phases)/sizeof(*check_phases),
        certnum, rootnum, verified_certnum, used_root);
}

uint_fast32_t safe_add (
    uint_fast32_t i, uint_fast32_t j)
{
    if (i > UINT_FAST32_MAX-j)
        return UINT_FAST32_MAX;
    return i+j;
}

uint_fast32_t get_min_certnum (
    const struct check_phase * check_phases,
    uint_fast32_t phasecnt)
{
    uint_fast32_t i, mincnt;

    trace("%s(phasecnt:%u)", __func__, (unsigned) phasecnt);

    assert((check_phases == NULL) == (phasecnt == 0));

    if (check_phases == NULL)
        return 0;

    for (mincnt = i = 0; i < phasecnt; ++i) {
        if (check_phases[i].expected_cnt == 0) {
            continue;
        } else if (check_phases[i].expected_cnt > 0) {
            mincnt = safe_add(mincnt, check_phases[i].expected_cnt);
        }

        if (check_phases[i].end == EITHER_WORKS ||
            check_phases[i].end == MUST_END)
            break;
    }

    debug("cert chain mincnt: %u", (unsigned) mincnt);
    return mincnt;
}

uint_fast32_t get_max_certnum (
    const struct check_phase * check_phases,
    uint_fast32_t phasecnt)
{
    uint_fast32_t i, last_either_works_indx, maxcnt;
    bool has_mustend;

    trace("%s(phasecnt:%u)", __func__, (unsigned) phasecnt);

    assert((check_phases == NULL) == (phasecnt == 0));

    if (check_phases == NULL)
        return 0;

    has_mustend = false;
    last_either_works_indx = UINT_FAST32_MAX;
    assert(phasecnt <= last_either_works_indx);

    for (i = 0; i < phasecnt; ++i) {
        if (check_phases[i].expected_cnt == 0)
            continue;

        if (check_phases[i].end == EITHER_WORKS) {
            last_either_works_indx = i;
        } else if (check_phases[i].end == MUST_END) {
            has_mustend = true;
            break;
        }
    }

    if (!has_mustend && last_either_works_indx == UINT_FAST32_MAX)
        return 0; // no end phase, we return 0 for max cert cnt

    for (maxcnt = i = 0; i < phasecnt; ++i) {
        if (check_phases[i].expected_cnt == 0)
            continue;

        if (check_phases[i].expected_cnt > 0) {
            maxcnt = safe_add(maxcnt, check_phases[i].expected_cnt);
        } else if (check_phases[i].expected_cnt < 0) {
            maxcnt = UINT_FAST32_MAX;
            break;
        }

        if (check_phases[i].end == MUST_END ||
            (!has_mustend && last_either_works_indx == i))
            break;
    }

    debug("cert chain maxcnt: %u", (unsigned) maxcnt);
    return maxcnt;
}