// xdescr_dattr.c
//

#include <pktlab.h>

#include <assert.h>
#include <ctype.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "linked_list.h"
#include "private.h"

//
// INTERNAL CONSTANTS
//

enum parse_st {
    IN_NAME,
    IN_VALUE
};

//
// INTERNAL TYPE DEFINITIONS
//

typedef bool (*good_value_char_t) (char c);

struct match_case {
    uint_fast32_t pcursor;
    uint_fast32_t scursor;
};

//
// INTERNAL FUNCTION DECLARATIONS
//

static bool is_good_name_char(char c);
static bool is_good_name(const char * name, uint_fast32_t len);
static bool is_good_value_char_xdescr(char c);
static bool is_good_value_char_dattr(char c);
static bool is_good_value (
    const char * value, uint_fast32_t len,
    good_value_char_t func);

static bool have_enough_remainlen (
    uint_fast32_t namelen, uint_fast32_t valuelen, uint_fast32_t remainlen);

static int find_next_pair (
    const char * ptr, uint_fast32_t len,
    uint_fast32_t * namelen, uint_fast32_t * valuelen,
    good_value_char_t func);
static int decode_xdescr_dattr (
    const char * s, uint_fast32_t slen,
    struct pktlab_xdescr_dattr_pair ** pairs, uint_fast32_t * paircnt,
    good_value_char_t func);
static int encode_xdescr_dattr (
    const struct pktlab_xdescr_dattr_pair * pairs, uint_fast32_t paircnt,
    char * buf, uint_fast32_t buflen, uint_fast32_t * used,
    good_value_char_t func);

//static bool recursive_pattern_match (
//    const char * pattern, uint_fast32_t patternlen,
//    const char * s, uint_fast32_t slen);

static bool pattern_match (
    const char * pattern, uint_fast32_t patternlen,
    const char * s, uint_fast32_t slen);

//
// EXPORTED FUNCTION DEFINITIONS
//

int pktlab_decode_xdescr (
    const char * xdescr, uint_fast32_t xdescrlen,
    struct pktlab_xdescr_dattr_pair ** pairs, uint_fast32_t * paircnt)
{
    trace("%s(xdescrlen:%" PRIuFAST32 ")", __func__, xdescrlen);

    return decode_xdescr_dattr(
        xdescr, xdescrlen, pairs, paircnt,
        is_good_value_char_xdescr);
}

int pktlab_encode_xdescr (
    const struct pktlab_xdescr_dattr_pair * pairs, uint_fast32_t paircnt,
    char * buf, uint_fast32_t buflen, uint_fast32_t * used)
{
    trace("%s(paircnt:%" PRIuFAST32 ",buflen:%" PRIuFAST32 ")",
        __func__, paircnt, buflen);

    return encode_xdescr_dattr(
        pairs, paircnt, buf, buflen, used,
        is_good_value_char_xdescr);
}

int pktlab_decode_dattr (
    const char * dattr, uint_fast32_t dattrlen,
    struct pktlab_xdescr_dattr_pair ** pairs, uint_fast32_t * paircnt)
{
    uint_fast32_t i, j;

    trace("%s(dattrlen:%" PRIuFAST32 ")", __func__, dattrlen);

    RETURN_IF(
        decode_xdescr_dattr(
            dattr, dattrlen, pairs, paircnt,
            is_good_value_char_dattr) < 0, -1);
    RETURN_IF(pairs == NULL, 0);

    // check no dup pairs
    for (i = 0; i < *paircnt; ++i) {
        for (j = i+1; j < *paircnt; ++j) {
            if((*pairs)[i].namelen == (*pairs)[j].namelen &&
               memcmp((*pairs)[i].name, (*pairs)[j].name, (*pairs)[i].namelen) == 0) {
                free(*pairs);
                return -1;
            }
        }
    }

    return 0;
}

int pktlab_encode_dattr (
    const struct pktlab_xdescr_dattr_pair * pairs, uint_fast32_t paircnt,
    char * buf, uint_fast32_t buflen, uint_fast32_t * used)
{
    trace("%s(paircnt:%" PRIuFAST32 ",buflen:%" PRIuFAST32 ")",
        __func__, paircnt, buflen);

    return encode_xdescr_dattr(
        pairs, paircnt, buf, buflen, used,
        is_good_value_char_dattr);
}

bool pktlab_xdescr_dattr_match (
    const char * xdescr, uint_fast32_t xdescrlen,
    const char * dattr, uint_fast32_t dattrlen)
{
    struct pktlab_xdescr_dattr_pair * xdescr_pairs;
    struct pktlab_xdescr_dattr_pair * dattr_pairs;
    uint_fast32_t xdescr_paircnt, dattr_paircnt, i, j;
    bool found, match, ret = false;

    trace("%s(xdescrlen:%" PRIuFAST32 ",dattrlen:%" PRIuFAST32 ")",
        __func__, xdescrlen, dattrlen);

    GOTO_IF(
        (xdescr == NULL) != (xdescrlen == 0) ||
        (dattr == NULL) != (dattrlen == 0), fail);

    // we return false if either string is malformed
    GOTO_IF(
        pktlab_decode_xdescr(
            xdescr, xdescrlen,
            &xdescr_pairs, &xdescr_paircnt) < 0, fail);
    GOTO_IF(
        pktlab_decode_dattr(
            dattr, dattrlen,
            &dattr_pairs, &dattr_paircnt) < 0, fail_xdescr_pairs);

    // start matching
    // a match when either one is empty
    GOTO_IF(xdescr == NULL || dattr == NULL, either_null_success);

    // both are non-null, need matching
    // can be further optimized
    for (i = 0; i < dattr_paircnt; ++i) {
        for (match = found = false, j = 0; j < xdescr_paircnt; ++j) {
            // find matching pairs with name
            // note there may be none, which we assume a match in value
            // for multiple, as along as one pair matches we are fine

            CONTINUE_IF(
                dattr_pairs[i].namelen != xdescr_pairs[j].namelen ||
                memcmp(
                    dattr_pairs[i].name, xdescr_pairs[j].name,
                    dattr_pairs[i].namelen) != 0);

            found = true;

            // match value
            if(pattern_match(
                    xdescr_pairs[j].value, xdescr_pairs[j].valuelen,
                    dattr_pairs[i].value, dattr_pairs[i].valuelen)) {
                match = true;
                break;
            }

            // no match, continue to test more pairs with same name
        }

        // xdescr pair for dattr pair name found
        // but no match ...
        GOTO_IF(found && !match, fail_dattr_pairs);
    }

either_null_success:
    ret = true;
fail_dattr_pairs:
    free(dattr_pairs);
fail_xdescr_pairs:
    free(xdescr_pairs);
fail:
    return ret;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

bool is_good_name_char(char c) {
    return isalnum(c) || c == '_' || c == '-'; // allow normal words
}

bool is_good_name(const char * name, uint_fast32_t len) {
    uint_fast32_t i;

    // name (if pair present) must be non-empty
    RETURN_IF(
        name == NULL || len == 0 ||
        len > PKTLAB_XDESCR_DATTR_NAMELEN_MAX, false);

    for (i = 0; i < len; ++i)
        RETURN_IF(!is_good_name_char(name[i]), false);

    return true;
}

bool is_good_value_char_xdescr(char c) {
    switch (c) {
    case '_':
    case '-':
        // normal words
    case '.':
    case ':':
    case '[':
    case ']':
        // address and port; [] for ipv6
    case ',':
        // easy string separation
    case '*':
    case '?':
        // wildcard chars; note [] for char matching is NOT supported
        return true;
    default: break;
    }

    return isalnum(c) != 0; // normal words
}

bool is_good_value_char_dattr(char c) {
    switch (c) {
    case '_':
    case '-':
        // normal words
    case '.':
    case ':':
    case '[':
    case ']':
        // address and port; [] for ipv6
    case ',':
        // easy string separation
        // NO wildcards allowed
        return true;
    default: break;
    }

    return isalnum(c) != 0; // normal words
}

bool is_good_value (
    const char * value, uint_fast32_t len,
    good_value_char_t func)
{
    uint_fast32_t i;

    assert(func != NULL);

    // value (if pair present) must be non-empty
    RETURN_IF(
        value == NULL || len == 0 ||
        len > PKTLAB_XDESCR_DATTR_VALUELEN_MAX, false);

    for (i = 0; i < len; ++i)
        RETURN_IF(!func(value[i]), false);

    return true;
}

bool have_enough_remainlen (
    uint_fast32_t namelen, uint_fast32_t valuelen, uint_fast32_t remainlen)
{
    assert(namelen != 0 && valuelen != 0);

    RETURN_IF(
        remainlen < 4 ||                        // X=X; -> at least 4 char
        namelen >= remainlen-2 ||               // no/just-enough space to fit name in
        valuelen > remainlen-2-namelen, false); // no space to fit value in

    return true;
}

int find_next_pair (
    const char * ptr, uint_fast32_t len,
    uint_fast32_t * namelen, uint_fast32_t * valuelen,
    good_value_char_t func)
{
    uint_fast32_t i;
    int st;

    assert(
        ptr != NULL && namelen != NULL &&
        valuelen != NULL && func != NULL);

    for (st = IN_NAME, i = 0; i < len; ++i) {
        if (st == IN_NAME) {
            if (ptr[i] == PKTLAB_XDESCR_DATTR_PAIR_SEP) {
                RETURN_IF(i == 0, -1);

                // reached = in pair
                st = IN_VALUE;
                *namelen = i;
                RETURN_IF(!is_good_name(ptr, *namelen), -1);
                continue;
            }

            RETURN_IF(ptr[i] == PKTLAB_XDESCR_DATTR_PAIR_END, -1);
        } else if (st == IN_VALUE) {
            if (ptr[i] == PKTLAB_XDESCR_DATTR_PAIR_END) {
                RETURN_IF(i == *namelen+1, -1);

                // reached ; in pair
                *valuelen = i-*namelen-1;
                RETURN_IF(!is_good_value(ptr+*namelen+1, *valuelen, func), -1);
                return 0;
            }

            RETURN_IF(ptr[i] == PKTLAB_XDESCR_DATTR_PAIR_SEP, -1);
        }
    }

    return -1;
}

int decode_xdescr_dattr (
    const char * s, uint_fast32_t slen,
    struct pktlab_xdescr_dattr_pair ** pairs, uint_fast32_t * paircnt,
    good_value_char_t func)
{
    uint_fast32_t cursor, cnt, namelen, valuelen, totalmem;

    assert(func != NULL);

    RETURN_IF(
        (s == NULL) != (slen == 0) ||
        pairs == NULL || paircnt == NULL, -1);

    if (slen == 0) {
        *pairs = NULL;
        *paircnt = 0;
        return 0;
    }

    // get number of pairs
    cnt = cursor = 0;
    while (true) {
        RETURN_IF(
            find_next_pair(
                s+cursor, slen-cursor,
                &namelen, &valuelen, func) < 0, -1);

        cursor += namelen+1+valuelen+1;
        ++cnt;

        assert(cursor <= slen);
        BREAK_IF(cursor >= slen);
    }

    // get pairs
    totalmem = sizeof(**pairs)*cnt;
    RETURN_IF(
        totalmem / sizeof(**pairs) != cnt ||
        totalmem > SIZE_MAX, -1);

    *pairs = safe_malloc(totalmem);
    *paircnt = cursor = 0;
    while (true) {
        if (find_next_pair(
                s+cursor, slen-cursor,
                &namelen, &valuelen, func) < 0) {
            free(*pairs);
            return -1;
        }

        (*pairs)[*paircnt].name = s+cursor;
        (*pairs)[*paircnt].namelen = namelen;
        (*pairs)[*paircnt].value = s+cursor+namelen+1;
        (*pairs)[*paircnt].valuelen = valuelen;

        cursor += namelen+1+valuelen+1;
        ++*paircnt;

        assert(cursor <= slen);
        BREAK_IF(cursor >= slen);
    }

    assert(*paircnt == cnt);
    return 0;
}

int encode_xdescr_dattr (
    const struct pktlab_xdescr_dattr_pair * pairs, uint_fast32_t paircnt,
    char * buf, uint_fast32_t buflen, uint_fast32_t * used,
    good_value_char_t func)
{
    uint_fast32_t i, remainlen;
    char * ptr;

    assert(func != NULL);

    RETURN_IF(
        (pairs == NULL) != (paircnt == 0) ||
        (buf == NULL) != (buflen == 0) ||
        (pairs != NULL && buf == NULL), -1);

    remainlen = buflen;
    ptr = buf;
    for (i = 0; i < paircnt; ++i) {
        RETURN_IF(
            !is_good_name(pairs[i].name, pairs[i].namelen) ||
            !is_good_value(pairs[i].value, pairs[i].valuelen, func) ||
            !have_enough_remainlen(
                pairs[i].namelen, pairs[i].valuelen,
                remainlen), -1);

        // good name and value pairs, copy to buf and continue
        memcpy(ptr, pairs[i].name, pairs[i].namelen);
        ptr[pairs[i].namelen] = PKTLAB_XDESCR_DATTR_PAIR_SEP;
        memcpy(ptr+pairs[i].namelen+1, pairs[i].value, pairs[i].valuelen);
        ptr[pairs[i].namelen+1+pairs[i].valuelen] = PKTLAB_XDESCR_DATTR_PAIR_END;

        remainlen -= pairs[i].namelen+1+pairs[i].valuelen+1;
        ptr += pairs[i].namelen+1+pairs[i].valuelen+1;
    }

    if (used != NULL)
        *used = buflen-remainlen;

    return 0;
}

// RECURSIVE BAD!!!!!
//bool recursive_pattern_match (
//    const char * pattern, uint_fast32_t patternlen,
//    const char * s, uint_fast32_t slen)
//{
//    uint_fast32_t i;
//
//    // all char matched
//    RETURN_IF(patternlen == 0 && slen == 0, true);
//
//    // used up pattern, can not match anymore
//    RETURN_IF(patternlen == 0 && slen != 0, false);
//
//    // still have pattern left
//    switch (pattern[0]) {
//    case '*':
//        // try skipping arbitrary no of s char
//        for (i = 0; i <= slen; ++i) {
//            RETURN_IF(
//                recursive_pattern_match(
//                    pattern+1, patternlen-1,
//                    s+i, slen-i), true);
//        }
//        break;
//    case '?':
//        // no match when no more char in s
//        RETURN_IF(slen == 0, false);
//
//        return
//            recursive_pattern_match(
//                pattern+1, patternlen-1,
//                s+1, slen-1);
//    default: // non-wildcard char
//        // no match when no more char in s left or char no match
//        RETURN_IF(
//            slen == 0 ||
//            pattern[0] != s[0], false);
//
//        return
//            recursive_pattern_match(
//                pattern+1, patternlen-1,
//                s+1, slen-1);
//    }
//
//    return false;
//}

bool pattern_match (
    const char * pattern, uint_fast32_t patternlen,
    const char * s, uint_fast32_t slen)
{
    struct linked_list * ll;
    struct match_case * c;
    uint_fast32_t i, pcursor, scursor;
    bool ret = false;

    assert(
        pattern != NULL && s != NULL &&
        patternlen != 0 && slen != 0);

    ll = create_ll();
    GOTO_IF(ll == NULL, fail);

    // root match case
    c = safe_malloc(sizeof(*c));
    c->pcursor = 0;
    c->scursor = 0;
    if (push_ll(c, ll) < 0) {
        free(c);
        goto fail_ll;
    }

    while (len_ll(ll) > 0) {
        // pop and process new match case
        GOTO_IF(remove_node(ll, 0, (void **) &c) < 0, fail_ll);

        pcursor = c->pcursor;
        scursor = c->scursor;
        free(c);

        assert(pcursor <= patternlen && scursor <= slen);

        // match case all char matched
        if (patternlen-pcursor == 0 && slen-scursor == 0) {
            ret = true;
            break;
        }

        // match case used up pattern, can not match anymore
        CONTINUE_IF(patternlen-pcursor == 0 && slen-scursor != 0);

        // match case still have pattern left
        switch (pattern[pcursor]) {
        case '*':
            // try skipping arbitrary numbers of remaining s char
            for (i = scursor; i <= slen; ++i) {
                c = safe_malloc(sizeof(*c));
                c->pcursor = pcursor+1;
                c->scursor = i;
                if (push_ll(c, ll) < 0) {
                    free(c);
                    goto fail_ll;
                }
            }
            continue;
        default: // non-wildcard char
            CONTINUE_IF(
                scursor == slen ||               // no more char in s
                pattern[pcursor] != s[scursor]); // char no match
            // fall through
        case '?':
            // no match when no more char in s
            CONTINUE_IF(scursor == slen);

            c = safe_malloc(sizeof(*c));
            c->pcursor = pcursor+1;
            c->scursor = scursor+1;
            if (push_ll(c, ll) < 0) {
                free(c);
                goto fail_ll;
            }
            continue;
        }
    }

fail_ll:
    cleanup_ll(ll, wrapped_free, NULL);
fail:
    return ret;
}