// c_reader.c
// Conservative reader.
// Read only needed bytes for current msg.
//

#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>

#include <pktctrl.h>
#include <pktlab.h>

#include "console.h"
#include "c_reader.h"
#include "util.h"

//
// INTERNAL MACROS
//

#define is_try_again(err) (err == EAGAIN || err == EWOULDBLOCK)

//
// INTERNAL CONSTANTS
//

enum c_reader_st {
    EMPTY,
    READ_HEADER,
    READ_DATA
};

//
// OPAQUE STRUCTURE DEFINITIONS
//

struct c_reader {
    struct pktctrl_obj * obj;
    uint_fast32_t readlen;
    uint_fast32_t totallen;
    struct pktlab_message * msg;
    uint8_t * buf;
    int st;
};

//
// INTERNAL FUNCTION DECLARATIONS
//

static int read_header(struct c_reader * reader);
static int read_data(struct c_reader * reader);
static void clear_reader(struct c_reader * reader);

//
// EXPORTED FUNCTION DEFINITIONS
//

struct c_reader * get_c_reader (
    struct pktctrl_obj * obj)
{
    struct c_reader * ret;

    if (obj == NULL)
        return NULL;

    ret = safe_malloc(sizeof(*ret));
    memset(ret, 0, sizeof(*ret));
    ret->obj = obj;
    ret->st = EMPTY;
    return ret;
}

int c_read_msg (
    struct c_reader * reader,
    struct pktlab_message ** msg)
{
    int rst;

    trace("%s(...)", __func__);

    if (reader == NULL || msg == NULL)
        return -1;

    switch (reader->st) {
    case EMPTY:
        reader->msg = safe_malloc(sizeof(**msg)+PKTLAB_HLEN);
        reader->buf = (uint8_t *) (reader->msg+1);
        reader->st = READ_HEADER;
    case READ_HEADER:
        rst = read_header(reader);
        break;
    case READ_DATA:
        rst = read_data(reader);
        break;
    default:
        assert(0); // bug
    }

    if (rst <= 0)
        return rst; // error or try again

    assert(rst == 1 && reader->readlen == reader->totallen);

    *msg = NULL;
    if (reader->msg != NULL) {
        rst = pktlab_decode_message(
            reader->msg, reader->buf, reader->totallen);
        assert(rst != 0);

        *msg = reader->msg;

        reader->msg = NULL; // don't free msg when clearing
        clear_reader(reader);
    }

    return 1;
}

void close_c_reader(struct c_reader * reader) {
    if (reader == NULL)
        return;

    clear_reader(reader);
    free(reader);
    return;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int read_header(struct c_reader * reader) {
    ssize_t rst;

    assert(
        reader != NULL &&
        reader->st == READ_HEADER &&
        reader->msg != NULL &&
        reader->buf != NULL &&
        reader->readlen < PKTLAB_HLEN);

    rst = pktctrl_read(
        reader->obj,
        reader->buf+reader->readlen,
        PKTLAB_HLEN-reader->readlen);
    if (rst <= 0) {
        debug("pktctrl_read failed with %d (errno: %s)",
            rst, strerror(errno));

        if (rst == 0) { // eof
            clear_reader(reader);
            return 1; // signal read completed
        } else if (rst < 0 && is_try_again(errno))
            return 0;
        return -1;
    }

    assert(rst <= PKTLAB_HLEN-reader->readlen);

    reader->readlen += rst;
    if (reader->readlen < PKTLAB_HLEN)
        return 0; // not yet ready

    // read enough bytes for header
    // set up to read payload
    reader->totallen = pktlab_get24n(reader->buf+1)+PKTLAB_HLEN;
    if (reader->totallen == reader->readlen) {
        return 1; // no payload case
    } else if (reader->totallen > SIZE_MAX-sizeof(*reader->msg)) {
        return -2; // overflow
    }

    reader->msg = safe_realloc(
        reader->msg, sizeof(*reader->msg)+reader->totallen);
    reader->buf = (uint8_t *) (reader->msg+1);
    reader->st = READ_DATA;
    return read_data(reader);
}

int read_data(struct c_reader * reader) {
    ssize_t rst;

    assert(
        reader != NULL &&
        reader->st == READ_DATA &&
        reader->msg != NULL &&
        reader->buf != NULL &&
        reader->readlen >= PKTLAB_HLEN &&
        reader->totallen > reader->readlen);

    rst = pktctrl_read(
        reader->obj,
        reader->buf+reader->readlen,
        reader->totallen-reader->readlen);
    if (rst <= 0) {
        debug("pktctrl_read failed with %d (errno: %s)",
            rst, strerror(errno));

        if (rst == 0) { // eof
            clear_reader(reader);
            return 1; // signal read completed
        } else if (rst < 0 && is_try_again(errno))
            return 0;
        return -1;
    }

    assert(rst <= reader->totallen-reader->readlen);

    reader->readlen += rst;
    if (reader->readlen < reader->totallen)
        return 0; // not yet ready

    // read enough bytes for whole msg
    return 1;
}

void clear_reader(struct c_reader * reader) {
    assert(reader != NULL);

    free(reader->msg);
    reader->msg = NULL;
    reader->buf = NULL;
    reader->readlen = 0;
    reader->totallen = 0;
    reader->st = EMPTY;
    return;
}