// config.c
//

// void * casts are horrible
// Used here in many places,
// but only to export const stuffs.

// todo: check const ptr field casting

// potential more options
// #TODO:
// # - cmdline get list of target endpoint ips
// # - add this as a new option to
// # Default publication controller address (`PubControllerAddr`)
// # Should be a string in the format of ADDR/ADDR:PORT for IPv4,
// # or ADDR/[ADDR]:PORT for IPv6.
// # Domain name is acceptible for ADDR as well, in which case
// # the IPv4 notation should be used.
// # Omitting PORT implicitly implies port 20556.
// # (e.g. example.org:20556, 127.0.0.1, [::1]:20556)
// # - This option is REQUIRED when enabling publication.
// #PubControllerAddr 127.0.0.1:20556

#include "config.h"

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "console.h"
#include "net.h"
#include "util.h"

#include <pktconf.h>
#include <pktctrl.h>

//
// INTERNAL CONSTANTS AND MACROS
//

enum ec_init_keys {
    ENABLE_EXP_PUB,
    BROKER_ADDR,
    PUB_AGENT_CERT_PATH,
    BROKER_CERT_ROOT_PATH,
    PUB_CONTROLLER_ADDR,
    PUB_PERIOD,
    PUB_EXP_PRIORITY,
    PUB_COUNTRY,
    PUB_LOCATION,
    PUB_END_GREET_PERIOD,
    CONTROLLER_PRIVKEY_PATH,
    PROGRAM_DIR_PATH,
    EXP_EXEC,
    ENABLE_PROXY_MODE,
    BACKLOG_CONN_AGENT_CERT_PATH,
    PROXY_CERT_ROOT_PATH,
    TARGET_BACKLOG_CONN,
    BACKLOG_CONN_TIMEOUT,
    UKN
};

enum exp_exec_fields {
    IS_DEFAULT,
    SNIKEY,
    EXPPRIVCHAIN_PATH,
    TRUSTEDEOP_PATH,
    PUB_CHANNEL,
    OVERRIDE_XDESCRNVP,
    EXP_EXEC_FIELDNO
    // hack for calculating the number of exp_exec fields
    // will only work if no value is specified
};

#define BUF_SIZ 4096

#define ENABLE_EXP_PUB_STR                "EnableExpPublication"
#define BROKER_ADDR_STR                   "BrokerAddr"
#define PUB_AGENT_CERT_PATH_STR           "PubAgentCertPath"
#define BROKER_CERT_ROOT_PATH_STR         "BrokerCertRootPath"
#define PUB_CONTROLLER_ADDR_STR           "PubControllerAddr"
#define PUB_PERIOD_STR                    "PubPeriod"
#define PUB_EXP_PRIORITY_STR              "PubExpPriority"
#define PUB_COUNTRY_STR                   "PubCountry"
#define PUB_LOCATION_STR                  "PubLocation"
#define PUB_END_GREET_PERIOD_STR          "GreetPeriodAfterPubEnd"
#define CONTROLLER_PRIVKEY_PATH_STR       "ControllerPrivkeyPath"
#define PROGRAM_DIR_PATH_STR              "ProgramDirPath"
#define ENABLE_PROXY_MODE_STR             "EnableProxyMode"
#define BACKLOG_CONN_AGENT_CERT_PATH_STR  "BacklogConnAgentCertPath"
#define PROXY_CERT_ROOT_PATH_STR          "ProxyCertRootPath"
#define TARGET_BACKLOG_CONN_STR           "TargetBacklogConn"
#define BACKLOG_CONN_TIMEOUT_STR          "BacklogConnTimeout"
#define EXP_EXEC_STR                      "ExpExec"

//
// INTERNAL TYPE DEFINITIONS
//

typedef int (*value_parser_t)(
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);

struct key_dict_entry{
    const char * const keyname;
    uint_fast32_t keylen;
    const enum ec_init_keys key;
};

//
// INTERNAL GLOBAL VARIABLES
//

static const struct key_dict_entry _key_dict[] = {
    { ENABLE_EXP_PUB_STR,
      sizeof(ENABLE_EXP_PUB_STR)-1,
      ENABLE_EXP_PUB },
    { BROKER_ADDR_STR,
      sizeof(BROKER_ADDR_STR)-1,
      BROKER_ADDR },
    { PUB_AGENT_CERT_PATH_STR,
      sizeof(PUB_AGENT_CERT_PATH_STR)-1,
      PUB_AGENT_CERT_PATH },
    { BROKER_CERT_ROOT_PATH_STR,
      sizeof(BROKER_CERT_ROOT_PATH_STR)-1,
      BROKER_CERT_ROOT_PATH },
    { PUB_CONTROLLER_ADDR_STR,
      sizeof(PUB_CONTROLLER_ADDR_STR)-1,
      PUB_CONTROLLER_ADDR },
    { PUB_PERIOD_STR,
      sizeof(PUB_PERIOD_STR)-1,
      PUB_PERIOD },
    { PUB_EXP_PRIORITY_STR,
      sizeof(PUB_EXP_PRIORITY_STR)-1,
      PUB_EXP_PRIORITY },
    { PUB_COUNTRY_STR,
      sizeof(PUB_COUNTRY_STR)-1,
      PUB_COUNTRY },
    { PUB_LOCATION_STR,
      sizeof(PUB_LOCATION_STR)-1,
      PUB_LOCATION },
    { PUB_END_GREET_PERIOD_STR,
      sizeof(PUB_END_GREET_PERIOD_STR)-1,
      PUB_END_GREET_PERIOD },
    { CONTROLLER_PRIVKEY_PATH_STR,
      sizeof(CONTROLLER_PRIVKEY_PATH_STR)-1,
      CONTROLLER_PRIVKEY_PATH },
    { PROGRAM_DIR_PATH_STR,
      sizeof(PROGRAM_DIR_PATH_STR)-1,
      PROGRAM_DIR_PATH },
    { ENABLE_PROXY_MODE_STR,
      sizeof(ENABLE_PROXY_MODE_STR)-1,
      ENABLE_PROXY_MODE },
    { BACKLOG_CONN_AGENT_CERT_PATH_STR,
      sizeof(BACKLOG_CONN_AGENT_CERT_PATH_STR)-1,
      BACKLOG_CONN_AGENT_CERT_PATH },
    { PROXY_CERT_ROOT_PATH_STR,
      sizeof(PROXY_CERT_ROOT_PATH_STR)-1,
      PROXY_CERT_ROOT_PATH },
    { TARGET_BACKLOG_CONN_STR,
      sizeof(TARGET_BACKLOG_CONN_STR)-1,
      TARGET_BACKLOG_CONN },
    { BACKLOG_CONN_TIMEOUT_STR,
      sizeof(BACKLOG_CONN_TIMEOUT_STR)-1,
      BACKLOG_CONN_TIMEOUT },
    { EXP_EXEC_STR,
      sizeof(EXP_EXEC_STR)-1,
      EXP_EXEC },
    { NULL, 0, UKN }
};

//
// EXPORTED GLOBAL VARIABLES
//

struct config _config;

//
// INTERNAL FUNCTION DECLARATIONS
//

static enum ec_init_keys check_dict (
    const char * keyname, uint_fast32_t keylen);
static int load_config (
    const char * key, uint_fast32_t keylen,
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);

static bool need_eop_key (
    const struct pktconf_sep_str_ll * ll);

//
// FIELD PARSERS
//

static int parse_bool (
    const char * value, uint_fast32_t valuelen,
    bool * field_set, bool * field);
static int parse_addr (
    const char * value, uint_fast32_t valuelen,
    const char ** field);
static int parse_filepath (
    const char * value, uint_fast32_t valuelen,
    const char ** field);
static int parse_filepath_ls (
    const char * value, uint_fast32_t valuelen,
    const struct pktconf_sep_str_ll ** field);
static int parse_dirpath (
    const char * value, uint_fast32_t valuelen,
    const char ** field);
static int parse_hex_ls (
    const char * value, uint_fast32_t valuelen,
    const struct pktconf_sep_str_ll ** field);

static int parse_enable_exp_pub (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_broker_addr (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_pub_agent_cert_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_broker_cert_root_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_pub_controller_addr (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_pub_period (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_pub_exp_priority (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_pub_country (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_pub_location (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_pub_end_greet_period (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_controller_privkey_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_program_dir_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_enable_proxy_mode (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_backlog_conn_agent_cert_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_proxy_cert_root_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_target_backlog_conn (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_backlog_conn_timeout (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);
static int parse_exp_exec (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno);

//
// EXPORTED FUNCTION DEFINITIONS
//

int config_init(void) {
    trace("%s()", __func__);

    if (_config.state != CONFIG_NOT_INIT)
        return -1; // only init when CONFIG_NOT_INIT

    memset(&_config, 0, sizeof(_config));
    _config.enable_exp_pub_specified = false;
    _config.pub_period_specified = false;
    _config.pub_exp_priority_specified = false;
    _config.pub_end_greet_period_specified = false;
    _config.enable_proxy_mode_specified = false;
    _config.target_backlog_conn_specified = false;
    _config.backlog_setup_timeout_specified = false;
    _config.state = CONFIG_INIT;

    return 0;
}

int config_read(const char * path, uint_fast32_t * errlineno) {
    struct pktconf_option_ll * ll;
    struct pktconf_opt_ll_node * node;
    uint_fast32_t i;

    trace("%s(path:%s)", __func__, safe_str(path));

    if (_config.state != CONFIG_INIT) {
        debug("Bad config state");
        return -1; // only read when in CONFIG_INIT state
    }

    if (path == NULL) {
        debug("No config path provided");
        return -2;
    }

    if (errlineno != NULL)
        *errlineno = 0; // 0 implies no line read

    if (pktconf_load_option(path, &ll) < 0) {
        debug("pktconf_load_option failed");
        return -3;
    }

    node = ll->head;
    for (i = 0; i < ll->len; ++i) {
        if (load_config(
                node->key, node->keylen,
                node->value, node->valuelen,
                node->lineno) < 0) {
            debug("load_config(key:%s,value:%s) failed at line %u",
                node->key, node->value, (unsigned) node->lineno);
            *errlineno = node->lineno;
            pktconf_cleanup_option_ll(ll);
            config_cleanup();
            return -4;
        }

        node = node->next;
    }

    _config.state = CONFIG_READY;
    pktconf_cleanup_option_ll(ll);
    return 0;
}

bool is_config_presence_correct(int * err) {
    bool is_default_set;
    const struct exp_exec_entry * entry;

    assert(err != NULL);

    *err = 0;

    if (_config.state != CONFIG_READY) {
        *err = -1;
        return false;
    }

    if (!_config.enable_exp_pub_specified) {
        *err = -2;
        return false;
    }

    if (_config.enable_exp_pub &&
        (_config.broker_addr == NULL ||
         _config.pub_agent_cert_path == NULL)) {
        *err = -3;
        return false;
    }

    if (_config.controller_privkey_path == NULL ||
        _config.exp_exec_ll == NULL) {
        *err = -4;
        return false;
    }

    is_default_set = false;
    for (entry = _config.exp_exec_ll; entry != NULL; entry = entry->next) {
        if (!entry->is_default)
            continue;

        if (is_default_set) {
            *err = -5;
            return false; // multiple is_default set
        }

        is_default_set = true;
    }

    return true;
}

void config_cleanup(void) {
    const struct exp_exec_entry * entry;

    trace("%s()", __func__);

    if (_config.state == CONFIG_NOT_INIT)
        return; // do nothing when CONFIG_NOT_INIT

    // free allocated const ptr needs explicit casting
    free((void *) (_config.broker_addr));
    free((void *) (_config.pub_controller_addr));
    free((void *) (_config.controller_privkey_path));
    free((void *) (_config.program_dir_path));
    free((void *) (_config.pub_agent_cert_path));
    free((void *) (_config.backlog_conn_agent_cert_path));

    pktconf_cleanup_sep_str_ll((void *) (_config.broker_cert_root_path_ll));
    pktconf_cleanup_sep_str_ll((void *) (_config.pub_country));
    pktconf_cleanup_sep_str_ll((void *) (_config.pub_location));
    pktconf_cleanup_sep_str_ll((void *) (_config.proxy_cert_root_path_ll));

    while (_config.exp_exec_ll != NULL) {
        entry = _config.exp_exec_ll->next;
        pktconf_cleanup_sep_str_ll(
            (void *) (_config.exp_exec_ll->exppriv_chain_path_ll));
        pktconf_cleanup_sep_str_ll(
            (void *) (_config.exp_exec_ll->trusted_eop_path_ll));
        free((void *) (_config.exp_exec_ll->override_xdescr_nvps));
        free((void *) (_config.exp_exec_ll));
        _config.exp_exec_ll = entry;
    }

    _config.state = CONFIG_NOT_INIT;
    return;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

enum ec_init_keys check_dict (
    const char * key, uint_fast32_t keylen)
{
    uint_fast32_t i;

    assert(key != NULL);

    for (i = 0; _key_dict[i].keyname != NULL; ++i)
        if (keylen == _key_dict[i].keylen &&
            strncmp(key, _key_dict[i].keyname, keylen) == 0)
            return _key_dict[i].key;

    return UKN;
}

int load_config (
    const char * key, uint_fast32_t keylen,
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    enum ec_init_keys k;
    static const value_parser_t parsers[] = {
        [ENABLE_EXP_PUB]               = &parse_enable_exp_pub,
        [BROKER_ADDR]                  = &parse_broker_addr,
        [PUB_AGENT_CERT_PATH]          = &parse_pub_agent_cert_path,
        [BROKER_CERT_ROOT_PATH]        = &parse_broker_cert_root_path,
        [PUB_CONTROLLER_ADDR]          = &parse_pub_controller_addr,
        [PUB_PERIOD]                   = &parse_pub_period,
        [PUB_EXP_PRIORITY]             = &parse_pub_exp_priority,
        [PUB_COUNTRY]                  = &parse_pub_country,
        [PUB_LOCATION]                 = &parse_pub_location,
        [CONTROLLER_PRIVKEY_PATH]      = &parse_controller_privkey_path,
        [PROGRAM_DIR_PATH]             = &parse_program_dir_path,
        [PUB_END_GREET_PERIOD]         = &parse_pub_end_greet_period,
        [ENABLE_PROXY_MODE]            = &parse_enable_proxy_mode,
        [BACKLOG_CONN_AGENT_CERT_PATH] = &parse_backlog_conn_agent_cert_path,
        [PROXY_CERT_ROOT_PATH]         = &parse_proxy_cert_root_path,
        [TARGET_BACKLOG_CONN]          = &parse_target_backlog_conn,
        [BACKLOG_CONN_TIMEOUT]         = &parse_backlog_conn_timeout,
        [EXP_EXEC]                     = &parse_exp_exec,
        [UKN]                          = NULL
    };
    value_parser_t parser;

    assert(key != NULL && value != NULL);

    trace(
        "%s(key:%s,keylen:%" PRIuFAST32
        ",value:%s,valuelen:%" PRIuFAST32 ")",
        __func__, key, keylen, value, valuelen);

    k = check_dict(key, keylen);

    assert(k >= 0 && k < sizeof(parsers)/sizeof(*parsers));

    parser = parsers[k];
    if (parser != NULL) {
        if (parser(value, valuelen, lineno) != 0) {
            fatal("Failed to parse field in config file "
                "(key:'%s',value:'%s') at line %" PRIuFAST32 ", aborting",
                key, value, lineno);
            goto fail;
        } else
            goto success;
    } else {
        warn("Unknown field in config file "
            "(key:'%s',value:'%s') at line %" PRIuFAST32 ", ignoring",
            key, value, lineno);
        goto success;
    }

success:
    return 0;
fail:
    return -1;
}

bool need_eop_key (
    const struct pktconf_sep_str_ll * ll)
{
    struct pktconf_sep_str_entry * entry;

    if (ll == NULL)
        return true;

    assert(ll->len != 0);

    for (entry = ll->head; entry != NULL; entry = entry->next)
        if (entry->entrylen == 0 ||
            entry->entrylen == PKTLAB_CHANNELID_LEN_MAX*2)
            return true;

    return false;
}

int parse_bool (
    const char * value, uint_fast32_t valuelen,
    bool * field_set, bool * field)
{
    assert(value != NULL && field_set != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")",
        __func__, value, valuelen);

    if (*field_set)
        return 0; // use last entry

    if (valuelen != 1)
        return -1;

    switch (*value) {
    case '0':
        *field_set = true;
        *field = false;
        break;
    case '1':
        *field_set = true;
        *field = true;
        break;
    default:
        return -1;
    }

    return 0;
}

int parse_addr (
    const char * value, uint_fast32_t valuelen,
    const char ** field)
{
    char * addr;

    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")",
        __func__, value, valuelen);

    if (*field != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no address specified, remain NULL

    if (valuelen >= SIZE_MAX)
        return -1; // overflow

    if (copy_str(
            value, valuelen, &addr) < 0) {
        debug("copy_str failed");
        return -1;
    }

    if (parse_n_getaddrinfo(addr, 0, NULL, NULL) < 0) {
        debug("parse_n_getaddrinfo failed, possibly bad address");
        free(addr);
        return -1;
    }

    *field = addr;
    return 0;
}

int parse_filepath (
    const char * value, uint_fast32_t valuelen,
    const char ** field)
{
    char * filepath;

    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")",
        __func__, value, valuelen);

    if (*field != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no path specified, remain NULL

    if (valuelen >= SIZE_MAX)
        return -1; // overflow

    if (copy_str(
            value, valuelen, &filepath) < 0) {
        debug("copy_str failed");
        return -1;
    }

    if (!is_valid_filepath(filepath)) {
        debug("Bad filepath: %s", filepath);
        free(filepath);
        return -1;
    }

    *field = filepath;
    return 0;
}

int parse_filepath_ls (
    const char * value, uint_fast32_t valuelen,
    const struct pktconf_sep_str_ll ** field)
{
    struct pktconf_sep_str_ll * ll;
    struct pktconf_sep_str_entry * entry;
    uint_fast32_t i;

    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")",
        __func__, value, valuelen);

    if (*field != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no path specified, remain NULL

    if (pktconf_parse_sep_str(
            value, valuelen, ',',
            UINT_FAST32_MAX, &ll) < 0) {
        debug("pktconf_parse_sep_str failed");
        return -1;
    }

    entry = ll->head;
    for (i = 0; i < ll->len; ++i) {
        assert(entry != NULL);
        if (!is_valid_filepath(entry->entry)) {
            debug("Bad filepath: %s", entry->entry);
            goto fail;
        }

        entry = entry->next;
    }

    *field = ll;
    return 0;
fail:
    pktconf_cleanup_sep_str_ll(ll);
    return -1;
}

int parse_dirpath (
    const char * value, uint_fast32_t valuelen,
    const char ** field)
{
    char * dirpath;

    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")",
        __func__, value, valuelen);

    if (*field != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no path specified, remain NULL

    if (valuelen >= SIZE_MAX)
        return -1; // overflow

    if (copy_str(
            value, valuelen, &dirpath) < 0) {
        debug("copy_str failed");
        return -1;
    }

    if (!is_valid_dirpath(dirpath)) {
        debug("Bad dirpath: %s", dirpath);
        free(dirpath);
        return -1;
    }

    // remove ending '/' if exists
    if (dirpath[valuelen-1] == '/')
        dirpath[valuelen-1] = '\0';

    *field = dirpath;
    return 0;
}

int parse_ull (
    const char * value, uint_fast32_t valuelen,
    bool * field_set, unsigned long long * field)
{
    unsigned long long u;
    char * endptr;

    assert(value != NULL && field_set != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")",
        __func__, value, valuelen);

    if (*field_set)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // nothing specified, remain 0

    errno = 0;
    u = strtoull(value, &endptr, 10);
    if (errno != 0 || endptr != value+valuelen) {
        debug("Bad ull: %s", value);
        return -1; // conversion failed/out-of-range
    }

    *field_set = true;
    *field = u;
    return 0;
}

int parse_pktlab_time (
    const char * value, uint_fast32_t valuelen,
    bool * field_set, pktlab_time_t * field)
{
    bool tmp_field_set;
    unsigned long long ull;
    int rst;

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")",
        __func__, value, valuelen);

    if (*field_set)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // nothing specified

    tmp_field_set = false;

    rst = parse_ull(
        value, valuelen,
        &tmp_field_set, &ull);
    if (rst < 0) {
        debug("parse_ull failed with %d", rst);
        return rst;
    }

    assert(tmp_field_set && rst == 0);

    if (ull > UINT64_MAX) {
        warn("Read time too large (%llu), "
            "capping at UINT64_MAX (%" PRIu64 ")",
            ull, UINT64_MAX);
        ull = UINT64_MAX;
    }

    *field_set = true;
    *field = ull;
    return 0;
}

int parse_hex_ls (
    const char * value, uint_fast32_t valuelen,
    const struct pktconf_sep_str_ll ** field)
{
    struct pktconf_sep_str_ll * ll;
    struct pktconf_sep_str_entry * entry;
    uint_fast32_t i;

    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")",
        __func__, value, valuelen);

    if (*field != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no path specified, remain NULL

    if (pktconf_parse_sep_str(
            value, valuelen, ',',
            UINT_FAST32_MAX, &ll) < 0) {
        debug("pktconf_parse_sep_str failed");
        return -1;
    }

    entry = ll->head;
    for (i = 0; i < ll->len; ++i) {
        assert(entry != NULL);
        if (entry->entrylen > 0 &&
            !is_valid_hexstr(entry->entry)) {
            debug("Bad hex str: %s", entry->entry);
            goto fail;
        }

        entry = entry->next;
    }

    *field = ll;
    return 0;
fail:
    pktconf_cleanup_sep_str_ll(ll);
    return -1;
}

int parse_enable_exp_pub (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_bool(
            value, valuelen,
            &_config.enable_exp_pub_specified,
            &_config.enable_exp_pub);
}

int parse_broker_addr (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_addr(
            value, valuelen,
            &_config.broker_addr);
}

int parse_pub_agent_cert_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_filepath(
            value, valuelen,
            &_config.pub_agent_cert_path);
}

int parse_broker_cert_root_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_filepath_ls(
            value, valuelen,
            &_config.broker_cert_root_path_ll);
}

int parse_pub_controller_addr (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_addr(
            value, valuelen,
            &_config.pub_controller_addr);
}

int parse_pub_period (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_pktlab_time(
            value, valuelen,
            &_config.pub_period_specified,
            &_config.pub_period);
}

int parse_pub_exp_priority (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_ull(
            value, valuelen,
            &_config.pub_exp_priority_specified,
            &_config.pub_exp_priority);
}

int parse_pub_country (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    struct pktconf_sep_str_ll * ll;
    struct pktconf_sep_str_entry * entry;
    uint_fast32_t i;

    assert(value != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    if (_config.pub_country != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no country specified, remain NULL

    if (pktconf_parse_sep_str(
            value, valuelen, ',',
            UINT_FAST32_MAX, &ll) < 0) {
        debug("pktconf_parse_sep_str failed");
        return -1;
    }

    entry = ll->head;
    for (i = 0; i < ll->len; ++i) {
        assert(entry != NULL);
        if (!is_valid_iso31661_alpha3(entry->entry)) {
            debug("Bad country: %s", entry->entry);
            goto fail;
        }

        entry = entry->next;
    }

    _config.pub_country = ll;
    return 0;
fail:
    pktconf_cleanup_sep_str_ll(ll);
    return -1;
}

int parse_pub_location (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    struct pktconf_sep_str_ll * ll;
    struct pktconf_sep_str_entry * entry;
    uint_fast32_t i;

    assert(value != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    if (_config.pub_location != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no location specified, remain NULL

    if (pktconf_parse_sep_str(
            value, valuelen, ',',
            UINT_FAST32_MAX, &ll) < 0) {
        debug("pktconf_parse_sep_str failed");
        return -1;
    }

    entry = ll->head;
    for (i = 0; i < ll->len; ++i) {
        assert(entry != NULL);
        if (!is_valid_maidenhead_w_wc(entry->entry)) {
            debug("Bad locator: %s", entry->entry);
            goto fail;
        }

        entry = entry->next;
    }

    _config.pub_location = ll;
    return 0;
fail:
    pktconf_cleanup_sep_str_ll(ll);
    return -1;
}

int parse_pub_end_greet_period (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_pktlab_time(
            value, valuelen,
            &_config.pub_end_greet_period_specified,
            &_config.pub_end_greet_period);
}

int parse_controller_privkey_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_filepath(
            value, valuelen,
            &_config.controller_privkey_path);
}

int parse_program_dir_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_dirpath(
            value, valuelen,
            &_config.program_dir_path);
}

int parse_enable_proxy_mode (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_bool(
            value, valuelen,
            &_config.enable_proxy_mode_specified,
            &_config.enable_proxy_mode);
}

int parse_backlog_conn_agent_cert_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_filepath(
            value, valuelen,
            &_config.backlog_conn_agent_cert_path);
}

int parse_proxy_cert_root_path (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_filepath_ls(
            value, valuelen,
            &_config.proxy_cert_root_path_ll);
}

int parse_target_backlog_conn (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    int rst;

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    rst = parse_ull(
        value, valuelen,
        &_config.target_backlog_conn_specified,
        &_config.target_backlog_conn);
    if (rst >= 0 &&
        _config.target_backlog_conn_specified &&
        _config.target_backlog_conn == 0)
        return -1; // target_backlog_conn should be > 0 int

    return 0;
}

int parse_backlog_conn_timeout (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    return \
        parse_pktlab_time(
            value, valuelen,
            &_config.backlog_setup_timeout_specified,
            &_config.backlog_setup_timeout);
}

int parse_exp_exec (
    const char * value, uint_fast32_t valuelen,
    uint_fast32_t lineno)
{
    struct pktconf_sep_str_ll * ll;
    struct pktconf_sep_str_entry * rawentry;
    struct exp_exec_entry * entry;
    uint_fast32_t i;
    size_t totalmem;
    char * ptr;
    bool is_default_set;
    int ret = -128;

    assert(value != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
        __func__, value, valuelen, lineno);

    if (valuelen == 0)
        return 0; // no exp_exec entry specified

    if (pktconf_parse_sep_str(
            value, valuelen, ':',
            EXP_EXEC_FIELDNO, &ll) < 0) {
        debug("pktconf_parse_sep_str failed");
        return -1;
    }

    if (ll->len != EXP_EXEC_FIELDNO) {
        debug(
            "exp_exec entry does not contain %d fields (got %u)",
            EXP_EXEC_FIELDNO, (unsigned) ll->len);
        ret = -2;
        goto fail;
    }

    // allocate new entry
    totalmem = sizeof(*entry);
    rawentry = ll->head;
    for (i = 0; i < ll->len; ++i) {
        assert(rawentry != NULL && rawentry->entry != NULL);

        if (i == SNIKEY || i == OVERRIDE_XDESCRNVP) {
            if (totalmem > SIZE_MAX-1 ||
                rawentry->entrylen > SIZE_MAX-totalmem-1) {
                debug("Overflow");
                ret = -3;
                goto fail;
            }

            totalmem += rawentry->entrylen+1;
        }

        rawentry = rawentry->next;
    }

    entry = safe_malloc(totalmem);
    memset(entry, 0, totalmem);
    ptr = (char *) (entry+1);
    entry->lineno = lineno;

    // set new entry
    rawentry = ll->head;
    for (i = 0; i < ll->len; ++i) {
        assert(rawentry != NULL && rawentry->entry != NULL);

        switch (i) {
        case IS_DEFAULT:
            is_default_set = false;
            if (parse_bool(
                    rawentry->entry, rawentry->entrylen,
                    &is_default_set, &entry->is_default) < 0) {
                ret = -4;
                goto fail_entry;
            }
            break;
        case SNIKEY: // SNIKEY
            if (rawentry->entrylen == 0)
                break; // empty snikey => NULL in entry

            if (!pktctrl_is_valid_snikey(rawentry->entry)) {
                ret = -5;
                goto fail_entry;
            }
            memcpy(ptr, rawentry->entry, rawentry->entrylen+1);
            entry->sni_key = ptr;
            ptr += rawentry->entrylen+1;
            break;
        case EXPPRIVCHAIN_PATH: // EXPPRIVCHAIN_PATH
            if (rawentry->entrylen == 0) {
                ret = -6;
                goto fail_entry; // must be non-empty
            }

            if (parse_filepath_ls(
                    rawentry->entry, rawentry->entrylen,
                    &entry->exppriv_chain_path_ll) < 0) {
                ret = -7;
                goto fail_entry;
            }
            break;
        case TRUSTEDEOP_PATH: // TRUSTEDEOP_PATH
            // entry is still NULL if valuelen 0
            if (parse_filepath_ls(
                    rawentry->entry, rawentry->entrylen,
                    &entry->trusted_eop_path_ll) < 0) {
                ret = -8;
                goto fail_entry;
            }
            break;
        case PUB_CHANNEL: // PUB_CHANNEL
            // entry is still NULL if valuelen 0
            if (parse_hex_ls(
                    rawentry->entry, rawentry->entrylen,
                    &entry->pub_channel_ll) < 0) {
                ret = -9;
                goto fail_entry;
            }

            // trusted eops required if key pub channel requested
            if (entry->trusted_eop_path_ll == NULL &&
                need_eop_key(entry->pub_channel_ll)) {
                ret = -10;
                goto fail_entry;
            }
            break;
        case OVERRIDE_XDESCRNVP: // OVERRIDE_XDESCRNVP
            if (rawentry->entrylen == 0)
                break; // empty override_xdescrnvp => NULL in entry

            memcpy(ptr, rawentry->entry, rawentry->entrylen+1);
            entry->override_xdescr = ptr;

            // void * cast to write into const field
            // todo: use non-const ptr and assign
            if (pktlab_decode_xdescr(
                    ptr, rawentry->entrylen,
                    (void *) (&entry->override_xdescr_nvps),
                    (void *) (&entry->override_xdescr_nvpcnt)) < 0) {
                entry->override_xdescr_nvps = NULL;
                ret = -11;
                goto fail_entry;
            }
            break;
        default: // bug
            assert(0);
        }

        rawentry = rawentry->next;
    }

    entry->next = _config.exp_exec_ll;
    _config.exp_exec_ll = entry;

    pktconf_cleanup_sep_str_ll(ll);
    return 0;

fail_entry:
    debug("Subfield parsing failed at phase %u",
        (unsigned) i);

    pktconf_cleanup_sep_str_ll(
        (void *) (entry->pub_channel_ll));
    pktconf_cleanup_sep_str_ll(
        (void *) (entry->exppriv_chain_path_ll));
    pktconf_cleanup_sep_str_ll(
        (void *) (entry->trusted_eop_path_ll));
    free((void *) (entry->override_xdescr_nvps));
    free(entry);
fail:
    pktconf_cleanup_sep_str_ll(ll);
    return ret;
}