// ge_deputy_active.c
// Active greet endpoint deputy
// Connect to proxy for incoming endpoints
// and create worker threads for exp running.
//

// todo: backlog exponential backoff

#include <pktlab.h>

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>

#include <netdb.h>
#include <pthread.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>

#include "console.h"
#include "c_reader.h"
#include "net.h"
#include "snikey.h"
#include "util.h"
#include "worker.h"

#include "ge_deputy.h"

//
// INTERNAL CONSTANTS
//

enum conn_setup_st {
    NOT_ASSIGNED    = 0,
    TCP_HANDSHAKE   = 1,
    TLS_HANDSHAKE   = 2,
    PROXY_REG_SEND  = 3,
    PROXY_REG_RECV  = 4,
    RECV_PREAMBLE   = 5,
    SEND_PREAMBLE   = 6,
    SETUP_DONE      = 7
};

enum ge_deputy_active_st {
    NOT_INIT     = 0,
    INIT         = 1,
    G_ONGOING    = 2,
    WRAP_UP      = 3,
    DONE         = 4
};

#define PROXY_PREAMABLE "PKTLAB"

//
// INTERNAL MACROS
//

#define safe_add_time(t1,t2) \
    (((t1) > UINT64_MAX-(t2)) ? UINT64_MAX:((t1)+(t2)))

//
// INTERNAL TYPE DEFINITIONS
//

struct backlog_ctr {
    const char * snikey;
    uint_fast32_t ongoing;   // still working with proxy on setting things up
    uint_fast32_t completed; // awaiting forwarded endpoint TLS traffic
};

struct conn_setup_ll_node {
    int st;
    int sock;
    pktlab_time_t stime;
    uint_fast32_t snikeyindx;
    const char * snikey;
    struct pktctrl_obj * obj;
    struct c_reader * reader;
    size_t preamble_match_indx;
    size_t preamble_sent_indx;
    struct conn_setup_ll_node * next;
};

struct worker_ll_node {
    pthread_t worker_id;
    int worker_fd;
    uint_fast32_t num_id;
    uint_fast32_t snikeyindx;
    bool handshake_complete;
    struct worker_ll_node * next;
};

//
// INTERNAL GLOBAL VARIABLES
//

static struct {
    enum ge_deputy_active_st st;

    char ** exp_prog_n_args;
    uint_fast32_t target_backlog;
    struct addrinfo * proxy_ai;
    struct pktlab_publickey ** trusted_proxy_root;
    uint_fast32_t rootno;
    struct snikey_ll_node * snikeys;
    pktlab_time_t proxy_timeout;
    struct pktlab_privatekey * controller_key;
    struct pktlab_certificate * agent_cert;
    struct pktctrl_keycert_store * greet_store;

    struct {
        struct conn_setup_ll_node * conn_ll;
    } backlog_setup;

    struct {
        struct backlog_ctr * ctrs;
        uint_fast32_t ctrno;
    } backlog_state;

    struct {
        uint_fast32_t max_worker_id;
        struct worker_ll_node * worker_ll;
    } worker_info;
} _st;

//
// EXPORTED INTERFACE FUNCTION DECLARATIONS
//

static int ge_deputy_init(
    char ** exp_prog_n_args,
    struct pktlab_privatekey * controller_key,
    struct pktctrl_keycert_store * greet_store,
    void * aux);
static int ge_deputy_start(void);

static int ge_deputy_prepare_select (
    int * nfds, fd_set * rset, fd_set * wset,
    pktlab_time_t * select_until);
static int ge_deputy_process_select (
    const fd_set * rset, const fd_set * wset,
    pktlab_time_t now);

static int ge_deputy_wrapup(void);
static int ge_deputy_cleanup(void);
static bool ge_deputy_is_done(void);

//
// INTERNAL FUNCTION DECLARATIONS
//

static int backlog_state_setup (
    struct snikey_ll_node * ll);

static int replenish_backlog(void);

static void push_new_backlog_conn (
    struct conn_setup_ll_node ** node,
    const char * snikey, uint_fast32_t snikeyindx);

static int backlog_conn_setup (
    struct conn_setup_ll_node * node);
static int start_backlog_tcp_handshake (
    struct conn_setup_ll_node * node);
static int finish_tcp_handshake_and_start_tls (
    struct conn_setup_ll_node * node);
static int finish_tls_handshake_and_send_req (
    struct conn_setup_ll_node * node);
static int finish_send_req_and_wait_rsp (
    struct conn_setup_ll_node * node);
static int finish_recv_rsp_and_close_obj (
    struct conn_setup_ll_node * node);
static int recv_until_preamble (
    struct conn_setup_ll_node * node);
static int send_preamble (
    struct conn_setup_ll_node * node);
static void free_conn_setup_ll_node (
    struct conn_setup_ll_node * node, bool close_sock);
static void free_n_update_conn_setup_ll_node (
    struct conn_setup_ll_node ** pre,
    struct conn_setup_ll_node ** node,
    bool close_sock);

static void free_all_conn_setup_ll_node(void);

static bool backlog_is_replenishable(void);
static bool conn_setup_ll_node_is_set (
    struct conn_setup_ll_node * node,
    const fd_set * rset, const fd_set * wset);

static void get_conn_setup_needed_rw (
    int st, bool * readable, bool * writable);

static int process_backlog_setup(
    struct conn_setup_ll_node ** pre,
    struct conn_setup_ll_node ** node);

static bool is_recoverable_connect_err(int err);

static int worker_setup (
    int endpoint_fd, uint_fast32_t snikeyindx);

static void free_worker_ll_node (
    struct worker_ll_node * node);
static void free_pre_tls_worker_ll_node(void);
static void free_all_worker_ll_node(void);

static void process_worker_readable (
    struct worker_ll_node ** pre,
    struct worker_ll_node ** node);

static struct pktctrl_keycert_store * get_client_keycert_store (
    struct pktlab_certificate * cert,
    struct pktlab_publickey ** trusted_root,
    uint_fast32_t rootno);

static inline const char * state_str(int st);

//
// EXPORTED FUNCTION DEFINITIONS
//

struct ge_methods get_active_ge_methods(void) {
    struct ge_methods ret = {
        .init = ge_deputy_init,
        .start = ge_deputy_start,
        .prepare_select = ge_deputy_prepare_select,
        .process_select = ge_deputy_process_select,
        .wrapup = ge_deputy_wrapup,
        .cleanup = ge_deputy_cleanup,
        .is_done = ge_deputy_is_done
    };

    return ret;
}

//
// EXPORTED INTERFACE FUNCTION DEFINITIONS
//

int ge_deputy_init(
    char ** exp_prog_n_args,
    struct pktlab_privatekey * controller_key,
    struct pktctrl_keycert_store * greet_store,
    void * aux)
{
    struct ge_active_aux * arg = aux; // aux owned by ge_deputy_init

    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (_st.st != NOT_INIT ||
        exp_prog_n_args == NULL || controller_key == NULL ||
        greet_store == NULL || aux == NULL ||
        arg->proxy_ai == NULL || arg->snikeys == NULL ||
        arg->target_backlog == 0)
        return -1;

    memset(&_st, 0, sizeof(_st));

    if (backlog_state_setup(arg->snikeys) < 0)
        return -2;

    // none of the passed in ptrs are owned by ge deputy
    _st.exp_prog_n_args = exp_prog_n_args;
    _st.proxy_ai = arg->proxy_ai;
    _st.agent_cert = arg->agent_cert;
    _st.trusted_proxy_root = arg->trusted_proxy_root;
    _st.rootno = arg->rootno;
    _st.target_backlog = arg->target_backlog;
    _st.snikeys = arg->snikeys;
    _st.proxy_timeout = arg->proxy_timeout;
    _st.controller_key = controller_key;
    _st.greet_store = greet_store;
    _st.st = INIT;

    free(aux);
    return 0;
}

int ge_deputy_start(void) {
    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (_st.st != INIT)
        return -1;

    if (replenish_backlog() < 0) {
        free_all_conn_setup_ll_node();
        return -2;
    }

    _st.st = G_ONGOING;
    return 0;
}

int ge_deputy_prepare_select (
    int * nfds, fd_set * rset, fd_set * wset,
    pktlab_time_t * select_until)
{
    int rst;
    struct conn_setup_ll_node * csnode;
    struct worker_ll_node * wnode;
    bool readable, writable;

    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (nfds == NULL || rset == NULL || wset == NULL ||
        select_until == NULL || _st.st <= INIT)
        return -1;

    switch (_st.st) {
    case G_ONGOING:
        // immediate replenishing in process select
        if (backlog_is_replenishable())
            *select_until = 0;

        for (csnode = _st.backlog_setup.conn_ll;
             csnode != NULL; csnode = csnode->next) {
            switch (csnode->st) {
            case TCP_HANDSHAKE:
            case SEND_PREAMBLE:
                set_fd_set(nfds, wset, csnode->sock);
                break;
            case TLS_HANDSHAKE:
            case PROXY_REG_SEND:
            case PROXY_REG_RECV:
                get_conn_setup_needed_rw(
                    csnode->st, &readable, &writable);

                rst = pktctrl_prepare_select(
                    nfds, rset, wset,
                    readable, writable, csnode->obj);
                assert(rst == PKTCTRL_SUCCESS);
                break;
            case RECV_PREAMBLE:
                set_fd_set(nfds, rset, csnode->sock);
                break;
            case SETUP_DONE:
                // immediate do process select for SETUP_DONE ones
                *select_until = 0;
                break;
            default:
                assert(0); // bug
            }

            // for other cases select until proxy timeout
            if (csnode->st != SETUP_DONE)
                *select_until = MIN(
                    *select_until,
                    safe_add_time(
                        csnode->stime, _st.proxy_timeout));
        }
        // fall through
    case WRAP_UP:
        for (wnode = _st.worker_info.worker_ll;
             wnode != NULL; wnode = wnode->next)
            set_fd_set(nfds, rset, wnode->worker_fd);

        if (*nfds == 0)
            *select_until = 0; // don't wait if nothing to wait for
        break;
    case DONE:
        *select_until = 0; // done, no need to wait
        break;
    default:
        assert(0); // bug
    }

    debug("ge_deputy_prepare_select select until %" PRIu64, *select_until);
    return 0;
}

int ge_deputy_process_select (
    const fd_set * rset, const fd_set * wset,
    pktlab_time_t now)
{
    struct conn_setup_ll_node * pre_csnode;
    struct conn_setup_ll_node * csnode;
    struct worker_ll_node * pre_wnode;
    struct worker_ll_node * wnode;
    int rst;

    trace("%s(_st.st:%s,now:%" PRIu64 ")",
        __func__, state_str(_st.st), now);

    if (rset == NULL || wset == NULL ||
        _st.st <= INIT)
        return -1;

    switch (_st.st) {
    case G_ONGOING:
        pre_csnode = NULL;
        csnode = _st.backlog_setup.conn_ll;
        while (csnode != NULL) {
            assert(now >= csnode->stime);

            if (csnode->st == SETUP_DONE ||
                conn_setup_ll_node_is_set(csnode, rset, wset)) {
                rst = process_backlog_setup(&pre_csnode, &csnode);
                if (rst < 0) {
                    debug("process_backlog_setup failed with %d", rst);
                    return -2; // fail immediately
                }

                continue;
            } else if (now-csnode->stime > _st.proxy_timeout) {
                // timed-out and not set, clean up node directly
                warn("Attempted backlog connection setup timed-out, "
                    "retrying later");
                debug("node:{sock:%d,st:%d} timed out",
                    csnode->sock, csnode->st);
                free_n_update_conn_setup_ll_node(&pre_csnode, &csnode, true);

                continue;
            }

            // wait a bit longer
            pre_csnode = csnode;
            csnode = csnode->next;
        }

        if (backlog_is_replenishable() &&
            replenish_backlog() < 0)
            return -3; // fail immediately
        // fall through
    case WRAP_UP:
        pre_wnode = NULL;
        wnode = _st.worker_info.worker_ll;
        while (wnode != NULL) {
            if (!FD_ISSET(wnode->worker_fd, rset)) {
                pre_wnode = wnode;
                wnode = wnode->next;
                continue;
            }

            process_worker_readable(&pre_wnode, &wnode);
        }
        // fall through
    case DONE: // do nothing
        break;
    default:
        assert(0); //bug
    }

    // G_ONGOING -> WRAP_UP triggered externally
    // WRAP_UP -> DONE done internally
    if (_st.st == WRAP_UP &&
        _st.worker_info.worker_ll == NULL)
            _st.st = DONE;

    return 0;
}

int ge_deputy_wrapup(void) {
    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (_st.st <= INIT)
        return -1;

    if (_st.st >= WRAP_UP)
        return 0; // already wrapping/wrapped up

    assert(_st.st == G_ONGOING);

    // clean up all non-experiment-running backlog connections
    free_all_conn_setup_ll_node();
    free_pre_tls_worker_ll_node();

    _st.st = WRAP_UP;
    return 0;
}

int ge_deputy_cleanup(void) {
    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (_st.st == NOT_INIT ||
        (_st.st >= G_ONGOING &&
         _st.st <= WRAP_UP))
        return -1;

    info("Cleaning up deputy");

    free_all_conn_setup_ll_node();
    free_all_worker_ll_node();

    free(_st.backlog_state.ctrs);

    memset(&_st, 0, sizeof(_st));
    _st.st = NOT_INIT;
    return 0;
}

bool ge_deputy_is_done(void) {
    trace("%s(_st.st:%s)", __func__, state_str(_st.st));
    return (_st.st == DONE);
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int backlog_state_setup (
    struct snikey_ll_node * ll)
{
    struct snikey_ll_node * node;
    uint_fast32_t totalmem, i;

    assert(ll != NULL);

    trace("%s(...)", __func__);

    for (node = ll; node != NULL; node = node->next) {
        assert(node->snikey != NULL);
        ++_st.backlog_state.ctrno;
    }

    totalmem = \
        sizeof(*_st.backlog_state.ctrs)*\
        _st.backlog_state.ctrno;
    if (totalmem > SIZE_MAX ||
        totalmem/sizeof(*_st.backlog_state.ctrs) !=
        _st.backlog_state.ctrno) {
        return -1; // overflow
    }

    _st.backlog_state.ctrs = safe_malloc(totalmem);
    memset(_st.backlog_state.ctrs, 0, totalmem);

    i = 0;
    for (node = ll; node != NULL; node = node->next) {
        _st.backlog_state.ctrs[i].snikey = node->snikey;
        ++i;
    }

    return 0;
}

int replenish_backlog(void) {
    uint_fast32_t i, j;
    struct conn_setup_ll_node * node;
    struct conn_setup_ll_node * next;
    int rst;

    assert(_st.st == INIT || _st.st == G_ONGOING);

    trace("%s(...)", __func__);

    for (i = 0; i < _st.target_backlog; ++i) {
        // ensure that in case backlog count is restricted on proxy
        // we have them spreaded among different snikeys
        for (j = 0; j < _st.backlog_state.ctrno; ++j) {
            assert(
                _st.backlog_state.ctrs[j].ongoing <=
                UINT_FAST32_MAX-_st.backlog_state.ctrs[j].completed);

            if (_st.backlog_state.ctrs[j].ongoing+
                _st.backlog_state.ctrs[j].completed > i)
                continue;

            push_new_backlog_conn(
                &node, _st.backlog_state.ctrs[j].snikey, j);

            rst = backlog_conn_setup(node);
            if (rst == -1) {
                warn(
                    "Encountered non-fatal error "
                    "when setting up backlog connect "
                    "for snikey: %s, retrying later",
                    _st.backlog_state.ctrs[j].snikey);

                next = node->next;
                free_conn_setup_ll_node(node, true);
                _st.backlog_setup.conn_ll = next;
                continue;
            } else if (rst == -2) {
                fatal(
                    "Encountered fatal error "
                    "when setting up backlog connect "
                    "for snikey: %s", _st.backlog_state.ctrs[j].snikey);
                return -1;
            }

            // conn setup in progress or successful
            ++_st.backlog_state.ctrs[j].ongoing;
        }
    }

    return 0;
}

void push_new_backlog_conn (
    struct conn_setup_ll_node ** node,
    const char * snikey, uint_fast32_t snikeyindx)
{
    assert(node != NULL);

    trace("%s(snikey:%s,snikeyindx:%" PRIuFAST32 ")",
        __func__, snikey, snikeyindx);

    *node = safe_malloc(sizeof(**node));
    memset(*node, 0, sizeof(**node));

    (*node)->st = NOT_ASSIGNED;
    (*node)->sock = -1;
    (*node)->stime = 0;
    (*node)->snikey = snikey;
    (*node)->snikeyindx = snikeyindx;
    (*node)->preamble_match_indx = 0;
    (*node)->preamble_sent_indx = 0;
    (*node)->next = _st.backlog_setup.conn_ll;
    _st.backlog_setup.conn_ll = *node;

    return;
}

int backlog_conn_setup (
    struct conn_setup_ll_node * node)
{
    int ret = -128;

    assert(node != NULL);

    trace("%s(node:{st:%d,sock:%d})",
        __func__, node->st, node->sock);

    //
    // Return value
    //       1: success
    //       0: try again later
    //      -1: recoverable-error, i.e. a new backlog_conn_setup call may work
    //      -2: non-recoverable-error, i.e. new backlog_conn_setups will fail
    //          not point in continuing backlog conn setup
    //

    switch (node->st) {
    case NOT_ASSIGNED:
        ret = start_backlog_tcp_handshake(node);
        break;
    case TCP_HANDSHAKE:
        ret = finish_tcp_handshake_and_start_tls(node);
        break;
    case TLS_HANDSHAKE:
        ret = finish_tls_handshake_and_send_req(node);
        break;
    case PROXY_REG_SEND:
        ret = finish_send_req_and_wait_rsp(node);
        break;
    case PROXY_REG_RECV:
        ret = finish_recv_rsp_and_close_obj(node);
        break;
    case RECV_PREAMBLE:
        ret = recv_until_preamble(node);
        break;
    case SEND_PREAMBLE:
        ret = send_preamble(node);
        break;
    case SETUP_DONE: // shouldn't call this function for done node
    default:
        assert(0); // bug
    }

    // set keepalive & close-on-exec after setup completed
    if (ret == 1) {
        if (set_keepalive(node->sock) < 0) {
            warn("Set backlog connection keepalive failed");
            return -1; // not sure why will this fail, soft fail for now
        }

        if (set_close_on_exec(node->sock) < 0) {
            warn("Set backlog connection close-on-exec failed");
            return -1; // not sure why will this fail, soft fail for now
        }
    }

    return ret;
}

int start_backlog_tcp_handshake (
    struct conn_setup_ll_node * node)
{
    int rst, sock, err;

    assert(node != NULL && node->st == NOT_ASSIGNED);

    trace("%s(node:{sock:%d})", __func__, node->sock);

    rst = setup_nonblock_connect_socket(
        _st.proxy_ai, &sock, &err);
    if (rst < 0) {
        debug("setup_nonblock_connect_socket failed with %d", rst);
        if (rst == -3) {
            warn("Attempted connect failed with: %s", strerror(err));
            if (is_recoverable_connect_err(err))
                return -1;
        }

        return -2;
    }

    // setup successful
    node->sock = sock;
    node->stime = pktlab_time_now();
    node->st = TCP_HANDSHAKE;

    // if connect is fast, try next step
    // give time as in immediate subsequent backlog_conn_setup call
    if (rst == 1)
        return finish_tcp_handshake_and_start_tls(node);

    return 0;
}

int finish_tcp_handshake_and_start_tls (
    struct conn_setup_ll_node * node)
{
    int rst, err, ret = -2;
    struct pktctrl_obj * obj;
    struct pktctrl_keycert_store * store;

    assert(node != NULL && node->st == TCP_HANDSHAKE);

    trace("%s(node:{sock:%d})", __func__, node->sock);

    rst = check_nonblock_connect_socket(
        node->sock, &err);
    if (rst < 0) {
        debug("check_nonblock_connect_socket failed with %d", rst);

        if (rst == -2) {
            warn("Attempted connect failed after wait with: %s",
                strerror(err));

            if (is_recoverable_connect_err(err))
                return -1;
        }

        return -2;
    }

    // check successful
    store = get_client_keycert_store(
        _st.agent_cert, _st.trusted_proxy_root, _st.rootno);
    assert(store != NULL);

    obj = pktctrl_create_obj();
    assert(obj != NULL);

    rst = pktctrl_connect(
        node->sock, PKTLAB_AUTH_AGENT, NULL,
        (_st.agent_cert != NULL) ?
            _st.controller_key:NULL, // provide key only when having cert
        store, obj, NULL, NULL);
    if (rst != PKTCTRL_IN_PROGRESS &&
        rst != PKTCTRL_SUCCESS) {
        debug("pktctrl_connect failed with %d", rst);

        switch (rst) {
        case PKTCTRL_BAD_CERT:
            fatal("Bad agent cert provided "
                "for proxy TLS connection setup");
            break;
        case PKTCTRL_BAD_CERTKEY:
            fatal("Controller private key "
                "does not match agent cert for proxy");
            break;
        case PKTCTRL_SSL_FAILURE:
            fatal("Backlog connection TLS setup failed");
            break;
        default:
            assert(0); // bug
        }

        goto fail_obj;
    }

    // setup successful
    pktctrl_cleanup_keycert_store(store);

    node->obj = obj;
    node->stime = pktlab_time_now();
    node->st = TLS_HANDSHAKE;

    // if TLS is fast, try next step
    if (rst == PKTCTRL_SUCCESS)
        return finish_tls_handshake_and_send_req(node);

    return 0;

fail_obj:
    pktctrl_cleanup_keycert_store(store);
    pktctrl_close(obj);
    return ret;
}

int finish_tls_handshake_and_send_req (
    struct conn_setup_ll_node * node)
{
    int rst;
    struct pktlab_message xreg_msg;

    assert(node != NULL && node->st == TLS_HANDSHAKE);

    trace("%s(node:{sock:%d})", __func__, node->sock);

    if (!pktctrl_is_established(node->obj)) {
        rst = pktctrl_connect(
            node->sock, PKTLAB_AUTH_AGENT,
            NULL, NULL, NULL, node->obj, NULL, NULL);
        if (rst != PKTCTRL_SUCCESS) {
            debug("pktctrl_connect failed with %d", rst);

            switch (rst) {
            case PKTCTRL_IN_PROGRESS:
                // TLS may still be in progress after select returns
                // wait a bit longer, but stime is not updated
                return 0;
            case PKTCTRL_SSL_FAILURE:
                fatal("Backlog connection TLS setup failed");
                return -2;
            default:
                assert(0); // bug
            }
        }
    }

    // setup successful
    xreg_msg.type = PKTLAB_XREG_MESSAGE;
    xreg_msg.xreg.snikeyptr = node->snikey;
    xreg_msg.xreg.snikeylen = strlen(node->snikey);

    rst = pktctrl_write_message(node->obj, &xreg_msg);
    if (rst < 0) {
        debug("pktctrl_write_message failed with %d (%s)",
            errno, strerror(errno));

        if (errno == EPIPE) {
            warn("Backlog connection interrupted "
                "during xreg request sending");
            return -1; // connection somehow interrupted
        }

        fatal("Unknown error during backlog connection "
            "xreg request sending");
        return -2; // unknown error
    } else if (rst == 0) {
        debug("pktctrl_write_message denied xreg");
        // wait a bit longer, but stime is not updated
        return 0;
    }

    // xreg msg accepted by writer
    node->stime = pktlab_time_now();
    node->st = PROXY_REG_SEND;

    // if xreg is flushed, try next step
    if (pktctrl_write_unsent(node->obj) == 0)
        return finish_send_req_and_wait_rsp(node);

    return 0;
}

int finish_send_req_and_wait_rsp (
    struct conn_setup_ll_node * node)
{
    int rst;

    assert(node != NULL && node->st == PROXY_REG_SEND);

    trace("%s(node:{sock:%d})", __func__, node->sock);

    while (pktctrl_write_unsent(node->obj) != 0) {
        rst = pktctrl_flush_write(node->obj);
        if (rst < 0) {
            debug("pktctrl_flush_write failed with %d (%s)",
                errno, strerror(errno));

            if (errno == EPIPE) {
                warn("Backlog connection interrupted "
                    "during xreg request sending");
                return -1;
            }

            fatal("Unknown error during backlog connection "
                "xreg request sending");
            return -2; // unknown error
        } else if (rst == 0) {
            // wait a bit longer, but stime is not updated
            return 0;
        }
    }

    // xreg msg sent
    node->reader = get_c_reader(node->obj);
    node->stime = pktlab_time_now();
    node->st = PROXY_REG_RECV;

    // always try and see if we can received stuff
    return finish_recv_rsp_and_close_obj(node);
}

int finish_recv_rsp_and_close_obj (
    struct conn_setup_ll_node * node)
{
    int rst;
    struct pktlab_message * recv_msg;

    assert(node != NULL && node->st == PROXY_REG_RECV);

    trace("%s(node:{sock:%d})", __func__, node->sock);

    rst = c_read_msg(node->reader, &recv_msg);
    if (rst < 0) {
        debug("c_read_msg failed with %d", rst);
        fatal("Unknown error during backlog connection "
            "xreg response receiving");
        return -2;
    } else if (rst == 0) {
        // wait a bit longer, but stime is not updated
        return 0;
    }

    assert(rst == 1);

    if (recv_msg == NULL) {
        warn("Unexpected EOF during backlog connection "
            "xreg response receiving");
        return -1;
    }

    // received response
    if (recv_msg->type != PKTLAB_RESULT_MESSAGE) {
        fatal("Unexpected response msg type (%d) "
            "during backlog connection "
            "xreg response receiving", (int) recv_msg->type);
        free(recv_msg);
        return -2;
    } else if (recv_msg->result.errid != PKTLAB_SUCCESS) {
        // soft error only, can change this to return strong error (-2)
        // based on errid content
        warn("Backlog connection xreg response received errid: %"
            PRIu16, recv_msg->result.errid);
        free(recv_msg);
        return -1;
    }

    // setup successful
    free(recv_msg);
    pktctrl_close(node->obj);
    close_c_reader(node->reader);
    node->obj = NULL;
    node->reader = NULL;
    node->stime = pktlab_time_now();
    node->st = RECV_PREAMBLE;

    // try if we can receive preamble already
    return recv_until_preamble(node);
}

int recv_until_preamble (
    struct conn_setup_ll_node * node)
{
    ssize_t rst;
    char buf;
    static const char preamble[] = PROXY_PREAMABLE;

    assert(node != NULL && node->st == RECV_PREAMBLE);

    trace("%s(node:{sock:%d})", __func__, node->sock);

    while (true) {
        rst = read(node->sock, &buf, sizeof(buf));
        if (rst < 0) {
            debug("read failed with %s", strerror(errno));
            if (errno == EAGAIN || errno == EWOULDBLOCK)
                return 0;

            fatal("Unknown error during proxy preamble receiving");
            return -2;
        } else if (rst == 0) {
            warn("Unexpected EOF during proxy preamble receiving");
            return -1;
        }

        assert(
            rst == sizeof(buf) &&
            node->preamble_match_indx < sizeof(preamble)-1);

        if (buf == preamble[node->preamble_match_indx]) {
            ++node->preamble_match_indx;

            if (node->preamble_match_indx == sizeof(preamble)-1)
                break;
        } else {
            node->preamble_match_indx = 0; // restart matching
        }
    }

    // setup successful
    node->preamble_match_indx = 0;
    node->stime = pktlab_time_now();
    node->st = SEND_PREAMBLE;

    // try if we can send preamble already
    return send_preamble(node);
}

int send_preamble (
    struct conn_setup_ll_node * node)
{
    ssize_t rst;
    static const char preamble[] = PROXY_PREAMABLE;

    assert(node != NULL && node->st == SEND_PREAMBLE);

    trace("%s(node:{sock:%d})", __func__, node->sock);

    while (node->preamble_sent_indx < sizeof(preamble)-1) {
        rst = write(
            node->sock, preamble+node->preamble_sent_indx,
            sizeof(preamble)-1-node->preamble_sent_indx);
        if (rst < 0) {
            debug("write failed with %s", strerror(errno));
            if (errno == EAGAIN || errno == EWOULDBLOCK)
                return 0;

            if (errno == EPIPE) {
                warn("Unexpected broken pipe during proxy preamble sending");
                return -1;
            }

            fatal("Unknown error during proxy preamble sending");
            return -2;
        }

        assert(rst != 0);
        node->preamble_sent_indx += rst;
    }

    // send preamble successful
    // cleanup setup stuffs
    debug("Backlog connection (sock:%d) established", node->sock);
    node->stime = 0; // don't keep track anymore
    node->preamble_sent_indx = 0;
    node->st = SETUP_DONE;
    return 1;
}

void free_conn_setup_ll_node (
    struct conn_setup_ll_node * node, bool close_sock)
{
    assert(node != NULL);

    trace("%s(node:{sock:%d,st:%d},close_sock:%d)",
        __func__, node->sock, node->st, close_sock);

    if (node->obj != NULL)
        pktctrl_close(node->obj);

    if (close_sock)
        close(node->sock);

    close_c_reader(node->reader);
    free(node);
    return;
}

void free_n_update_conn_setup_ll_node (
    struct conn_setup_ll_node ** pre,
    struct conn_setup_ll_node ** node,
    bool close_sock)
{
    uint_fast32_t snikeyindx;
    struct conn_setup_ll_node * next;

    assert(pre != NULL && node != NULL && *node != NULL);

    trace("%s(pre:%p,node:{sock:%d,st:%d},close_sock:%d)",
        __func__, *pre, (*node)->sock, (*node)->st, close_sock);

    snikeyindx = (*node)->snikeyindx;

    if (*pre != NULL) {
        next = (*pre)->next = (*node)->next;
    } else {
        next = _st.backlog_setup.conn_ll = (*node)->next;
    }

    free_conn_setup_ll_node(*node, close_sock);
    --_st.backlog_state.ctrs[snikeyindx].ongoing;
    *node = next; // pre not updated for free case
    return;
}

void free_all_conn_setup_ll_node(void) {
    uint_fast32_t i;
    struct conn_setup_ll_node * node;

    trace("%s()", __func__);

    while (_st.backlog_setup.conn_ll != NULL) {
        node = _st.backlog_setup.conn_ll->next;
        free_conn_setup_ll_node(
            _st.backlog_setup.conn_ll, true);
        _st.backlog_setup.conn_ll = node;
    }

    // reset backlog state if possible
    if (_st.backlog_state.ctrs != NULL) {
        for (i = 0; i < _st.backlog_state.ctrno; ++i)
            _st.backlog_state.ctrs[i].ongoing = 0;
    }

    return;
}

bool backlog_is_replenishable(void) {
    uint_fast32_t i;

    assert( _st.st == G_ONGOING);

    trace("%s()", __func__);

    for (i = 0; i < _st.backlog_state.ctrno; ++i) {
        assert(
            _st.backlog_state.ctrs[i].ongoing <=
            UINT_FAST32_MAX-_st.backlog_state.ctrs[i].completed);

        debug("Backlog state (snikey:%s): "
            "ongoing:%" PRIuFAST32 ", completed: %" PRIuFAST32,
            _st.backlog_state.ctrs[i].snikey,
            _st.backlog_state.ctrs[i].ongoing,
            _st.backlog_state.ctrs[i].completed);

        if (_st.backlog_state.ctrs[i].ongoing+
            _st.backlog_state.ctrs[i].completed < _st.target_backlog)
            return true;
    }

    return false;
}

bool conn_setup_ll_node_is_set (
    struct conn_setup_ll_node * node,
    const fd_set * rset, const fd_set * wset)
{
    int rst;
    bool readable, writeable;

    // NOT_ASSIGNED and SETUP_DONE should not be called

    assert(
        node != NULL && rset != NULL && wset != NULL &&
        node->st != NOT_ASSIGNED && node->st != SETUP_DONE);

    trace("%s(node:{sock:%d,st:%d})",
        __func__, node->sock, node->st);

    if (node->st == TCP_HANDSHAKE || node->st == SEND_PREAMBLE) {
        return FD_ISSET(node->sock, wset);
    } else if (node->st == RECV_PREAMBLE) {
        return FD_ISSET(node->sock, rset);
    }

    assert(node->obj != NULL);

    rst = pktctrl_process_select(
        rset, wset, node->obj, &readable, &writeable);
    assert(rst == PKTCTRL_SUCCESS);

    debug("pktctrl_process_select(sock:%d) "
        "returned readable:%d writable:%d",
        node->sock, readable, writeable);

    switch (node->st) {
    case TLS_HANDSHAKE:
        return (readable || writeable);
    case PROXY_REG_SEND:
        return writeable;
    case PROXY_REG_RECV:
        return readable;
    default:
        assert(0); // bug
    }
}

void get_conn_setup_needed_rw (
    int st, bool * readable, bool * writable)
{
    assert(readable != NULL && writable != NULL);

    trace("%s(st:%d)", __func__, st);

    switch (st) {
    case TLS_HANDSHAKE:
        *readable = true;
        *writable = true;
        break;
    case PROXY_REG_SEND:
        *readable = false;
        *writable = true;
        break;
    case PROXY_REG_RECV:
        *readable = true;
        *writable = false;
        break;
    default:
        assert(0); // bug
    }

    return;
}

int process_backlog_setup(
    struct conn_setup_ll_node ** pre,
    struct conn_setup_ll_node ** node)
{
    int rst;
    uint_fast32_t snikeyindx;

    assert(pre != NULL && node != NULL && *node != NULL);

    trace("%s(node:{sock:%d,st:%d})",
        __func__, (*node)->sock, (*node)->st);

    // backlog connection done case
    if ((*node)->st == SETUP_DONE) {
        snikeyindx = (*node)->snikeyindx;

        rst = worker_setup(
            (*node)->sock, snikeyindx);
        debug("worker_setup returned %d", rst);

        free_n_update_conn_setup_ll_node(pre, node, (rst < 0));

        if (rst < 0) { // worker not created
            if (rst == -6)
                return -1; // fail immediately
        } else
            ++_st.backlog_state.ctrs[snikeyindx].completed;

        return 0;
    }

    // backlog connection setup case
    rst = backlog_conn_setup(*node);
    debug("backlog_conn_setup returned %d", rst);

    if (rst < 0) {
        free_n_update_conn_setup_ll_node(pre, node, true);

        if (rst == -2) // serious error
            return -2; // fail immediately
    } else {
        *pre = *node;
        *node = (*node)->next;
    }

    // we deal with rst == 1 in next process select
    return 0;
}

bool is_recoverable_connect_err(int err) {
    trace("%s(err:%s)", __func__, strerror(err));

    switch (err) {
    case ECONNREFUSED:
    case ECONNRESET:
    case EHOSTUNREACH:
    case EINTR:
    case ENETDOWN:
    case ENETUNREACH:
    case ETIMEDOUT:
        return true; // we return remote down/refuse as recoverable
    case ENOBUFS:
        return false;
    case EINPROGRESS: // we should catch this before calling
    default:
        assert(0); // bug
    }
}

int worker_setup(int endpoint_fd, uint_fast32_t snikeyindx) {
    int rst, err, ret = -128, fds[2];
    struct worker_arg * wa;
    sigset_t oset;
    pthread_t worker_id;
    struct worker_ll_node * node;
    uint_fast32_t worker_num_id;

    assert(_st.st == G_ONGOING);

    trace("%s(endpoint_fd:%d,snikeyindx:%" PRIuFAST32 ")",
        __func__, endpoint_fd, snikeyindx);

    info("Set up new worker awaiting endpoint TLS setup");

    // spawn worker thread
    if (get_stream_socketpair(fds, &err) < 0) {
        warn("Failed to create manager-worker socket pair (%s), "
            "skipping worker create", strerror(err));
        return -1;
    }

    if (set_close_on_exec(fds[0]) < 0 ||
        set_close_on_exec(fds[1]) < 0) {
        warn("Failed to set internal socket pair close-on-exec (%s), "
            "skipping worker create", strerror(errno));
        ret = -2;
        goto fail_fds;
    }

    if (set_nonblock(fds[0], true) < 0 ||
        set_nonblock(fds[1], true) < 0) {
        warn("Failed to set internal socket pair non-blocking (%s), "
            "skipping worker create", strerror(errno));
        ret = -3;
        goto fail_fds;
    }

    worker_num_id = _st.worker_info.max_worker_id+1;
    _st.worker_info.max_worker_id = worker_num_id;
    // update _st here, so if fail afterwards id is not retaken

    wa = safe_malloc(sizeof(*wa));
    wa->endpoint_fd = endpoint_fd;
    wa->manager_fd = fds[1]; // fds[1] is for worker
    wa->worker_id = worker_num_id;
    wa->exp_prog_n_args = _st.exp_prog_n_args;
    wa->k = _st.controller_key;
    wa->store = _st.greet_store;

    if (set_sigmask_block_sigint(&oset) < 0) {
        warn("Failed to block SIGINT, skipping worker create");
        ret = -4;
        goto fail_wa;
    }

    rst = pthread_create(
        &worker_id, NULL, worker_run, wa);
    if (rst != 0) {
        warn("Failed to create worker thread (%s), "
            "skipping worker create", strerror(rst));
        ret = -5;
        goto fail_wa;
    }

    if (restore_sigmask(&oset) < 0) {
        fatal("Failed to restore sigmask, aborting");
        ret = -6;
        goto fail_post_create;
    }

    rst = pthread_detach(worker_id); // no joining afterwards
    if (rst != 0) {
        warn("Failed to detach worker thread (%s), "
            "early worker termination", strerror(rst));
        ret = -7;
        // we only "signal" thread should die here
        goto fail_post_create;
    }

    node = safe_malloc(sizeof(*node));
    node->worker_id = worker_id;
    node->worker_fd = fds[0]; // fds[0] is for manager
    node->num_id = worker_num_id;
    node->handshake_complete = false;
    node->snikeyindx = snikeyindx;
    node->next = _st.worker_info.worker_ll;
    _st.worker_info.worker_ll = node;

    info("Worker %" PRIuFAST32 " created; awaiting endpoint",
        worker_num_id);
    return 0;

// for failing until pthread_create
// as deputy still owns everything

fail_wa:
    free(wa);
fail_fds:
    close(fds[0]);
    close(fds[1]);
    return ret;

// for failing after pthread_create
// only free the part deputy owns.
// the others are freed by worker

fail_post_create:
    close(fds[0]);
    return ret;
}

void free_worker_ll_node (
    struct worker_ll_node * node)
{
    assert(node != NULL);

    trace("%s(...)", __func__);

    close(node->worker_fd);
    free(node);
    return;
}

void free_pre_tls_worker_ll_node(void) {
    struct worker_ll_node * pre;
    struct worker_ll_node * node;
    struct worker_ll_node * next;

    trace("%s()", __func__);

    pre = NULL;
    node = _st.worker_info.worker_ll;
    while (node != NULL) {
        if (!node->handshake_complete) {
            if (pre == NULL) {
                next = _st.worker_info.worker_ll = node->next;
            } else {
                next = pre->next = node->next;
            }

            free_worker_ll_node(node);
            node = next; // pre is the same
            continue;
        }

        pre = node;
        node = node->next;
    }

    return;
}

void free_all_worker_ll_node(void) {
    struct worker_ll_node * node;

    trace("%s()", __func__);

    while (_st.worker_info.worker_ll != NULL) {
        node = _st.worker_info.worker_ll->next;
        free_worker_ll_node(_st.worker_info.worker_ll);
        _st.worker_info.worker_ll = node;
    }

    return;
}

void process_worker_readable (
    struct worker_ll_node ** pre,
    struct worker_ll_node ** node)
{
    uint8_t msg;
    ssize_t rst;
    struct worker_ll_node * next;

    assert(_st.st == G_ONGOING || _st.st == WRAP_UP);
    assert(pre != NULL && node != NULL && *node != NULL);

    trace("%s(...)", __func__);

    rst = recv_worker_msg((*node)->worker_fd, &msg);
    switch (rst) {
    case 1: // received msg or EOF
        if (msg == WORKER_TLS_COMPLETED) {
            assert(!(*node)->handshake_complete);

            info("Worker %" PRIuFAST32 " TLS handshake completed",
                (*node)->num_id);

            (*node)->handshake_complete = true;
            --_st.backlog_state.ctrs[(*node)->snikeyindx].completed;

            *pre = *node;
            *node = (*node)->next;
            return; // worker notification only, carry on
        }

        assert(msg == WORKER_EOF);
        info("Worker %" PRIuFAST32 " socket EOF, "
            "cleaning up worker", (*node)->num_id);
        break; // free worker node
    case 0: // try again
        *pre = *node; // for next node
        *node = (*node)->next;
        return;
    default:
        assert(rst < 0); // error
        warn("Unexpected error (%s) when reading worker (%" PRIuFAST32
            ") socket, cleaning up worker",
            strerror(errno), (*node)->num_id);
        break; // free worker node
    }

    if (*pre == NULL) {
        next = _st.worker_info.worker_ll = (*node)->next;
    } else {
        next = (*pre)->next = (*node)->next;
    }

    if (!(*node)->handshake_complete)
        --_st.backlog_state.ctrs[(*node)->snikeyindx].completed;

    free_worker_ll_node(*node);
    *node = next;
    return; // pre stays the same
}

struct pktctrl_keycert_store * get_client_keycert_store (
    struct pktlab_certificate * cert,
    struct pktlab_publickey ** trusted_root,
    uint_fast32_t rootno)
{
    struct pktctrl_keycert_store * store;

    store = pktctrl_create_keycert_store();

    if (pktctrl_set_client_keycert(
            store,
            (cert == NULL) ? NULL:&cert,
            (cert == NULL) ? 0:1,
            trusted_root, rootno) != PKTCTRL_SUCCESS) {
        debug("pktctrl_set_client_keycert failed");
        goto fail;
    }

    return store;

fail:
    pktctrl_cleanup_keycert_store(store);
    return NULL;
}

const char * state_str(int st) {
    static const char * sstr[] = {
        [NOT_INIT]     = "NOT_INIT",
        [INIT]         = "INIT",
        [G_ONGOING]    = "G_ONGOING",
        [WRAP_UP]      = "WRAP_UP",
        [DONE]         = "DONE"
    };

    assert(st >= 0 && st < sizeof(sstr)/sizeof(*sstr));
    return safe_str(sstr[st]);
}