// prog.c
// Program helper module
//

#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <pktlab.h>
#include <pktctrl.h>

#include "console.h"
#include "util.h"

//
// INTERNAL CONSTANTS
//

#define FILEPATH_MAX 0x10000

#define EBPF_PROG_FILEEXT ".ebpf"

enum module_state {
    PREINIT_ST = 0,
    INIT_ST    = 1,
    LOADED_ST  = 2
};

enum submit_state {
    SENDMSG_ST,
    RECVRSP_ST,
    ERR_ST
};

enum submit_target {
    SENDMON,
    RECVMON,
    FILTER
};

//
// INTERNAL TYPE DEFINITIONS
//

struct loaded_prog {
    void * encoded_prog;
    uint_fast32_t encoded_proglen;
    uint8_t digest[PKTLAB_SHA256_DIGEST_LEN];
    struct loaded_prog * next;
};

//
// INTERNAL GLOBAL VARIABLES
//

static struct {
    enum module_state state;
    const char * dirpath;
    struct loaded_prog prog; // sentinel node
} _st;

//
// EXPORTED TYPE DEFINITIONS
//

struct prog_submit_obj {
    struct pktctrl_obj * obj;

    // for recording progress
    enum submit_target target;
    enum submit_state st;
    uint_fast32_t current_prog_indx;

    // data reference
    uint8_t * send_mon_digests;
    uint_fast32_t send_mon_digest_cnt;
    uint8_t * recv_mon_digests;
    uint_fast32_t recv_mon_digest_cnt;
    uint8_t * filter_digests;
    uint_fast32_t filter_digest_cnt;
};

//
// INTERNAL FUNCTION DECLARATIONS
//

static enum pktlab_prog_type get_prog_type(const char * filename);

static int load_prog (
    const char * filepath, enum pktlab_prog_type type,
    uint_fast32_t * errlineno);
static void prog_ll_cleanup(void);

static int set_n_set_prog_digests (
    const uint8_t * prog_digests, uint_fast32_t prog_cnt,
    uint8_t ** digests_field, uint_fast32_t * cnt_field);

static const char * get_submit_target_hint (
    enum submit_target target);
static uint8_t * get_submit_target_digest (
    struct prog_submit_obj * submit_obj);
static uint_fast32_t get_submit_target_digest_cnt (
    struct prog_submit_obj * submit_obj);

static int send_cdata_prog (
    struct prog_submit_obj * submit_obj);
static int recv_cdata_rsp (
    struct prog_submit_obj * submit_obj,
    struct pktlab_message ** bad_rsp);

static void warn_digest (
    const char * prefix, const uint8_t * digest);

//
// EXPORTED FUNCTION DEFINITIONS
//

int prog_init(const char * prog_dir_path) {
    trace("%s(_st.state:%d,prog_dir_path:%s)",
        __func__, _st.state, safe_str(prog_dir_path));

    if (_st.state != PREINIT_ST)
        return -1; // bad state

    memset(&_st, 0, sizeof(_st));

    _st.state = INIT_ST;
    _st.dirpath = prog_dir_path;
    return 0;
}

int prog_load(void) {
    DIR * dp;
    struct dirent * dptr;
    char filepath[FILEPATH_MAX];
    size_t dirpathlen, filepathlen;
    uint_fast32_t errlineno;
    int type, ret = -128, rst;

    trace("%s(_st.state:%d)", __func__, _st.state);

    if (_st.state != INIT_ST)
        return -1; // bad state

    if (_st.dirpath == NULL) {
        _st.state = LOADED_ST;
        return 0; // load nothing
    }

    dp = opendir(_st.dirpath);
    if (dp == NULL) {
        fatal("Open program directory (%s) failed with %s",
            _st.dirpath, strerror(errno));
        return -2;
    }

    dirpathlen = strlen(_st.dirpath);
    if (dirpathlen > sizeof(filepath)-2) {
        fatal("Directory path (%zu) too long (max allowed: %zu)",
            dirpathlen, sizeof(filepath)-2);
        ret = -3;
        goto fail;
    }

    strncpy(filepath, _st.dirpath, dirpathlen);
    filepath[dirpathlen] = '/'; // dir /
    *(filepath+dirpathlen+1) = '\0';

    // try loading all programs from dir
    for (errno = 0, dptr = readdir(dp);
         dptr != NULL; errno = 0, dptr = readdir(dp)) {
        if (dptr->d_name[0] == '.')
            continue; // ignore hidden files

        type = get_prog_type(dptr->d_name);
        if (type < 0) {
            warn("Unknown program file extension for %s,"
                " skipping", dptr->d_name);
            continue;
        }

        filepathlen = strlen(dptr->d_name);
        if (filepathlen > sizeof(filepath)-dirpathlen-2) {
            warn("File name (%zu for %s) too long (max allowed: %zu),"
                " skipping", filepathlen, dptr->d_name,
                sizeof(filepath)-dirpathlen-2);
            continue;
        }

        strncpy(filepath+dirpathlen+1, dptr->d_name, filepathlen);
        *(filepath+dirpathlen+1+filepathlen) = '\0';

        rst = load_prog(filepath, type, &errlineno);
        if (rst < 0) {
            debug("load_prog failed with %d", rst);

            if (errlineno != 0) {
                warn("Bad line %" PRIuFAST32, errlineno);
            } else if (rst == -5) {
                warn("File too big for cdata msg");
            }

            warn("Failed to load program %s, skipping", dptr->d_name);
            continue;
        }
    }

    if (errno != 0) {
        fatal("Get directory entry failed with %s",
            strerror(errno));
        prog_ll_cleanup();
        ret = -4;
        goto fail;
    }

    _st.state = LOADED_ST;
    ret = 0;

fail:
    closedir(dp);
    return ret;
}

void prog_cleanup(void) { // clean up and reset things back to PREINIT
    if (_st.state == PREINIT_ST)
        return;

    prog_ll_cleanup();

    memset(&_st, 0, sizeof(_st));
    _st.state = PREINIT_ST;
    return;
}

int prog_lookup_program (
    const uint8_t * prog_digest,
    const void ** encoded_prog, uint_fast32_t * encoded_proglen)
{
    struct loaded_prog * node;

    if (_st.state != LOADED_ST)
        return -1;

    if (prog_digest == NULL ||
        encoded_prog == NULL || encoded_proglen == NULL)
        return -2; // bad arg

    for (node = _st.prog.next;
         node != NULL; node = node->next) {
        if (memcmp(
                prog_digest, node->digest,
                sizeof(node->digest)) != 0)
            continue;

        // found program
        *encoded_prog = node->encoded_prog;
        *encoded_proglen = node->encoded_proglen;
        return 0;
    }

    return -3; // program not found
}

int prog_submit_programs_init_hash (
    struct pktctrl_obj * obj,
    const uint8_t * send_mon_digests, uint_fast32_t send_mon_digest_cnt,
    const uint8_t * recv_mon_digests, uint_fast32_t recv_mon_digest_cnt,
    const uint8_t * filter_digests, uint_fast32_t filter_digest_cnt,
    struct prog_submit_obj ** submit_obj)
{
    struct prog_submit_obj * sobj;
    int ret = -128;

    trace("%s(send_mon_digest_cnt:%" PRIuFAST32
        ",recv_mon_digest_cnt:%" PRIuFAST32
        ",filter_digest_cnt:%" PRIuFAST32 ")",
        __func__, send_mon_digest_cnt,
        recv_mon_digest_cnt, filter_digest_cnt);

    if (_st.state != LOADED_ST)
        return -1; // bad module state

    if (obj == NULL || submit_obj == NULL)
        return -2; // bad arg

    sobj = safe_malloc(sizeof(*sobj));
    memset(sobj, 0, sizeof(*sobj));
    sobj->obj = obj;
    sobj->st = SENDMSG_ST;
    sobj->target = SENDMON; // always start at send monitor
    sobj->current_prog_indx = 0;

    if (set_n_set_prog_digests(
            send_mon_digests, send_mon_digest_cnt,
            &sobj->send_mon_digests,
            &sobj->send_mon_digest_cnt) < 0) {
        ret = -3;
        goto fail_sobj;
    }

    if (set_n_set_prog_digests(
            recv_mon_digests, recv_mon_digest_cnt,
            &sobj->recv_mon_digests,
            &sobj->recv_mon_digest_cnt) < 0) {
        ret = -4;
        goto fail_send_mon;
    }

    if (set_n_set_prog_digests(
            filter_digests, filter_digest_cnt,
            &sobj->filter_digests,
            &sobj->filter_digest_cnt) < 0) {
        ret = -5;
        goto fail_recv_mon;
    }

    *submit_obj = sobj;
    return 0;

fail_recv_mon:
    free(sobj->recv_mon_digests);
fail_send_mon:
    free(sobj->send_mon_digests);
fail_sobj:
    free(sobj);
    return ret;
}

int prog_submit_programs_init_store (
    struct pktctrl_obj * obj, const char * snikey,
    struct pktctrl_keycert_store * store,
    struct prog_submit_obj ** submit_obj)
{
    int rst, ret = -128;
    struct pktlab_certificate ** cert_ls;
    uint_fast32_t certnum;
    uint8_t * ftdigests;
    uint8_t * mtdigests;
    uint_fast32_t ftcnt, mtcnt;
    uint8_t subable_channel[PKTLAB_CHANNELID_LEN_MAX];

    rst = pktctrl_get_server_keycert(
        store, snikey, &cert_ls, &certnum, NULL, NULL);
    if (rst != PKTCTRL_SUCCESS) {
        debug("pktctrl_get_server_keycert(snikey:%s) failed with %d",
            safe_str(snikey), rst);
        return -1;
    }

    rst = pktlab_parse_controller_exp_cert_chain(
        cert_ls, certnum, NULL, NULL,
        &ftdigests, &ftcnt, &mtdigests, &mtcnt,
        NULL, NULL, subable_channel);
    if (rst != PKTLAB_PARSE_CHAIN_SUCCESS) {
        debug("pktlab_parse_controller_exp_cert_chain failed with %d", rst);
        return -2;
    }

    // todo: add support for recv monitor
    rst = prog_submit_programs_init_hash(
        obj, mtdigests, mtcnt, NULL, 0,
        ftdigests, ftcnt, submit_obj);
    if (rst < 0) {
        debug("prog_submit_programs_init_hash failed with %d", rst);
        ret = -3;
        goto fail_parse;
    }

    ret = 0;

fail_parse:
    free(ftdigests);
    free(mtdigests);
    return ret;
}

int prog_submit_programs_prepare_select (
    int * nfds, fd_set * rset, fd_set * wset,
    pktlab_time_t * select_until,
    struct prog_submit_obj * submit_obj)
{
    int rst;

    trace("%s(mstate:%d)", __func__, _st.state);

    if (_st.state != LOADED_ST)
        return -1; // bad module state

    if (nfds == NULL || rset == NULL || wset == NULL ||
        select_until == NULL || submit_obj == NULL ||
        submit_obj->st == ERR_ST)
        return -2; // bad arg

    debug("submit_obj.target:%d,submit_obj.st:%d,"
        "submit_obj.current_prog_indx:%" PRIuFAST32,
        submit_obj->target, submit_obj->st,
        submit_obj->current_prog_indx);

    if (submit_obj->target == FILTER &&
        submit_obj->st == SENDMSG_ST &&
        submit_obj->current_prog_indx >= \
        get_submit_target_digest_cnt(submit_obj) &&
        pktctrl_write_unsent(submit_obj->obj) == 0) {
        *select_until = 0; // immediately timeout select
        return 1;
    }

    rst = pktctrl_prepare_select(
        nfds, rset, wset,
        (submit_obj->st == RECVRSP_ST),
        (submit_obj->st == SENDMSG_ST), submit_obj->obj);
    if (rst != PKTCTRL_SUCCESS) {
        debug("pktctrl_prepare_select failed with %d", rst);
        submit_obj->st = ERR_ST;
        return -3;
    }

    // select_until not modified if actually selecting
    return 0;
}

int prog_submit_programs_process_select (
    const fd_set * rset, const fd_set * wset,
    struct prog_submit_obj * submit_obj)
{
    int rst;
    bool readable, writable;
    struct pktlab_message * bad_rsp;

    trace("%s(mstate:%d)", __func__, _st.state);

    if (_st.state != LOADED_ST)
        return -1; // bad module state

    if (rset == NULL || wset == NULL ||
        submit_obj == NULL || submit_obj->st == ERR_ST)
        return -2; // bad arg

    debug("submit_obj.target:%d,submit_obj.st:%d,"
        "submit_obj.current_prog_indx:%" PRIuFAST32,
        submit_obj->target, submit_obj->st,
        submit_obj->current_prog_indx);

    if (submit_obj->target == FILTER &&
        submit_obj->st == SENDMSG_ST &&
        submit_obj->current_prog_indx >= \
        get_submit_target_digest_cnt(submit_obj) &&
        pktctrl_write_unsent(submit_obj->obj) == 0)
        return 1; // immediate return as already done

    rst = pktctrl_process_select(
        rset, wset, submit_obj->obj, &readable, &writable);
    if (rst != PKTCTRL_SUCCESS) {
        debug("pktctrl_process_select failed with %d", rst);
        submit_obj->st = ERR_ST;
        return -3;
    }

    if (submit_obj->st == SENDMSG_ST && writable) {
        // submit_obj->st updated within
        rst = send_cdata_prog(submit_obj);
        switch (rst) {
        case 1:
            return 1; // done
        case 0:
            break; // continue on
        case -1:
        case -3:
            debug("send_cdata_prog send failed with %d", rst);
            return -4;
        default:
            debug("send_cdata_prog failed with %d", rst);
            return -5;
        }
    } else if (submit_obj->st == RECVRSP_ST && readable) {
        // submit_obj->st updated within
        rst = recv_cdata_rsp(submit_obj, &bad_rsp);
        if (rst < 0) {
            debug("recv_cdata_rsp failed with %d", rst);

            if (rst == -1) { // bad rsp
                if (bad_rsp == NULL) {
                    info("Endpoint EOF detected");
                } else if (bad_rsp->type == PKTLAB_RESULT_MESSAGE) {
                    warn("Result message reporting error: %" PRIu16,
                        bad_rsp->result.errid);
                } else {
                    warn("Unexpected message type: %d", bad_rsp->type);
                }

                free(bad_rsp);
            }

            return -6;
        }
    }

    return 0;
}

void prog_submit_programs_cleanup (
    struct prog_submit_obj * submit_obj)
{
    if (submit_obj == NULL)
        return;

    free(submit_obj->send_mon_digests);
    free(submit_obj->recv_mon_digests);
    free(submit_obj->filter_digests);
    free(submit_obj);
    return;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

enum pktlab_prog_type get_prog_type(const char * filename) {
    const char * ptr;

    assert(filename != NULL);

    ptr = strrchr(filename, '.');
    if (ptr == NULL)
        return PKTLAB_EBPF; // file with no extension treated as ebpf progs

    if (strcmp(EBPF_PROG_FILEEXT, ptr) == 0)
        return PKTLAB_EBPF;

    return -1; // unknown
}

int load_prog (
    const char * filepath, enum pktlab_prog_type type,
    uint_fast32_t * errlineno)
{
    uint8_t * content;
    size_t contentlen;
    void * raw_prog;
    uint_fast32_t raw_proglen;
    void * encoded_prog;
    uint_fast32_t encoded_proglen;
    struct loaded_prog * new_prog;
    int ret = -128, rst;

    assert(
        _st.state == INIT_ST &&
        filepath != NULL && errlineno != NULL);

    *errlineno = 0;

    if (read_whole_file(
            filepath, &content, &contentlen) < 0) {
        debug("Failed to read %s content", filepath);
        return -1;
    }

    if (contentlen > UINT_FAST32_MAX) {
        ret = -2; // overflow
        goto fail_content;
    }

    rst = pktlab_load_program(
        (char *) content, contentlen, type,
        &raw_prog, &raw_proglen, errlineno);
    if (rst < 0) {
        debug("Failed to load program %s content (%d)",
            filepath, rst);
        ret = -3;
        goto fail_content;
    }

    *errlineno = 0;

    rst = pktlab_encode_program(
        type, raw_prog, raw_proglen,
        &encoded_prog, &encoded_proglen);
    if (rst < 0) {
        debug("Failed to encode program %s raw bytes (%d)",
            filepath, rst);
        ret = -4;
        goto fail_raw_prog;
    }

    if (encoded_proglen > UINT32_MAX) {
        debug("Encoded program too long for cdata msg");
        ret = -5;
        goto fail_encoded_prog;
    }

    new_prog = safe_malloc(sizeof(*new_prog));
    memset(new_prog, 0, sizeof(*new_prog));

    new_prog->encoded_prog = encoded_prog;
    new_prog->encoded_proglen = encoded_proglen;
    new_prog->next = _st.prog.next;

    rst = pktlab_get_program_digest(
        encoded_prog, encoded_proglen, new_prog->digest);
    if (rst < 0) {
        debug("Failed to get encoded program %s digest (%d)",
            filepath, rst);
        ret = -6;
        goto fail_new_prog;
    }

    // actual ll push
    _st.prog.next = new_prog;

    free(raw_prog);
    free(content);
    return 0;

fail_new_prog:
    free(new_prog);
fail_encoded_prog:
    free(encoded_prog);
fail_raw_prog:
    free(raw_prog);
fail_content:
    free(content);
    return ret;
}

void prog_ll_cleanup(void) {
    struct loaded_prog * node;

    assert(_st.state != PREINIT_ST);

    while (_st.prog.next != NULL) {
        node = _st.prog.next;
        _st.prog.next = node->next;

        free(node->encoded_prog);
        free(node);
    }

    return;
}

int set_n_set_prog_digests (
    const uint8_t * prog_digests, uint_fast32_t prog_cnt,
    uint8_t ** digests_field, uint_fast32_t * cnt_field)
{
    uint8_t * actual_prog_digests;
    uint_fast32_t actual_prog_cnt;
    size_t totalmem;

    assert(
        _st.state == LOADED_ST &&
        digests_field != NULL && cnt_field != NULL);

    actual_prog_digests = NULL;
    actual_prog_cnt = 0;

    // only send programs if we have them
    if (prog_cnt != 0) {
        totalmem = PKTLAB_SHA256_DIGEST_LEN*prog_cnt;
        if (totalmem/PKTLAB_SHA256_DIGEST_LEN != prog_cnt)
            return -1; // overflow

        actual_prog_digests = safe_malloc(totalmem);
        memcpy(actual_prog_digests, prog_digests, totalmem);
        actual_prog_cnt = prog_cnt;
    }

    *digests_field = actual_prog_digests;
    *cnt_field = actual_prog_cnt;
    return 0;
}

const char * get_submit_target_hint (
    enum submit_target target)
{
    switch (target) {
    case SENDMON:
        return PKTLAB_CDATA_HINT_SEND_MONITOR;
    case RECVMON:
        return PKTLAB_CDATA_HINT_RECV_MONITOR;
    case FILTER:
        return PKTLAB_CDATA_HINT_FILTER;
    default:
        assert(0); // bug
    }
}

uint8_t * get_submit_target_digest (
    struct prog_submit_obj * submit_obj)
{
    assert(submit_obj != NULL);

    switch (submit_obj->target) {
    case SENDMON:
        return submit_obj->send_mon_digests;
    case RECVMON:
        return submit_obj->recv_mon_digests;
    case FILTER:
        return submit_obj->filter_digests;
    default:
        assert(0); // bug
    }
}

uint_fast32_t get_submit_target_digest_cnt (
    struct prog_submit_obj * submit_obj)
{
    assert(submit_obj != NULL);

    switch (submit_obj->target) {
    case SENDMON:
        return submit_obj->send_mon_digest_cnt;
    case RECVMON:
        return submit_obj->recv_mon_digest_cnt;
    case FILTER:
        return submit_obj->filter_digest_cnt;
    default:
        assert(0); // bug
    }
}

int send_cdata_prog (
    struct prog_submit_obj * submit_obj)
{
    struct pktlab_message cdata_msg;
    size_t offset;
    const void * encoded_prog;
    uint_fast32_t encoded_proglen;
    int rst;

    assert(
        _st.state == LOADED_ST &&
        submit_obj != NULL && submit_obj->st == SENDMSG_ST);

    // flush out unsent stuffs
    if (pktctrl_write_unsent(submit_obj->obj) != 0) {
        if (pktctrl_flush_write(submit_obj->obj) < 0) {
            debug("pktctrl_flush_write failed with %s", strerror(errno));
            submit_obj->st = ERR_ST;
            return -1;
        }

        // recv rsp if buffer cleaned
        // otherwise retry
        if (pktctrl_write_unsent(submit_obj->obj) == 0)
            submit_obj->st = RECVRSP_ST;

        return 0;
    }

    // no buffered stuffs left
    // no unsent digest left case
    if (submit_obj->current_prog_indx >= \
        get_submit_target_digest_cnt(submit_obj)) {
        // reach submission end
        if (submit_obj->target == FILTER)
            return 1; // done

        // submit next target
        ++submit_obj->target;
        submit_obj->current_prog_indx = 0;
        return 0;
    }

    // unsent digest exists case
    offset = \
        PKTLAB_SHA256_DIGEST_LEN*
        submit_obj->current_prog_indx;
    if (offset/PKTLAB_SHA256_DIGEST_LEN !=
        submit_obj->current_prog_indx) {
        submit_obj->st = ERR_ST;
        return -2; // overflow
    }

    rst = prog_lookup_program(
        get_submit_target_digest(submit_obj)+offset,
        &encoded_prog, &encoded_proglen);
    if (rst < 0) {
        assert(rst == -3);

        // only warn when program dir specified
        if (_st.dirpath != NULL) {
            warn("Skipping as cannot find program for following digest:");
            warn_digest(NULL, get_submit_target_digest(submit_obj)+offset);
        }

        ++submit_obj->current_prog_indx;
        return 0;
    }

    // ensured during loading
    assert(encoded_proglen <= UINT32_MAX);

    cdata_msg.type = PKTLAB_CDATA_MESSAGE;
    cdata_msg.cdata.hintptr = \
        get_submit_target_hint(submit_obj->target);
    cdata_msg.cdata.hintlen = \
        strlen(get_submit_target_hint(submit_obj->target));
    cdata_msg.cdata.dataptr = encoded_prog;
    cdata_msg.cdata.datalen = encoded_proglen;

    rst = pktctrl_write_message(
        submit_obj->obj, &cdata_msg);
    switch (rst) {
    case 1:
        // if nothing buffered wait for rsp directly
        // otherwise try flush
        if (pktctrl_write_unsent(submit_obj->obj) == 0)
            submit_obj->st = RECVRSP_ST;
        break;
    case 0:
        return 0; // retry later
    case -1:
    default:
        debug("pktctrl_write_message failed with %d (%s)",
            rst, strerror(errno));
        submit_obj->st = ERR_ST;
        return -3;
    }

    return 0;
}

int recv_cdata_rsp (
    struct prog_submit_obj * submit_obj,
    struct pktlab_message ** bad_rsp)
{
    struct pktlab_message * recv_msg;
    int rst, ret = -128;

    assert(
        _st.state == LOADED_ST &&
        submit_obj != NULL && submit_obj->st == RECVRSP_ST &&
        bad_rsp != NULL);

    rst = pktctrl_read_message(submit_obj->obj, &recv_msg);
    switch (rst) {
    case 1:
        if (recv_msg == NULL ||
            recv_msg->type != PKTLAB_RESULT_MESSAGE ||
            recv_msg->result.errid != PKTLAB_SUCCESS) {
            ret = -1; // bad rsp
            *bad_rsp = recv_msg;
            submit_obj->st = ERR_ST;
            break;
        }

        ++submit_obj->current_prog_indx;
        submit_obj->st = SENDMSG_ST;
        ret = 0; // continue on
        free(recv_msg);
        break;
    case 0:
        ret = 0; // retry later
        break;
    case -1:
    default:
        debug("pktctrl_read_message failed with %d (%s)",
            rst, strerror(errno));
        submit_obj->st = ERR_ST;
        ret = -2; // failed to read msg
        break;
    }

    return ret;
}

void warn_digest (
    const char * prefix, const uint8_t * digest)
{
    char digest_str[PKTLAB_SHA256_DIGEST_LEN*3];
    int i;

    assert(digest != NULL);

    for (i = 0; i < PKTLAB_SHA256_DIGEST_LEN; ++i)
        snprintf(
            digest_str+i*3, PKTLAB_SHA256_DIGEST_LEN*3-i*3,
            "%02X ", digest[i]);

    warn("%s%s", (prefix == NULL) ? "":prefix, digest_str);
}