// worker.c
//

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <limits.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include <pktlab.h>
#include <pktctrl.h>

#include "console.h"
#include "net.h"
#include "prog.h"
#include "util.h"
#include "worker.h"

//
// INTERNAL CONSTANTS
//

enum start_session_st {
    START_SEND_ST,
    FLUSH_SEND_ST,
    RECVRSP_ST
};

//
// INTERNAL FUNCTION DECLARATIONS
//

static int complete_TLS_handshake (
    int manager_fd, int endpoint_fd,
    uint_fast32_t worker_id,
    struct pktlab_privatekey * k,
    struct pktctrl_keycert_store * store,
    struct pktctrl_obj ** endpointobj, char ** client_snikey);
static int submit_programs (
    int manager_fd, uint_fast32_t worker_id,
    struct pktctrl_keycert_store * store,
    struct pktctrl_obj * endpointobj, const char * client_snikey);
static int start_session (
    int manager_fd, uint_fast32_t worker_id,
    struct pktctrl_obj * endpointobj);

static char ** create_child_environ(uint_fast32_t worker_id, int endpoint_fd);

//
// EXPORTED FUNCTION DEFINITIONS
//

void * worker_run(void * arg) {
    struct worker_arg * wa = arg;
    int endpoint_fd, manager_fd, rst, fds[2], err, nsel;
    uint_fast32_t worker_id;
    char ** exp_prog_envp;
    char ** exp_prog_n_args;
    char * exp_prog;
    char * ptr;
    char * client_snikey;
    struct pktlab_privatekey * k;
    struct pktctrl_keycert_store * store;
    struct pktctrl_obj * endpointobj;
    struct pktctrl_obj * expprogobj;
    pid_t pid;
    fd_set rset, wset;
    struct pktlab_message * endpoint2expprog;
    struct pktlab_message * expprog2endpoint;
    bool read_from_endpoint, write_to_endpoint;
    bool read_from_expprog, write_to_expprog;
    bool endpoint_readable, endpoint_writable;
    bool expprog_readable, expprog_writable;

    assert(arg != NULL);

    trace("%s(arg:{endpoint_fd:%d,manager_fd:%d,worker_id:%"
        PRIuFAST32 ",exp_prog:%s})", __func__,
        wa->endpoint_fd, wa->manager_fd,
        wa->worker_id, safe_str(wa->exp_prog_n_args[0]));

    // extract arguments
    endpoint_fd = wa->endpoint_fd;
    manager_fd = wa->manager_fd;
    worker_id = wa->worker_id;
    exp_prog_n_args = wa->exp_prog_n_args;
    exp_prog = exp_prog_n_args[0];
    k = wa->k;
    store = wa->store;
    free(arg);
    arg = NULL;
    assert(
        endpoint_fd >= 0 && manager_fd >= 0 &&
        exp_prog_n_args != NULL && exp_prog != NULL &&
        k != NULL && store != NULL);

    // set up pktctrl
    rst = complete_TLS_handshake(
        manager_fd, endpoint_fd, worker_id,
        k, store, &endpointobj, &client_snikey);
    if (rst < 0) {
        debug("complete_TLS_handshake failed with %d", rst);
        goto fail;
    }

    // submit programs
    rst = submit_programs(
        manager_fd, worker_id, store,
        endpointobj, client_snikey);
    if (rst < 0) {
        debug("submit_programs failed with %d", rst);
        goto fail_obj;
    }

    // signal start session
    rst = start_session(
        manager_fd, worker_id, endpointobj);
    if (rst < 0) {
        debug("start_session failed with %d", rst);
        goto fail_obj;
    }

    // signal pktctrl setup success
    rst = send_worker_msg(
        manager_fd, WORKER_TLS_COMPLETED);
    if (rst != 1) {
        debug("send_worker_msg returned %d", rst);
        assert(rst <= 0);

        if (rst <= 0) {
            warn("Worker %" PRIuFAST32
                " failed to notify manager on TLS connection completed,"
                " skipping endpoint", worker_id);
            debug("send_worker_msg err: %s",
                (rst == 0) ? "n/a":strerror(errno));
        }

        goto fail_obj;
    }

    // set up exp prog socket pair
    // fds[0] is worker end
    // fds[1] is exp prog end
    if (get_stream_socketpair(fds, &err) < 0) {
        warn("Worker %" PRIuFAST32
            " failed to create worker-expprog socket pair (%s),"
            " skipping this endpoint", worker_id, strerror(err));
        goto fail_obj;
    }

    pid = 0; // for skipping child waiting in case of early cleanup
    expprogobj = pktctrl_create_obj();
    assert(expprogobj != NULL);

    if (pktctrl_raw_session(
            fds[0], expprogobj) != PKTCTRL_SUCCESS) {
        warn("Worker %" PRIuFAST32
            " failed create pktctrl_obj for exp program (%s),"
            " skipping this endpoint", worker_id, strerror(errno));
        close(fds[1]);
        goto fail_expprogobj;
    }

    pid = fork();
    if (pid == 0) { // child
        // Create child environment. We pass the endpoint address to the child
        // as an ASCII string in EPADDR.

        exp_prog_envp = create_child_environ(worker_id, endpoint_fd);

        // move fd and exec
        close(fds[0]);

        if (dup2(fds[1], 0) < 0) {
            warn("Worker %" PRIuFAST32
                " failed to move worker descriptor to 0 (%s),"
                " skipping this endpoint", worker_id, strerror(errno));
            exit(EXIT_FAILURE);
        }

        // modification to exp_prog_n_args here is fine as it is a new copy
        exp_prog = exp_prog_n_args[0];
        ptr = strrchr(exp_prog_n_args[0], '/');
        if (ptr != NULL)
            exp_prog_n_args[0] = ptr+1; // remove string prior to '/'

        // note it is possible for new exp_prog_n_args[0] to be 0 len
        // bad bin though, will fail next step

        if (execve(exp_prog, exp_prog_n_args, exp_prog_envp) < 0) {
            warn("Worker %" PRIuFAST32
                " cannot run %s (%s), skipping this endpoint",
                worker_id, exp_prog, strerror(errno));
            exit(EXIT_FAILURE);
        }
    }

    // parent forward traffic, and clean up when EOF
    close(fds[1]);

    info("Worker %" PRIuFAST32 " setup complete, start message forwarding",
        worker_id);

    // we only do single message buffer for each direction
    // should be NULL if no msg buffered or msg ptr
    endpoint2expprog = NULL;
    expprog2endpoint = NULL;

    while (true) {
        FD_ZERO(&rset);
        FD_ZERO(&wset);
        nsel = 0;

        if (endpoint2expprog == NULL) {
            read_from_endpoint = true;
            write_to_expprog = false;
        } else {
            read_from_endpoint = false;
            write_to_expprog = true;
        }

        if (expprog2endpoint == NULL) {
            read_from_expprog = true;
            write_to_endpoint = false;
        } else {
            read_from_expprog = false;
            write_to_endpoint = true;
        }

        if (pktctrl_write_unsent(endpointobj) > 0)
            write_to_endpoint = true;

        if (pktctrl_write_unsent(expprogobj) > 0)
            write_to_expprog = true;

        if (pktctrl_prepare_select(
                &nsel, &rset, &wset,
                read_from_endpoint,
                write_to_endpoint,
                endpointobj) != PKTCTRL_SUCCESS) {
            warn("Worker %" PRIuFAST32
                " failed to select pktctrl_obj for endpoint,"
                " skipping this endpoint", worker_id);
            goto fail_msgbuf;
        }

        if (pktctrl_prepare_select(
                &nsel, &rset, &wset,
                read_from_expprog,
                write_to_expprog,
                expprogobj) != PKTCTRL_SUCCESS) {
            warn("Worker %" PRIuFAST32
                " failed to select pktctrl_obj for exp program,"
                " skipping this endpoint", worker_id);
            goto fail_msgbuf;
        }

        FD_SET(manager_fd, &rset); // for check eof only

        rst = select(nsel, &rset, &wset, NULL, NULL); // block indefinitely
        if (rst < 0) {
            if (errno == EINTR)
                continue; // re-select again

            warn("Worker %" PRIuFAST32
                " select failed (%s), skipping this endpoint",
                worker_id, strerror(errno));
            goto fail_msgbuf;
        }

        if (pktctrl_process_select(
                &rset, &wset, endpointobj,
                &endpoint_readable,
                &endpoint_writable) != PKTCTRL_SUCCESS) {
            warn("Worker %" PRIuFAST32
                " failed to select pktctrl_obj for endpoint,"
                " skipping this endpoint", worker_id);
            goto fail_msgbuf;
        }

        if (pktctrl_process_select(
                &rset, &wset, expprogobj,
                &expprog_readable,
                &expprog_writable) != PKTCTRL_SUCCESS) {
            warn("Worker %" PRIuFAST32
                " failed to select pktctrl_obj for exp program,"
                " skipping this endpoint", worker_id);
            goto fail_msgbuf;
        }

        if (endpoint_readable) {
            assert(endpoint2expprog == NULL);
            rst = pktctrl_read_message(
                endpointobj, &endpoint2expprog);
            if (rst < 0) {
                warn("Worker %" PRIuFAST32
                    " read message from endpoint failed,"
                    " skipping this endpoint", worker_id);
                goto fail_msgbuf;
            } else if (rst == 0) {
                assert(endpoint2expprog == NULL);
                debug("Worker %" PRIuFAST32
                    " endpoint message temporarily unavailable",
                    worker_id);
                // retry later
            } else {
                if (endpoint2expprog == NULL) { // eof
                    info("Worker %" PRIuFAST32
                        " got endpoint EOF, ending this session",
                        worker_id);
                    break;
                }
            }
        }

        if (endpoint_writable) {
            if (expprog2endpoint == NULL) {
                rst = pktctrl_flush_write(endpointobj);
                if (rst < 0) {
                    warn("Worker %" PRIuFAST32
                        " flush message to endpoint failed,"
                        " skipping this endpoint", worker_id);
                    goto fail_msgbuf;
                }
            } else {
                rst = pktctrl_write_message(
                    endpointobj, expprog2endpoint);
                if (rst < 0) {
                    warn("Worker %" PRIuFAST32
                        " write message to endpoint failed,"
                        " skipping this endpoint", worker_id);
                    goto fail_msgbuf;
                } else if (rst == 0) {
                    debug("Worker %" PRIuFAST32
                        " write message to endpoint failed temporarily",
                        worker_id);
                    // retry later
                } else {
                    assert(rst == 1);
                    free(expprog2endpoint);
                    expprog2endpoint = NULL;
                }
            }
        }

        if (expprog_readable) {
            assert(expprog2endpoint == NULL);
            rst = pktctrl_read_message(
                expprogobj, &expprog2endpoint);
            if (rst < 0) {
                warn("Worker %" PRIuFAST32
                    " read message from exp program failed,"
                    " skipping this endpoint", worker_id);
                goto fail_msgbuf;
            } else if (rst == 0) {
                assert(expprog2endpoint == NULL);
                debug("Worker %" PRIuFAST32
                    " exp program message temporarily unavailable",
                    worker_id);
                // retry
            } else {
                if (expprog2endpoint == NULL) { // eof
                    info("Worker %" PRIuFAST32
                        " got exp program EOF, ending this session",
                        worker_id);
                    break;
                }
            }
        }

        if (expprog_writable) {
            if (endpoint2expprog == NULL) {
                rst = pktctrl_flush_write(expprogobj);
                if (rst < 0) {
                    warn("Worker %" PRIuFAST32
                        " flush message to exp program failed,"
                        " skipping this endpoint", worker_id);
                    goto fail_msgbuf;
                }
            } else {
                rst = pktctrl_write_message(
                    expprogobj, endpoint2expprog);
                if (rst < 0) {
                    warn("Worker %" PRIuFAST32
                        " write message to exp program failed,"
                        " skipping this endpoint", worker_id);
                    goto fail_msgbuf;
                } else if (rst == 0) {
                    debug("Worker %" PRIuFAST32
                        " write message to exp program failed temporarily",
                        worker_id);
                    // retry later
                } else {
                    assert(rst == 1);
                    free(endpoint2expprog);
                    endpoint2expprog = NULL;
                }
            }
        }

        if (FD_ISSET(manager_fd, &rset)) { // eof
            info("Worker %" PRIuFAST32
                " detect manager EOF, ending this session",
                worker_id);
            break;
        }
    }

fail_msgbuf:
    free(endpoint2expprog);
    free(expprog2endpoint);
fail_expprogobj:
    pktctrl_close(expprogobj);
    close(fds[0]); // fds[1] already closed
    if (pid != 0) // signal exp prog by closing sock, then wait
        waitpid(pid, NULL, 0); // don't get child exit status
fail_obj:
    pktctrl_close(endpointobj);
    free(client_snikey);
fail:
    close(endpoint_fd);
    close(manager_fd);
    return NULL;
}

int send_worker_msg(int sock, uint8_t msg) {
    ssize_t rst;

    assert(sock >= 0);

    trace("%s(sock:%d,msg:%u)",
        __func__, sock, (unsigned) msg);

    rst = write(sock, &msg, sizeof(msg));
    if (rst < 0) {
        if (errno == EAGAIN || errno == EWOULDBLOCK)
            return 0; // signal should retry
        return -1;
    }

    debug("Wrote worker msg %u into %d",
        (unsigned) msg, sock);

    assert(rst == sizeof(msg));
    return 1; // success
}

int recv_worker_msg(int sock, uint8_t * msg) {
    ssize_t rst;

    assert(sock >= 0 && msg != NULL);

    trace("%s(sock:%d)", __func__, sock);

    rst = read(sock, msg, sizeof(*msg));
    if (rst < 0) {
        if (errno == EAGAIN || errno == EWOULDBLOCK)
            return 0; // signal should retry
        return -1; // return error
    } else if (rst == 0) {
        debug("sock %d EOF", sock);
        *msg = WORKER_EOF;
        return 1;
    }

    debug("Read worker msg %u into %d",
        (unsigned) *msg, sock);

    assert(rst == sizeof(*msg));
    return 1; // success
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int complete_TLS_handshake (
    int manager_fd, int endpoint_fd,
    uint_fast32_t worker_id,
    struct pktlab_privatekey * k,
    struct pktctrl_keycert_store * store,
    struct pktctrl_obj ** endpointobj, char ** client_snikey)
{
    fd_set rset, wset;
    int nsel, rst, ret = -128;
    bool endpoint_readable, endpoint_writable;

    assert(
        manager_fd >= 0 && endpoint_fd >= 0 &&
        k != NULL && store != NULL &&
        endpointobj != NULL && client_snikey != NULL);

    *endpointobj = pktctrl_create_obj();
    assert(*endpointobj != NULL);

    rst = pktctrl_accept(
        endpoint_fd, PKTLAB_CONTROLLER_AUTH_ENDPOINT,
        k, store, *endpointobj, client_snikey, NULL, NULL);
    if (rst != PKTCTRL_SUCCESS && rst != PKTCTRL_IN_PROGRESS) {
        warn("Worker %" PRIuFAST32
            " failed to establish TLS connection with endpoint,"
            " skipping endpoint", worker_id);
        debug("pktctrl_accept failed with %d", rst);
        ret = -1;
        goto fail;
    }

    if (rst != PKTCTRL_SUCCESS) { // in progress
        while (true) {
            FD_ZERO(&rset);
            FD_ZERO(&wset);

            FD_SET(manager_fd, &rset); // for checking eof
            nsel = manager_fd+1;

            if (pktctrl_prepare_select(
                    &nsel, &rset, &wset,
                    true, true, *endpointobj) != PKTCTRL_SUCCESS) {
                warn("Worker %" PRIuFAST32
                    " failed to select pktctrl_obj for endpoint,"
                    " skipping this endpoint", worker_id);
                ret = -2;
                goto fail;
            }

            rst = select(nsel, &rset, &wset, NULL, NULL); // block indefinitely
            if (rst < 0) {
                if (errno == EINTR)
                    continue; // re-select again

                warn("Worker %" PRIuFAST32
                    " select failed (%s), skipping this endpoint",
                    worker_id, strerror(errno));
                ret = -3;
                goto fail;
            }

            if (pktctrl_process_select(
                    &rset, &wset, *endpointobj,
                    &endpoint_readable,
                    &endpoint_writable) != PKTCTRL_SUCCESS) {
                warn("Worker %" PRIuFAST32
                    " failed to select pktctrl_obj for endpoint,"
                    " skipping this endpoint", worker_id);
                ret = -4;
                goto fail;
            }

            if (endpoint_readable || endpoint_writable) {
                rst = pktctrl_accept(
                    endpoint_fd, PKTLAB_CONTROLLER_AUTH_ENDPOINT,
                    NULL, NULL, *endpointobj, client_snikey, NULL, NULL);
                if (rst != PKTCTRL_SUCCESS && rst != PKTCTRL_IN_PROGRESS) {
                    warn("Worker %" PRIuFAST32
                        " failed to establish TLS connection with endpoint,"
                        " skipping endpoint", worker_id);
                    debug("pktctrl_accept failed with %d", rst);
                    ret = -5;
                    goto fail;
                }

                if (rst == PKTCTRL_SUCCESS)
                    break;
            }

            if (FD_ISSET(manager_fd, &rset)) { // eof
                info("Worker %" PRIuFAST32
                    " detect manager EOF, ending this session prematurely",
                    worker_id);
                ret = -6;
                goto fail;
            }
        }
    }

    return 0; // success!

fail:
    pktctrl_close(*endpointobj);
    return ret;
}

int submit_programs(
    int manager_fd, uint_fast32_t worker_id,
    struct pktctrl_keycert_store * store,
    struct pktctrl_obj * endpointobj, const char * client_snikey)
{
    struct prog_submit_obj * submit_obj;
    pktlab_time_t select_until, now;
    fd_set rset, wset;
    int nsel, rst, ret = -128;
    struct timeval tv;

    assert(
        manager_fd >= 0 && store != NULL &&
        endpointobj != NULL && client_snikey != NULL);

    rst = prog_submit_programs_init_store(
        endpointobj, client_snikey, store, &submit_obj);
    if (rst < 0) {
        debug("prog_submit_programs_init_store failed with %d", rst);

        if (rst == -2)
            warn("Worker %" PRIuFAST32
                " bad certificate chain for snikey: %s",
                worker_id, safe_str(client_snikey));

        warn("Worker %" PRIuFAST32
            " failed to init program submission process,"
            " skipping endpoint", worker_id);
        return -1;
    }

    while (true) {
        FD_ZERO(&rset);
        FD_ZERO(&wset);

        select_until = PKTLAB_TIME_MAX;

        FD_SET(manager_fd, &rset); // for checking eof
        nsel = manager_fd+1;

        if (prog_submit_programs_prepare_select(
                &nsel, &rset, &wset, &select_until, submit_obj) < 0) {
            warn("Worker %" PRIuFAST32
                " failed to prepare select for endpoint program submission,"
                " skipping this endpoint", worker_id);
            ret = -2;
            goto fail;
        }

        now = pktlab_time_now();
        pktlab_time_to_timeval(
            (now > select_until) ? 0:select_until-now, &tv);

        // if select time longer than 31 days, cap at 31
        if (tv.tv_sec >= MAX_SELECT_SEC) {
            tv.tv_sec = MAX_SELECT_SEC;
            tv.tv_usec = 0;
        }

        rst = select(nsel, &rset, &wset, NULL, &tv);
        if (rst < 0) {
            if (errno == EINTR)
                continue; // re-select again

            warn("Worker %" PRIuFAST32
                " select failed (%s), skipping this endpoint",
                worker_id, strerror(errno));
            ret = -3;
            goto fail;
        }

        if (FD_ISSET(manager_fd, &rset)) { // eof
            info("Worker %" PRIuFAST32
                " detect manager EOF, ending this session prematurely",
                worker_id);
            ret = -4;
            goto fail;
        }

        rst = prog_submit_programs_process_select(
                &rset, &wset, submit_obj);
        if (rst > 0) {
            break; // done
        } else if (rst == 0) {
            continue; // continue/retry
        } else {
            warn("Worker %" PRIuFAST32
                " endpoint program submission failed,"
                " skipping this endpoint", worker_id);

            ret = -5;
            goto fail;
        }
    }

    ret = 0; // success

fail:
    prog_submit_programs_cleanup(submit_obj);
    return ret;
}

int start_session(
    int manager_fd, uint_fast32_t worker_id,
    struct pktctrl_obj * endpointobj)
{
    struct pktlab_message start_msg = { .type=PKTLAB_START_MESSAGE };
    struct pktlab_message * recv_msg;
    fd_set rset, wset;
    int nsel, rst, ret = -128;
    enum start_session_st st;
    bool endpoint_readable, endpoint_writable;

    assert(manager_fd >= 0 && endpointobj != NULL);

    st = START_SEND_ST;
    recv_msg = NULL;

    while (true) {
        FD_ZERO(&rset);
        FD_ZERO(&wset);

        FD_SET(manager_fd, &rset); // for checking eof
        nsel = manager_fd+1;

        if (pktctrl_prepare_select(
                &nsel, &rset, &wset,
                (st == RECVRSP_ST),
                (st == START_SEND_ST ||
                 st == FLUSH_SEND_ST),
                endpointobj) != PKTCTRL_SUCCESS) {
            warn("Worker %" PRIuFAST32
                " failed to select pktctrl_obj for endpoint,"
                " skipping this endpoint", worker_id);
            ret = -1;
            goto fail;
        }

        rst = select(nsel, &rset, &wset, NULL, NULL); // block indefinitely
        if (rst < 0) {
            if (errno == EINTR)
                continue; // re-select again

            warn("Worker %" PRIuFAST32
                " select failed (%s), skipping this endpoint",
                worker_id, strerror(errno));
            ret = -2;
            goto fail;
        }

        if (FD_ISSET(manager_fd, &rset)) { // eof
            info("Worker %" PRIuFAST32
                " detect manager EOF, ending this session prematurely",
                worker_id);
            ret = -3;
            goto fail;
        }

        if (pktctrl_process_select(
                &rset, &wset, endpointobj,
                &endpoint_readable,
                &endpoint_writable) != PKTCTRL_SUCCESS) {
            warn("Worker %" PRIuFAST32
                " failed to select pktctrl_obj for endpoint,"
                " skipping this endpoint", worker_id);
            ret = -4;
            goto fail;
        }

        if (st == START_SEND_ST && endpoint_writable) {
            rst = pktctrl_write_message(endpointobj, &start_msg);
            if (rst < 0) {
                debug("pktctrl_write_message failed with %d (%s)",
                    rst, strerror(errno));
                warn("Worker %" PRIuFAST32
                    " failed to send start msg to endpoint,"
                    " skipping endpoint", worker_id);
                ret = -5;
                goto fail;
            } else if (rst > 0)
                st = FLUSH_SEND_ST;

            // retry when rst == 0
        } else if (st == FLUSH_SEND_ST && endpoint_writable) {
            if (pktctrl_write_unsent(endpointobj) == 0) {
                st = RECVRSP_ST;
            } else {
                rst = pktctrl_flush_write(endpointobj);
                if (rst < 0) {
                    debug("pktctrl_flush_write failed with %d (%s)",
                        rst, strerror(errno));
                    warn("Worker %" PRIuFAST32
                        " failed to flush start msg to endpoint,"
                        " skipping endpoint", worker_id);
                    ret = -6;
                    goto fail;
                }
            }
        } else if (st == RECVRSP_ST && endpoint_readable) {
            rst = pktctrl_read_message(endpointobj, &recv_msg);
            if (rst < 0) {
                debug("pktctrl_read_message failed with %d (%s)",
                    rst, strerror(errno));
                warn("Worker %" PRIuFAST32
                    " failed to receive start msg response from endpoint,"
                    " skipping endpoint", worker_id);
                recv_msg = NULL;
                ret = -7;
                goto fail;
            } else if (rst > 0) {
                if (recv_msg == NULL) {
                    info("Worker %" PRIuFAST32
                        " got endpoint EOF, ending this session",
                        worker_id);
                    ret = -8;
                    goto fail;
                } else if (recv_msg->type != PKTLAB_RESULT_MESSAGE) {
                    warn("Worker %" PRIuFAST32
                        " unexpected message type (%d) from endpoint,"
                        " skipping endpoint", worker_id, recv_msg->type);
                    ret = -9;
                    goto fail;
                } else if (recv_msg->result.errid != PKTLAB_SUCCESS) {
                    warn("Worker %" PRIuFAST32
                        " result message from endpoint reported error (%" PRIu16
                        "), skipping endpoint", worker_id, recv_msg->result.errid);
                    ret = -10;
                    goto fail;
                } else
                    break; // good rsp
            }

            // retry when rst == 0
        }
    }

    ret = 0; // success

fail:
    free(recv_msg);
    return ret;
}

char ** create_child_environ(uint_fast32_t worker_id, int endpoint_fd) {
    struct sockaddr_storage sockaddr;
    struct sockaddr_in6 * sa6;
    struct sockaddr_in * sa4;
    socklen_t len = sizeof(sockaddr);
    char strbuf[INET6_ADDRSTRLEN];
    const char * addrstr;
    char ** envarr;
    int result;

    // Get endpoint address and store it in EPADDR (as ASCII string).

    result = getpeername(endpoint_fd, (struct sockaddr *)&sockaddr, &len);

    if (result != 0)
        goto fail;

    switch (sockaddr.ss_family) {
    case AF_INET:
        sa4 = (struct sockaddr_in *)&sockaddr;
        addrstr = inet_ntop(AF_INET, &sa4->sin_addr, strbuf, sizeof(strbuf));
        break;
    case AF_INET6:
        sa6 = (struct sockaddr_in6 *)&sockaddr;
        addrstr = inet_ntop(AF_INET6, &sa6->sin6_addr, strbuf, sizeof(strbuf));
        break;
    default:
        goto fail;
    }

    // Allocate and fill in environ array.

    envarr = safe_calloc(2, sizeof(char*));
    safe_asprintf(&envarr[0], "EPADDR=%s", addrstr);

    return envarr;

fail:
    warn("Worker %" PRIuFAST32
       " failed to get endpoint address (%s),"
       " skipping this endpoint", worker_id, strerror(errno));
    exit(EXIT_FAILURE);
}
