// xdescr.c
//

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <pktlab.h>
#include <pktconf.h>

#include "config.h"
#include "console.h"
#include "net.h"
#include "util.h"
#include "xdescr.h"

//
// INTERNAL CONSTANTS
//

#define XDESCR_SNI_KEY_NAME           "sni_key"
#define XDESCR_CONTROLLER_ADDR_NAME   "controller_addr"
#define XDESCR_PUB_END_NAME           "pub_end"
#define XDESCR_HOSTID_NAME            "hostid"
#define XDESCR_EXP_PRIORITY_NAME      "exp_priority"
#define XDESCR_ENDPOINT_COUNTRY_NAME  "endpoint_country"
#define XDESCR_ENDPOINT_LOCATION_NAME "endpoint_location"

#define XDESCR_SNI_KEY_MAXCNT            0
#define XDESCR_CONTROLLER_ADDR_MAXCNT    1
#define XDESCR_PUB_END_MAXCNT            1
#define XDESCR_EXP_PRIORITY_MAXCNT       1
#define XDESCR_HOSTID_MAXCNT             1
#define XDESCR_ENDPOINT_COUNTRY_MAXCNT   (UINT_FAST32_MAX-1)
#define XDESCR_ENDPOINT_LOCATION_MAXCNT  (UINT_FAST32_MAX-1)

//
// INTERNAL TYPE DEFINITIONS
//

struct xdescr_predef_nvp {
    const char * name;
    uint_fast32_t namelen;
    uint_fast32_t max_appearcnt;
    uint_fast32_t appearcnt;
};

struct nvp_ll_node {
    const struct pktlab_xdescr_dattr_pair * nvp;
    struct nvp_ll_node * next;
};

//
// INTERNAL FUNCTION DECLARATIONS
//

static bool is_same_name (
    const char * n1, uint_fast32_t l1,
    const char * n2, uint_fast32_t l2);
static int check_override_nvps (
    const struct pktlab_xdescr_dattr_pair * nvps,
    uint_fast32_t nvpcnt,
    struct xdescr_predef_nvp * pd_nvps,
    uint_fast32_t pd_nvps_len,
    pktlab_time_t * max_pub_end);
static void push_all_ls_nvp (
    const struct pktlab_xdescr_dattr_pair * nvps,
    uint_fast32_t nvpcnt, struct nvp_ll_node ** head);
static void push_nvp_ll (
    struct nvp_ll_node ** head,
    const struct pktlab_xdescr_dattr_pair * nvp);
static void push_all_sepstr_ll_nvp (
    const char * name, uint_fast32_t namelen,
    const struct pktconf_sep_str_ll * ll,
    struct nvp_ll_node ** head,
    struct pktlab_xdescr_dattr_pair * mem);
static int nvp_ll_2_xdescr (
    struct nvp_ll_node * ll, char ** xdescr);

static void free_nvp_ll(struct nvp_ll_node * ll);

//
// EXPORTED FUNCTION DEFINITIONS
//

int create_xdescr (
    const char * pub_controller_addr,
    const char * sni_key,
    const char * hostid,
    bool pub_period_specified,
    pktlab_time_t pub_period,
    bool pub_exp_priority_specified,
    unsigned long long pub_exp_priority,
    const struct pktconf_sep_str_ll * pub_country,
    const struct pktconf_sep_str_ll * pub_location,
    const struct exp_exec_entry * ee_entry, // sni_key here not used
    char ** xdescr, pktlab_time_t * max_pub_end)
{
    struct nvp_ll_node * nvp_ll = NULL;
    struct pktlab_xdescr_dattr_pair snikey_nvp;
    struct pktlab_xdescr_dattr_pair pub_controller_addr_nvp;
    struct pktlab_xdescr_dattr_pair pub_end_nvp;
    struct pktlab_xdescr_dattr_pair exp_priority_nvp;
    struct pktlab_xdescr_dattr_pair hostid_nvp;
    struct pktlab_xdescr_dattr_pair * endpoint_country_nvps = NULL;
    struct pktlab_xdescr_dattr_pair * endpoint_location_nvps = NULL;
    pktlab_time_t override_max_pub_end;
    char pub_end_str[64];
    char exp_priority_str[64];
    size_t totalmem;
    int rst, ret = -128;
    struct xdescr_predef_nvp pd_nvps[] = {
        { XDESCR_SNI_KEY_NAME,
          sizeof(XDESCR_SNI_KEY_NAME)-1,
          XDESCR_SNI_KEY_MAXCNT, 0 },
        { XDESCR_CONTROLLER_ADDR_NAME,
          sizeof(XDESCR_CONTROLLER_ADDR_NAME)-1,
          XDESCR_CONTROLLER_ADDR_MAXCNT, 0 },
        { XDESCR_PUB_END_NAME,
          sizeof(XDESCR_PUB_END_NAME)-1,
          XDESCR_PUB_END_MAXCNT, 0 },
        { XDESCR_EXP_PRIORITY_NAME,
          sizeof(XDESCR_EXP_PRIORITY_NAME)-1,
          XDESCR_EXP_PRIORITY_MAXCNT, 0 },
        { XDESCR_ENDPOINT_COUNTRY_NAME,
          sizeof(XDESCR_ENDPOINT_COUNTRY_NAME)-1,
          XDESCR_ENDPOINT_COUNTRY_MAXCNT, 0 },
        { XDESCR_ENDPOINT_LOCATION_NAME,
          sizeof(XDESCR_ENDPOINT_LOCATION_NAME)-1,
          XDESCR_ENDPOINT_LOCATION_MAXCNT, 0 },
        { XDESCR_HOSTID_NAME,
          sizeof(XDESCR_HOSTID_NAME)-1,
          XDESCR_HOSTID_MAXCNT, 0 },
    };

    assert(
        pub_controller_addr != NULL &&
        sni_key!= NULL && ee_entry != NULL &&
        xdescr != NULL && max_pub_end != NULL);

    // go through list of override nvps first
    rst = check_override_nvps(
        ee_entry->override_xdescr_nvps,
        ee_entry->override_xdescr_nvpcnt,
        pd_nvps, sizeof(pd_nvps)/sizeof(*pd_nvps),
        &override_max_pub_end);
    if (rst < 0) {
        debug("check_override_nvps returned %d", rst);
        ret = -1;
        goto fail;
    }

    // create nvp ll
    push_all_ls_nvp( // push all override nvps
        ee_entry->override_xdescr_nvps,
        ee_entry->override_xdescr_nvpcnt,
        &nvp_ll);

    // sni_key
    snikey_nvp.name = pd_nvps[0].name;
    snikey_nvp.namelen = pd_nvps[0].namelen;
    snikey_nvp.value = sni_key;
    snikey_nvp.valuelen = strlen(sni_key);
    push_nvp_ll(&nvp_ll, &snikey_nvp);

    // for non-overridden ones, add default
    // controller_addr
    if (pd_nvps[1].appearcnt == 0) {
        pub_controller_addr_nvp.name = pd_nvps[1].name;
        pub_controller_addr_nvp.namelen = pd_nvps[1].namelen;
        pub_controller_addr_nvp.value = pub_controller_addr;
        pub_controller_addr_nvp.valuelen = strlen(pub_controller_addr);
        push_nvp_ll(&nvp_ll, &pub_controller_addr_nvp);
    }

    // pub_end
    *max_pub_end = override_max_pub_end;
    if (pd_nvps[2].appearcnt == 0 &&
        pub_period_specified) {
        *max_pub_end = ceil_add_time(pktlab_time_now(), pub_period);
        if (snprintf(
                pub_end_str, sizeof(pub_end_str),
                "%" PRIu64, *max_pub_end) < 0) {
            debug("snprintf returned %s", strerror(errno));
            ret = -2;
            goto fail;
        }

        pub_end_nvp.name = pd_nvps[2].name;
        pub_end_nvp.namelen = pd_nvps[2].namelen;
        pub_end_nvp.value = pub_end_str;
        pub_end_nvp.valuelen = strlen(pub_end_str);
        push_nvp_ll(&nvp_ll, &pub_end_nvp);
    } // else don't put pub_end in xdescr

    // exp_priority
    if (pd_nvps[3].appearcnt == 0 &&
        pub_exp_priority_specified) {
        if (snprintf(
                exp_priority_str, sizeof(exp_priority_str),
                "%llu", pub_exp_priority) < 0) {
            debug("snprintf returned %s", strerror(errno));
            ret = -3;
            goto fail;
        }

        exp_priority_nvp.name = pd_nvps[3].name;
        exp_priority_nvp.namelen = pd_nvps[3].namelen;
        exp_priority_nvp.value = exp_priority_str;
        exp_priority_nvp.valuelen = strlen(exp_priority_str);
        push_nvp_ll(&nvp_ll, &exp_priority_nvp);
    } // else don't put exp_priority in xdescr

    // endpoint_country
    if (pd_nvps[4].appearcnt == 0 &&
        pub_country != NULL &&
        pub_country->len != 0) {
        totalmem = sizeof(*endpoint_country_nvps)*pub_country->len;
        if (totalmem/sizeof(*endpoint_country_nvps) != pub_country->len) {
            debug("Overflow");
            ret = -4;
            goto fail;
        }

        endpoint_country_nvps = safe_malloc(totalmem);
        push_all_sepstr_ll_nvp(
            pd_nvps[4].name, pd_nvps[4].namelen,
            pub_country, &nvp_ll,
            endpoint_country_nvps);
    }

    // endpoint_location
    if (pd_nvps[5].appearcnt == 0 &&
        pub_location != NULL &&
        pub_location->len != 0) {
        totalmem = sizeof(*endpoint_location_nvps)*pub_location->len;
        if (totalmem/sizeof(*endpoint_location_nvps) != pub_location->len) {
            debug("Overflow");
            ret = -5;
            goto fail;
        }

        endpoint_location_nvps = safe_malloc(totalmem);
        push_all_sepstr_ll_nvp(
            pd_nvps[5].name, pd_nvps[5].namelen,
            pub_location, &nvp_ll,
            endpoint_location_nvps);
    }

    // hostid
    if (pd_nvps[6].appearcnt == 0 && hostid != NULL) {
        hostid_nvp.name = pd_nvps[6].name;
        hostid_nvp.namelen = pd_nvps[6].namelen;
        hostid_nvp.value = hostid;
        hostid_nvp.valuelen = strlen(hostid);
        push_nvp_ll(&nvp_ll, &hostid_nvp);
    }

    // convert ll to xdescr
    rst = nvp_ll_2_xdescr(nvp_ll, xdescr);
    if (rst < 0) {
        debug("nvp_ll_2_xdescr returned %d", rst);
        ret = -6;
        goto fail;
    }

    ret = 0;

fail:
    // free ll
    free(endpoint_country_nvps);
    free(endpoint_location_nvps);
    free_nvp_ll(nvp_ll);
    return ret;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

bool is_same_name (
    const char * n1, uint_fast32_t l1,
    const char * n2, uint_fast32_t l2)
{
    if (l1 != l2 || strncmp(n1, n2, l1) != 0)
        return false;

    return true;
}

int check_override_nvps (
    const struct pktlab_xdescr_dattr_pair * nvps,
    uint_fast32_t nvpcnt,
    struct xdescr_predef_nvp * pd_nvps,
    uint_fast32_t pd_nvps_len,
    pktlab_time_t * max_pub_end)
{
    uint_fast32_t i, j;
    char * tmp_field;
    pktlab_time_t tmp_pe, max_pe;

    assert(
        (nvps == NULL) == (nvpcnt == 0) &&
        pd_nvps != NULL && pd_nvps_len == 7 &&
        max_pub_end != NULL);

    max_pe = 0;
    for (i = 0; i < nvpcnt; ++i) {
        for (j = 0; j < pd_nvps_len; ++j) {
            if (!is_same_name(
                    nvps[i].name, nvps[i].namelen,
                    pd_nvps[j].name, pd_nvps[j].namelen))
                continue;

            // check appearance cnt
            ++pd_nvps[j].appearcnt;
            if (pd_nvps[j].appearcnt > pd_nvps[j].max_appearcnt)
                return -1; // bad list of override nvps

            // check field validity
            if (nvps[i].valuelen > SIZE_MAX-1 ||
                copy_str(
                    nvps[i].value, nvps[i].valuelen+1,
                    &tmp_field) < 0)
                return -2;

            switch (j) {
            case 1: // controller_addr
                if (parse_n_getaddrinfo(
                        tmp_field, 1, NULL, NULL) < 0) {
                    free(tmp_field);
                    return -3;
                }
                break;
            case 2: // pub_end
                if (!is_valid_pktlab_time(tmp_field, &tmp_pe)) {
                    free(tmp_field);
                    return -4;
                }

                if (tmp_pe > max_pe)
                    max_pe = tmp_pe;
                break;
            case 3: // exp_priority
                if (!is_valid_ull(tmp_field, NULL)) {
                    free(tmp_field);
                    return -5;
                }
                break;
            case 4: // endpoint_country
                if (!is_valid_iso31661_alpha3(tmp_field)) {
                    free(tmp_field);
                    return -6;
                }
                break;
            case 5: // endpoint_location
                if (!is_valid_maidenhead_w_wc(tmp_field)) {
                    free(tmp_field);
                    return -7;
                }
                break;
            default:
                debug("Unexpected j: %u", (unsigned) j);
                assert(0); // bug
            }

            free(tmp_field);
        }
    }

    *max_pub_end = max_pe;
    return 0;
}

void push_all_ls_nvp (
    const struct pktlab_xdescr_dattr_pair * nvps,
    uint_fast32_t nvpcnt, struct nvp_ll_node ** head)
{
    uint_fast32_t i;

    assert(
        (nvps == NULL) == (nvpcnt == 0) &&
        head != NULL);

    for (i = 0; i < nvpcnt; ++i)
        push_nvp_ll(head, &nvps[i]);

    return;
}

void push_nvp_ll (
    struct nvp_ll_node ** head,
    const struct pktlab_xdescr_dattr_pair * nvp)
{
    struct nvp_ll_node * node;

    assert(nvp != NULL && head != NULL);

    node = safe_malloc(sizeof(*node));
    node->nvp = nvp;
    node->next = *head;
    *head = node;
    return;
}

void push_all_sepstr_ll_nvp (
    const char * name, uint_fast32_t namelen,
    const struct pktconf_sep_str_ll * ll,
    struct nvp_ll_node ** head,
    struct pktlab_xdescr_dattr_pair * mem)
{
    uint_fast32_t i;
    struct pktconf_sep_str_entry * entry;

    if (ll == NULL)
        return;

    assert(
        name != NULL && namelen != 0 &&
        head != NULL && (ll->len == 0 || mem != NULL));

    entry = ll->head;
    for (i = 0; i < ll->len; ++i) {
        mem[i].name = name;
        mem[i].namelen = namelen;
        mem[i].value = entry->entry;
        mem[i].valuelen = entry->entrylen;
        push_nvp_ll(head, &mem[i]);
        entry = entry->next;
    }

    return;
}

int nvp_ll_2_xdescr (
    struct nvp_ll_node * ll, char ** xdescr)
{
    struct nvp_ll_node * node;
    struct pktlab_xdescr_dattr_pair * pairs;
    uint_fast32_t i, paircnt, used;
    size_t nvplen, xdescrlen, pairsmem;

    assert(xdescr != NULL);

    if (ll == NULL) {
        *xdescr = NULL;
        return 0;
    }

    // count paircnt and get xdescr len
    xdescrlen = paircnt = 0;
    for (node = ll; node != NULL; node = node->next) {
        if (paircnt == UINT_FAST32_MAX)
            return -1; // overflow
        ++paircnt;

        if (node->nvp->namelen > SIZE_MAX-2 ||
            node->nvp->valuelen > SIZE_MAX-2-node->nvp->namelen)
            return -1; // overflow
        nvplen = 2+node->nvp->namelen+node->nvp->valuelen;

        if (xdescrlen > SIZE_MAX-nvplen)
            return -1; //overflow

        xdescrlen += nvplen;
    }

    if (xdescrlen == SIZE_MAX || xdescrlen > UINT_FAST32_MAX-1)
        return -1;

    // alloc and set up
    pairsmem = sizeof(*pairs)*paircnt;
    if (pairsmem/sizeof(*pairs) != paircnt)
        return -1; // overflow

    pairs = safe_malloc(pairsmem);
    *xdescr = safe_malloc(xdescrlen+1);

    node = ll;
    for (i = 0; i < paircnt; ++i) {
        memcpy(&pairs[i], node->nvp, sizeof(pairs[i]));
        node = node->next;
    }

    // encode_xdescr
    if (pktlab_encode_xdescr(
            pairs, paircnt,
            *xdescr, xdescrlen+1, &used) < 0) {
        debug("pktlab_encode_xdescr failed");
        free(pairs);
        free(*xdescr);
        return -2;
    }
    assert(xdescrlen == used);

    (*xdescr)[xdescrlen] = '\0';
    free(pairs);
    return 0;
}

void free_nvp_ll(struct nvp_ll_node * ll) {
    struct nvp_ll_node * node;

    while (ll != NULL) {
        node = ll->next;
        free(ll);
        ll = node;
    }

    return;
}