// broker.c
//

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <limits.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#include <pktconf.h>
#include <pktctrl.h>
#include <pktlab.h>

#include "broker.h"
#include "cert.h"
#include "channel.h"
#include "config.h"
#include "console.h"
#include "key.h"
#include "net.h"
#include "util.h"

//
// INTERNAL CONSTANTS
//

#define RECONNECT_DELAY 5 // 5 sec
#define ADDRESS_TTL 3600  // 1 hour

enum broker_state {
    ST_UNKNOWN,
    ST_INIT,
    ST_SLEEP,
    ST_HANDSHAKE,
    ST_SUBSCRIBING,
    ST_RUNNING,
    ST_ERROR
};

//
// INTERNAL GLOBAL VARIABLES
//

static enum broker_state _state;
static const char *_hoststr = NULL;
static const char *_portstr = NULL;
static struct addrinfo *_addrinfo = NULL;
static pktlab_time_t _addrinfo_expires;
static pktlab_time_t _time_now;
static char _specstr[48];
static int _sockfd = -1;
static pktlab_time_t _sleep_expires = 0;

static struct pktctrl_obj *_obj = NULL;

//
// INTERNAL FUNCTION DECLARATIONS
//

static void parse_host(const char *broker_spec);
static void refresh_addrinfo(void);
static void socket_connect(void);
static inline const char *state_name(enum broker_state state);
static void socket_set_nonblock_keepalive(void);
static char *channel_printable(const uint8_t *chan);
static void install_sighandler(void);
static void sighandler(int sig);
static void install_exithandler(void);
static void exithandler(void);

static int parse_exppriv_chain(struct pktlab_privatekey *pk,
                               struct pktlab_certificate ***cert_ls,
                               uint_fast32_t *certnum);
static int encode_channels(const struct pktconf_sep_str_ll *ll,
                           const struct private_channel *privchs,
                           uint8_t **chlistptr, uint_fast32_t *chlistlen);

static int record_new_subable_privch(struct pktlab_certificate *cert,
                                     struct private_channel **head);
static void free_privchs(struct private_channel *head);

//
// EXPORTED FUNCTION DEFINITIONS
//

void broker_main(int mgrfd, int depfd, const char *broker_spec, const char *snikey,
                 struct pktlab_privatekey *pk,
                 struct pktlab_publickey ** rooteopls, size_t rooteopnum,
                 struct pktctrl_keycert_store *store) {

    trace("%s(depfd:%d,broker_spec:\"%s\",snikey:\"%s\")", __func__, depfd,
          broker_spec, snikey);

    int rst;
    struct pktlab_certificate **peer_cert_ls;
    struct pktlab_certificate **subpriv_cert_ls = NULL;
    uint_fast32_t subpriv_cert_num;
    uint_fast32_t peer_cert_num;
    int nfds;
    bool readable, writable;
    fd_set wset, rset;
    struct timeval tv;
    struct pktlab_message *msg_send, *msg_recv;

    _state = ST_INIT;

    install_exithandler();
    install_sighandler();

    parse_host(broker_spec);
    if (parse_exppriv_chain(pk, &subpriv_cert_ls, &subpriv_cert_num) < 0) {
        debug("Failed to load sub privilege chain");
    }

    while (true) {

        _time_now = pktlab_time_now();

        switch (_state) {
        case ST_SLEEP:
            if (_sleep_expires < _time_now)
                _state = ST_HANDSHAKE;
            else
                usleep((_sleep_expires - _time_now) / 1000);
            break;

        case ST_HANDSHAKE:
            // make tcp socket connection with the broker
            if (_sockfd == -1) {
                socket_connect();
            }

            if (_state != ST_HANDSHAKE)
                break;

            if (_obj == NULL) {
                _obj = pktctrl_create_obj();
                if (_obj == NULL)
                    fatal("Broker failed object creation");
            }

            rst = pktctrl_connect(_sockfd, PKTLAB_AUTH_AGENT, snikey, pk, store,
                                  _obj, &peer_cert_ls, &peer_cert_num);

            if (rst == PKTCTRL_SUCCESS) {
                trace("broker handshake finished");
                _state = ST_SUBSCRIBING;
            } else if (rst != PKTCTRL_IN_PROGRESS) {
                fatal("Broker failed pktctrl_connect with %d", rst);
            }
            break;
        case ST_SUBSCRIBING:
            trace("broker subscribing");
            struct pktlab_message xsub;
            xsub.type = PKTLAB_XSUB_MESSAGE;
            struct pktlab_message xcert;
            xcert.type = PKTLAB_XCERT_MESSAGE;
            struct pktlab_certificate *cert;
            int sendidx = 0;
            int recvidx = 0;
            uint32_t certnum = _config.subpriv_chain_path_ll->len;
            uint32_t msgnum = certnum + 1;
            struct pktconf_sep_str_entry *cur_certpath_entry =
                _config.subpriv_chain_path_ll->head;
            void *der;
            uint_fast32_t derlen;
            struct private_channel *channelids = NULL;

            // info("Broker trying to subscribe to: %s",
            //      channel_printable(parse_channel()));
            while (recvidx < msgnum) {
                nfds = 0;
                FD_ZERO(&rset);
                FD_ZERO(&wset);
                tv.tv_sec = 10;
                tv.tv_usec = 0;
                // want_read if haven't received all results for subscribing
                // want_write if haven't write all xsub messages or unsent
                // message in obj
                if (pktctrl_prepare_select(
                        &nfds, &rset, &wset, (recvidx < msgnum),
                        (sendidx < msgnum) || pktctrl_write_unsent(_obj),
                        _obj) != PKTCTRL_SUCCESS) {
                    fatal("Broker prepare select failed");
                }

                // Poll our manager fd. This lets us know if the manager is
                // still running.

                FD_SET(mgrfd, &rset);
                nfds = MAX(nfds, mgrfd + 1);

                rst = select(nfds, &rset, &wset, NULL, &tv);
                if (rst <= 0) {
                    if (rst < 0) {
                        warn("Broker select failed (%s), exiting", strerror(errno));
                        exit(EXIT_FAILURE);
                    } else
                        continue;
                }

                // Manager will not write, so if mgrfd is readable, it's an EOF
                // and the manager process terminated. Terminate broker also.

                if (FD_ISSET(mgrfd, &rset)) {
                    info("Broker detect manager EOF (readable), exiting");
                    exit(EXIT_SUCCESS);
                }

                if (pktctrl_process_select(&rset, &wset, _obj, &readable,
                                           &writable) != PKTCTRL_SUCCESS) {
                    fatal("Broker process select failed");
                }

                if (writable) {
                    trace("object writable");
                    if (sendidx < certnum) {
                        if (load_cert(cur_certpath_entry->entry, &cert) < 0) {
                            fatal("Failed to load subpriv cert");
                        }

                        if (_config.auto_extract_channel &&
                            record_new_subable_privch(cert, &channelids) < 0) {
                            fatal(
                                "Failed to extract private channel from cert");
                        }

                        if (pktlab_get_cert_der(cert, &der, &derlen) < 0) {
                            fatal("pktlab_get_cert_der failed");
                        }

                        xcert.type = PKTLAB_XCERT_MESSAGE;
                        xcert.xcert.derptr = der;
                        xcert.xcert.derlen = derlen;
                        if (sendidx+1 != certnum) // last cert not freed
                            pktlab_cleanup_cert(cert);

                        msg_send = &xcert;
                        rst = pktctrl_write_message(_obj, msg_send);
                        free(der);
                        if (rst < 0) {
                            fatal("Broker pktctrl_write_message failed with "
                                  "errno (%d)",
                                  errno);
                        } else if (rst == 0) {
                            continue;
                        }

                        cur_certpath_entry = cur_certpath_entry->next;
                        sendidx++;
                        msg_send = NULL;
                    } else if (sendidx == certnum) {
                        uint8_t *chlistptr;
                        uint_fast32_t chlistlen, keyindx;
                        uint8_t raweopkey[PKTLAB_PUBLICKEY_LEN] = {};

                        if (certnum != 0) { // need to provide raw eop key
                            if (find_signing_key(
                                    rooteopls, rooteopnum,
                                    cert, &keyindx) < 0)
                                fatal("Cannot find eop key for subpriv chain root");
                            pktlab_cleanup_cert(cert);

                            assert(keyindx < rooteopnum);
                            if (pktlab_get_publickey_bytes(rooteopls[keyindx],
                                                           raweopkey,
                                                           sizeof(raweopkey)) < 0)
                                fatal("Cannot get public key bytes");
                        }

                        rst = encode_channels(_config.broker_channel_ll,
                                              channelids, &chlistptr, &chlistlen);
                        if (rst < 0) {
                            debug("encode_channels failed with %d", rst);
                            fatal("Failed to encode channel list");
                        }

                        free_privchs(channelids);

                        xsub.type = PKTLAB_XSUB_MESSAGE;
                        xsub.xsub.chlistptr = chlistptr;
                        xsub.xsub.chlistlen = chlistlen;
                        xsub.xsub.eopkey = raweopkey;
                        xsub.xsub.dattrptr = _config.endpoint_attr;
                        xsub.xsub.dattrlen = (_config.endpoint_attr == NULL) ? \
                            0:strlen(_config.endpoint_attr);
                        msg_send = &xsub;
                        rst = pktctrl_write_message(_obj, msg_send);
                        free(chlistptr);

                        if (rst < 0) {
                            fatal("Broker pktctrl_write_message failed with "
                                  "errno (%d: %s)", errno, strerror(errno));
                        } else if (rst == 0)
                            continue;

                        sendidx++;
                    }

                    if (pktctrl_write_unsent(_obj) != 0) {
                        if (pktctrl_flush_write(_obj) < 0)
                            fatal("Broker pktctrl_flush_write failed with "
                                  "errno (%d: %s)", errno, strerror(errno));
                    }
                }

                if (readable) {
                    trace("object readable");
                    rst = pktctrl_read_message(_obj, &msg_recv);
                    if (rst < 0) { // bad stuff happened
                        fatal("Broker pktctrl_read_message failed with "
                              "errno (%d)",
                              errno);
                        break;
                    } else if (rst == 0) { // would block
                        continue;
                    }

                    if (msg_recv == NULL) {
                        fatal("Broker pktctrl_read_message EOF prematurely");
                    }

                    if (msg_recv->type != PKTLAB_RESULT_MESSAGE) {
                        fatal("Broker unexpected msg type during subscribing");
                    }
                    if (msg_recv->result.errid != PKTLAB_SUCCESS) {
                        info("Broker failed subscription");
                    }
                    info("Broker subscription finished");
                    recvidx++;
                    free(msg_recv);
                }
            }
            _state = ST_RUNNING;
            break;
        case ST_RUNNING:;
            const char *xdescr = NULL;
            size_t xdescr_len = 0;
            while (true) {
                nfds = 0;
                FD_ZERO(&rset);
                FD_ZERO(&wset);
                tv.tv_sec = 10;
                tv.tv_usec = 0;

                if (pktctrl_prepare_select(&nfds, &rset, &wset, true, false,
                                           _obj) != PKTCTRL_SUCCESS) {
                    fatal("Broker prepare select failed");
                }

                // Poll our manager fd. This lets us know if the manager is
                // still running.

                FD_SET(mgrfd, &rset);
                nfds = MAX(nfds, mgrfd+1);

                if ((rst = select(nfds, &rset, &wset, NULL, &tv)) <= 0) {
                    if (rst == -1) {
                        warn("Broker select failed (%s), exiting", strerror(errno));
                        exit(EXIT_FAILURE);
                    } else {
                        trace("Broker loop timedout", errno);
                    }
                    continue;
                }

                if (FD_ISSET(mgrfd, &rset)) {
                    info("Broker detect manager EOF (readable), exiting");
                    exit(EXIT_SUCCESS);
                }

                if (pktctrl_process_select(&rset, &wset, _obj, &readable,
                                           &writable) != PKTCTRL_SUCCESS) {
                    fatal("Broker process select failed");
                }

                // recv xnotify from the broker
                if (readable) {
                    rst = pktctrl_read_message(_obj, &msg_recv);
                    if (rst < 0) { // bad stuff happened
                        fatal("Broker pktctrl_read_message failed with "
                              "errno (%d)",
                              errno);
                        break;
                    } else if (rst == 0) { // would block
                        continue;
                    }

                    // broker EOF
                    if (msg_recv == NULL) {
                        warn("broker EOF");
                        if (_obj != NULL)
                            pktctrl_close(_obj);
                        _sleep_expires =
                            _time_now + pktlab_time_sec(RECONNECT_DELAY);
                        _state = ST_SLEEP;
                    }

                    if (msg_recv->type != PKTLAB_XNOTIFY_MESSAGE) {
                        warn("Broker unexpected msg type");
                        free(msg_recv);
                        continue;
                    }

                    info("Broker recv xnotify from chan: %s",
                         channel_printable(
                             (const uint8_t *)msg_recv->xnotify.channelptr));

                    xdescr = msg_recv->xnotify.xdescrptr;
                    xdescr_len = msg_recv->xnotify.xdescrlen;

                    sendxdescr(depfd, xdescr, xdescr_len);
                    free(msg_recv);
                }
            }

            break;
        case ST_ERROR:
            trace("broker state error");
            if (_obj != NULL)
                pktctrl_close(_obj);
        default:
            trace("broker mainloop exited");
            break;
        }
    }
}

void parse_host(const char *broker_spec) {
    const char *sptr;
    char *hoststr;
    size_t hostlen;

    trace("broker:%s(state:%s,broker_spec:%s)", __func__, state_name(_state),
          broker_spec);
    assert(_state == ST_INIT);

    sptr = strchr(broker_spec, ':');

    if (sptr != NULL) {
        hostlen = sptr - broker_spec;
        hoststr = safe_malloc(hostlen + 1);
        memcpy(hoststr, broker_spec, hostlen);
        hoststr[hostlen] = '\0';
        _hoststr = hoststr;
        _portstr = sptr + 1;
    } else {
        _hoststr = broker_spec;
        _portstr = PKTLAB_DEFAULT_SUB_PORT_STR;
    }
    _state = ST_HANDSHAKE;
}

void socket_connect(void) {
    char hoststr[42], portstr[6];
    struct addrinfo *next;
    int result;

    trace("broker:%s(state:%s)", __func__, state_name(_state));
    assert(_state == ST_HANDSHAKE);

    if (_addrinfo == NULL || _addrinfo_expires < _time_now)
        refresh_addrinfo();

    while (_addrinfo != NULL) {
        result = getnameinfo(_addrinfo->ai_addr, _addrinfo->ai_addrlen, hoststr,
                             sizeof(hoststr), portstr, sizeof(portstr),
                             NI_NUMERICHOST | NI_NUMERICSERV);

        if (result != 0)
            fatal("%s", gai_strerror(errno));

        snprintf(_specstr, sizeof(_specstr), "%s,%s", hoststr, portstr);
        _sockfd = socket(_addrinfo->ai_family, _addrinfo->ai_socktype,
                         _addrinfo->ai_protocol);

        if (_sockfd < 0)
            fatal("%s", strerror(errno));
        result = connect(_sockfd, _addrinfo->ai_addr, _addrinfo->ai_addrlen);

        if (result >= 0) {
            info("%s: Broker socket connected", _specstr);
            socket_set_nonblock_keepalive();
            debug("Connected to %s", _specstr);
            return;
        }

        switch (errno) {
        case ETIMEDOUT:
        case ECONNREFUSED:
        case ENETUNREACH:
        case EHOSTUNREACH:
        case ENETDOWN:
            warn("%s: %s", _specstr, strerror(errno));
            break; // and continue loop
        default:
            fatal("%s: %s", _specstr, strerror(errno));
            break;
        }

        close(_sockfd);
        _sockfd = -1;

        next = _addrinfo->ai_next;
        _addrinfo->ai_next = NULL;
        freeaddrinfo(_addrinfo);
        _addrinfo = next;
    }

    warn("Could not connect to %s,%s; sleeping.", _hoststr, _portstr);
    _sleep_expires = _time_now + pktlab_time_sec(RECONNECT_DELAY);
    _state = ST_SLEEP;
}

void socket_set_nonblock_keepalive(void) {
    int result;
    int optval = 1;

    // can cause cleanup issue, so fatal
    if (set_nonblock(_sockfd) < 0)
        fatal("Set broker socket nonblocking failed");

    // keepalive failing is mostly fine as broker does this as well
    result = setsockopt(_sockfd, SOL_SOCKET, SO_KEEPALIVE,
                        (void *)&optval, (socklen_t)sizeof(optval));
    if (result < 0)
        warn("Set broker socket keepalive failed");
}

void refresh_addrinfo(void) {
    struct addrinfo hints;
    int result;

    trace("broker:%s(state:%s)", __func__, state_name(_state));

    // Free current sockaddr_list (if not empty).

    if (_addrinfo != NULL) {
        freeaddrinfo(_addrinfo);
        _addrinfo = NULL;
    }

    // Resolve the host name using getaddrinfo.

    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_ADDRCONFIG | AI_NUMERICSERV;
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    result = getaddrinfo(_hoststr, _portstr, &hints, &_addrinfo);

    if (result < 0 || _addrinfo == NULL)
        fatal("%s,%s: %s", _hoststr, _portstr, gai_strerror(result));

    _addrinfo_expires = _time_now + pktlab_time_sec(ADDRESS_TTL);
}

inline const char *state_name(enum broker_state state) {
    static const char *const names[] = {
        [ST_UNKNOWN] = "UNKNOWN", [ST_INIT] = "INIT",
        [ST_SLEEP] = "SLEEP",     [ST_HANDSHAKE] = "HANDSHAKE",
        [ST_RUNNING] = "RUNNING", [ST_SUBSCRIBING] = "SUBSCRIBING",
        [ST_ERROR] = "ERROR"};

    const char *name = NULL;

    if (state < sizeof(names) / sizeof(names[0]))
        name = names[state];

    return (name != NULL) ? name : names[ST_UNKNOWN];
}

int sendxdescr(int depfd, const char *xdescr, size_t len) {
    ssize_t size, off;
    char term = '\0';

    trace("%s(depfd:%d,len:%d)", __func__, depfd, len);
    off = 0;
    while (off < len) {
        size = write(depfd, xdescr + off, len - off);
        if (size < 0)
            fatal("%s write failure", __func__);
        off += size;
    }
    while ((size = write(depfd, &term, 1)) != 1) {
        if (size < 0)
            fatal("%s write failure", __func__);
    }
    return len;
}

int recvxdescr(int brokerfd, char *buf, size_t bufsz) {
    int len;
    ssize_t rdlen;
    char discard_c;

    assert(buf != NULL && bufsz != 0);

    trace("%s(brokerfd:%d,bufsz:%zu)", __func__, brokerfd, bufsz);

    len = 0;
    while (len < bufsz) {
        rdlen = read(brokerfd, buf+len, 1);
        if (rdlen == 0) {
            warn("Broker process EOF detected");
            return 0;
        }

        if (rdlen < 0) {
            warn("Read xdescr from broker process failed: %s", strerror(errno));
            return -1;
        }

        assert(rdlen == 1);
        len += rdlen;

        if (buf[len-1] == '\0') {
            debug("Received xdescr from broker process: %s", buf);
            return len;
        }

        if (len == INT_MAX) {
            warn("xdescr too long (> INT_MAX long), ignoring");
            break;
        }
    }

    if (len >= bufsz && len != INT_MAX)
        warn("xdescr too long (> supplied buffer len), ignoring");

    do {
        rdlen = read(brokerfd, &discard_c, sizeof(discard_c));
        if (rdlen == 0) {
            warn("Broker process EOF detected");
            return 0;
        }

        if (rdlen < 0) {
            warn("Read xdescr from broker process failed: %s", strerror(errno));
            return -1;
        }
    } while (discard_c != '\0');

    return -2;
}

char *channel_printable(const uint8_t *chan) {
    static char buf[256];
    snprintf(buf, sizeof(buf), "%02x %02x %02x %02x.... %02x %02x %02x %02x",
             chan[0], chan[1], chan[2], chan[3], chan[PKTLAB_CHANNELID_LEN_MAX - 4],
             chan[PKTLAB_CHANNELID_LEN_MAX - 3], chan[PKTLAB_CHANNELID_LEN_MAX - 2],
             chan[PKTLAB_CHANNELID_LEN_MAX - 1]);
    return buf;
}

void install_sighandler(void) {
    struct sigaction sa;
    int result;

    memset(&sa, 0, sizeof(sa));
    sa.sa_handler = &sighandler;
    sa.sa_flags = SA_RESETHAND;

    result = sigemptyset(&sa.sa_mask);
    perror_exit_if(result != 0, __func__);

    result = sigaddset(&sa.sa_mask, SIGTERM);
    perror_exit_if(result != 0, __func__);

    result = sigaddset(&sa.sa_mask, SIGINT);
    perror_exit_if(result != 0, __func__);

    result = sigaddset(&sa.sa_mask, SIGHUP);
    perror_exit_if(result != 0, __func__);

    result = sigaddset(&sa.sa_mask, SIGQUIT);
    perror_exit_if(result != 0, __func__);

    result = sigaction(SIGTERM, &sa, NULL);
    perror_exit_if(result != 0, __func__);

    result = sigaction(SIGINT, &sa, NULL);
    perror_exit_if(result != 0, __func__);

    result = sigaction(SIGHUP, &sa, NULL);
    perror_exit_if(result != 0, __func__);

    result = sigaction(SIGQUIT, &sa, NULL);
    perror_exit_if(result != 0, __func__);
}

void sighandler(int sig) {
    trace("%s(%s)", __func__, strsignal(sig));
    exit(EXIT_SUCCESS);
}

void install_exithandler(void) {
    int rst;

    rst = atexit(exithandler);
    if (rst != 0)
        warn("Broker failed to install exit handler");
}

void exithandler(void) {
    trace("%s()", __func__);

    if (_obj != NULL) {
        pktctrl_close(_obj);
        _obj = NULL;
    }

    // This is the important part: close the broker socket to generate FIN so
    // that the broker closes the connection.

    if (0 <= _sockfd) {
        close(_sockfd);
        _sockfd = -1;
    }
}

int parse_exppriv_chain(struct pktlab_privatekey *pk,
                        struct pktlab_certificate ***cert_ls,
                        uint_fast32_t *certnum) {
    assert(cert_ls && certnum);
    if (ll_load_cert_chain(_config.subpriv_chain_path_ll, cert_ls, certnum) <
        0) {
        debug("Failed to load sub privilege chain");
    }

    if (reorder_certs(pk, *cert_ls, *certnum) < 0) {
        debug("Failed to reorder sub privilege chain");
        free_certs(*cert_ls, *certnum);
        return -1;
    }
    return 0;
}

// this function encodes the linked list of channels into a
// DNS like byte array. Every channel start with a byte representing
// the channel length followed by that many bytes for the channel.
// The list terminates at a zero for channel length
int encode_channels(const struct pktconf_sep_str_ll *ll,
                    const struct private_channel *privchs, uint8_t **chlistptr,
                    uint_fast32_t *chlistlen) {
    struct pktlab_channel *channels;
    int ret = -128, rst;

    assert(chlistptr != NULL && chlistlen != NULL);

    trace("%s(...)", __func__);

    rst = get_pktlab_channels(ll, privchs, &channels);
    if (rst < 0) {
        debug("get_pktlab_channels failed with %d", rst);
        return -1;
    }

    rst = pktlab_encode_chlist(channels, (void **)chlistptr, chlistlen);
    if (rst < 0) {
        debug("pktlab_encode_chlist failed with %d", rst);
        ret = -2;
        goto fail_channels;
    }

    ret = 0;

fail_channels:
    free_pktlab_channels(channels);
    return ret;
}

int record_new_subable_privch(struct pktlab_certificate *cert,
                              struct private_channel **head) {
    struct pktlab_cert_info *info;
    struct private_channel *ch;

    assert(cert != NULL && head != NULL);

    if (pktlab_decode_certificate(cert, NULL, &info) != PKTLAB_DECODE_SUCCESS)
        return -1;

    ch = safe_malloc(sizeof(*ch));
    ch->next = NULL;
    memcpy(ch->channelid, info->authority_key_id, sizeof(ch->channelid));

    // append to ll
    if (*head != NULL)
        ch->next = *head;

    *head = ch;

    free(info);
    return 0;
}

void free_privchs(struct private_channel *head) {
    struct private_channel *node;

    while (head != NULL) {
        node = head->next;
        free(head);
        head = node;
    }

    return;
}
