// control.c
//

// todo: check state assertion

#include "control.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <netdb.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <unistd.h>

#include <pktctrl.h>
#include <pktconf.h>

#include "comm.h"
#include "config.h"
#include "console.h"
#include "filter.h"
#include "net.h"
#include "network.h"
#include "network_util.h"
#include "nqueue.h"
#include "tls_util.h"
#include "util.h"
#include "vmem.h"
#include "worker.h"
#include "rlim.h"
#include "limitations.h"

//
// COMPILE-TIME PARAMETERS
//

#define CONNECT_TIMEOUT    3    // 3 sec
#define RECONNECT_DELAY    1    // 1 sec
#define ADDRESS_TTL        3600 // 1 hour
#define STATE_SET_UP_TIME  10   // 10 sec

#define MAX_PROG_NUM       100
#define MAX_MSGQ_LEN       100

//
// INTERNAL CONSTANTS
//

enum state {
    ST_UNKNOWN,
    ST_INIT,
    ST_SLEEP,
    ST_SETUP,
    ST_IDLE
};

enum cdata_hint {
    HINT_UNK = -1,
    HINT_FILTER,
    HINT_MONITOR_SEND,
    HINT_MONITOR_RECV,
};

//
// INTERNAL MACROS
//

//
// INTERNAL TYPE DEFINITIONS
//

struct queued_msg {
    struct pktlab_message * msg;
    send_msg_cb_t flushed_cb;
    void * aux;
};

struct recorded_progams {
    uint8_t * digests;
    uint_fast32_t digestcnt;
    uint8_t seen_digests[MAX_PROG_NUM];
    struct raw_prog progll; // sentinel node
};

//
// INTERNAL GLOBAL VARIABLES
//

static enum state _state;

static const char *_hoststr = NULL;
static const char *_portstr = NULL;
static const char *_snikey = NULL;
static struct addrinfo *_addrinfo = NULL;
static pktlab_time_t _addrinfo_expires;
static char _specstr[48];
static int _mgrfd = -1;

static int _sockfd = -1;
static struct pktctrl_obj *_ctrlobj = NULL;
static struct pktctrl_keycert_store *_store = NULL;
static struct pktlab_privatekey *_privk = NULL;
static uint32_t _exp_priority = 0;
static struct nqueue _msgq; // for request reply and notification

static struct recorded_progams _ftprogs = \
    {.progll = RAW_PROG_INITIALIZER};
static struct recorded_progams _smtprogs = \
    {.progll = RAW_PROG_INITIALIZER};
static struct recorded_progams _rmtprogs = \
    {.progll = RAW_PROG_INITIALIZER}; // todo: unused

static union {
    struct {
        pktlab_time_t expires; // next target try again time
        pktlab_time_t timeout; // try again give up time
    } sleep;

    struct {
        pktlab_time_t setup_expires;
    } setup;
} _st;

//
// INTERNAL FUNCTION DECLARATIONS
//

static int try_connect_until (
    struct addrinfo * ai, pktlab_time_t timeout_time);
static int try_pktctrl_connect_until (
    pktlab_time_t timeout_time,
    struct pktlab_certificate *** peer_cert_ls,
    uint_fast32_t * peer_cert_num);

static void socket_connect(void);
static void socket_connected(void);
static void socket_tryread(void);
static void socket_trywrite(void);
static void socket_eof(void);
static void socket_read_error(int err);
static void socket_write_error(int err);

static void handle_message(struct pktlab_message *msg);
static void handle_setup_message(struct pktlab_message *msg);
static void control_handle_start(struct pktlab_message *msg);
static void control_handle_cdata(struct pktlab_message *msg);
static void control_handle_yield(struct pktlab_message *msg);

static bool is_same_hint (
    const void * hint0, size_t hint0len,
    const void * hint1, size_t hint1len);
static enum cdata_hint cdata_hint_match (
    uint8_t hintlen, const void * hintptr);

static void setup_timeout(void);

static struct queued_msg * create_queued_msg (
    struct pktlab_message * msg, send_msg_cb_t cb, void * aux);
static bool call_flushed_cb_filt(void * elt, uintptr_t aux);

static void flush_msgq(void);
static void flush_unsent(void);

// bypasses _msgq, use caution
static void force_flush(const struct pktlab_message *msg);
static void force_flush_n_reset_errid(enum pktlab_status errid);

static void refresh_addrinfo(void);

static int record_program (
    struct recorded_progams * rprogs, const uint8_t * digest,
    const void * decoded_prog, uint_fast32_t decoded_proglen);
static void recorded_program_cleanup (
    struct recorded_progams * rprogs);

static inline const char *state_name(enum state state);

//
// EXPORTED FUNCTION DEFINITIONS
//

void control_initialize(void) {
    assert(_state < ST_INIT);

    trace("%s()", __func__);

    rlim_init(_config.rlim_db_path);

    nqinit(&_msgq);
    _state = ST_INIT;
    return;
}

// The control_start function resolves the host name of the control server
// and connects to each address in turn. If all fail, control_start causes
// the program to exit. In the future, we may want to handle this by
// sleeping for some time and then trying again.

void control_start(const char *ctlspec, const char *snikey,
                   int mgrfd, uint_fast32_t sleep_sec, uint_fast32_t sleep_timeout,
                   struct pktctrl_keycert_store *store,
                   struct pktlab_privatekey *privk,
                   uint32_t exp_priority) {
    const char *sptr;
    char *hoststr;
    size_t hostlen;

    assert(_state == ST_INIT && ctlspec != NULL);

    trace("%s(state:%s,ctlspec:%s)", __func__, state_name(_state), ctlspec);

    sptr = strrchr(ctlspec, ':');

    if (sptr != NULL) {
        hostlen = sptr - ctlspec;
        hoststr = safe_malloc(hostlen + 1);
        memcpy(hoststr, ctlspec, hostlen);
        hoststr[hostlen] = '\0';
        _hoststr = hoststr;
        _portstr = sptr + 1;
    } else {
        _hoststr = ctlspec;
        _portstr = PKTLAB_DEFAULT_EXP_PORT_STR;
    }

    _snikey = snikey;
    _store = store;
    _privk = privk;
    _exp_priority = exp_priority;

    rlim_vacuum(pktlab_time_now());

    _st.sleep.expires = pktlab_time_now() + pktlab_time_sec(sleep_sec);
    _st.sleep.timeout = pktlab_time_now() + pktlab_time_sec(sleep_timeout);
    _state = ST_SLEEP;
    _mgrfd = mgrfd;
    return;
}

void control_reset(void) {
    assert(_state >= ST_INIT);

    trace("%s(state:%s)", __func__, state_name(_state));

    if (_state == ST_INIT)
        return; // already reset

    // clean up control connection
    if (_ctrlobj != NULL)
        pktctrl_close(_ctrlobj);
    _ctrlobj = NULL;

    if (_sockfd >= 0)
        close(_sockfd);
    _sockfd = -1;

    info("%s: Controller connection closed", _specstr);

    // clean up msg queue content
    // only reset network after finishing _msgq cleanup

    nqfilter(&_msgq, call_flushed_cb_filt, (uintptr_t)NULL);
    nqvacuum(&_msgq);

    network_reset();

    // clean up recored programs and digests
    recorded_program_cleanup(&_ftprogs);
    recorded_program_cleanup(&_smtprogs);
    recorded_program_cleanup(&_rmtprogs);

    // spec
    memset(_specstr, 0, sizeof(_specstr));

    // clean up controller addrinfo
    if (_addrinfo != NULL)
        freeaddrinfo(_addrinfo);
    _addrinfo = NULL;

    _addrinfo_expires = 0;

    if (_state >= ST_SETUP) {
        rlim_commit(_time_now);
        limitations_reset();
    }

    _state = ST_INIT;
    return;
}

void control_suspend(void) {
    static const struct pktlab_message suspd_msg = \
        {.type = PKTLAB_SUSPD_MESSAGE};

    assert(_state >= ST_SETUP);

    trace("%s(state:%s)", __func__, state_name(_state));

    force_flush(&suspd_msg); // may fail, but carry on
    control_reset();
    return;
}

bool control_running(void) {
    trace("%s(state:%s): %s", __func__, state_name(_state),
          (_state >= ST_SLEEP) ? "True" : "False");
    return (_state >= ST_SLEEP);
}

// Control communication.

void control_push_msgq_cb(struct pktlab_message *msg,
                          send_msg_cb_t cb, void * aux) {
    assert((_state == ST_SETUP || _state == ST_IDLE) && msg != NULL);

    trace("%s(state:%s,msg{type:%d})",
          __func__, state_name(_state), msg->type);

    nqput(&_msgq, create_queued_msg(msg, cb, aux));
    return;
}

void control_push_msgq(struct pktlab_message *msg) {
    assert((_state == ST_SETUP || _state == ST_IDLE) && msg != NULL);

    trace("%s(state:%s,msg{type:%d})",
          __func__, state_name(_state), msg->type);

    nqput(&_msgq, create_queued_msg(msg, NULL, NULL));
    return;
}

void control_report(enum pktlab_status errid) {
    assert(_state == ST_SETUP || _state == ST_IDLE);

    trace("%s(state:%s,errid:%d)", __func__, state_name(_state), errid);

    control_push_msgq(pktlab_create_result_message(errid));
    return;
}

// Select handlers

int control_prepare_select(fd_set *rset, fd_set *wset,
                           pktlab_time_t *select_until) {
    int nfds, rst;
    bool want_read, want_write;

    assert(_state >= ST_INIT);

    trace("%s(state:%s,len(_msgq):%zu)", __func__,
          state_name(_state), nqlen(&_msgq));

    nfds = 0;
    switch (_state) {
    case ST_INIT:
        return 0;

    case ST_SLEEP:
        *select_until = MIN(*select_until, _st.sleep.expires);
        *select_until = MIN(*select_until, _st.sleep.timeout);
        return 0;

    case ST_SETUP:
        if (_st.setup.setup_expires < *select_until)
            *select_until = _st.setup.setup_expires;
        // normal msg read and write but with timeout
        // fall-through

    case ST_IDLE:
        // well, time is up, time to die ^_^
        // immediately go to process select
        if (!check_exp_endtime(_time_now)) {
            *select_until = 0;
            return 0;
        }

        want_read = want_write = false;

        // Make sure we don't have too much unsent outgoing messages
        // before accepting another one.
        if (nqlen(&_msgq) < MAX_MSGQ_LEN) {
            want_read = true;

            if (pktctrl_read_message(_ctrlobj, NULL))
                *select_until = _time_now;
        }

        if (nqlen(&_msgq) != 0 ||
            pktctrl_write_unsent(_ctrlobj) != 0)
            want_write = true;

        if (want_read || want_write) {
            rst = pktctrl_prepare_select(&nfds, rset, wset, want_read,
                                         want_write, _ctrlobj);
            assert(rst == PKTCTRL_SUCCESS);
        }

        return nfds;
    default:
        // not reached
        return 0;
    }
}

int control_process_select(int nsel, const fd_set *rset, const fd_set *wset) {
    int rst;
    bool readable, writable;

    assert(_state >= ST_INIT);

    trace("%s(state:%s,nsel:%d)", __func__, state_name(_state), nsel);

    switch (_state) {
    case ST_INIT:
        return 0;

    case ST_SLEEP:
        if (_st.sleep.timeout < _time_now) {
            control_reset(); // retry time up, give up
            return 0;
        }

        if (_st.sleep.expires < _time_now)
            socket_connect(); // may die
        return 0;

    case ST_SETUP:
        if (_st.setup.setup_expires < _time_now) {
            setup_timeout();
            return 0;
        }
        // fall-through for normal read write otherwise

    case ST_IDLE:
        // no more exp time left, reset immediately
        if (!check_exp_endtime(_time_now)) {
            warn("%s: exp time exhaustion detected, disconnecting", _specstr);
            control_reset();
            return 0;
        }

        rst = pktctrl_process_select(rset, wset, _ctrlobj, &readable, &writable);
        assert(rst == PKTCTRL_SUCCESS);

        if (readable && nqlen(&_msgq) < MAX_MSGQ_LEN) {
            socket_tryread();
            if (!control_running()) // socket_tryread may reset
                break;
        }

        if (writable &&
            (nqlen(&_msgq) != 0 ||
             pktctrl_write_unsent(_ctrlobj) != 0)) {
            socket_trywrite();
            if (!control_running()) // socket_trywrite may reset
                break;
        }
        // no break;

    default:
        break;
    }

    if (_sockfd >= 0)
        return (FD_ISSET(_sockfd, rset) != 0)+\
               (FD_ISSET(_sockfd, wset) != 0);

    return 0;
}

// Program access handlers

struct raw_prog * control_fetch_cert_filters(void) {
    assert(_state >= ST_IDLE);
    trace("%s(state:%s)", __func__, state_name(_state));
    debug("_ftprogs.digestcnt:%" PRIuFAST32, _ftprogs.digestcnt);
    return _ftprogs.progll.next;
}

struct raw_prog * control_fetch_cert_send_monitors(void) {
    assert(_state >= ST_IDLE);
    trace("%s(state:%s)", __func__, state_name(_state));
    debug("_smtprogs.digestcnt:%" PRIuFAST32, _smtprogs.digestcnt);
    return _smtprogs.progll.next;
}

struct raw_prog * control_fetch_cert_recv_monitors(void) {
    assert(_state >= ST_IDLE);
    trace("%s(state:%s)", __func__, state_name(_state));
    debug("_rmtprogs.digestcnt:%" PRIuFAST32, _rmtprogs.digestcnt);
    return _rmtprogs.progll.next;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int try_connect_until (
    struct addrinfo * ai, pktlab_time_t timeout_time)
{
    int sock, result;
    struct timeval tv;
    fd_set wfds;
    pktlab_time_t now;
    bool should_try;

    assert(ai != NULL);

    sock = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
    if (sock < 0)
        fatal("Failed to get socket with %s", strerror(errno));

    if (set_nonblock(sock) < 0)
        fatal("Failed to set socket non-blocking");

    should_try = true;

    while (true) {
        if (should_try) {
            result = connect(sock, _addrinfo->ai_addr, _addrinfo->ai_addrlen);
            if (result < 0) {
                // We treat timeout, connection refused, or network unreachable
                // as non-fatal error and ask for another addrinfo. For inprogress/
                // already we see if there is time left and select again
                // (if timeout it is also soft error). All other errors are fatal.

                switch (errno) {
                case EINPROGRESS:
                case EALREADY:
                    break; // need to retry
                case ETIMEDOUT:
                case ECONNREFUSED:
                case ENETUNREACH:
                case EHOSTUNREACH:
                case ENETDOWN:
                    warn("%s: %s", _specstr, strerror(errno));
                    close(sock);
                    return -1; // soft connect error
                default:
                    // strong connect error just dies
                    fatal("%s: %s", _specstr, strerror(errno));
                }
            } else
                return sock; // success!
        }

        // cannot connect yet, we select and retry
        FD_ZERO(&wfds);
        FD_SET(sock, &wfds);
        should_try = false;

        now = pktlab_time_now();
        if (timeout_time <= now) {
            warn("%s: TCP connect timed out (non-OS)", _specstr);
            close(sock);
            return -2; // connect timeout is soft error too
        }

        pktlab_time_to_timeval(timeout_time-now, &tv);

        result = select(sock+1, NULL, &wfds, NULL, &tv);
        if (result < 0) {
            if (errno == EINTR)
                continue;

            fatal("select failed with %s", strerror(errno));
        }

        if (FD_ISSET(sock, &wfds))
            should_try = true;
    }
}

void socket_connect(void) {
    char hoststr[42], portstr[6];
    struct addrinfo *next;
    int result;

    assert(_state == ST_SLEEP);

    trace("control:%s(state:%s)", __func__, state_name(_state));

    // STEP 1. If our address list is more than ADDR_FRESHNESS seconds old,
    // resolve the host name again.

    if (_addrinfo == NULL || _addrinfo_expires < _time_now)
        refresh_addrinfo();

    while (_addrinfo != NULL) {
        // STEP 2: Get string representation of current address and store it
        // in _specstr for error reporting.
        result = getnameinfo(_addrinfo->ai_addr, _addrinfo->ai_addrlen, hoststr,
                             sizeof(hoststr), portstr, sizeof(portstr),
                             NI_NUMERICHOST | NI_NUMERICSERV);
        if (result != 0)
            fatal("%s", gai_strerror(errno));

        snprintf(_specstr, sizeof(_specstr), "%s:%s", hoststr, portstr);

        // STEP 3: Connect to address.
        result = try_connect_until(
            _addrinfo, pktlab_time_now()+pktlab_time_sec(CONNECT_TIMEOUT));
        if (result >= 0) {
            info("%s: Controller socket connected", _specstr);

            _sockfd = result;
            socket_connected();
            return;
        }

        // we reach here if it is soft error
        // try another addrinfo
        next = _addrinfo->ai_next;
        _addrinfo->ai_next = NULL;
        freeaddrinfo(_addrinfo);
        _addrinfo = next;
    }

    // We get here if we go through all the addresses without making a
    // successful connection.

    warn("Could not connect to %s:%s; sleeping.", _hoststr, _portstr);

    _st.sleep.expires = _time_now + pktlab_time_sec(RECONNECT_DELAY);
    _state = ST_SLEEP;
    return;
}

int try_pktctrl_connect_until (
    pktlab_time_t timeout_time,
    struct pktlab_certificate *** peer_cert_ls,
    uint_fast32_t * peer_cert_num)
{
    int result, nfds;
    struct timeval tv;
    pktlab_time_t now;
    bool readable, writable;
    fd_set rset, wset;

    assert(peer_cert_num != NULL && peer_cert_ls != NULL);

    _ctrlobj = pktctrl_create_obj();
    assert(_ctrlobj != NULL);

    readable = writable = true;

    while (true) {
        if (readable || writable) {
            result = pktctrl_connect(_sockfd, PKTLAB_ENDPOINT_AUTH_CONTROLLER,
                                     _snikey, _privk, _store, _ctrlobj,
                                     peer_cert_ls, peer_cert_num);
            if (result == PKTCTRL_SUCCESS) {
                return 0;
            } else if (result != PKTCTRL_IN_PROGRESS)
                fatal("%s: TLS handshake failed: %d", _specstr, result);
        }

        // TLS handshake ongoing
        FD_ZERO(&rset);
        FD_ZERO(&wset);
        nfds = 0;
        readable = writable = false;

        if (pktctrl_prepare_select(
                &nfds, &rset, &wset,
                true, true, _ctrlobj) != PKTCTRL_SUCCESS)
            fatal("Failed to prepare for select");

        now = pktlab_time_now();
        if (timeout_time <= now)
            fatal("%s: TLS handshake timed out (non-OS)", _specstr);

        pktlab_time_to_timeval(timeout_time-now, &tv);

        result = select(nfds, &rset, &wset, NULL, &tv);
        if (result < 0) {
            if (errno == EINTR)
                continue;

            fatal("select failed with %s", strerror(errno));
        }

        if (pktctrl_process_select(
                &rset, &wset, _ctrlobj,
                &readable, &writable) != PKTCTRL_SUCCESS)
            fatal("Failed to process select result");
    }
}

void socket_connected(void) {
    int result;
    struct pktlab_certificate **peer_cert_ls;
    uint_fast32_t peer_cert_num;
    enum pktlab_parse_cert_chain_rst cert_chain_rst;
    // uint8_t *mtdigests = NULL;
    // uint_fast32_t mtcnt = 0;
    uint8_t subable_channel[PKTLAB_CHANNELID_LEN_MAX];

    assert(_state == ST_SLEEP);

    trace("control:%s(state:%s)", __func__, state_name(_state));

    // Try setting up TLS
    _ctrlobj = pktctrl_create_obj();
    assert(_ctrlobj != NULL);

    result = try_pktctrl_connect_until(
        pktlab_time_now()+pktlab_time_sec(CONNECT_TIMEOUT),
        &peer_cert_ls, &peer_cert_num);
    assert(
        result == 0 &&
        (peer_cert_ls != NULL) == (peer_cert_num != 0));

    info("%s: Controller TLS handshake completed", _specstr);

    if (peer_cert_ls != NULL && peer_cert_num != 0) {
        rlim_begin(peer_cert_ls, peer_cert_num, _time_now);
        limitations_init();

        if (!check_exp_priority(_exp_priority)) {
            warn("%s: Controller published invalid experiment priority: %"
                PRIu32 ", disconnecting", _specstr, _exp_priority);
            control_reset(); // no rlim commit here is fine
            return;
        }

        cert_chain_rst = \
            pktlab_parse_controller_exp_cert_chain(
                peer_cert_ls, peer_cert_num, NULL, NULL,
                &_ftprogs.digests, &_ftprogs.digestcnt,
                &_smtprogs.digests, &_smtprogs.digestcnt,
                NULL, NULL, subable_channel);
        if (cert_chain_rst != PKTLAB_PARSE_CHAIN_SUCCESS)
            fatal("%s: Controller certificate chain parsing failed with %d",
                  _specstr, cert_chain_rst);

        if ((_ftprogs.digestcnt > \
             UINT_FAST32_MAX-_smtprogs.digestcnt) ||
            (_rmtprogs.digestcnt > \
             UINT_FAST32_MAX-_smtprogs.digestcnt-_ftprogs.digestcnt) ||
            (_rmtprogs.digestcnt+_smtprogs.digestcnt+_ftprogs.digestcnt > \
             MAX_PROG_NUM))
            fatal("%s: Too many program hashes (smtcnt:%" PRIuFAST32
                  ",rmtcnt:%" PRIuFAST32 ",ftcnt:%" PRIuFAST32 ")",
                  _specstr, _smtprogs.digestcnt,
                  _rmtprogs.digestcnt, _ftprogs.digestcnt);
    }

    // always go into setup state even when no digests
    _st.setup.setup_expires = _time_now + pktlab_time_sec(STATE_SET_UP_TIME);
    _state = ST_SETUP;

    // set socket tcp keepalive and non-blocking
    if (set_keepalive(_sockfd))
        warn("%s: set TCP keepalive failed: %s", _specstr, strerror(errno));

    // done
    debug("TLS Connected to %s", _specstr);
    return;
}

void socket_tryread(void) {
    struct pktlab_message *msg;
    int result;

    assert(_state >= ST_SETUP && nqlen(&_msgq) < MAX_MSGQ_LEN);

    trace("control:%s(state:%s,len(_msgq):%zu)",
          __func__, state_name(_state), nqlen(&_msgq));

    result = pktctrl_read_message(_ctrlobj, &msg);
    if (result < 0) {
        socket_read_error(errno);
        return;
    } else if (result == 0)
        return; // retry

    // result > 0
    if (msg == NULL) { // msg NULL signify EOF
        socket_eof();
        return;
    }

    if (_state == ST_SETUP) {
        handle_setup_message(msg);
    } else
        handle_message(msg);

    return;
}

void socket_trywrite(void) {
    assert(_state >= ST_SETUP);

    trace("control:%s(state:%s,len(_msgq):%zu)",
          __func__, state_name(_state), nqlen(&_msgq));

    if (nqlen(&_msgq) != 0) {
        flush_msgq();
    } else if (pktctrl_write_unsent(_ctrlobj) != 0)
        flush_unsent();

    return;
}

void socket_eof(void) {
    info("%s: Controller socket read EOF, resetting worker", _specstr);
    control_reset();
    return;
}

void socket_read_error(int err) {
    switch (err) {
    default:
        warn("%s: Control obj read failed: %s, resetting worker",
            _specstr, strerror(err));
        break;
    }

    control_reset();
    return;
}

void socket_write_error(int err) {
    switch (err) {
    default:
        warn("%s: Control obj write failed: %s, resetting worker",
            _specstr, strerror(err));
        break;
    }

    control_reset();
    return;
}

bool is_same_hint (
    const void * hint0, size_t hint0len,
    const void * hint1, size_t hint1len)
{
    return (hint0len == hint1len &&
            memcmp(hint0, hint1, hint0len) == 0);
}

enum cdata_hint cdata_hint_match (
    uint8_t hintlen, const void * hintptr)
{
    static const char match_filter[] = PKTLAB_CDATA_HINT_FILTER;
    static const char match_monitor_send[] = PKTLAB_CDATA_HINT_SEND_MONITOR;
    static const char match_monitor_recv[] = PKTLAB_CDATA_HINT_RECV_MONITOR;

    assert(hintptr != NULL);

    if (is_same_hint(
            match_filter, sizeof(match_filter)-1,
            hintptr, hintlen)) {
        return HINT_FILTER;
    } else if (is_same_hint(
                    match_monitor_send, sizeof(match_monitor_send)-1,
                    hintptr, hintlen)) {
        return HINT_MONITOR_SEND;
    } else if (is_same_hint(
                    match_monitor_recv, sizeof(match_monitor_recv)-1,
                    hintptr, hintlen)) {
        return HINT_MONITOR_RECV;
    }

    return HINT_UNK;
}

void handle_setup_message(struct pktlab_message *msg) {
    assert(_state == ST_SETUP && msg != NULL);

    trace("control:%s(state:%s,msg:{type:%d})",
          __func__, state_name(_state), msg->type);

    switch (msg->type) {
    case PKTLAB_CDATA_MESSAGE:
        control_handle_cdata(msg);
        return;
    case PKTLAB_START_MESSAGE:
        control_handle_start(msg);
        return;
    case PKTLAB_UNDEF_MESSAGE:
        if (msg->raw.malformed) {
            warn("%s: Controller supplied malformed msg"
                " during setup, resetting worker", _specstr);
            force_flush_n_reset_errid(PKTLAB_ECTLBADMSG);
            break;
        }
        // fall-through, for unknown msg
    default:
        warn("%s: Controller supplied unsupported msg"
            " (type:%d) during setup, rejecting",
            _specstr, msg->type);
        control_report(PKTLAB_ECTLNOTSUP);
        break;
    }

    free(msg);
    return;
}

void handle_message(struct pktlab_message *msg) {
    assert(_state == ST_IDLE && msg != NULL);

    trace("control:%s(state:%s,msg:{type:%d})",
          __func__, state_name(_state), msg->type);

    switch (msg->type) {
    case PKTLAB_MREAD_MESSAGE:
        vmem_handle_mread(msg);
        return;
    case PKTLAB_MWRITE_MESSAGE:
        vmem_handle_mwrite(msg);
        return;
    case PKTLAB_NOPEN_MESSAGE:
        network_handle_nopen(msg);
        return;
    case PKTLAB_NCLOSE_MESSAGE:
        network_handle_nclose(msg);
        return;
    case PKTLAB_NSEND_MESSAGE:
        network_handle_nsend(msg);
        return;
    case PKTLAB_NCAP_MESSAGE:
        network_handle_ncap(msg);
        return;
    case PKTLAB_NCTL_MESSAGE:
        network_handle_nctl(msg);
        return;
    case PKTLAB_YIELD_MESSAGE:
        control_handle_yield(msg);
        return;
    case PKTLAB_UNDEF_MESSAGE:
        if (msg->raw.malformed) {
            warn("%s: Controller supplied malformed msg,"
                " resetting worker", _specstr);
            force_flush_n_reset_errid(PKTLAB_ECTLBADMSG);
            break;
        }
        // fall-through, for unknown msg
    default:
        warn("%s: Controller supplied unsupported msg"
            " (type:%d), rejecting", _specstr, msg->type);
        control_report(PKTLAB_ECTLNOTSUP);
        break;
    }

    free(msg);
    return;
}

void flush_msgq(void) {
    int result;
    struct queued_msg * msg;

    assert(_state >= ST_SETUP);

    trace("control:%s(state:%s,len(_msgq):%zu)",
          __func__, state_name(_state), nqlen(&_msgq));

    while (nqlen(&_msgq) > 0) {
        msg = nqpeek(&_msgq);
        assert(msg != NULL && msg->msg != NULL);

        result = pktctrl_write_message(_ctrlobj, msg->msg);
        if (result < 0) {
            socket_write_error(errno);
            return;
        } else if (result == 0)
            return; // retry next time

        nqdrop(&_msgq);
        call_flushed_cb_filt(msg, (uintptr_t)NULL);
    }

    return;
}

void flush_unsent(void) {
    int result;

    assert(_state >= ST_SETUP);

    trace("control:%s(state:%s)", __func__, state_name(_state));

    result = pktctrl_flush_write(_ctrlobj);
    if (result < 0)
        socket_write_error(errno);

    return;
}

void refresh_addrinfo(void) {
    struct addrinfo hints;
    int result;

    assert(_state == ST_SLEEP);

    trace("control:%s(state:%s)", __func__, state_name(_state));

    // Free current sockaddr_list (if not empty).

    if (_addrinfo != NULL) {
        freeaddrinfo(_addrinfo);
        _addrinfo = NULL;
    }

    // Resolve the host name using getaddrinfo.

    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_ADDRCONFIG | AI_NUMERICSERV;
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    result = getaddrinfo(_hoststr, _portstr, &hints, &_addrinfo);

    if (result < 0 || _addrinfo == NULL)
        fatal("%s:%s:%s", _hoststr, _portstr, gai_strerror(result));

    _addrinfo_expires = _time_now + pktlab_time_sec(ADDRESS_TTL);
    return;
}

inline const char *state_name(enum state state) {
    static const char *const names[] = {
        [ST_UNKNOWN] = "UNKNOWN", [ST_INIT] = "INIT",
        [ST_SLEEP] = "SLEEP",     [ST_SETUP] = "SETUP", [ST_IDLE] = "IDLE"};
    const char *name = NULL;

    if (state < sizeof(names) / sizeof(names[0]))
        name = names[state];

    return (name != NULL) ? name : names[ST_UNKNOWN];
}

void force_flush(const struct pktlab_message *msg) {
    int result, err;

    assert(_state >= ST_SETUP && msg != NULL);

    trace("control:%s(state:%s,msg:{type:%d})",
          __func__, state_name(_state), msg->type);

    result = pktctrl_write_message(_ctrlobj, msg);

    if (result < 0) {
        err = errno;
        warn("%s: Force flushing msg with type %d failed,"
            " resetting worker", _specstr, msg->type);
        socket_write_error(err);
    } else if (result == 0) {
        warn("%s: Force flushing msg with type %d"
            " temporary not possible", _specstr, msg->type);
    } else if (pktctrl_write_unsent(_ctrlobj) != 0) // result > 0
        warn("%s: Force flushing msg with type %d"
            " only done partially", _specstr, msg->type);

    return;
}

void force_flush_n_reset_errid(enum pktlab_status errid) {
    struct pktlab_message msg = {
        .type=PKTLAB_RESULT_MESSAGE, .result.errid=errid};

    assert(_state >= ST_SETUP);

    trace("control:%s(state:%s,msg:{type:%d})",
          __func__, state_name(_state), errid);

    force_flush(&msg);
    control_reset(); // force_flush may already reset, but redoing is fine
    return;
}

void control_handle_cdata(struct pktlab_message *msg) {
    enum pktlab_status errid;
    int hint_match, rst;
    enum pktlab_prog_type decoded_prog_type;
    void * decoded_prog;
    uint_fast32_t decoded_proglen;
    uint8_t digest[PKTLAB_SHA256_DIGEST_LEN];
    struct recorded_progams * dst_progs;
    const char * dst_progs_name;

    // TODO: Support actual monitors

    assert(
        _state == ST_SETUP && msg != NULL &&
        msg->type == PKTLAB_CDATA_MESSAGE);

    trace("%s(msg:{hint:%.*s})", __func__,
          msg->cdata.hintlen, (char *) msg->cdata.hintptr);

    errid = PKTLAB_EUNKFAULT;

    hint_match = cdata_hint_match(msg->cdata.hintlen, msg->cdata.hintptr);
    if (hint_match != HINT_FILTER &&
        hint_match != HINT_MONITOR_SEND) { // match filters and send monitors for now
        warn("%s: Controller cdata contains unsupported hint: %.*s,"
            " rejecting", _specstr, msg->cdata.hintlen,
            (const char *) msg->cdata.hintptr);
        errid = PKTLAB_ECRTBADHINT;
        goto fail;
    }

    rst = pktlab_get_program_digest(
        msg->cdata.dataptr, msg->cdata.datalen, digest);
    if (rst < 0) {
        debug("pktlab_get_program_digest failed with %d", rst);
        warn("%s: Cannot compute digest for cdata program (len:%" PRIu32 "),"
            " rejecting", _specstr, msg->cdata.datalen);
        errid = PKTLAB_ECRTBADHDAT;
        goto fail;
    }

    rst = pktlab_decode_program(
        msg->cdata.dataptr, msg->cdata.datalen,
        &decoded_prog_type, &decoded_prog, &decoded_proglen);
    if (rst < 0) {
        debug("pktlab_decode_program failed with %d", rst);
        warn("%s: cdata program cannot be decoded, rejecting", _specstr);
        errid = PKTLAB_ECRTBADHDAT;
        goto fail;
    }

    if (decoded_prog_type != PKTLAB_EBPF ||
        decoded_proglen > SIZE_MAX) {
        warn("%s: Bad cdata program type/len"
            " (type:%d,len:%" PRIuFAST32 "), rejecting",
            _specstr, decoded_prog_type, decoded_proglen);
        errid = PKTLAB_ECRTBADHDAT;
        goto fail_decoded_prog;
    }

    // only check actual validity when on platform supporting ebpf
    rst = filter_verify_ebpf_filter(decoded_prog, decoded_proglen);
    if (rst < 0 && rst != -128) { // -128 means unsupported
        warn("%s: Bad cdata program (len:%" PRIuFAST32 "),"
            " rejecting", _specstr, msg->cdata.datalen);
        errid = PKTLAB_ECRTBADHDAT;
        goto fail_decoded_prog;
    }

    dst_progs = &_ftprogs;
    dst_progs_name = "recv filter digests";
    if (hint_match == HINT_MONITOR_SEND) {
        dst_progs = &_smtprogs;
        dst_progs_name = "send filter digests";
    }

    rst = record_program(
        dst_progs, digest,
        decoded_prog, decoded_proglen);
    if (rst < 0) {
        if (rst == -1) {
            warn("%s: Cannot find matching digest"
                " (first 4 bytes: %02x%02x%02x%02x)"
                " for cdata program in %s, rejecting",
                _specstr, digest[0], digest[1], digest[2],
                digest[3], dst_progs_name);
        } else if (rst == -2)
            warn("%s: cdata program in %s loading failed,"
                " rejecting", _specstr, dst_progs_name);

        errid = PKTLAB_ECRTBADHDAT;
        goto fail_decoded_prog;
    }

    errid = PKTLAB_SUCCESS;

fail_decoded_prog:
    free(decoded_prog);
fail:
    assert(errid != PKTLAB_EUNKFAULT);

    control_report(errid);
    free(msg);
    return;
}

void control_handle_start(struct pktlab_message *msg) {
    uint_fast32_t i;
    enum pktlab_status errid;

    assert(
        _state == ST_SETUP && msg != NULL &&
        msg->type == PKTLAB_START_MESSAGE);

    trace("%s(...)", __func__);

    errid = PKTLAB_ECRTNOHDAT;

    for (i = 0; i < _ftprogs.digestcnt; ++i)
        if (_ftprogs.seen_digests[i] != 1)
            goto fail;

    for (i = 0; i < _smtprogs.digestcnt; ++i)
        if (_smtprogs.seen_digests[i] != 1)
            goto fail;

    for (i = 0; i < _rmtprogs.digestcnt; ++i)
        if (_rmtprogs.seen_digests[i] != 1)
            goto fail;

    errid = PKTLAB_SUCCESS;

fail:

    if (errid == PKTLAB_SUCCESS) {
        control_report(errid);
        _state = ST_IDLE;
    } else { // errid == PKTLAB_ECRTNOHDAT
        warn("%s: Controller didn't supply all required programs,"
            " resetting worker", _specstr);
        force_flush_n_reset_errid(errid);
    }

    free(msg);
    return;
}

void control_handle_yield(struct pktlab_message *msg) {
    assert(
        _state == ST_IDLE && msg != NULL &&
        msg->type == PKTLAB_YIELD_MESSAGE);

    trace("%s(msg:{rhint:%" PRIu64 "})", __func__, msg->yield.rhint);

    // todo: fix racing condition
    // notify yield to deputy
    send_ctrlmsg(_mgrfd, CTRLMSG_TYPE_YIELD, msg->yield.rhint);
    // reply result to controller and reset
    force_flush_n_reset_errid(PKTLAB_SUCCESS);
    free(msg);
    return;
}

void setup_timeout(void) {
    assert(_state == ST_SETUP);

    trace("%s(state:%s)", __func__, state_name(_state));

    // Note we directly disconnects when a setup timeout happens
    warn("%s: Controller setup timeout, resetting worker", _specstr);
    control_reset();
    return;
}

struct queued_msg * create_queued_msg (
    struct pktlab_message * msg, send_msg_cb_t cb, void * aux)
{
    struct queued_msg * ret;

    assert(msg != NULL);

    ret = safe_malloc(sizeof(*ret));
    ret->msg = msg;
    ret->flushed_cb = cb;
    ret->aux = aux;
    return ret;
}

// for _msgq msg cleanup
bool call_flushed_cb_filt(void * elt, uintptr_t aux) {
    struct queued_msg * m;

    assert(elt != NULL);

    (void)aux;

    m = elt;
    if (m->flushed_cb != NULL)
        m->flushed_cb(m->aux);

    free(m->msg);
    free(m);

    // always queue into nqueue free queue for nqvacuum
    return false;
}

int record_program (
    struct recorded_progams * rprogs, const uint8_t * digest,
    const void * decoded_prog, uint_fast32_t decoded_proglen)
{
    struct raw_prog * prog;
    static const struct raw_prog prog_init = RAW_PROG_INITIALIZER;
    bool found;
    uint_fast32_t i;
    uint8_t * ptr;

    assert(
        _state == ST_SETUP &&
        rprogs != NULL && digest != NULL &&
        decoded_proglen <= SIZE_MAX);

    found = false;
    for (ptr = rprogs->digests, i = 0;
         i < rprogs->digestcnt;
         ptr += PKTLAB_SHA256_DIGEST_LEN, ++i) {
        if (memcmp(digest, ptr, PKTLAB_SHA256_DIGEST_LEN) != 0)
            continue;

        found = true;

        if (rprogs->seen_digests[i] != 0)
            continue; // already set before, skip

        prog = safe_malloc(sizeof(*prog));
        *prog = prog_init;
        prog->is_ebpf = true;
        prog->proglen = decoded_proglen;

        prog->progptr = NULL;
        if (decoded_proglen != 0) {
            prog->progptr = safe_malloc(decoded_proglen);
            memcpy(prog->progptr, decoded_prog, decoded_proglen);
        }

        // note we do not load program here
        // prog_fd remains un-init (-1)

        // actual append to progll
        prog->next = rprogs->progll.next;
        rprogs->progll.next = prog;
        rprogs->seen_digests[i] = 1;
    }

    if (!found)
        return -1;

    return 0;
}

void recorded_program_cleanup (
    struct recorded_progams * rprogs)
{
    struct raw_prog * node;

    assert(rprogs != NULL);

    while (rprogs->progll.next != NULL) {
        node = rprogs->progll.next->next;

        if (rprogs->progll.next->prog_fd >= 0)
            close(rprogs->progll.next->prog_fd);

        free(rprogs->progll.next->progptr);
        free(rprogs->progll.next);

        rprogs->progll.next = node;
    }

    free(rprogs->digests);
    memset(rprogs, 0, sizeof(*rprogs));
    return;
}