// filter.c
// The eBPF filter module.
//

#include <assert.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdlib.h>

#include <ubpf.h>
#include <pktlab.h>

#include "console.h"
#include "raw.h"
#include "util.h"

#include "filter.h"

//
// INTERNAL MACROS
//

#define MAX_RETVAL  0x20000
#define MIN_RETVAL -0x20000

//
// INTERNAL TYPE DEFINITIONS
//

enum filter_state {
    ST_UNKNOWN, // == 0
    ST_INIT
};

//
// INTERNAL GLOBAL VARIABLES
//

static enum filter_state _st;
static struct ubpf_vm * _vm;

//
// INTERNAL FUNCTION DECLARATIONS
//

static int vm_load_ebpf_filter (
    const void * filter, uint_fast32_t filterlen);
static uint64_t safe_retval(uint64_t retval);

//
// EXPORTED FUNCTION DEFINITIONS
//

int filter_init(void) {
    trace("%s(_st:%d)", __func__, _st);

    if (_st != ST_UNKNOWN)
        return -1;

    assert(_vm == NULL);

    _vm = ubpf_create();
    if (_vm == NULL)
        return -2;

    _st = ST_INIT;
    return 0;
}

void filter_cleanup(void) {
    trace("%s(_st:%d)", __func__, _st);

    if (_st != ST_INIT)
        return;

    assert(_vm != NULL);
    ubpf_destroy(_vm);
    _vm = NULL;
    _st = ST_UNKNOWN;
    return;
}

int filter_verify_ebpf_filter (
    const void * filter, uint_fast32_t filterlen)
{
    int rst;

    trace("%s(_st:%d,filterlen:%" PRIuFAST32 ")",
        __func__, _st, filterlen);

    if (_st == ST_UNKNOWN)
        return -1;

    // verify by loading and seeing the result

    rst = vm_load_ebpf_filter(filter, filterlen);
    if (rst < 0) {
        debug("vm_load_ebpf_filter returned %" PRIuFAST32, filterlen);
        return -2;
    }

    return 0;
}

int filter_run_ebpf_filter (
    const void * filter, uint_fast32_t filterlen,
    void * ctx, uint_fast32_t ctxlen,
    void * data, uint_fast32_t datalen,
    uint64_t * acceptlen)
{
    int rst;
    uint64_t retval;

    trace("%s(_st:%d,filterlen:%" PRIuFAST32
        ",ctxlen:%" PRIuFAST32 ",datalen:%" PRIuFAST32 ")",
        __func__, _st, filterlen, ctxlen, datalen);

    if (_st == ST_UNKNOWN ||
        ctxlen > SIZE_MAX || (ctx == NULL && ctxlen != 0) ||
        datalen > SIZE_MAX || (data == NULL && datalen != 0) ||
        acceptlen == NULL)
        return -1;

    rst = vm_load_ebpf_filter(filter, filterlen);
    if (rst != 0) {
        debug("vm_load_filter failed with %d", rst);
        return -2;
    }

    rst = ubpf_exec(_vm, ctx, ctxlen, data, datalen, &retval);
    if (rst < 0) {
        debug("ubpf_exec failed with %d", rst);
        return -3; // note code is already loaded
    }

    *acceptlen = safe_retval(retval);

    debug("retval: %" PRIu64, retval);
    debug("acceptlen: %" PRIu64, *acceptlen);
    return 0;
}

int filter_run_ebpf_filters (
    struct raw_prog * filters,
    void * ctx, uint_fast32_t ctxlen,
    void * data, uint_fast32_t datalen,
    uint64_t * acceptlen)
{
    struct raw_prog * prog;
    uint64_t retval;
    int rst;

    // Return (via acceptlen) shortest read length allowed
    // Since stealing packets is not allowed, only accept zero and negative
    // Maximum return value is datalen

    trace("%s(_st:%d,filters:%p,ctxlen:%" PRIuFAST32
        ",datalen:%" PRIuFAST32 ")",
        __func__, _st, filters, ctxlen, datalen);

    // ctx data validity checked within filter_run_ebpf_filter call
    if (_st == ST_UNKNOWN || acceptlen == NULL)
        return -1;

    *acceptlen = datalen;
    for (prog = filters; prog != NULL; prog = prog->next) {
        rst = filter_run_ebpf_filter(
            prog->progptr, prog->proglen,
            ctx, ctxlen, data, datalen, &retval);

        if (rst < 0) { // cannot run filter
            debug("filter_run_ebpf_filter failed with %d", rst);
            return -2;
        } else if (retval < *acceptlen) {
            *acceptlen = retval;

            if (retval == 0)
                break;
        }
    }

    return 0;
}

void filter_filter_cache_cleanup(void) {
    trace("%s(_st:%d)", __func__, _st);

    if (_st == ST_UNKNOWN)
        return;

    ubpf_unload_code(_vm);
    return;
}


//
// INTERNAL FUNCTION DEFINITIONS
//

int vm_load_ebpf_filter (
    const void * filter, uint_fast32_t filterlen)
{
    char * errmsg;

    assert(_st == ST_INIT);

    trace("%s(_st:%d,filterlen:%" PRIuFAST32 ")",
        __func__, _st, filterlen);

    if (filter == NULL || filterlen == 0)
        return -1; // bad filter

    if (filterlen > UINT32_MAX)
        return -2; // overflow

    // IMPORTANT: one should make sure that the ubpf ebpf instruction struct
    // IS THE SAME definition as struct pktlab_ebpf_instr

    ubpf_unload_code(_vm); // always unload before reloading

    if (ubpf_load(_vm, filter, filterlen, &errmsg) != 0) {
        debug("ubpf_load failed with %s\n", safe_str(errmsg));
        free(errmsg);
        return -3;
    }

    return 0;
}

uint64_t safe_retval(uint64_t retval) {
    int32_t signed_retval;

    // for transforming retval to int

    // assuming that filter program returns int32_t
    // there must be a better way to do this ...
    signed_retval = (int32_t) retval;
    if (signed_retval < MIN_RETVAL) {
        signed_retval = MIN_RETVAL;
    } else if (signed_retval > MAX_RETVAL) {
        signed_retval = MAX_RETVAL;
    }

    return (signed_retval < 0) ? -signed_retval:signed_retval;
}