// key.c
//

#include <assert.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pktlab.h>
#include <pktconf.h>

#include "key.h"
#include "console.h"
#include "util.h"

//
// EXPORTED FUNCTION DEFINITIONS
//

int ll_load_pubkeys (
    const struct pktconf_sep_str_ll * sep_list,
    struct pktlab_publickey *** pubkeys,
    uint_fast32_t * pubkeyno)
{
    struct pktconf_sep_str_entry * node;
    uint_fast32_t i;
    size_t totalmem;
    int ret = -128;

    assert(pubkeys != NULL && pubkeyno != NULL);

    trace("%s(sep_list->len:%u)", __func__,
        (unsigned) sep_list->len);

    if (sep_list == NULL) {
        *pubkeys = NULL;
        *pubkeyno = 0;
        return 0;
    }

    totalmem = sizeof(**pubkeys)*sep_list->len;
    if (totalmem / sizeof(**pubkeys) != sep_list->len) {
        debug("Overflow");
        return -1;
    }

    *pubkeys = safe_malloc(totalmem);
    *pubkeyno = 0;
    node = sep_list->head;
    for (i = 0; i < sep_list->len; ++i) {
        assert(node->entry != NULL);

        if (node->entrylen == 0) {
            debug("Bad key path ll entry content (%s,%u)",
                node->entry, (unsigned) node->entrylen);
            ret = -2;
            goto fail;
        }

        if (load_pubkey(node->entry, &(*pubkeys)[i]) < 0) {
            debug("Failed to load from %s", node->entry);
            ret = -3;
            goto fail;
        }

        ++*pubkeyno;
        node = node->next;
    }

    return 0;

fail:
    free_pubkeys(*pubkeys, *pubkeyno);
    return ret;
}

void free_pubkeys (
    struct pktlab_publickey ** pubkeys,
    uint_fast32_t pubkeyno)
{
    uint_fast32_t i;

    assert(pubkeys != NULL);

    for (i = 0; i < pubkeyno; ++i)
        pktlab_cleanup_publickey(pubkeys[i]);

    free(pubkeys);
}

int load_pubkey (
    const char * path,
    struct pktlab_publickey ** pubkey)
{
    char * rawpubkey;
    size_t rawpubkeylen;
    int ret = -128;

    assert(path != NULL && pubkey != NULL);

    trace("%s(path:%s)", __func__, path);

    if (read_whole_file(
            path, (uint8_t **) &rawpubkey, &rawpubkeylen) < 0) {
        debug("Cannot read whole file content for %s", path);
        ret = -1;
        goto fail;
    }

    if (rawpubkeylen > UINT_FAST32_MAX) {
        debug("Overflow");
        ret = -2;
        goto fail_raw;
    }

    *pubkey = pktlab_create_publickey();

    if (pktlab_load_publickey_pem(
            rawpubkey, rawpubkeylen, *pubkey) < 0) {
        debug("Failed to load public key");
        ret = -3;
        goto fail_pubkey;
    }

    free(rawpubkey);
    return 0;

fail_pubkey:
    pktlab_cleanup_publickey(*pubkey);
fail_raw:
    free(rawpubkey);
fail:
    return ret;
}

int load_privkey (
    const char * path, const char * passphrase,
    struct pktlab_privatekey ** privkey)
{
    char * rawprivkey;
    size_t rawprivkeylen;
    int ret = -128;

    assert(path != NULL && privkey != NULL);

    trace("%s(path:%s,passphrase:%s)",
        __func__, path,
        (passphrase == NULL) ? "(NULL)":"(REDACTED)");

    if (read_whole_file(
            path, (uint8_t **) &rawprivkey, &rawprivkeylen) < 0) {
        debug("Cannot read whole file content for %s", path);
        ret = -1;
        goto fail;
    }

    if (rawprivkeylen > UINT_FAST32_MAX) {
        debug("Overflow");
        ret = -2;
        goto fail_raw;
    }

    *privkey = pktlab_create_privatekey();

    if (pktlab_load_privatekey_pem(
            rawprivkey, rawprivkeylen,
            passphrase, *privkey) < 0) {
        debug("Failed to load private key");
        ret = -3;
        goto fail_privkey;
    }

    free(rawprivkey);
    return 0;

fail_privkey:
    pktlab_cleanup_privatekey(*privkey);
fail_raw:
    free(rawprivkey);
fail:
    return ret;
}

int load_key_w_prompt (
    const char * path, struct pktlab_privatekey ** privkey)
{
    int rst;
    char prompt[1024];
    char * passphrase;

    assert(path != NULL && privkey != NULL);

    trace("%s(path:%s)", __func__, path);

    // first try load key without passphrase
    rst = load_privkey(path, NULL, privkey);
    switch (rst) {
    case 0:
        return 0;
    case -1:
    case -2:
        debug("load_privkey unfixable error");
        return rst;
    case -3:
        break;
    default:
        assert(0); // should never happen
    }

    // load failed, prompt for passphrase and retry
    debug("load_privkey failed without passphrase");
    debug("Retry with passphrase");

    snprintf(
        prompt, sizeof(prompt),
        "Enter passphrase for private key at %s:", path);
    if (posix_getpass(prompt, &passphrase) < 0) {
        debug("Get passphrase failed");
        return -1;
    }

    rst = load_privkey(path, passphrase, privkey);
    if (rst != 0) {
        debug("load_privkey failed with passphrase");
        free(passphrase);
        return -2;
    }

    free(passphrase);
    return 0;
}

int find_signing_key (
    struct pktlab_publickey ** keyls, uint_fast32_t keynum,
    struct pktlab_certificate * cert, uint_fast32_t * skindx)
{
    struct pktlab_cert_info * info;
    uint8_t keyid[PKTLAB_KEYID_LEN];
    uint_fast32_t i;
    int rst, ret = -128;

    assert(
        (keyls == NULL) == (keynum == 0) &&
        cert != NULL && skindx != NULL);

    rst = pktlab_decode_certificate(cert, NULL, &info);
    if (rst != PKTLAB_DECODE_SUCCESS) {
        debug("pktlab_decode_certificate failed with %d", rst);
        return -1;
    }

    for (i = 0; i < keynum; ++i) {
        if (pktlab_get_key_identifier(
                keyls[i], keyid, sizeof(keyid)) < 0) {
            debug("pktlab_get_key_identifier failed");
            goto fail_info;
        }

        if (memcmp(
                info->authority_key_id,
                keyid, sizeof(keyid)) != 0)
            continue;

        // found match
        *skindx = i;
        ret = 0;
        goto success;
    }

    // no match
    debug("No matching signing key found");
    ret = -2;

success:
fail_info:
    free(info);
    return ret;
}