// limitations.c
//

#include <assert.h>
#include <inttypes.h>
#include <limits.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <pktlab.h>

#include "console.h"
#include "network.h"
#include "network_util.h"
#include "rlim.h"
#include "util.h"

#include "limitations.h"

// todo: combine rlim and limitations

//
// INTERNAL TYPE DEFINITIONS
//

enum limit_st {
    ST_UNKNOWN = 0,
    ST_INIT
};

struct pktlab_hard_code_limits {
    int_fast32_t exp_max_num_priority;
    struct pktlab_valid_dest * valid_dsts; // linked list of valid destinations
    bool src_ip_spoofing;                  // false: disallow spoofing, true: spoofing
};

//
// INTERNAL GLOBAL VARIABLES
//

static enum limit_st _st;
static struct pktlab_hard_code_limits _limits;

static struct burst_rate_data * _burst_data_ll;
static struct burst_rate_data * _burst_data_ll_tail;
static struct burst_rate_control * _burst_byte_ctrl;
static struct burst_rate_control * _burst_pkt_ctrl;
static pktlab_time_t _max_win;

// the max pkt byte len is limited by the lowest byte cnt of
// all byte-based burst rates
static uint64_t _max_pkt_byte = UINT64_MAX;
static bool _can_send = true; // correspond to whether pkt burst rate contains 0 cnt

//
// INTERNAL FUNCTION DECLARATIONS
//

static pktlab_time_t get_max_win(void);
static uint64_t get_max_pkt_byte(void);
static bool get_can_send(void);

static void burst_rate_reset(void);
static void burst_rate_init(void);
static void shift_burst_ctrl (
    pktlab_time_t current_time,
    struct burst_rate_control *ctrl, bool is_pkt);
static struct burst_rate_data * new_burst_rate_data (
    uint64_t size, pktlab_time_t timestamp);
static void move_burst_data_window (
    pktlab_time_t cur_time, pktlab_time_t time_window_size);
static void update_burst_rate_ctrl (
    struct rlim_rate * list, struct burst_rate_control ** ctrl);

//
// EXPORTED FUNCTION DEFINITIONS
//

// module state transition functions
//

void limitations_init(void) {
    trace("%s(_st:%d)", __func__, _st);

    if (_st == ST_INIT)
        return;

    memset(&_limits, 0, sizeof(_limits));

    // hard code for now
    _limits.exp_max_num_priority = INT_FAST32_MAX;
    _limits.valid_dsts = NULL;
    _limits.src_ip_spoofing = false;

    // load from rlim module
    burst_rate_init();

    _st = ST_INIT;
}

void limitations_reset(void) {
    trace("%s(_st:%d)", __func__, _st);

    if (_st == ST_UNKNOWN)
        return;

    // free all nodes in linked list
    burst_rate_reset();

    _st = ST_UNKNOWN;
}

// export limitation value functions
//

uint64_t get_burst_rate_max_pkt_byte_len(void) {
    assert(_st == ST_INIT);
    if (!_can_send)
        return 0;

    return _max_pkt_byte;
}

// check functions
//

bool check_socket_cnt(void) {
    trace("%s(rlim.caps.sktcnt:%" PRIu16 ")",
        __func__, rlim.caps.sktcnt);
    return (rlim.caps.sktcnt > 0);
}

bool check_exp_endtime(pktlab_time_t current_time) {
    trace("%s(current_time:%" PRIu64 ",rlim.endtime:%" PRIu64 ")",
        __func__, current_time, rlim.endtime);
    return (current_time <= rlim.endtime);
}

// TODO: change to new struct
bool check_exp_priority(uint32_t exp_priority) {
    trace("%s(exp_priority:%" PRIu32
        ",_limits.exp_max_num_priority:%" PRId32 ")",
        __func__, exp_priority, _limits.exp_max_num_priority);
    return (_limits.exp_max_num_priority < 0 ||
            exp_priority <= _limits.exp_max_num_priority);
}

// TODO: change to new struct
bool check_src_limit(const uint8_t * src_ip, uint8_t prfam) {
    trace("%s(prfam:%" PRIu8 ")", __func__, prfam);
    return (_limits.src_ip_spoofing ||
            is_exported_laddr(prfam, src_ip));
}

// TODO: change to new struct
bool check_dst_limit(const uint8_t * dst_ip, uint8_t prfam) {
    struct pktlab_valid_dest * cur;
    bool ip_match;
    int i, addrlen;

    trace("%s(prfam:%" PRIu8 ")", __func__, prfam);

    if (_limits.valid_dsts == NULL)
        return true;

    for (cur = _limits.valid_dsts;
         cur != NULL; cur = cur->next) {
        if (cur->prfam != prfam)
            continue;

        ip_match = true;
        addrlen = ADDRLEN(prfam);

        for (i = 0; i < addrlen; i++) {
            if ((dst_ip[i] & cur->network_mask[i]) != cur->ip_address[i]) {
                ip_match = false;
                break;
            }
        }

        if (ip_match)
            return true;
    }

    return false;
}

bool check_send_limit (
    uint64_t size, pktlab_time_t cur_time,
    pktlab_time_t * wait_until_time)
{
    int rst = 0;

    assert(wait_until_time != NULL);

    trace("%s(size:%" PRIu64 ",cur_time:%" PRIu64 ")",
        __func__, size, cur_time);

    if (size > rlim.accs.sndbytecnt ||
        rlim.accs.sndpktcnt < 1 ||
        size > _max_pkt_byte || !_can_send) {
        *wait_until_time = PKTLAB_TIME_MAX; // never available
        return false;
    }

    debug("rlim.accs.sndbytecnt:%" PRIu64
        ",rlim.accs.sndpktcnt:%" PRIu64
        ",_max_pkt_byte:%" PRIu64 ",_can_send:%d",
        rlim.accs.sndbytecnt, rlim.accs.sndpktcnt,
        _max_pkt_byte, _can_send);

    // previous checks passed, now check burst rate, already update burst rate data
    // if burst rate is exceeded, but able to schedule send, don't add the data
    rst = get_burst_wait_time(size, cur_time, wait_until_time);
    assert(rst >= 0);

    debug("wait_until_time:%" PRIu64, *wait_until_time);

    return (*wait_until_time == 0);
}

bool check_recv_limit(uint64_t size) {
    trace("%s(size:%" PRIu64 ",rlim.accs.rcvbytecnt:%" PRIu64
        ",rlim.accs.rcvpktcnt:%" PRIu64 ")", __func__, size,
        rlim.accs.rcvbytecnt, rlim.accs.rcvpktcnt);

    return (rlim.accs.rcvbytecnt >= size &&
            rlim.accs.rcvpktcnt >= 1);
}

int get_burst_wait_time (
    uint64_t data_size, pktlab_time_t current_time,
    pktlab_time_t * wait_until_time)
{
    struct burst_rate_control * ctrl_node;
    struct burst_rate_data * data_node;
    uint64_t min_byte_quota, future_quota;
    pktlab_time_t byte_unlimit_time, pkt_unlimit_time;

    assert(wait_until_time != NULL);

    trace("%s(data_size:%" PRIu64 ",current_time:%" PRIu64
        ",_max_pkt_byte:%" PRIu64 ",_can_send:%d)", __func__,
        data_size, current_time, _max_pkt_byte, _can_send);

    if (data_size > _max_pkt_byte)
        return -1; // data_size too big

    if (!_can_send)
        return -2; // pkt burst rate does not allow sending

    // always try cleanup first
    shift_burst_ctrl(current_time, _burst_byte_ctrl, false);
    shift_burst_ctrl(current_time, _burst_pkt_ctrl, true);
    move_burst_data_window(current_time, _max_win);

    // identify if byte burst rate is limiting
    min_byte_quota = UINT64_MAX;

    for (ctrl_node = _burst_byte_ctrl;
         ctrl_node != NULL; ctrl_node = ctrl_node->next) {

        assert(
            data_size <= ctrl_node->cnt &&
            ctrl_node->wincnt <= ctrl_node->cnt);

        if (ctrl_node->cnt - ctrl_node->wincnt < min_byte_quota)
            min_byte_quota = ctrl_node->cnt - ctrl_node->wincnt;
    }

    debug("min_byte_quota:%" PRIu64, min_byte_quota);

    // is limiting, find max time when limit releases
    byte_unlimit_time = 0;
    if (min_byte_quota < data_size) {
        for (ctrl_node = _burst_byte_ctrl;
             ctrl_node != NULL; ctrl_node = ctrl_node->next) {

            assert(
                data_size <= ctrl_node->cnt &&
                ctrl_node->wincnt <= ctrl_node->cnt);

            if (ctrl_node->cnt - ctrl_node->wincnt != min_byte_quota)
                continue;

            assert(
                (ctrl_node->wincnt != 0) == \
                (ctrl_node->burst_data != NULL));

            future_quota = min_byte_quota;
            for (data_node = ctrl_node->burst_data;
                 data_node != NULL; data_node = data_node->next) {

                assert(future_quota < data_size);
                if (data_node->size < data_size - future_quota) {
                    future_quota += data_node->size;
                    continue;
                }

                // future quota from this point on enough!
                // see if this happens much later than known
                if (data_node->timestamp > PKTLAB_TIME_MAX-ctrl_node->win) {
                    byte_unlimit_time = PKTLAB_TIME_MAX;
                } else if (byte_unlimit_time < data_node->timestamp+ctrl_node->win)
                    byte_unlimit_time = data_node->timestamp+ctrl_node->win;

                break; // try next burst rate
            }

            // such node must exist, and provide enough quota
            assert(
                data_node != NULL &&
                data_node->size >= data_size - future_quota);
        }
    }

    // byte_unlimit_time from this point onward is the latest time
    // all byte burst rate will allow the data_size to be sent

    // identify pkt burst rate unlimit time
    pkt_unlimit_time = 0;
    for (ctrl_node = _burst_pkt_ctrl;
         ctrl_node != NULL; ctrl_node = ctrl_node->next) {

        assert(
            ctrl_node->cnt != 0 &&
            ctrl_node->wincnt <= ctrl_node->cnt);

        if (ctrl_node->cnt - ctrl_node->wincnt != 0)
            continue; // skip non-limiting ones

        // found limiting burst rate

        assert(ctrl_node->burst_data != NULL);

        if (ctrl_node->burst_data->timestamp > PKTLAB_TIME_MAX - ctrl_node->win) {
            pkt_unlimit_time = PKTLAB_TIME_MAX;
        } else {
            pkt_unlimit_time = MAX(
                pkt_unlimit_time,
                ctrl_node->burst_data->timestamp+ctrl_node->win);
        }
    }

    // pkt_unlimit_time from this point onward is the latest time
    // all pkt burst rate will allow the data_size to be sent

    *wait_until_time = MAX(byte_unlimit_time, pkt_unlimit_time);
    return 0;
}

// update functions
//

void consume_socket_cnt(void) {
    assert(check_socket_cnt());
    trace("%s(rlim.caps.sktcnt:%" PRIu16 ")",
        __func__, rlim.caps.sktcnt);
    --rlim.caps.sktcnt;
}

void consume_send_quota(uint64_t size, pktlab_time_t cur) {
    struct burst_rate_data * new_data_node;
    struct burst_rate_control * ctrl_node;
    pktlab_time_t wait_until;

    (void) wait_until; // suppress warning

    assert(
        get_burst_wait_time(
            size, cur, &wait_until) == 0 &&
        wait_until == 0 &&
        rlim.accs.sndbytecnt >= size &&
        rlim.accs.sndpktcnt >= 1);

    trace("%s(size:%" PRIu64 ",cur:%" PRIu64 ")",
        __func__, size, cur);

    rlim.accs.sndbytecnt -= size;
    rlim.accs.sndpktcnt -= 1;

    // append to burst data ll
    new_data_node = new_burst_rate_data(size, cur);

    if (_burst_data_ll == NULL) {
        _burst_data_ll_tail = _burst_data_ll = new_data_node;
    } else {
        _burst_data_ll_tail->next = new_data_node;
        _burst_data_ll_tail = new_data_node;
    }

    // update ctrl_node wincnt and burst data pointer
    for (ctrl_node = _burst_byte_ctrl;
         ctrl_node != NULL; ctrl_node = ctrl_node->next) {
        assert(
            size <= ctrl_node->cnt &&
            ctrl_node->wincnt <= ctrl_node->cnt - size);

        ctrl_node->wincnt += size;
        if (ctrl_node->burst_data == NULL)
            ctrl_node->burst_data = new_data_node;
    }

    for (ctrl_node = _burst_pkt_ctrl;
         ctrl_node != NULL; ctrl_node = ctrl_node->next) {
        assert(
            1 <= ctrl_node->cnt &&
            ctrl_node->wincnt <= ctrl_node->cnt - 1);

        ++ctrl_node->wincnt;
        if (ctrl_node->burst_data == NULL)
            ctrl_node->burst_data = new_data_node;
    }

    return;
}

void consume_recv_quota(uint64_t size) {
    assert(
        rlim.accs.rcvbytecnt >= size &&
        rlim.accs.rcvpktcnt >= 1);

    trace("%s(size:%" PRIu64 ")", __func__, size);

    rlim.accs.rcvbytecnt -= size;
    rlim.accs.rcvpktcnt -= 1;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

void burst_rate_reset(void) {
    struct burst_rate_data * data_node;
    struct burst_rate_control * ctrl_node;

    while (_burst_data_ll != NULL) {
        data_node = _burst_data_ll;
        _burst_data_ll = _burst_data_ll->next;
        free(data_node);
    }
    _burst_data_ll_tail = NULL;

    while (_burst_byte_ctrl != NULL) {
        ctrl_node = _burst_byte_ctrl;
        _burst_byte_ctrl = _burst_byte_ctrl->next;
        free(ctrl_node);
    }

    while (_burst_pkt_ctrl != NULL) {
        ctrl_node = _burst_pkt_ctrl;
        _burst_pkt_ctrl = _burst_pkt_ctrl->next;
        free(ctrl_node);
    }

    _max_win = 0;
    _max_pkt_byte = UINT64_MAX;
    _can_send = true;
}

void burst_rate_init(void) {
    burst_rate_reset();
    update_burst_rate_ctrl(rlim.rates.sndbytelist, &_burst_byte_ctrl);
    update_burst_rate_ctrl(rlim.rates.sndpktlist, &_burst_pkt_ctrl);

    _max_win = get_max_win();
    _max_pkt_byte = get_max_pkt_byte();
    _can_send = get_can_send();
}

// copy burst rate from rlim rate list
void update_burst_rate_ctrl (
    struct rlim_rate * list, struct burst_rate_control ** ctrl)
{
    struct rlim_rate * cur;
    struct burst_rate_control * new_ctrl;

    assert(ctrl != NULL && *ctrl == NULL);

    for (cur = list; cur != NULL; cur = cur->next) {
        new_ctrl = safe_malloc(sizeof(*new_ctrl));
        memset(new_ctrl, 0, sizeof(*new_ctrl));

        new_ctrl->cnt = cur->cnt;
        new_ctrl->win = cur->win;
        new_ctrl->burst_data = NULL;
        new_ctrl->next = *ctrl;
        *ctrl = new_ctrl;
    }
}

struct burst_rate_data * new_burst_rate_data (
    uint64_t size, pktlab_time_t timestamp)
{
    struct burst_rate_data * new_data_node;

    new_data_node = safe_malloc(sizeof(*new_data_node));
    memset(new_data_node, 0, sizeof(*new_data_node));

    new_data_node->size = size;
    new_data_node->timestamp = timestamp;
    new_data_node->next = NULL;
    return new_data_node;
}

// update rate ctrl burst data pointer to reflect advanced window
void shift_burst_ctrl (
    pktlab_time_t current_time,
    struct burst_rate_control * ctrl, bool is_pkt)
{
    struct burst_rate_control * cur_ctrl;
    struct burst_rate_data * current;
    pktlab_time_t window_start;

    for (cur_ctrl = ctrl; cur_ctrl != NULL;
         cur_ctrl = cur_ctrl->next) {

        if (current_time < cur_ctrl->win)
            continue; // window start extends beyond epoch

        window_start = current_time - cur_ctrl->win;

        for (current = cur_ctrl->burst_data;
             current != NULL; current = current->next) {
            if (current->timestamp >= window_start)
                break;

            if (is_pkt) {
                assert(cur_ctrl->wincnt >= 1);
                --cur_ctrl->wincnt;
            } else {
                assert(cur_ctrl->wincnt >= current->size);
                cur_ctrl->wincnt -= current->size;
            }

            cur_ctrl->burst_data = current->next;
        }
    }
}

// update rate burst data ll to reflect advanced window
// i.e. remove non-relevant old recorded burst_rate_data
void move_burst_data_window (
    pktlab_time_t cur_time, pktlab_time_t time_window_size)
{
    struct burst_rate_data * next_node;

    if (cur_time < time_window_size)
        return; // window start extends beyond epoch

    while (_burst_data_ll != NULL) {
        if (_burst_data_ll->timestamp < cur_time - time_window_size) {
            next_node = _burst_data_ll->next;
            free(_burst_data_ll);
            _burst_data_ll = next_node;
        } else
            break;
    }

    // in case everything data freed
    if (_burst_data_ll == NULL)
        _burst_data_ll_tail = NULL;
}

pktlab_time_t get_max_win(void) {
    struct burst_rate_control * node;
    pktlab_time_t max_win;

    max_win = 0;

    for (node = _burst_byte_ctrl;
         node != NULL; node = node->next)
        max_win = MAX(node->win, max_win);

    for (node = _burst_pkt_ctrl;
         node != NULL; node = node->next)
        max_win = MAX(node->win, max_win);

    return max_win;
}

uint64_t get_max_pkt_byte(void) {
    struct burst_rate_control * node;
    uint64_t max_pkt_byte;

    max_pkt_byte = UINT64_MAX;

    for (node = _burst_byte_ctrl;
         node != NULL; node = node->next)
        max_pkt_byte = MIN(node->cnt, max_pkt_byte);

    return max_pkt_byte;
}

bool get_can_send(void) {
    struct burst_rate_control * node;

    for (node = _burst_pkt_ctrl;
         node != NULL; node = node->next) {
        if (node->cnt == 0)
            return false;
    }

    return true;
}