// network.c
//

#include "network.h"

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <inttypes.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include <pktlab.h>

#include "config.h"
#include "console.h"
#include "control.h"
#include "filter.h"
#include "network_util.h"
#include "nqueue.h"
#include "raw.h"
#include "util.h"
#include "vmem.h"
#include "worker.h"
#include "net.h"
#include "limitations.h"

// TODO Fix error return order
// TODO Implement buffer usage accounting
// We currently do not account for buffer use, which could lead to memory
// exhaustion. To implement this, we need to do the following:
//
// 1. Initialize maximum number of bytes used for buffers (_bufmax) at
//    start of experiment.
// 2. Account for all buffer allocations by incrementing _bufused.
// 3. Account for all receive buffer allocations by incrementing
//    skt->rbufused.
// 4. Do not receive on socket if rbufmax <= rbufused.
// 5. Account for all buffer releases by decrementing _bufused and
//    skt->rbufused (receive buffers only).

//
// COMPILE-TIME CONSTANTS
//

#ifndef CONNECT_TIMEOUT_MS
#define CONNECT_TIMEOUT_MS 60000
#endif

//
// INTERNAL MACROS
//

#define SKT_CNT      PKTLAB_MAX_SKT_CNT
#define CTLFL_CNT    0x80
#define TSTP_CNT     0x10000 // remove?
#define VMEM_SKT_LEN PKTLAB_VMEMADDR_SKTINFO_BLKLEN

#define NSEND_DELAY_LOG_THRES 1000000 // in ticks

#define DEFAULT_BUFMAX 0x800000

// 20 for potential extra book keeping overhead
#define SEND_BUF_OVERHEAD     (sizeof(struct pktlab_message)+sizeof(struct txq_msg)+20)
#define SEND_RAW_BUF_OVERHEAD (SEND_BUF_OVERHEAD+sizeof(struct txq_msg_aux))
#define RECV_BUF_OVERHEAD     (sizeof(struct pktlab_message)+24+20) // sizeof(struct queued_msg) is 24

#define FILT_BUFSZ 0x10000

#define UINT24_MAX UINT32_C(0xffffff)

//
// INTERNAL TYPE DEFINITIONS
//

struct socket {
    enum pktlab_socket_state state;
    uint8_t netproto;
    uint8_t transproto;
    uint8_t addrlen;
    int wfd, rfd, err;
    struct nqueue txq;
    struct nqueue notifq;
    uint32_t msgoff;
    uint32_t rbufmax;
    uint32_t rbufused;
    pktlab_time_t open_until;
    uint8_t locaddr[PKTLAB_ADDRLEN_MAX];
    uint8_t remaddr[PKTLAB_ADDRLEN_MAX];
    uint16_t locport, remport;
    struct {
        uint32_t bytecnt;
        uint32_t pktcnt;
    } dropstats;
    enum pktlab_nsend_err nsenderr;
    uint16_t nsenderr_tag;
    uint16_t notifmask;
    // uint16_t errtidx;
    uint8_t ctlfl[CTLFL_CNT];
    bool wfin;
    pktlab_time_t wclose_from;
    struct raw_socket_obj * robj;
    pktlab_time_t filter_until;
};

enum state { ST_UNKNOWN, ST_INIT, ST_IDLE };

struct nopen_addr {
    uint8_t laddrlen;
    uint8_t lportlen;
    uint8_t raddrlen;
    uint8_t rportlen;
    const void *laddrptr;
    const void *lportptr;
    const void *raddrptr;
    const void *rportptr;
};

struct notif_aux {
    struct pktlab_message *msg;
    uint16_t notifmask;
    int_fast32_t glob_used;
    int_fast32_t skt_used;
    send_msg_cb_t memupdate;
};

struct txq_msg {
    struct pktlab_message *msg;
    void * aux;
};

// for storing aux info for a msg
struct txq_msg_aux {
    bool truncated;
};

//
// INTERNAL GLOBAL VARIABLES
//

static const struct raw_socket_method_t * raw_socket_module;
static struct socket _skt[SKT_CNT];
// static struct nqueue _rxq;
static enum state _state;
static bool _dropfl; // true if at least one sock has dropped packets
static pktlab_time_t _send_tstp[TSTP_CNT];

static uint32_t _bufmax;
static uint32_t _bufused;

static bool _ip4_sup;
static bool _ip6_sup;
static bool _raw_sup;
static bool _tcp_sup;
static bool _udp_sup;
static bool _raw_lkm;

//
// INTERNAL FUNCTION DECLARATIONS
//

// static void poll_flush(void);

static void state_change(enum state state);

static void raw_socket_open(uint_fast8_t sktid, uint_fast8_t netproto,
                            uint_fast32_t rbufsz, struct nopen_addr *addr);

static void trans_socket_open(uint_fast8_t sktid, uint_fast8_t netproto,
                              uint_fast32_t rbufsz, uint_fast8_t transproto,
                              struct nopen_addr *addr);

static void socket_close(uint_fast8_t sktid);

static void socket_readable(uint_fast8_t sktid);
static void socket_writable(uint_fast8_t sktid);
static void socket_connected(uint_fast8_t sktid);

static void socket_state_change(unsigned int sktid,
                                enum pktlab_socket_state state);
// static void inject_null_ndata(uint_fast8_t sktid);
static void pop_skt_txq_n_update_buf_usage (
    uint_fast8_t sktid, pktlab_time_t time, enum pktlab_status errid);
static void drain_skt_txq(uint_fast8_t sktid);

static void clear_dropstats(void);
// static struct pktlab_message *create_nstat(void);

static bool free_txq_msg_filt(void *ptr, uintptr_t aux);
// static bool freesktid_filt(void *ptr, uintptr_t sktid);

static inline const char *state_name(enum state state);

static void skt_read(uint8_t sktid, void *buf);
static void
comp_sktwrite_param(uint_fast32_t field_start, uint_fast32_t field_len,
                    uint_fast32_t write_start, uint_fast32_t write_end,
                    uint_fast32_t *write_off, uint_fast32_t *buf_off,
                    uint_fast32_t *write_len);

static bool fin_socket(uint8_t sktid);

// static void show_skt(void);

// nsend and ndata buffer size should not be larger than 2 ^ 31 - 1 bytes
static int32_t nsend_mem_cal(const struct pktlab_message *restrict msg);
static int32_t ndata_mem_cal(const struct pktlab_message *restrict msg);

// FIXME: do we need to account for global buffer usage of ntag
static int32_t ntag_mem_cal(const struct pktlab_message *restrict msg);
static int32_t nstat_mem_cal(const struct pktlab_message *restrict msg);
static int32_t ndrop_mem_cal(const struct pktlab_message *restrict msg);

static uint_fast8_t get_notif_msg_sktid(struct pktlab_message *msg);
static void skt_mem_update(void * aux);

static struct txq_msg *
create_txq_msg(struct pktlab_message *msg, bool truncated);

// notification related
static struct notif_aux *
create_notif_aux(struct pktlab_message *msg, uint16_t notifmask,
                 int_fast32_t glob_used, int_fast32_t skt_used,
                 send_msg_cb_t memupdate);
static void push_notif(uint8_t sktid, struct pktlab_message *msg,
                       int_fast32_t glob_used, int_fast32_t skt_used,
                       send_msg_cb_t memupdate);
static bool notif_filt(void *elt, uintptr_t aux);
static void flush_notif(uint8_t sktid, uint16_t notifmask);

static bool have_space(uint_fast8_t sktid, uint32_t * spaceleft);
static uint32_t limited_add(const uint32_t target, const ssize_t incre);

static bool is_invalid_netproto(uint_fast8_t netproto);
static bool is_invalid_transproto(uint_fast8_t transproto);

static void check_proto_sup(uint_fast8_t netproto, uint_fast8_t transproto,
                            bool * netproto_sup, bool * transproto_sup);

static int expand_all0_addr(uint_fast8_t prfam,
                            const uint8_t * passed_addr, uint8_t * actual_laddr);

//
// EXPORTED FUNCTION DEFINITIONS
//

void network_initialize(void) {
    long long conf_bufmax;
    int rst;

    trace("%s()", __func__);
    assert(_state < ST_INIT);

    conf_bufmax = _config.network_bufmax; // should be <= UINT32_MAX and >= -1
    if (conf_bufmax < 0)
        conf_bufmax = DEFAULT_BUFMAX;

    _bufmax = conf_bufmax; // should not overflow
    info("Exporting bufmax: %" PRIu32, _bufmax);

    _ip4_sup = !vmem_indx_get_laddr(PKTLAB_IP4_PROTO, 0, NULL);
    _ip6_sup = !vmem_indx_get_laddr(PKTLAB_IP6_PROTO, 0, NULL);
    _raw_sup = _config.raw_sup;
    _tcp_sup = _config.tcp_sup;
    _udp_sup = _config.udp_sup;
    _raw_lkm = _config.raw_lkm;

    if (_raw_sup)
        info("Using kernel module: %s", (_raw_lkm) ? "T" : "F");

    info("Supporting netprotos:%s%s", (_ip4_sup) ? " IPv4" : "",
         (_ip6_sup) ? " IPv6" : "");
    info("Supporting transprotos:%s%s%s", (_raw_sup) ? " Raw" : "",
         (_tcp_sup) ? " TCP" : "", (_udp_sup) ? " UDP" : "");

    // nqinit(&_rxq);
    _state = ST_INIT;

    // still init for send filter support
    rst = filter_init();
    if (rst < 0 && rst != -128)
        fatal("Failed to init eBPF VM");

    // TODO: May have different implementations later on
    raw_socket_module = NULL;
    if (_raw_sup)
        raw_socket_module = _raw_lkm ? \
            raw_socket_steal_method():raw_socket_nosteal_method();
}

void network_start(void) {
    trace("%s()", __func__);
    assert(_state == ST_INIT);
    state_change(ST_IDLE);
}

void network_reset(void) {
    struct socket *skt;
    int sktid;

    assert(_state >= ST_INIT);

    trace("network:%s(state:%s)", __func__, state_name(_state));

    if (_state == ST_INIT)
        return;

    // Close all sockets.
    for (sktid = 0; skt = _skt + sktid, sktid < SKT_CNT; sktid++)
        if (skt->state != PKTLAB_SKTST_FREE)
            socket_close(sktid);

    // Free all buffered messages.

    // while (!nqempty(&_rxq))
    //     free(nqget(&_rxq));

    // Reinit module global variables
    memset(_send_tstp, 0, sizeof(_send_tstp));

    _dropfl = false;
    _bufused = 0;
    // _bufmax, sup flags remains the same

    // reuse filter ebpf VM but clean up loaded stuffs
    filter_filter_cache_cleanup();

    state_change(ST_INIT);
}

int network_prepare_select(fd_set *rset, fd_set *wset,
                           pktlab_time_t *select_until) {
    struct txq_msg *tmsg;
    struct socket *skt;
    unsigned int sktid;
    pktlab_time_t limit_wait_until;
    bool sendable;
    int nfds = 0;

    trace("%s(state:%s)", __func__, state_name(_state));
    assert(_state > ST_INIT);

    for (sktid = 0; skt = _skt + sktid, sktid < SKT_CNT; sktid++) {
        if (skt->state <= PKTLAB_SKTST_FREE)
            continue;

        debug("SKT[%d]:{wfd:%d, rfd:%d, txqlen:%zu, state:%s}",
              sktid, skt->wfd, skt->rfd, nqlen(&skt->txq),
              pktlab_sktstate_name(skt->state));

        switch (skt->state) {
        case PKTLAB_SKTST_OPENING:
            FD_SET(skt->wfd, wset);
            nfds = MAX(nfds, skt->wfd + 1);
            break;

        case PKTLAB_SKTST_OPEN:
        case PKTLAB_SKTST_WFIN:
            // Prepare skt read
            if (skt->transproto != PKTLAB_TCP_PROTO || have_space(sktid, NULL)) {
                debug("fu:%llu, to:%llu", skt->filter_until, _time_now);
                if (skt->transproto != PKTLAB_RAW_PROTO ||
                    _time_now < skt->filter_until) {
                    FD_SET(skt->rfd, rset);
                    nfds = MAX(nfds, skt->rfd + 1);
                }
            }
            // no break

        case PKTLAB_SKTST_EOF:
            debug("rbuf (%" PRIu32 "/%" PRIu32 "),"
                  " buf (%" PRIu32 "/%" PRIu32 ")",
                  skt->rbufused, skt->rbufmax, _bufused, _bufmax);

            if (skt->state == PKTLAB_SKTST_WFIN)
                break;

            // Prepare skt write
            // Case with outstanding nsend
            if ((tmsg = nqpeek(&skt->txq))) {
                assert(tmsg->msg->type == PKTLAB_NSEND_MESSAGE);

                // check limit and see if we can actually send
                sendable = check_send_limit(
                    tmsg->msg->nsend.len, _time_now,
                    &limit_wait_until);

                if (sendable) {
                    if (tmsg->msg->nsend.time <= _time_now) {
                        FD_SET(skt->wfd, wset);
                        nfds = MAX(nfds, skt->wfd + 1);
                    } else
                        *select_until = MIN(*select_until, tmsg->msg->nsend.time);
                } else {
                    if (limit_wait_until == PKTLAB_TIME_MAX) {
                        // forever cannot send case,
                        // we immediate go to process select to drain txq
                        *select_until = _time_now;
                    } else {
                        // select until either nsend is ready or limit releases,
                        // whatever happens later
                        *select_until = \
                            MIN(MAX(limit_wait_until, tmsg->msg->nsend.time),
                                *select_until);
                    }
                }
            }

            // Case with requested nfin, still do socket_writable
            // for shutting down write end
            if (skt->wfin) {
                if (skt->wclose_from <= _time_now) {
                    FD_SET(skt->wfd, wset);
                    nfds = MAX(nfds, skt->wfd + 1);
                } else {
                    if (skt->wclose_from < *select_until)
                        *select_until = skt->wclose_from;
                }
            }
            // no break

        case PKTLAB_SKTST_END:
        default:
            break;
        }
    }

    return nfds;
}

int network_process_select(int nsel, const fd_set *rset, const fd_set *wset) {
    struct socket *skt;
    unsigned int sktid;
    struct txq_msg *tmsg;
    pktlab_time_t wait_until;
    int nhandled = 0;

    trace("%s(state:%s)", __func__, state_name(_state));
    assert(_state > ST_INIT);

    for (sktid = 0; skt = _skt + sktid, sktid < SKT_CNT; sktid++) {
        if (skt->state <= PKTLAB_SKTST_FREE)
            continue;

        if (FD_ISSET(skt->rfd, rset)) {
            socket_readable(sktid);
            nhandled += 1;
        }

        if (FD_ISSET(skt->wfd, wset)) {
            switch (skt->state) {
            case PKTLAB_SKTST_OPENING:
                socket_connected(sktid);
                nhandled += 1;
                break;
            case PKTLAB_SKTST_OPEN:
            case PKTLAB_SKTST_EOF:
                socket_writable(sktid);
                nhandled += 1;
                break;
            default: // should not happen
                debug("SKT %u descriptor %d is writable in state %s "
                      "(should not happen)",
                      sktid, skt->wfd, pktlab_sktstate_name(skt->state));
                break;
            }
        }

        // check special case where we forever cannot send anymore
        if ((tmsg = nqpeek(&skt->txq))) {
            assert(tmsg->msg->type == PKTLAB_NSEND_MESSAGE);
            if (!check_send_limit(
                    tmsg->msg->nsend.len, _time_now, &wait_until) &&
                wait_until == PKTLAB_TIME_MAX)
                drain_skt_txq(sktid);
        }
    }

    return nhandled;
}

// Message handlers.
//

void network_handle_nopen(struct pktlab_message *msg) {
    const uint_fast8_t sktid = msg->nopen.sktid;
    const uint_fast8_t netproto = msg->nopen.family;
    const uint_fast32_t rbufsz = msg->nopen.rbufsz;
    const uint_fast8_t transproto = msg->nopen.proto;
    struct nopen_addr addr = {.laddrlen = msg->nopen.laddrlen,
                              .lportlen = msg->nopen.lportlen,
                              .raddrlen = msg->nopen.raddrlen,
                              .rportlen = msg->nopen.rportlen,
                              .laddrptr = msg->nopen.laddrptr,
                              .lportptr = msg->nopen.lportptr,
                              .raddrptr = msg->nopen.raddrptr,
                              .rportptr = msg->nopen.rportptr};
    struct socket *const skt = _skt + sktid;
    bool prfam_sup, proto_sup;

    trace("network_handle_nopen(msg:{sktid:%d,proto:0x%02x})",
          (int)msg->nopen.sktid, (int)msg->nopen.proto);
    assert(msg->type == PKTLAB_NOPEN_MESSAGE);
    assert(_state == ST_IDLE);

    if (!check_socket_cnt()) {
        control_report(PKTLAB_ENORES);
        return;
    }

    if (skt->state != PKTLAB_SKTST_FREE) {
        control_report(PKTLAB_ESKTINUSE);
        return;
    }

    // check if not-defined prfam/proto values
    if (is_invalid_netproto(netproto)) {
        control_report(PKTLAB_ESKTNOPRFAM);
        return;
    }

    if (is_invalid_transproto(transproto)) {
        control_report(PKTLAB_ESKTNOPROTO);
        return;
    }

    // check support config
    check_proto_sup(netproto, transproto, &prfam_sup, &proto_sup);
    if (!prfam_sup) {
        control_report(PKTLAB_ESKTNOPRFAM);
        return;
    }

    if (!proto_sup) {
        control_report(PKTLAB_ESKTNOPROTO);
        return;
    }

    // todo: add ESYSNOBUFS error return? discuss this

    if (transproto == PKTLAB_RAW_PROTO) {
        // ops for raw sock
        if (!is_exported_laddr(netproto, addr.laddrptr)) {
            control_report(PKTLAB_ESKTNOADDR);
            return;
        }

        raw_socket_open(sktid, netproto, rbufsz, &addr);
        return;
    } else if (transproto == PKTLAB_TCP_PROTO &&
               msg->nopen.raddrptr != NULL) {

        // dst check for tcp connect
        if (!check_dst_limit(
                msg->nopen.raddrptr, msg->nopen.family)) {
            control_report(PKTLAB_ESKTOPNDENY);
            return;
        }
    }

    // ops for transport sock
    trans_socket_open(sktid, netproto, rbufsz, transproto, &addr);
}

void network_handle_nclose(struct pktlab_message *msg) {
    const uint_fast8_t sktid = msg->nopen.sktid;
    struct socket *const skt = _skt + sktid;

    trace("network_handle_nclose(msg:{sktid:%d})", (int)sktid);
    assert(msg->type == PKTLAB_NCLOSE_MESSAGE);
    assert(_state == ST_IDLE);

    if (skt->state == PKTLAB_SKTST_FREE) {
        control_report(PKTLAB_ESKTNOTOPEN);
        return;
    } else
        control_report(PKTLAB_SUCCESS);

    socket_close(sktid);
}

void network_handle_nsend(struct pktlab_message *msg) {
    const uint_fast8_t sktid = msg->nsend.sktid;
    struct socket *const skt = _skt + sktid;
    enum pktlab_status errid;
    void * copied_pkt;
    size_t copiedlen;
    int32_t msg_bufused;
    uint64_t acceptlen;
    bool truncated;
    const uint8_t * src_addr;
    const uint8_t * dst_addr;
    uint8_t prfam;
    struct txq_msg * tmsg;
    pktlab_time_t wait_until_time;
    int rst;

    assert(msg->type == PKTLAB_NSEND_MESSAGE && _state == ST_IDLE);

    trace("%s(msg:{sktid:%d})", __func__, (int)msg->nsend.sktid);

    errid = PKTLAB_EUNKFAULT;

    if (skt->state == PKTLAB_SKTST_FREE || skt->state == PKTLAB_SKTST_WFIN ||
        skt->state == PKTLAB_SKTST_END) {
        errid = PKTLAB_ESKTNOTOPEN;
        goto fail;
    }

    if (skt->state < PKTLAB_SKTST_FREE) {
        debug("skt state %d", skt->state);
        errid = PKTLAB_ESKTNOTOPEN; // FIXME?
        goto fail;
    }

    if ((skt->netproto != msg->nsend.family) |
        (skt->transproto != msg->nsend.proto)) {
        errid = PKTLAB_ESKTNOMATCH;
        goto fail;
    }

    // temporary: filter sent packet for raw sockets using send monitors
    truncated = false;

    if (skt->transproto == PKTLAB_RAW_PROTO) {
        // note we discard any changes the filter may have to the pkt
        // though currently the filter module does not allow writing
        copiedlen = (msg->nsend.len > SIZE_MAX) ? SIZE_MAX:msg->nsend.len;
        copied_pkt = safe_malloc(copiedlen);
        memcpy(copied_pkt, msg->nsend.ptr, copiedlen);

        rst = filter_run_ebpf_filters(
            control_fetch_cert_send_monitors(), NULL, 0,
            copied_pkt, copiedlen, &acceptlen);
        free(copied_pkt);

        if (rst < 0) {
            assert(rst == -2);

            debug("filter_run_ebpf_filters exec failed, rejecting nsend msg");
            errid = PKTLAB_ESKTPKTDENY;
            goto fail;
        }

        if (acceptlen < msg->nsend.len) {
            debug("pkt truncation detected");
            msg->nsend.len = acceptlen;
            truncated = true;
        }

        // check src address spoofing
        rst = get_src_addr(
            msg->nsend.ptr, msg->nsend.len,
            &prfam, &src_addr);
        if (rst < 0) {
            debug("get_src_addr failed with %d", rst);
            errid = PKTLAB_ESKTPKTDENY; // we deny pkt that we cannot get src
            goto fail;
        }

        if (!check_src_limit(src_addr, prfam)) {
            debug("Invalid src address");
            errid = PKTLAB_ESKTPKTDENY; // src spoofing pkt disallowed
            goto fail;
        }

        // check dst address validity for raw
        rst = get_dst_addr(
            msg->nsend.ptr, msg->nsend.len,
            &prfam, &dst_addr);
        if (rst < 0) {
            debug("get_dst_addr failed with %d", rst);
            errid = PKTLAB_ESKTPKTDENY; // we deny pkt that we cannot get dst
            goto fail;
        }

        if (!check_dst_limit(dst_addr, prfam)) {
            debug("Invalid dst address");
            errid = PKTLAB_ESKTPKTDENY; // bad dst
            goto fail;
        }
    } else if (skt->transproto == PKTLAB_UDP_PROTO) {
        // check dst address validity for udp
        if (!check_dst_limit(
                msg->nsend.udp.raddrptr, msg->nsend.family)) {
            debug("Invalid dst address");
            errid = PKTLAB_ESKTPKTDENY; // bad dst
            goto fail;
        }
    }

    // check pkt len and see if send limit is forever
    if (!check_send_limit(
            msg->nsend.len, _time_now, &wait_until_time) &&
        wait_until_time == PKTLAB_TIME_MAX) {
        debug("send limit prevents pkt from sending forever");
        errid = PKTLAB_ESKTPKTDENY; // pkt len too big
        goto fail;
    }

    msg_bufused = nsend_mem_cal(msg);
    if (msg_bufused < 0) {
        errid = PKTLAB_ESKTNOPROTO;
        goto fail;
    }

    if (((uint_fast64_t)_bufused + msg_bufused) > _bufmax) {
        errid = PKTLAB_ESYSNOBUFS;
        goto fail;
    }

    errid = PKTLAB_SUCCESS;

    tmsg = create_txq_msg(msg, truncated);
    nqput(&skt->txq, tmsg);
    _bufused += msg_bufused;

    debug("After new nsend req, buf (%" PRIu32 "/%" PRIu32 ")",
          _bufused, _bufmax);
    assert(_bufused <= _bufmax);
    control_report(errid);
    return;

fail:
    assert(errid != PKTLAB_EUNKFAULT);

    control_report(errid);
    free(msg);
    return;
}

void network_handle_ncap(struct pktlab_message *msg) {
    struct socket * skt = _skt + msg->ncap.sktid;
    struct raw_prog prog = RAW_PROG_INITIALIZER;
    enum pktlab_status ret = PKTLAB_EUNKFAULT;
    enum pktlab_prog_type prog_type;
    int rst;
    uint8_t * filter;
    uint_fast32_t filterlen;

    trace("%s(msg:{filtlen:%" PRIu32 ", endtime:%" PRIu64 "})",
        __func__ , msg->ncap.filterlen, msg->ncap.endtime);

    assert(msg->type == PKTLAB_NCAP_MESSAGE);
    assert(_state == ST_IDLE);

    if (skt->state <= PKTLAB_SKTST_FREE || skt->state >= PKTLAB_SKTST_EOF) {
        ret = PKTLAB_ESKTNOTOPEN;
        goto fail;
    }

    if (skt->netproto != msg->ncap.family ||
        skt->transproto != msg->ncap.proto) {
        ret = PKTLAB_ESKTNOMATCH;
        goto fail;
    }

    if (msg->ncap.filterlen > FILT_BUFSZ) {
        ret = PKTLAB_ESKTBADFILT;
        goto fail;
    }

    if (pktlab_decode_program(
            msg->ncap.filter, msg->ncap.filterlen,
            &prog_type, (void**) &filter, &filterlen) != 0) {
        ret = PKTLAB_ESKTBADFILT;
        goto fail;
    }

    assert(filterlen % PKTLAB_EBPF_INSTR_LEN == 0);

    if (prog_type != PKTLAB_EBPF) {
        ret = PKTLAB_ESKTBADFILT;
        goto fail_filter;
    }

    // Only ebpf supported
    prog.is_ebpf = true;
    prog.progptr = filter; // in pktlab_ebpf_instr fmt (can be NULL)
    prog.proglen = filterlen;

    // prog fd unset
    rst = raw_socket_module->raw_socket_set_recvfilter(
        &prog, skt->robj);

    if (rst != RAW_SUCCESS) {
        if (rst == RAW_BADFILT) {
            ret = PKTLAB_ESKTBADFILT;
            goto fail_filter;
        }

        debug("raw_socket_set_recvfilter failed with %d", rst);

        socket_state_change(msg->ncap.sktid, PKTLAB_SKTST_UNKFAULT);
        ret = PKTLAB_EUNKFAULT;
        goto fail_filter;
    }

    skt->filter_until = msg->ncap.endtime;
    ret = PKTLAB_SUCCESS;

fail_filter:
    free(filter);
fail:
    control_report(ret);
    free(msg);
    return;
}

void network_handle_nctl(struct pktlab_message *msg) {
    const uint_fast8_t sktid = msg->nctl.sktid;
    struct socket *const skt = _skt + sktid;

    trace("%s(msg:{sktid:%" PRIu8 "})", __func__, msg->nctl.sktid);
    assert(msg->type == PKTLAB_NCTL_MESSAGE);
    assert(_state == ST_IDLE);

    if ((skt->netproto != msg->nctl.family) |
        (skt->transproto != msg->nctl.proto)) {
        control_report(PKTLAB_ESKTNOMATCH);
        return;
    }

    if (skt->transproto != PKTLAB_TCP_PROTO) {
        control_report(PKTLAB_ECTLNOTSUP);
        return;
    }

    switch (msg->nctl.tcp.command) {
    case PKTLAB_NCTL_TCP_FIN:
        switch (skt->state) {
        case PKTLAB_SKTST_FREE:
            control_report(PKTLAB_ESKTNOTOPEN);
            return;
        case PKTLAB_SKTST_OPENING:
        case PKTLAB_SKTST_OPEN:
        case PKTLAB_SKTST_EOF:
        case PKTLAB_SKTST_WFIN:
        case PKTLAB_SKTST_END:
            if (!skt->wfin) { // prevent updating wclose_from after set
                skt->wfin = true;
                skt->wclose_from = msg->nctl.time;
            }

            control_report(PKTLAB_SUCCESS);
            return;
        default: // skt in err state
            assert(skt->state < PKTLAB_SKTST_FREE);
            control_report(PKTLAB_ESKTNOTOPEN);
            return;
        }
        break;

    default:
        assert(skt->state < PKTLAB_SKTST_FREE);
        control_report(PKTLAB_ECTLBADMSG);
        return;
        break;
    }
}

// void network_handle_nfin(struct pktlab_message *msg) {
//     const uint_fast8_t sktid = msg->nfin.sktid;
//     struct socket *const skt = _skt + sktid;

//     trace("%s(msg:{sktid:%" PRIu8 "})", __func__, msg->nfin.sktid);
//     assert(msg->type == PKTLAB_NFIN_MESSAGE);
//     assert(_state == ST_IDLE);

//     if (skt->transproto != PKTLAB_TCP_PROTO &&
//         skt->transproto != PKTLAB_UDP_PROTO) {
//         control_report(PKTLAB_BADSKT);
//         return;
//     }

//     switch (skt->state) {
//     case PKTLAB_SKTST_FREE:
//         control_report(PKTLAB_BADSKT);
//         return;
//     case PKTLAB_SKTST_OPENING:
//     case PKTLAB_SKTST_OPEN:
//     case PKTLAB_SKTST_EOF:
//     case PKTLAB_SKTST_WFIN:
//     case PKTLAB_SKTST_END:
//         if (!skt->wfin) { // prevent updating wclose_from after set
//             skt->wfin = true;
//             skt->wclose_from = msg->nfin.time;
//         }

//         control_report(PKTLAB_SUCCESS);
//         return;
//     default: // skt in err state
//         assert(skt->state < PKTLAB_SKTST_FREE);
//         control_report(PKTLAB_SKTERR);
//         return;
//     }
// }

void network_bufmax_read(const struct pktlab_vmem_region *rgn,
                         uint_fast32_t off, uint_fast32_t len, void *buf) {
    uint32_t tmp_bufmax = pktlab_hton32(_bufmax);

    assert(_state >= ST_INIT);

    memcpy(buf, (char *)&tmp_bufmax + off, len);
}

void network_bufused_read(const struct pktlab_vmem_region *rgn,
                          uint_fast32_t off, uint_fast32_t len, void *buf) {
    uint32_t tmp_bufused = pktlab_hton32(_bufused);

    assert(_state >= ST_INIT);

    memcpy(buf, (char *)&tmp_bufused + off, len);
}

void network_transproto_sup_read(const struct pktlab_vmem_region *restrict rgn,
                                 uint_fast32_t off, uint_fast32_t len,
                                 void *restrict buf) {
    uint32_t tmp_sup = 0;

    assert(_state >= ST_INIT);

    if (_raw_sup)
        tmp_sup = pktlab_setbit32(tmp_sup, PKTLAB_RAW_SUP_BIT_INDX);
    if (_tcp_sup)
        tmp_sup = pktlab_setbit32(tmp_sup, PKTLAB_TCP_SUP_BIT_INDX);
    if (_udp_sup)
        tmp_sup = pktlab_setbit32(tmp_sup, PKTLAB_UDP_SUP_BIT_INDX);

    tmp_sup = pktlab_hton32(tmp_sup);
    memcpy(buf, (char *)&tmp_sup + off, len);
}

void network_skt_read(const struct pktlab_vmem_region *restrict rgn,
                      uint_fast32_t off, uint_fast32_t len,
                      void *restrict buf) {
    unsigned int first_skt;
    unsigned int last_skt;
    unsigned int sktid;
    unsigned int read_start;
    unsigned int read_end;
    unsigned int read_off;
    unsigned int read_len;
    char skt_buf[VMEM_SKT_LEN];

    assert(_state >= ST_INIT);

    first_skt = off / VMEM_SKT_LEN;
    last_skt =
        (off + len) / VMEM_SKT_LEN + (((off + len) % VMEM_SKT_LEN) ? 1 : 0);

    debug("network_skt_read:{first_skt:%u, last_skt:%u}", first_skt, last_skt);

    read_off = 0;

    for (sktid = first_skt; sktid < last_skt; ++sktid) {

        read_start = (sktid == first_skt) ? off % VMEM_SKT_LEN : 0;
        read_end = (sktid == last_skt - 1) && ((off + len) % VMEM_SKT_LEN)
                       ? (off + len) % VMEM_SKT_LEN
                       : VMEM_SKT_LEN;
        read_len = read_end - read_start;

        debug("network_skt_read:{sktid:%u, start:%u, end:%u, len:%u, off:%u}",
              sktid, read_start, read_end, read_len, read_off);

        if (read_len == VMEM_SKT_LEN)
            skt_read(sktid, buf + read_off);
        else {
            memset(skt_buf, 0, VMEM_SKT_LEN);
            skt_read(sktid, skt_buf);
            memcpy(buf + read_off, skt_buf + read_start, read_len);
        }

        read_off += read_len;
    }
}

void network_skt_write(const struct pktlab_vmem_region *restrict rgn,
                       uint_fast32_t off, uint_fast32_t len,
                       const void *restrict buf) {
    static uint16_t notifmasks[4] = {
        PKTLAB_NOTIFMASK_NTAG, PKTLAB_NOTIFMASK_NSTAT, PKTLAB_NOTIFMASK_NDATA,
        PKTLAB_NOTIFMASK_NDROP};
    unsigned int first_skt;
    unsigned int last_skt;
    unsigned int sktid;
    uint_fast32_t write_start;
    uint_fast32_t write_end;
    uint_fast32_t write_len;
    uint_fast32_t write_off;
    uint_fast32_t buf_off;
    uint32_t tmp_rbufsize;
    uint16_t tmp_errtag;
    uint16_t old_notifmask;
    uint16_t tmp_notifmask;
    uint16_t notifmask;

    assert(_state >= ST_INIT);

    first_skt = off / VMEM_SKT_LEN;
    last_skt =
        (off + len) / VMEM_SKT_LEN + (((off + len) % VMEM_SKT_LEN) ? 1 : 0);

    debug("network_skt_write:{first_skt:%u, last_skt:%u}", first_skt, last_skt);

    for (sktid = first_skt; sktid < last_skt; ++sktid) {

        if (_skt[sktid].state < PKTLAB_SKTST_OPENING)
            continue; // allow write only when socket is working normally

        write_start = (sktid == first_skt) ? off % VMEM_SKT_LEN : 0;
        write_end = (sktid == last_skt - 1) && ((off + len) % VMEM_SKT_LEN)
                        ? (off + len) % VMEM_SKT_LEN
                        : VMEM_SKT_LEN;

        // write skt rbufsiz
        comp_sktwrite_param(PKTLAB_VMEMADDR_SKTINFO_RBUFSZ,
                            PKTLAB_VMEMADDR_SKTINFO_RBUFSZ_FL, write_start,
                            write_end, &write_off, &buf_off, &write_len);

        tmp_rbufsize = pktlab_hton32(_skt[sktid].rbufmax);
        memcpy(((char *)&tmp_rbufsize) + write_off, buf + buf_off, write_len);
        _skt[sktid].rbufmax = pktlab_ntoh32(tmp_rbufsize);

        // write nsenderr
        comp_sktwrite_param(PKTLAB_VMEMADDR_SKTINFO_NSENDERR,
                            PKTLAB_VMEMADDR_SKTINFO_NSENDERR_FL, write_start,
                            write_end, &write_off, &buf_off, &write_len);

        memcpy(((char *)&_skt[sktid].nsenderr) + write_off, buf + buf_off,
               write_len);

        // write nsend error tidx
        comp_sktwrite_param(PKTLAB_VMEMADDR_SKTINFO_NSENDERRTAG,
                            PKTLAB_VMEMADDR_SKTINFO_NSENDERRTAG_FL, write_start,
                            write_end, &write_off, &buf_off, &write_len);

        tmp_errtag = pktlab_hton16(_skt[sktid].nsenderr_tag);
        memcpy(((char *)&tmp_errtag) + write_off, buf + buf_off, write_len);
        _skt[sktid].nsenderr_tag = pktlab_ntoh16(tmp_errtag);

        // write notification event mask
        comp_sktwrite_param(PKTLAB_VMEMADDR_SKTINFO_NOTIFMASK,
                            PKTLAB_VMEMADDR_SKTINFO_NOTIFMASK_FL, write_start,
                            write_end, &write_off, &buf_off, &write_len);

        old_notifmask = _skt[sktid].notifmask;
        tmp_notifmask = pktlab_hton16(_skt[sktid].notifmask);
        memcpy(((char *)&tmp_notifmask) + write_off, buf + buf_off, write_len);
        _skt[sktid].notifmask = pktlab_ntoh16(tmp_notifmask);

        // flush buffered notification is re-enabled
        if (old_notifmask != _skt[sktid].notifmask) {
            for (int notifmask_idx = 0;
                 notifmask_idx < sizeof(notifmasks) / sizeof(notifmasks[0]);
                 ++notifmask_idx) {
                notifmask = notifmasks[notifmask_idx];
                // used to be inhibited, now enabled
                if ((old_notifmask & notifmask) >
                    (_skt[sktid].notifmask & notifmask)) {
                    flush_notif(sktid, notifmask);
                }
            }
        }

        // write ctlfl
        comp_sktwrite_param(PKTLAB_VMEMADDR_SKTINFO_CTFL,
                            PKTLAB_VMEMADDR_SKTINFO_CTFL_FL, write_start,
                            write_end, &write_off, &buf_off, &write_len);

        memcpy(_skt[sktid].ctlfl + write_off, buf + buf_off, write_len);
    }
}

void network_tstp_read(const struct pktlab_vmem_region *restrict rgn,
                       uint_fast32_t off, uint_fast32_t len,
                       void *restrict buf) {
    assert(_state >= ST_INIT);

    memcpy(buf, (char *)_send_tstp + off, len);
}

//
// INTERNAL FUNCTION DEFINITIONS
//

void state_change(enum state state) {
    debug("network: state changed from %s to %s", state_name(_state),
          state_name(state));
    _state = state;
}

void raw_socket_open(uint_fast8_t sktid, uint_fast8_t netproto,
                     uint_fast32_t rbufsz, struct nopen_addr *addr) {
    struct socket *const skt = _skt + sktid;
    uint8_t laddr[PKTLAB_ADDRLEN_MAX];
    enum raw_error rst;

    trace("network:%s(msg:{sktid:%d,netproto:0x%02x,rbufsz:%u})",
        __func__, (int)sktid, (int)netproto, (unsigned)rbufsz);

    assert(
        skt->state == PKTLAB_SKTST_FREE &&
        addr != NULL && addr->laddrptr != NULL &&
        (netproto == PKTLAB_IP4_PROTO ||
         netproto == PKTLAB_IP6_PROTO));

    skt->robj = create_raw_socket_obj();
    if (skt->robj == NULL) {
        control_report(PKTLAB_EUNKFAULT);
        return;
    }

    if (netproto == PKTLAB_IP4_PROTO) {
        assert(addr->laddrlen == ADDRLEN(netproto));

        // get real addr (expand all 0)
        if (expand_all0_addr(netproto, addr->laddrptr, laddr) < 0) {
            warn("Cannot get first exported local address for all 0 address, "
                "sending PKTLAB_EUNKFAULT");
            control_report(PKTLAB_EUNKFAULT); // potential bad config
            return;
        }

        // TODO: Pass in certificate filters
        // BUGS:
        //   raw_socket_setup_v4 for linux bare will load all cert progs without freeing
        //   FILTER translation from pktlab_ebpf_instr to OS ebpf instr not done yet
        // NEED TO FIX!!!!
        // Check laddr for single socket per interface
        rst = raw_socket_module->raw_socket_setup_v4(laddr, control_fetch_cert_filters(), skt->robj);

        if (rst != RAW_SUCCESS) {
            switch (rst) {
            case RAW_PERM: // endpoint does not have permission
                warn("Endpoint does not have permission to open raw socket, "
                    "sending PKTLAB_EUNKFAULT");
                break;
            case RAW_INVAL: // user exported bad laddr
                warn("Cannot find intf having address %u.%u.%u.%u, "
                    "sending PKTLAB_EUNKFAULT",
                    (unsigned) laddr[0], (unsigned) laddr[1],
                    (unsigned) laddr[2], (unsigned) laddr[3]);
                break;
            default:
                debug("raw_socket_setup_v4 failed with %d", rst);
            }

            control_report(PKTLAB_EUNKFAULT);
            goto fail;
        }

        skt->addrlen = addr->laddrlen;
        memcpy(skt->locaddr, laddr, addr->laddrlen);
    } else {
        assert(addr->laddrlen == PKTLAB_IPV6_ADDR_LEN);
        warn("Endpoint implementation does not support IPv6 yet, "
            "sending PKTLAB_EUNKFAULT");
        control_report(PKTLAB_EUNKFAULT);
        goto fail;
    }

    // socket is gonna get created,
    // consume resource!
    consume_socket_cnt();

    skt->netproto = netproto;
    skt->transproto = PKTLAB_RAW_PROTO;
    skt->rfd = skt->robj->rfd;
    skt->wfd = skt->robj->wfd;
    skt->rbufmax = rbufsz;
    socket_state_change(sktid, PKTLAB_SKTST_OPEN);
    control_report(PKTLAB_SUCCESS);
    return;

fail:
    free(skt->robj);
    return;
}

bool is_nopen_tcp_listen(uint_fast8_t transproto,
                         struct nopen_addr *addr) {
    // must have local address and port
    // must not have remote address or port
    if (transproto == PKTLAB_TCP_PROTO && addr->laddrlen && addr->lportlen &&
        !addr->raddrlen && !addr->rportlen)
        return true;
    return false;
}

bool is_nopen_tcp_connect(uint_fast8_t transproto,
                          struct nopen_addr *addr) {
    // must have remote address and port
    // remote port must not be zero

    assert(addr != NULL);

    trace("%s(laddrlen:%d,lportlen:%d,raddrlen:%d,rportlen:%d)",
          __func__, addr->laddrlen, addr->lportlen,
          addr->raddrlen, addr->rportlen);

    if (transproto == PKTLAB_TCP_PROTO && addr->raddrlen && addr->rportlen &&
        *(uint16_t *)addr->rportptr != 0) {
        debug("tcp connect");
        return true;
    }

    debug("not tcp connect");
    return false;
}

bool is_nopen_udp(uint_fast8_t transproto, struct nopen_addr *addr) {
    // must have local address and port
    // remote address will be ignored during nopen
    if (transproto == PKTLAB_UDP_PROTO && addr->laddrlen && addr->lportlen)
        return true;
    return false;
}

void trans_socket_open(uint_fast8_t sktid, uint_fast8_t netproto,
                       uint_fast32_t rbufsz, uint_fast8_t transproto,
                       struct nopen_addr *addr) {
    struct socket *const skt = _skt + sktid;
    const void *locaddrptr = NULL;
    struct sockaddr_storage sa_mem;
    // uint8_t loc_addr[PKTLAB_ADDRLEN_MAX] = {};
    enum pktlab_status status;
    socklen_t sa_len;
    int sockdomain;
    int socktype;
    // int sockproto;
    bool inprogress = false;
    int sockfd;
    int result;
    uint_fast16_t real_locport;

    const uint_fast8_t laddrlen = addr->laddrlen;
    const void *const laddrptr = addr->laddrptr;
    const uint_fast8_t lportlen = addr->lportlen;
    const void *const lportptr = addr->lportptr;
    const uint_fast8_t raddrlen = addr->raddrlen;
    const void *const raddrptr = addr->raddrptr;
    const uint_fast8_t rportlen = addr->rportlen;
    const void *const rportptr = addr->rportptr;
    uint16_t locport, remport;

    uint_fast8_t expect_addrlen;
    uint_fast8_t expect_portlen;
    uint_fast8_t addrlen;

    // todo: expand all0 addr for transport socket types as well

    trace("network:%s(msg:{sktid:%d,netproto:0x%02x})", __func__, (int)sktid,
          (int)netproto);
    debug("%d %d %d %d", laddrlen, lportlen, raddrlen, rportlen);
    // verify addrlen and portlen
    expect_addrlen = (netproto == PKTLAB_IP4_PROTO) ? 4 : 16;
    expect_portlen = 2;
    addrlen = expect_addrlen;

    if ((laddrlen && laddrlen != expect_addrlen) ||
        (raddrlen && raddrlen != expect_addrlen)) {
        control_report(PKTLAB_ESKTNOADDR);
        goto fail0;
    }

    if ((lportlen && lportlen != expect_portlen) ||
        (rportlen && rportlen != expect_portlen)) {
        control_report(PKTLAB_ESKTNOADDR);
        goto fail0;
    }
    memcpy(&locport, lportptr, lportlen);
    memcpy(&remport, rportptr, rportlen);

    // caller function "network_handle_nopen" performs netproto check
    assert((netproto == PKTLAB_IP4_PROTO) || (netproto == PKTLAB_IP6_PROTO));
    assert((transproto == PKTLAB_TCP_PROTO) ||
           (transproto == PKTLAB_UDP_PROTO));

    sockdomain = netproto == PKTLAB_IP4_PROTO ? AF_INET : AF_INET6;
    socktype = transproto == PKTLAB_TCP_PROTO ? SOCK_STREAM : SOCK_DGRAM;

    sockfd = socket(sockdomain, socktype, 0);

    if (sockfd < 0) {
        debug("socket(PF_INET%s,SOCK_%s,0): %s",
              (sockdomain == PF_INET) ? "" : "6",
              (socktype == SOCK_STREAM) ? "STREAM" : "DGRAM", strerror(errno));
        fatal("%s", strerror(errno));
        control_report(PKTLAB_EUNKFAULT);
        goto fail0;
    }

    // udp and tcp_listen always supply local addr & port
    // tcp_connect may omit local addr, thus bypass binding
    if (is_nopen_udp(transproto, addr) ||
        is_nopen_tcp_listen(transproto, addr) ||
        (is_nopen_tcp_connect(transproto, addr) && addr->laddrlen != 0)) {
        sa_len = mksockaddr((void *)&sa_mem, locport, laddrptr, expect_addrlen);
        debug("lport %d laddr %08x", locport, *(const uint32_t *)laddrptr);
        result = bind(sockfd, (void *)&sa_mem, sa_len);
        debug("succesful bind");

        if (result < 0) {
            debug("bind(%" PRIuFAST16 "): %s", pktlab_ntoh16(locport),
                  strerror(errno));

            switch (errno) {
            case EADDRNOTAVAIL:
                status = PKTLAB_ESKTNOADDR;
                break;
            case EADDRINUSE:
                status = (transproto == PKTLAB_TCP_PROTO) ? PKTLAB_ETCPNOPORT
                                                          : PKTLAB_EUDPNOPORT;
                break;
            default:
                status = PKTLAB_EUNKFAULT;
                break;
            }

            control_report(status);
            goto fail1;
        }
    }

    fcntl(sockfd, F_SETFL, fcntl(sockfd, F_GETFL) | O_NONBLOCK);

    // only tcp connection calls connect
    if (is_nopen_tcp_connect(transproto, addr)) {
        sa_len = mksockaddr((void *)&sa_mem, remport, raddrptr, expect_addrlen);
        debug("remport %d raddr %08x", remport, *(const uint32_t *)raddrptr);
        result = connect(sockfd, (void *)&sa_mem, sa_len);
        if (result < 0) {
            // FIXME: more useful debug msg
            debug("connect(addrlen:%d): %s", (int)expect_addrlen,
                  strerror(errno));

            switch (errno) {
            case EINPROGRESS:
                inprogress = true;
                break;
            default:
                control_report(PKTLAB_EUNKFAULT);
                goto fail1;
            }
        } else {
            inprogress = false;

            if (transproto == PKTLAB_TCP_PROTO)
                if (set_keepalive(sockfd) < 0)
                    warn("SKT[%" PRIuFAST8 "] setsockopt keepalive failed: %s",
                         sktid, strerror(errno));
        }
    } else if (is_nopen_tcp_listen(transproto, addr)) {
        // TODO: tcp_listen
        fatal("unimplemented");
    }

    // Get local address.

    sa_len = sizeof(sa_mem);
    result = getsockname(sockfd, (void *)&sa_mem, &sa_len);

    if (result < 0)
        fatal("getsockname(%d): %s", sockfd, strerror(errno));

    switch (netproto) {
    case PKTLAB_IP4_PROTO:
        locaddrptr = &((struct sockaddr_in *)&sa_mem)->sin_addr;
        real_locport = ((struct sockaddr_in *)&sa_mem)->sin_port;
        break;
    case PKTLAB_IP6_PROTO:
        locaddrptr = &((struct sockaddr_in6 *)&sa_mem)->sin6_addr;
        real_locport = ((struct sockaddr_in6 *)&sa_mem)->sin6_port;
        break;
    default:
        goto fail1;
    }
    debug("locport %d real_locport %d", locport, real_locport);
    assert(locport == 0 || locport == real_locport);

    // will create socket at this point; consume resource
    consume_socket_cnt();

    // Initialize socket structure.

    memset(skt, 0, sizeof(struct socket));

    skt->netproto = netproto;
    skt->transproto = transproto;
    skt->rbufmax = rbufsz;
    skt->rfd = skt->wfd = sockfd;
    memcpy(skt->locaddr, locaddrptr, MIN(sizeof(skt->locaddr), addrlen));
    skt->locport = real_locport;

    if (is_nopen_tcp_connect(transproto, addr)) {
        memcpy(skt->remaddr, addr->raddrptr,
               MIN(sizeof(skt->remaddr), addrlen));
        skt->addrlen = addrlen;
        skt->remport = remport;
    }

    socket_state_change(sktid, (inprogress) ? PKTLAB_SKTST_OPENING
                                            : PKTLAB_SKTST_OPEN);
    control_report(PKTLAB_SUCCESS);
    return;

fail1:
    close(sockfd);
fail0:
    return;
}

void socket_close(uint_fast8_t sktid) {
    struct socket *skt = _skt + sktid;

    trace("network:%s(sktid:%d)", __func__, (int)sktid);

    // nqfilter(&_rxq, freesktid_filt, sktid);
    // TODO: possibly call control to filter the notifq
    nqfilter(&skt->txq, free_txq_msg_filt, 0);
    nqvacuum(&skt->txq);

    // TODO: free the socket notifq

    // TODO: Free filter chain?
    if (skt->transproto == PKTLAB_RAW_PROTO) {
        raw_socket_module->raw_socket_close(skt->robj);
    } else if (skt->wfd == skt->rfd) {
        close(skt->wfd);
    } else {
        close(skt->wfd);
        close(skt->rfd);
    }

    socket_state_change(sktid, PKTLAB_SKTST_FREE);
    memset(skt, 0, sizeof(struct socket));
}

void socket_readable(uint_fast8_t sktid) {
    uint32_t avail_buf;
    struct socket *skt = _skt + sktid;
    struct pktlab_message *msg;
    void *ptr;
    uint_fast8_t netproto = skt->netproto;
    uint_fast8_t transproto = skt->transproto;
    ssize_t readcnt;
    pktlab_time_t recv_time;
    enum raw_error err;
    enum pktlab_socket_state sktst;
    int32_t msg_bufused;
    struct sockaddr srcaddr;
    socklen_t addrlen = sizeof(srcaddr);
    struct raw_prog empty_filter = RAW_PROG_INITIALIZER;
    char *addrptr, *portptr;
    bool truncated;

    trace("network:%s(sktid:%d)", __func__, (int)sktid);

    //todo: fix readcnt may overflow realloc

    // When buffer full,
    // TCP socket: do nothing
    // Other socket: drop all received
    if (have_space(sktid, &avail_buf)) {
        avail_buf = MIN(avail_buf, PKTLAB_NDATA_MAX);
    } else if (transproto == PKTLAB_TCP_PROTO) {
        return;
    }

    msg = safe_malloc(sizeof(struct pktlab_message) + PKTLAB_NDATA_MAX);
    ptr = msg + 1; // points to end of pktlab_message structure

    assert(transproto == PKTLAB_TCP_PROTO || transproto == PKTLAB_UDP_PROTO ||
           transproto == PKTLAB_RAW_PROTO);

    switch (transproto) {
    case PKTLAB_RAW_PROTO:
        readcnt = raw_socket_module->raw_socket_recv(ptr, PKTLAB_NDATA_MAX, 0, skt->robj, &err);
        recv_time = skt->robj->time;
        break;
    case PKTLAB_TCP_PROTO:
        readcnt = read(skt->rfd, ptr, avail_buf);
        recv_time = pktlab_time_now();
        break;
    case PKTLAB_UDP_PROTO:
        readcnt = recvfrom(skt->rfd, ptr, PKTLAB_NDATA_MAX, 0, &srcaddr, &addrlen);
        recv_time = pktlab_time_now();
        break;
    default:
        warn("Unknown transproto: 0x%X", transproto);
        readcnt = 0; // should not happen, will return 0 length ndata message
        recv_time = 0;
        break;
    }

    if (skt->transproto == PKTLAB_RAW_PROTO) {
        if (readcnt <= 0) { // discard 0 len pkt
            debug("raw_read(%d,%u): %s", skt->rfd, PKTLAB_NDATA_MAX,
                  raw_strerror(err));

            if (readcnt < 0 && err != RAW_TRYAGAIN) {
                skt->err = err;
                socket_state_change(sktid, PKTLAB_SKTST_UNKFAULT);
            }

            free(msg);
            return;
        } else if (skt->filter_until < recv_time) { // only recv in filter time
            err = raw_socket_module->raw_socket_set_recvfilter(&empty_filter, skt->robj);
            if (err != RAW_SUCCESS) {
                skt->err = err;
                socket_state_change(sktid, PKTLAB_SKTST_UNKFAULT);
            }

            free(msg);
            return;
        }
    } else {
        if (readcnt <= 0) {
            struct pktlab_message *nstat_msg = safe_malloc(sizeof(*nstat_msg));
            nstat_msg->type = PKTLAB_NSTAT_MESSAGE;
            nstat_msg->nstat.sktid = sktid;
            nstat_msg->nstat.family = netproto;
            nstat_msg->nstat.proto = transproto;
            nstat_msg->nstat.time = recv_time;
            if (readcnt < 0) {
                debug("read(%d,%u): %s", skt->rfd, PKTLAB_NDATA_MAX,
                      strerror(errno));

                skt->err = errno;
                switch (skt->err) {
                case ETIMEDOUT:
                    sktst = PKTLAB_SKTST_TIMEDOUT;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_TIMEOUT;
                    break;
                case ECONNRESET:
                    sktst = PKTLAB_SKTST_RESET;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_TIMEOUT;
                    break;
                case ENETUNREACH:
                case EHOSTUNREACH:
                    sktst = PKTLAB_SKTST_UNREACH;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_UNKNOWN;
                    break;
                default:
                    sktst = PKTLAB_SKTST_UNKFAULT;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_UNKNOWN;
                    break;
                }
            } else {
                switch (skt->state) {
                case PKTLAB_SKTST_OPEN:
                    sktst = PKTLAB_SKTST_EOF;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_PEERCLOSED;
                    break;
                case PKTLAB_SKTST_WFIN:
                    sktst = PKTLAB_SKTST_END;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_PEERCLOSED;
                    break;
                default: // should not happen
                    warn("Unexpected skt state: %d", skt->state);
                    sktst = PKTLAB_SKTST_UNKFAULT;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_UNKNOWN;
                    break;
                }
            }

            push_notif(sktid, nstat_msg, 0, 0, NULL);
            if (nstat_msg->nstat.code != PKTLAB_TCPEV_PEERCLOSED) {
                drain_skt_txq(sktid);
            }

            socket_state_change(sktid, sktst);
            free(msg);
            return;
        }
    }

    truncated = false;
    if (avail_buf < readcnt) {
        // drop data, will not happen for TCP
        _dropfl = true;
        skt->dropstats.pktcnt = limited_add(skt->dropstats.pktcnt, 1);
        skt->dropstats.bytecnt =
            limited_add(skt->dropstats.bytecnt, readcnt - avail_buf);
        readcnt = avail_buf;
        truncated = true;

        if (avail_buf == 0) {
            free(msg);
            return;
        }
    }

    // check if lim allows us to receive
    if (!check_recv_limit(readcnt)) {
        debug("limit prevent us from receiving");

        // TODO: ndrop

        if (!truncated)
            skt->dropstats.pktcnt = limited_add(skt->dropstats.pktcnt, 1);

        skt->dropstats.bytecnt =
            limited_add(skt->dropstats.bytecnt, readcnt);

        free(msg);
        return;
    }

    consume_recv_quota(readcnt);

    if (skt->transproto == PKTLAB_UDP_PROTO) {
        msg = safe_realloc(msg, sizeof(*msg) + readcnt +
                           ADDRLEN(skt->netproto) + PKTLAB_PORTLEN_MAX);
    } else {
        msg = safe_realloc(msg, sizeof(*msg) + readcnt);
    }

    msg->type = PKTLAB_NDATA_MESSAGE;
    msg->ndata.ptr = msg + 1;
    msg->ndata.len = readcnt;
    msg->ndata.sktid = sktid;
    msg->ndata.family = skt->netproto;
    msg->ndata.proto = skt->transproto;
    msg->ndata.time = recv_time;
    if (skt->transproto == PKTLAB_UDP_PROTO) {
        if (skt->netproto == PKTLAB_IP4_PROTO) {
            addrptr = (char *)(msg + 1) + readcnt;
            portptr = (char *)(msg + 1) + readcnt + ADDRLEN(skt->netproto);
            memcpy(addrptr, &((struct sockaddr_in *)&srcaddr)->sin_addr.s_addr,
                   ADDRLEN(skt->netproto));
            memcpy(portptr, &((struct sockaddr_in *)&srcaddr)->sin_port, PKTLAB_PORTLEN_MAX);

            msg->ndata.udp.raddrptr = addrptr;
            msg->ndata.udp.raddrlen = ADDRLEN(skt->netproto);
            msg->ndata.udp.rportptr = portptr;
            msg->ndata.udp.rportlen = PKTLAB_PORTLEN_MAX;
        } else {
            // TODO: ipv6 udp
            fatal("not implemented");
        }
    }

    msg_bufused = ndata_mem_cal(msg);

    // nqput(&_rxq, msg);

    skt->rbufused += readcnt;
    _bufused += msg_bufused;
    push_notif(sktid, msg, msg_bufused, readcnt, skt_mem_update);

    debug("After new ndata msg, rbuf (%" PRIu32 "/%" PRIu32 "), "
          "buf (%" PRIu32 "/%" PRIu32 ")",
          skt->rbufused, skt->rbufmax, _bufused, _bufmax);
    assert(_bufused <= _bufmax && _skt->rbufused <= _skt->rbufmax);
}

uint_fast8_t get_notif_msg_sktid(struct pktlab_message *msg) {
    switch (msg->type) {
    case PKTLAB_NTAG_MESSAGE:
        return msg->ntag.sktid;
    case PKTLAB_NSTAT_MESSAGE:
        return msg->nstat.sktid;
    case PKTLAB_NDATA_MESSAGE:
        return msg->ndata.sktid;
    case PKTLAB_NDROP_MESSAGE:
        return msg->ndrop.sktid;
    default:
        assert(0); // unexpected msg type, bug
    }
}

void skt_mem_update(void * aux) {
    struct notif_aux *na = aux;
    uint_fast8_t sktid;
    struct socket *skt;

    assert(aux != NULL);

    sktid = get_notif_msg_sktid(na->msg);
    skt = _skt + sktid;

    assert(
        na->skt_used <= skt->rbufused &&
        na->glob_used <= _bufused);

    skt->rbufused -= na->skt_used;
    _bufused -= na->glob_used;
    free(aux);
    return;
}

void pop_skt_txq_n_update_buf_usage (
    uint_fast8_t sktid, pktlab_time_t time, enum pktlab_status errid)
{
    struct socket *skt = _skt + sktid;
    struct txq_msg *tmsg;
    struct pktlab_message *ntag_msg;
    int32_t msg_bufused;

    trace("network:%s(sktid:%" PRIuFAST8 ")", __func__, sktid);

    tmsg = nqget(&skt->txq);
    if (tmsg == NULL)
        return;

    assert(
        tmsg->msg->type == PKTLAB_NSEND_MESSAGE &&
        tmsg->msg->nsend.sktid == sktid);

    // non-zero tag: controller expects ntag
    if (tmsg->msg->nsend.tag) {
        ntag_msg = safe_malloc(sizeof(*ntag_msg));
        ntag_msg->type = PKTLAB_NTAG_MESSAGE;
        ntag_msg->ntag.sktid = sktid;
        ntag_msg->ntag.tag = tmsg->msg->nsend.tag;
        ntag_msg->ntag.errid = errid;
        ntag_msg->ntag.time = time;
        push_notif(sktid, ntag_msg, 0, 0, NULL);
    }

    msg_bufused = nsend_mem_cal(tmsg->msg);
    assert(msg_bufused >= 0);
    _bufused -= msg_bufused;

    free(tmsg->msg);
    free(tmsg);

    debug("Removed nsend req size %" PRId32 ", current buf (%" PRIu32
        "/%" PRIu32 ")", msg_bufused, _bufused, _bufmax);
    assert(_bufused <= _bufmax);
    return;
}

void drain_skt_txq(uint_fast8_t sktid) {
    struct socket *skt = _skt + sktid;

    trace("network:%s(sktid:%" PRIuFAST8 ")", __func__, sktid);

    while (nqpeek(&skt->txq) != NULL)
        pop_skt_txq_n_update_buf_usage(
            sktid, pktlab_time_now(), PKTLAB_EPKTNOTSENT);

    return;
}

void socket_writable(uint_fast8_t sktid) {
    struct socket *skt = _skt + sktid;
    struct txq_msg *tmsg;
    struct txq_msg_aux *tmsg_aux;
    struct pktlab_message *msg;
    uint_fast32_t off = skt->msgoff;
    enum pktlab_socket_state sktst;
    pktlab_time_t wait_until;
    ssize_t writecnt;
    size_t len = 0, salen;
    const uint_fast8_t transproto = skt->transproto;
    pktlab_time_t msg_time, send_time;
    uint_fast8_t msg_fam, msg_proto;
    uint_fast16_t msg_tag;
    enum raw_error err;
    struct sockaddr_storage dest;
    struct sockaddr_in6 *dest6 = (struct sockaddr_in6 *) &dest;
    struct sockaddr_in *dest4 = (struct sockaddr_in *) &dest;

    // TODO divide tcp udp logic from network module

    trace("network:%s(sktid:%d)", __func__, (int)sktid);

    while ((tmsg = nqpeek(&skt->txq))) {
        assert(tmsg->msg->type == PKTLAB_NSEND_MESSAGE);
        //show_skt();

        msg = tmsg->msg;
        msg_fam = msg->nsend.family;
        msg_proto = msg->nsend.proto;
        msg_time = msg->nsend.time;
        msg_tag = msg->nsend.tag;

        assert(
            msg->nsend.sktid == sktid &&
            msg->nsend.family == skt->netproto &&
            msg->nsend.proto == transproto);

        // Time for nfin ops
        // Last clause only allows for sending when msg_time < wclose_from
        // msg will be discarded when wclose_from == msg_time <= _time_now
        if (skt->wfin && skt->wclose_from <= _time_now &&
            skt->wclose_from <= msg_time)
            break;

        // limit prevent us to send anymore
        if (!check_send_limit(
                msg->nsend.len, _time_now, &wait_until))
            return;

        if (msg_time > _time_now)
            return; // too early for sending

        switch (transproto) {
        case PKTLAB_RAW_PROTO: // raw only perform send once per nsend
            len = msg->nsend.len;
            writecnt = raw_socket_module->raw_socket_send((const void *)msg->nsend.ptr, len, 0, skt->robj, &err);
            send_time = pktlab_time_now();
            break;
        case PKTLAB_TCP_PROTO:
            len = msg->nsend.len;
            writecnt = write(skt->wfd, msg->nsend.ptr + off, len - off);
            send_time = pktlab_time_now();
            break;
        case PKTLAB_UDP_PROTO: // TODO: think about partial write case
            len = msg->nsend.len;
            if (msg_fam == PKTLAB_IP4_PROTO) {
                salen = sizeof(*dest4);
                dest4->sin_family = AF_INET;
                assert(msg->nsend.udp.raddrlen == 4);
                assert(msg->nsend.udp.rportlen == 2);
                memcpy(&dest4->sin_addr.s_addr, msg->nsend.udp.raddrptr,
                       msg->nsend.udp.raddrlen);
                memcpy(&dest4->sin_port, msg->nsend.udp.rportptr,
                       msg->nsend.udp.rportlen);
            } else if (msg_fam == PKTLAB_IP6_PROTO) {
                salen = sizeof(*dest6);
                dest6->sin6_family = AF_INET6;
                // TODO: ipv6 unimplemented
                assert(false);
            }
            writecnt = sendto(skt->wfd, msg->nsend.ptr + off, len - off, 0,
                              (struct sockaddr *) &dest, salen);
            send_time = pktlab_time_now();
            break;
        default:
            assert(0); // bug
        }

        switch (transproto) {
        case PKTLAB_RAW_PROTO:
            if (writecnt < 0) {
                debug("raw_socket_send(%d,%u): %s", skt->wfd, len,
                    raw_strerror(err));

                if (err == RAW_TRYAGAIN)
                    return; // recorded off is always 0

                skt->err = err;
                skt->nsenderr = err; // todo: fix nsenderr

                // only a single pkt send fail
                if (err == RAW_BADPKT || err == RAW_INVAL) {
                    debug("Single raw pkt sending failed, discard and try next pkt");
                    pop_skt_txq_n_update_buf_usage(sktid, send_time, PKTLAB_EPKTNOTSENT);
                    continue; // try next pkt
                }

                // nonrecoverable error, fail all remaining sends
                drain_skt_txq(sktid);
                socket_state_change(sktid, PKTLAB_SKTST_UNKFAULT);
                return;
            }

            tmsg_aux = tmsg->aux;
            assert(tmsg_aux != NULL);

            pop_skt_txq_n_update_buf_usage(
                sktid, send_time, (tmsg_aux->truncated) ? \
                PKTLAB_EPKTTRUNC:PKTLAB_SUCCESS);
            consume_send_quota(writecnt, send_time);
            break;
        case PKTLAB_TCP_PROTO:
        case PKTLAB_UDP_PROTO:
            if (writecnt < 0) {
                debug("write(%d,%u): %s", skt->wfd, len, strerror(errno));
                if (errno == EWOULDBLOCK)
                    goto try_later;

                skt->err = errno;
                // skt->errtidx = msg->nsend.tidx;
                struct pktlab_message *nstat_msg =
                    safe_malloc(sizeof(*nstat_msg));
                nstat_msg->type = PKTLAB_NSTAT_MESSAGE;
                nstat_msg->nstat.sktid = sktid;
                nstat_msg->nstat.family = msg_fam;
                nstat_msg->nstat.proto = msg_proto;
                nstat_msg->nstat.time = send_time;

                switch (skt->err) {
                case ETIMEDOUT:
                    sktst = PKTLAB_SKTST_TIMEDOUT;
                    skt->nsenderr = PKTLAB_NSEND_TIMEDOUT;
                    skt->nsenderr_tag = msg_tag;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_TIMEOUT;
                    break;
                case ECONNRESET:
                    sktst = PKTLAB_SKTST_RESET;
                    skt->nsenderr = PKTLAB_NSEND_RESET;
                    skt->nsenderr_tag = msg_tag;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_RESET;
                    break;
                case ENETUNREACH:
                case EHOSTUNREACH:
                    sktst = PKTLAB_SKTST_UNREACH;
                    skt->nsenderr = PKTLAB_NSEND_UNREACH;
                    skt->nsenderr_tag = msg_tag;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_UNKNOWN;
                    break;
                default:
                    sktst = PKTLAB_SKTST_UNKFAULT;
                    skt->nsenderr = PKTLAB_NSEND_UNKFAULT;
                    skt->nsenderr_tag = msg_tag;
                    nstat_msg->nstat.code = PKTLAB_TCPEV_UNKNOWN;
                    break;
                }

                push_notif(sktid, nstat_msg, 0, 0, NULL);

                // drain the txq and send ntag for all msgs because
                // connection doens't exist
                drain_skt_txq(sktid);
                socket_state_change(sktid, sktst);
                return;
            }

            if (writecnt + off < len) {
                off += writecnt;
                consume_send_quota(writecnt, send_time);
            } else {
                // send ntag when all bytes pushed out
                // TODO: think about UDP truncate case
                pop_skt_txq_n_update_buf_usage(sktid, send_time, PKTLAB_SUCCESS);
                off = 0;

                consume_send_quota(writecnt, send_time);
            }
            break;
        default:
            assert(0); // bug
        }
    }

    // we get here when transmit queue drained, or nfin ops condition met, or
    // both
    if (skt->wfin && skt->wclose_from <= _time_now)
        fin_socket(sktid);

    // for tryagain the offset is updated; for nfin ops this does not matter
try_later:
    skt->msgoff = off;
}

void socket_connected(uint_fast8_t sktid) {
    struct socket *const skt = _skt + sktid;
    int result, sockerr = 0;
    socklen_t optlen;

    trace("network:%s(msg:{sktid:%d})", __func__, (int)sktid);
    assert(skt->transproto != PKTLAB_RAW_PROTO);

    // Use socket error to determine outcome of open.
    // wfd is the same as rfd for non raw sockets

    optlen = sizeof(sockerr);
    result = getsockopt(skt->wfd, SOL_SOCKET, SO_ERROR, &sockerr, &optlen);

    if (result < 0)
        fatal("%s", strerror(errno));

    switch (sockerr) {
    case 0:;
        struct pktlab_message *msg = safe_malloc(sizeof(*msg));
        msg->type = PKTLAB_NSTAT_MESSAGE;
        msg->nstat.sktid = sktid;
        msg->nstat.family = skt->netproto;
        msg->nstat.proto = skt->transproto;
        msg->nstat.time = pktlab_time_now();
        msg->nstat.code = PKTLAB_TCPEV_ESTABLISHED;
        msg->nstat.tcp_established.raddrptr = skt->remaddr;
        msg->nstat.tcp_established.raddrlen = skt->addrlen;
        msg->nstat.tcp_established.rportptr = &skt->remport;
        msg->nstat.tcp_established.rportlen = sizeof(skt->remport);
        push_notif(sktid, msg, 0, 0, NULL);

        socket_state_change(sktid, PKTLAB_SKTST_OPEN);

        if (skt->transproto == PKTLAB_TCP_PROTO)
            if (set_keepalive(skt->wfd) < 0)
                warn("SKT[%" PRIuFAST8 "] setsockopt keepalive failed: %s",
                     sktid, strerror(errno));
        break;
    case ECONNREFUSED:
        socket_state_change(sktid, PKTLAB_SKTST_REFUSED);
        break;
    case ETIMEDOUT:
        socket_state_change(sktid, PKTLAB_SKTST_TIMEDOUT);
        break;
    case EHOSTUNREACH:
        socket_state_change(sktid, PKTLAB_SKTST_UNREACH);
        break;
    default:
        warn("SKT %d: %s.", (int)sktid, strerror(sockerr));
        socket_state_change(sktid, PKTLAB_SKTST_UNKFAULT);
        break;
    }
}

void socket_state_change(unsigned int sktid, enum pktlab_socket_state state) {
    debug("network: socket %u state changed from %s to %s", sktid,
          pktlab_sktstate_name(_skt[sktid].state), pktlab_sktstate_name(state));
    _skt[sktid].state = state;
}

void clear_dropstats(void) {
    unsigned int sktid;

    for (sktid = 0; sktid < SKT_CNT; sktid++)
        memset(&_skt[sktid].dropstats, 0, sizeof(_skt[sktid].dropstats));
}

bool free_txq_msg_filt(void *ptr, uintptr_t aux) {
    struct txq_msg * tmsg;

    assert(ptr != NULL);

    tmsg = ptr;
    free(tmsg->msg);
    free(ptr);
    return false;
}

// bool freesktid_filt(void *ptr, uintptr_t sktid) {
//     struct pktlab_message *msg = ptr;
//     if (msg->ndata.sktid == sktid) {
//         free(ptr);
//         return false;
//     } else
//         return true;
// }

inline const char *state_name(enum state state) {
    static const char *const names[] = {
        [ST_UNKNOWN] = "UNKNOWN", [ST_INIT] = "INIT", [ST_IDLE] = "IDLE"};
    const char *name = NULL;

    if (state < sizeof(names) / sizeof(names[0]))
        name = names[state];

    return (name != NULL) ? name : names[ST_UNKNOWN];
}

void skt_read(uint8_t sktid, void *buf) {
    char *ptr = buf;
    struct socket *skt = _skt + sktid;
    int8_t state, nsenderr;

    memcpy(ptr, &skt->netproto, sizeof(uint8_t));
    ptr += 0x1;

    memcpy(ptr, &skt->transproto, sizeof(uint8_t));
    ptr += 0x1;

    state = _skt[sktid].state;
    memcpy(ptr, &state, sizeof(int8_t));
    ptr += 0x1;

    ptr += 0xD; // padding

    memcpy(ptr, _skt[sktid].locaddr, PKTLAB_ADDRLEN_MAX);
    ptr += PKTLAB_ADDRLEN_MAX;

    memcpy(ptr, _skt[sktid].remaddr, PKTLAB_ADDRLEN_MAX);
    ptr += PKTLAB_ADDRLEN_MAX;

    memcpy(ptr, &(_skt[sktid].locport), sizeof(uint16_t));
    ptr += sizeof(uint16_t);

    memcpy(ptr, &(_skt[sktid].remport), sizeof(uint16_t));
    ptr += sizeof(uint16_t);

    ptr += 0xC; // padding

    pktlab_set32n(ptr, _skt[sktid].rbufmax);
    ptr += sizeof(uint32_t);

    pktlab_set32n(ptr, _skt[sktid].rbufused);
    ptr += sizeof(uint32_t);

    ptr += 0x8; // padding

    pktlab_set32n(ptr, _skt[sktid].dropstats.bytecnt);
    ptr += sizeof(uint32_t);
    pktlab_set32n(ptr, _skt[sktid].dropstats.pktcnt);
    ptr += sizeof(uint32_t);

    ptr += 0x8; // padding

    nsenderr = _skt[sktid].nsenderr;
    memcpy(ptr, &nsenderr, sizeof(int8_t));
    pktlab_set16n(ptr + sizeof(int8_t), _skt[sktid].nsenderr_tag);
    ptr += sizeof(int8_t);
    ptr += sizeof(uint16_t);

    ptr += 0xD; // padding

    pktlab_set16n(ptr, _skt[sktid].notifmask);
    ptr += sizeof(uint16_t);

    ptr += 0x8E; // padding

    memcpy(ptr, _skt[sktid].ctlfl, CTLFL_CNT);
    ptr += CTLFL_CNT;

    memset(ptr, 0, VMEM_SKT_LEN - (ptr - (char *)buf));
}

void comp_sktwrite_param(uint_fast32_t field_start, uint_fast32_t field_len,
                         uint_fast32_t write_start, uint_fast32_t write_end,
                         uint_fast32_t *write_off, uint_fast32_t *buf_off,
                         uint_fast32_t *write_len) {
    *write_off = (write_start > field_start) ? write_start - field_start : 0;
    *buf_off = (field_start > write_start) ? field_start - write_start : 0;
    *write_len =
        MIN(field_start + field_len, write_end) > MAX(field_start, write_start)
            ? MIN(field_start + field_len, write_end) -
                  MAX(field_start, write_start)
            : 0;
}

bool fin_socket(uint8_t sktid) {
    struct socket *const skt = _skt + sktid;
    enum pktlab_socket_state status;
    bool ret;

    assert(skt->wfin && skt->wclose_from <= _time_now);

    assert(skt->state == PKTLAB_SKTST_OPEN || skt->state == PKTLAB_SKTST_EOF ||
           skt->state == PKTLAB_SKTST_WFIN || skt->state == PKTLAB_SKTST_END);

    assert(skt->transproto == PKTLAB_TCP_PROTO ||
           skt->transproto == PKTLAB_UDP_PROTO);

    ret = false;
    if (shutdown(skt->wfd, SHUT_WR) < 0) {
        skt->err = errno;
        status = PKTLAB_SKTST_UNKFAULT;
        ret = true;
    } else {
        status = skt->state;
        switch (skt->state) {
        case PKTLAB_SKTST_OPEN:
            status = PKTLAB_SKTST_WFIN;
            break;
        case PKTLAB_SKTST_EOF:
            status = PKTLAB_SKTST_END;
            break;
        default:
            break; // do nothing for WFIN & END
        }
    }

    // Free all queued nsends
    skt->nsenderr = PKTLAB_NSEND_WFIN;
    drain_skt_txq(sktid);
    socket_state_change(sktid, status);
    return ret;
}

// void show_skt(void) {
//     int sktid;
//     struct socket *skt;
//     for (sktid = 0; skt = _skt + sktid, sktid < SKT_CNT; sktid++) {
//         if (skt->state <= PKTLAB_SKTST_FREE)
//             continue;
//
//         debug("SKT[%d]:{rfd:%d, wfd:%d, txqlen:%zu, state:%s}", sktid, skt->rfd,
//               skt->wfd, nqlen(&skt->txq), pktlab_sktstate_name(skt->state));
//     }
// }

int32_t nsend_mem_cal(const struct pktlab_message *restrict msg) {
    int32_t msglen;
    uint_fast8_t transproto;

    assert(msg->type == PKTLAB_NSEND_MESSAGE);

    transproto = msg->nsend.proto;
    debug("%s: transproto (%" PRIuFAST8 ")", __func__, transproto);

    // nsend common msg fmt:
    // sktid 1B, family 1B, proto 1B, time 8B, tag 2B, addr/port len if
    // applicable and len of data

    switch (transproto) {
    case PKTLAB_RAW_PROTO:
        msglen = SEND_RAW_BUF_OVERHEAD + PKTLAB_HLEN + \
                 1 + 1 + 1 + 8 + 2 + msg->nsend.len;
        break;
    case PKTLAB_TCP_PROTO:
        msglen = SEND_BUF_OVERHEAD + PKTLAB_HLEN + \
                 1 + 1 + 1 + 8 + 2 + msg->nsend.len;
        break;
    case PKTLAB_UDP_PROTO:
        msglen = SEND_BUF_OVERHEAD + PKTLAB_HLEN + \
                 1 + 1 + 1 + 8 + 2 + msg->nsend.udp.raddrlen + \
                 msg->nsend.udp.rportlen + msg->nsend.len;
        break;
    default:
        warn("%s: unknown transproto: 0x%X", __func__, transproto);
        return -1;
    }

    debug("%s: transproto: %d msg_bufused: %" PRId32
          " datalen: %" PRId32, __func__,
          (int) transproto, msglen, msg->nsend.len);
    return msglen;
}

int32_t ndata_mem_cal(const struct pktlab_message *restrict msg) {
    int32_t ret;

    assert(msg->type == PKTLAB_NDATA_MESSAGE);

    if (msg->ndata.proto == PKTLAB_UDP_PROTO) {
        ret = RECV_BUF_OVERHEAD + msg->ndata.len + \
              msg->ndata.udp.raddrlen + msg->ndata.udp.rportlen;
    } else
        ret = RECV_BUF_OVERHEAD + msg->ndata.len;

    debug("%s: (is UDP: %d) msg_bufused: %" PRId32 " datalen: %" PRIu32,
        __func__, (msg->ndata.proto == PKTLAB_UDP_PROTO),
        ret, msg->ndata.len);
    return ret;
}

int32_t ntag_mem_cal(const struct pktlab_message *restrict msg) {
    assert(msg->type == PKTLAB_NTAG_MESSAGE);
    return RECV_BUF_OVERHEAD;
}

int32_t nstat_mem_cal(const struct pktlab_message *restrict msg) {
    assert(msg->type == PKTLAB_NSTAT_MESSAGE);
    return RECV_BUF_OVERHEAD;
}

int32_t ndrop_mem_cal(const struct pktlab_message *restrict msg) {
    assert(msg->type == PKTLAB_NDROP_MESSAGE);
    return RECV_BUF_OVERHEAD;
}

bool have_space(uint_fast8_t sktid, uint32_t * spaceleft) {
    struct socket *skt = _skt + sktid;
    uint32_t min_msg_sz;

    assert(skt->state > PKTLAB_SKTST_FREE);

    // TODO: change to use ndata_mem_cal
    min_msg_sz = RECV_BUF_OVERHEAD;
    if (skt->transproto == PKTLAB_UDP_PROTO)
        min_msg_sz += ADDRLEN(skt->netproto)+PKTLAB_PORTLEN_MAX;

    if (spaceleft != NULL)
        *spaceleft = 0;

    if (skt->rbufmax > skt->rbufused && _bufmax - _bufused > min_msg_sz){
        if (spaceleft != NULL)
            *spaceleft = MIN(
                skt->rbufmax-skt->rbufused, _bufmax-_bufused-min_msg_sz);
        return true;
    }

    return false;
}

uint32_t limited_add(const uint32_t target, const ssize_t incre) {
    assert(incre >= 0);
    if ((uintmax_t)target + incre > UINT32_MAX) {
        info("Wrap over detected for %" PRIu32 " and %zd", target, incre);
        return UINT32_MAX;
    }

    return target + incre;
}

bool is_invalid_netproto(uint_fast8_t netproto) {
    switch (netproto) {
    case PKTLAB_IP4_PROTO:
    case PKTLAB_IP6_PROTO:
        return false;
    default:
        return true;
    }
}

bool is_invalid_transproto(uint_fast8_t transproto) {
    switch (transproto) {
    case PKTLAB_RAW_PROTO:
    case PKTLAB_TCP_PROTO:
    case PKTLAB_UDP_PROTO:
        return false;
    default:
        return true;
    }
}

void check_proto_sup(uint_fast8_t netproto, uint_fast8_t transproto,
                     bool * netproto_sup, bool * transproto_sup) {
    assert(netproto_sup != NULL && transproto_sup != NULL);

    *netproto_sup = true;
    switch (netproto) {
    case PKTLAB_IP4_PROTO:
        if (!_ip4_sup)
            *netproto_sup = false;
        break;
    case PKTLAB_IP6_PROTO:
        if (!_ip6_sup)
            *netproto_sup = false;
        break;
    default:
        assert(0); // bug
    }

    *transproto_sup = true;
    switch (transproto) {
    case PKTLAB_RAW_PROTO:
        if (!_raw_sup)
            *transproto_sup = false;
        break;
    case PKTLAB_TCP_PROTO:
        if (!_tcp_sup)
            *transproto_sup = false;
        break;
    case PKTLAB_UDP_PROTO:
        if (!_udp_sup)
            *transproto_sup = false;
        break;
    default:
        assert(0); // bug
    }

    return;
}

struct txq_msg *
create_txq_msg(struct pktlab_message *msg, bool truncated) {
    size_t totalmem;
    struct txq_msg * tm;
    struct txq_msg_aux * tma;

    assert(msg != NULL && msg->type == PKTLAB_NSEND_MESSAGE);

    // truncated currently ignored for non nsend raw proto msg
    // may want to add for UDP as well

    totalmem = sizeof(*tm);
    if (msg->nsend.proto == PKTLAB_RAW_PROTO)
        totalmem += sizeof(*tma);

    tm = safe_malloc(totalmem);
    tm->msg = msg;
    tm->aux = NULL;

    if (msg->nsend.proto == PKTLAB_RAW_PROTO) {
        tma = tm->aux = (struct txq_msg_aux *) (tm+1);
        tma->truncated = truncated;
    }

    return tm;
}

struct notif_aux *
create_notif_aux(struct pktlab_message *msg, uint16_t notifmask,
                 int_fast32_t glob_used, int_fast32_t skt_used,
                 send_msg_cb_t memupdate) {
    struct notif_aux * aux;

    assert(msg != NULL);

    aux = safe_malloc(sizeof(*aux));
    aux->msg = msg;
    aux->notifmask = notifmask;
    aux->glob_used = glob_used;
    aux->skt_used = skt_used;
    aux->memupdate = memupdate;
    return aux;
}

void push_notif(uint8_t sktid, struct pktlab_message *msg,
                int_fast32_t glob_used, int_fast32_t skt_used,
                send_msg_cb_t memupdate) {
    struct socket *skt = _skt + sktid;
    uint16_t notifmask = 0;
    struct notif_aux * aux;

    assert(msg != NULL);

    switch (msg->type) {
    case PKTLAB_NTAG_MESSAGE:
        notifmask = PKTLAB_NOTIFMASK_NTAG;
        break;
    case PKTLAB_NSTAT_MESSAGE:
        notifmask = PKTLAB_NOTIFMASK_NSTAT;
        break;
    case PKTLAB_NDATA_MESSAGE:
        notifmask = PKTLAB_NOTIFMASK_NDATA;
        break;
    case PKTLAB_NDROP_MESSAGE:
        notifmask = PKTLAB_NOTIFMASK_NDROP;
        break;
    default:
        // non-notification msg type
        fatal("unexpected notification type: %d\n", msg->type);
        return;
    }

    aux = create_notif_aux(
        msg, notifmask, glob_used, skt_used, memupdate);

    if (skt->notifmask & notifmask) { // notification masked, queue
        nqput(&skt->notifq, aux);
    } else if (memupdate == NULL) {
        control_push_msgq(msg); // no accounting case
        free(aux);
    } else
        control_push_msgq_cb(msg, memupdate, aux);

    return;
}

bool notif_filt(void *elt, uintptr_t aux) {
    struct notif_aux * na = elt;
    uint16_t notifmask = aux;

    if (na->notifmask != notifmask)
        return true; // notification still buffered

    if (na->memupdate == NULL) {
        control_push_msgq(na->msg);
        free(elt);
    } else {
        // elt/na should be freed by memupdate
        control_push_msgq_cb(na->msg, na->memupdate, elt);
    }

    return false;
}

// flush all buffered notification of a certain type
void flush_notif(uint8_t sktid, uint16_t notifmask) {
    struct socket *skt = _skt + sktid;

    nqfilter(&skt->notifq, notif_filt, (uintptr_t)notifmask);
    nqvacuum(&skt->notifq);
}

bool is_exported_laddr(uint_fast8_t prfam, const void * ptr) {
    uint_fast16_t i;
    uint8_t laddr[PKTLAB_ADDRLEN_MAX];
    static const uint8_t all0[PKTLAB_ADDRLEN_MAX] = {};

    assert((prfam == PKTLAB_IP4_PROTO || prfam == PKTLAB_IP6_PROTO) && ptr != 0);

    if (memcmp(ptr, all0, ADDRLEN(prfam)) == 0)
        return !vmem_indx_get_laddr(prfam, 0, laddr);

    for (i = 0; i < PKTLAB_MAX_LADDR_CNT; ++i) {
        if (vmem_indx_get_laddr(prfam, i, laddr))
            return false;

        if (memcmp(ptr, laddr, ADDRLEN(prfam)) == 0)
            return true;
    }

    return false; // no match among all exported laddr
}

int expand_all0_addr(uint_fast8_t prfam,
                     const uint8_t * passed_addr, uint8_t * actual_laddr)
{
    static const uint8_t all0[PKTLAB_ADDRLEN_MAX] = {};

    assert(
        (prfam == PKTLAB_IP4_PROTO || prfam == PKTLAB_IP6_PROTO) &&
        passed_addr != NULL && actual_laddr != NULL);

    if (memcmp(passed_addr, all0, ADDRLEN(prfam)) == 0) {
        if (vmem_indx_get_laddr(prfam, 0, actual_laddr)) {
            debug("Failed to get addr at indx 0 for all 0 addr");
            return -1;
        }
    } else {
        memcpy(actual_laddr, passed_addr, ADDRLEN(prfam));
    }

    return 0;
}