// raw_bare_linux.c
// pktlab raw socket linux platform impl. without stealing
// Using only kernel mechanisms
//

// TODO: Fix IPv6 does not support send raw packets

// POSIX-compatible include headers
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <netinet/in.h>
#include <limits.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#include <pktlab.h>

#include "console.h"
#include "filter.h"
#include "net.h"
#include "network_util.h"
#include "raw.h"
#include "util.h"

// Linux dependent headers
#include <ifaddrs.h>
#include <linux/bpf.h>
#include <linux/filter.h>
#include <linux/if_packet.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/syscall.h>

#include "raw_linux.h"

//
// INTERNAL MACROS
//

#define LOG_BUF_SZ 0x10000

#define MIN_IP4HDR_LEN 20
#define MIN_IP6HDR_LEN 40

#define bpf(cmd,attr,size) syscall(SYS_bpf, (cmd), (attr), (size))

//
// INTERNAL GLOBAL VARIABLE
//

static struct ifindex_ll * if_ll;

//
// INTERNAL FUNCTION DECLARATIONS
//

static int bare_raw_socket_setup_v4 (
    const uint8_t * intfaddr,
    struct raw_prog * cert_prog_ll,
    struct raw_socket_obj * obj);

static ssize_t bare_raw_socket_send (
    const void * buf, size_t len, int flags,
    struct raw_socket_obj * obj, enum raw_error * err);
static ssize_t bare_raw_socket_recv (
    void * buf, size_t len, int flags,
    struct raw_socket_obj * obj, enum raw_error * err);

static int bare_raw_socket_set_recvfilter (
    struct raw_prog * filter,
    struct raw_socket_obj * obj);

static int bare_clean_recvbuf(struct raw_socket_obj * obj);

static void bare_raw_socket_close(struct raw_socket_obj * obj);

static bool is_for_self (
    const uint8_t * pkt, size_t pktlen,
    struct raw_socket_obj_aux * aux);

static int load_ebpf_filter (
    const void * filter, uint_fast32_t filterlen);

//
// EXPORTED FUNCTION DEFINITIONS
//

const struct raw_socket_method_t * raw_socket_nosteal_method(void) {
    static const struct raw_socket_method_t ret = {
        .raw_socket_setup_v4 = bare_raw_socket_setup_v4,
        .raw_socket_setup_v6 = NULL, // ipv6 not supported yet
        .raw_socket_close = bare_raw_socket_close,
        .raw_socket_send = bare_raw_socket_send,
        .raw_socket_recv = bare_raw_socket_recv,
        .raw_socket_set_recvfilter = bare_raw_socket_set_recvfilter,
    };

    return &ret;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int bare_raw_socket_setup_v4 (
    const uint8_t * intfaddr,
    struct raw_prog * cert_prog_ll,
    struct raw_socket_obj * obj)
{
    int sendfd, recvfd, ifindex, rst;
    struct sockaddr_ll sll = {
        .sll_family=AF_PACKET, .sll_protocol=htons(ETH_P_ALL)
    };
    struct sockaddr_in sin = { .sin_family=AF_INET };
    struct ifindex_ll * iftra;
    struct raw_socket_obj tmpobj = {};
    struct raw_socket_obj_aux * aux;
    enum raw_error ret;

    trace("%s(intfaddr:%p,cert_prog_ll:%p)",
        __func__, intfaddr, cert_prog_ll);

    if (intfaddr == NULL ||
        obj == NULL || obj->st >= ST_RAWIDLE)
        return RAW_INVAL;

    debug("intfaddr:%u.%u.%u.%u",
        (unsigned) intfaddr[0], (unsigned) intfaddr[1],
        (unsigned) intfaddr[2], (unsigned) intfaddr[3]);

    ifindex = addr2ifindex(AF_INET, intfaddr);
    if (ifindex <= 0) {
        debug("addr2ifindex failed with %d", ifindex);
        return RAW_INVAL;
    }

    // shouldn't be repeating netifs
    for (iftra = if_ll; iftra != NULL; iftra = iftra->next) {
        if (ifindex == iftra->ifindex) {
            return RAW_INVAL;
        }
    }

    aux = safe_malloc(sizeof(*aux));
    aux->ifindex = ifindex;
    memcpy(aux->intfaddr, intfaddr, PKTLAB_IP4_ADDR_LEN);

    // set up send socket (layer 3 raw socket)
    sendfd = safe_socket(AF_INET, SOCK_RAW, IPPROTO_RAW, &ret);
    if (sendfd < 0)
        goto fail_aux;

    if (set_nonblock(sendfd) < 0) {
        debug("set_nonblock failed for sendfd (%d)", sendfd);
        ret = RAW_UNKFAULT;
        goto fail_sendfd;
    }

    memcpy(
        &sin.sin_addr.s_addr, intfaddr,
        sizeof(sin.sin_addr.s_addr));

    if (bind(sendfd, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
        debug("bind failed for sendfd (%d) with %s",
            sendfd, strerror(errno));
        ret = RAW_UNKFAULT;
        goto fail_sendfd;
    }

    // set up recv socket (layer 2 dgram socket)
    recvfd = safe_socket(AF_PACKET, SOCK_DGRAM, htons(ETH_P_ALL), &ret);
    if (recvfd < 0) {
        debug("rawrecv:init failed with %d", recvfd);
        ret = RAW_UNKFAULT;
        goto fail_sendfd;
    }

    if (set_nonblock(recvfd) < 0) {
        debug("set_nonblock failed for recvfd (%d)", recvfd);
        ret = RAW_UNKFAULT;
        goto fail_recvfd;
    }

    sll.sll_ifindex = ifindex;

    if (bind(recvfd, (struct sockaddr *) &sll, sizeof(sll)) < 0) {
        debug("bind failed for recvfd (%d) with %s",
            recvfd, strerror(errno));
        ret = RAW_UNKFAULT;
        goto fail_recvfd;
    }

    tmpobj.af = AF_INET;
    tmpobj.rfd = recvfd;
    tmpobj.wfd = -1;
    tmpobj.aux = aux;
    tmpobj.st = ST_RAWIDLE;
    rst = bare_clean_recvbuf(&tmpobj);
    if (rst < 0) {
        debug("clean_recvbuf failed with %d", rst);
        ret = RAW_UNKFAULT;
        goto fail_recvfd;
    }

    obj->af = AF_INET;
    obj->rfd = recvfd;
    obj->wfd = sendfd;
    obj->aux = aux;
    obj->st = ST_RAWIDLE;
    obj->cert_prog_ll = cert_prog_ll;

    // prepend netif to if_ll linkedlist
    if (if_ll == NULL) {
        if_ll = safe_malloc(sizeof(*if_ll));
        if_ll->ifindex = ifindex;
        if_ll->next = NULL;
    } else {
        iftra = safe_malloc(sizeof(*if_ll));
        iftra->ifindex = ifindex;
        iftra->next = if_ll;
        if_ll = iftra;
    }

    return RAW_SUCCESS;

fail_recvfd:
    close(recvfd);
fail_sendfd:
    close(sendfd);
fail_aux:
    free(aux);
    return ret;
}

ssize_t bare_raw_socket_send (
    const void * buf, size_t len, int flags,
    struct raw_socket_obj * obj, enum raw_error * err)
{
    const uint8_t * ptr = buf;
    struct sockaddr_in sin = {};
    struct sockaddr * saddr;
    socklen_t socklen;
    ssize_t ret;
    enum raw_error reterr = RAW_SUCCESS;

    // TODO: send filters

    trace("%s(len:%zu,flags:%d)",
        __func__, len, flags);

    if (buf == NULL || len == 0 ||
        obj == NULL || obj->st != ST_RAWIDLE) {
        if (err != NULL)
            *err = RAW_INVAL;
        return -1;
    }

    switch (obj->af) {
    case AF_INET:
        sin.sin_family = obj->af;
        memcpy(&sin.sin_addr.s_addr, ptr+16, IP4_ALEN);
        saddr = (struct sockaddr *) &sin;
        socklen = sizeof(sin);
        break;
    case AF_INET6:
        // should not reach here
        debug("raw_send IPv6 support not implemented yet");
        assert(0); // bug
    default:
        debug("only supports AF_INET");
        obj->st = ST_RAWERROR;
        reterr = RAW_UNKFAULT;
    }

    ret = sendto(obj->wfd, buf, len, flags, saddr, socklen);

    if (ret < 0) {
        debug("sendto failed with %s", strerror(errno));

        if (try_again(errno)) {
            reterr = RAW_TRYAGAIN;
        } else {
            switch(errno) {
            case EACCES:
            case EINVAL:
            case EMSGSIZE:
                reterr = RAW_BADPKT;
                break;
            case EINTR:
                reterr = RAW_TRYAGAIN;
                break;
            case EFAULT:
            case EOPNOTSUPP:
                reterr = RAW_INVAL;
                break;
            case ENOBUFS:
            case ENOMEM:
                obj->st = ST_RAWERROR;
                reterr = RAW_NORES;
                break;
            default:
                obj->st = ST_RAWERROR;
                reterr = RAW_UNKFAULT;
            }
        }
    }

    debug("sendto returned %d", (int) ret);

    if (err != NULL)
        *err = reterr;
    return ret;
}

ssize_t bare_raw_socket_recv (
    void * buf, size_t len, int flags,
    struct raw_socket_obj * obj, enum raw_error * err)
{
    struct sockaddr_ll sll;
    socklen_t socklen;
    ssize_t ret;
    enum raw_socket_state state;
    enum raw_error reterr;
    uint_fast32_t acceptlen;
    pktlab_time_t recvtime;
    int rst;

    assert(buf != NULL && len != 0 && obj != NULL);

    trace("raw:%s(len:%zu,flags:%d,obj.st:%d)",
        __func__, len, flags, obj->st);

    if (buf == NULL || len == 0 ||
        obj == NULL || obj->st != ST_RAWIDLE) {
        if (err != NULL)
            *err = RAW_INVAL;
        return -1;
    }

    ret = recvfrom(
        obj->rfd, buf, len, flags,
        (struct sockaddr *) &sll, &socklen);

    debug("recvfrom ret: %d", (int) ret);

    recvtime = pktlab_time_now(); // as early as possible

    state = obj->st;
    reterr = RAW_SUCCESS;

    if (ret < 0) {
        debug("recvfrom failed with %d (%s)",
            (int) ret, strerror(errno));

        if (try_again(errno)) {
            reterr = RAW_TRYAGAIN;
        } else {
            switch(errno) {
            case EINTR:
                reterr = RAW_TRYAGAIN;
                break;
            case EFAULT:
            case EINVAL:
                reterr = RAW_INVAL;
                break;
            case ENOMEM:
                state = ST_RAWERROR;
                reterr = RAW_NORES;
                break;
            default:
                state = ST_RAWERROR;
                reterr = RAW_UNKFAULT;
            }
        }

        goto fail;
    } else if (ret == 0 ||
               (sll.sll_pkttype == PACKET_OUTGOING &&
                !is_for_self(buf, ret, obj->aux))) {
        // discard looped back sent packet or didn't receive anything
        // use try again to mask these cases
        reterr = RAW_TRYAGAIN;
        ret = -1; // -1 as for ssize_t only -1 is defined for negative
        goto fail;
    }

    obj->time = recvtime;

    // Run certificate filters
    if (ret > SIZE_MAX)
        ret = SIZE_MAX;

    rst = filter_run_ebpf_filters(
        obj->cert_prog_ll, NULL, 0, buf, ret, &acceptlen);
    if (rst < 0) {
        debug("filter_run_ebpf_filters failed with %d", rst);
        state = ST_RAWERROR;
        reterr = RAW_BADFILT;
        ret = -1;
        goto fail;
    }

    // cert filter running successful
    assert(acceptlen <= ret);
    ret = acceptlen;

fail:
    obj->st = state;
    if (err != NULL)
        *err = reterr;
    return ret;
}

int bare_raw_socket_set_recvfilter (
    struct raw_prog * filter,
    struct raw_socket_obj * obj)
{
    int prog_fd, rst;

    trace("%s(filter:%p,obj:%p)", __func__, filter, obj);

    // only ebpf is supported
    if (filter == NULL ||
        (filter->progptr == NULL) != (filter->proglen == 0) ||
        obj == NULL || obj->st != ST_RAWIDLE)
        return RAW_INVAL;

    debug("filter->progptr:%p filter->proglen:%u",
        filter->progptr, (unsigned) filter->proglen);

    if (bare_clean_recvbuf(obj) < 0) {
        debug("bare_clean_recvbuf failed");
        obj->st = ST_RAWERROR;
        return RAW_UNKFAULT;
    }

    if (filter->progptr == NULL)
        return RAW_SUCCESS; // empty filter is to stop capture

    // only the first filter is considered
    if (!filter->is_ebpf ||
        filter->proglen % PKTLAB_EBPF_INSTR_LEN != 0 ||
        filter->proglen / PKTLAB_EBPF_INSTR_LEN > INT_MAX)
        return RAW_BADFILT;

    // load the filter and attach
    prog_fd = load_ebpf_filter(filter->progptr, filter->proglen);
    if (prog_fd < 0) {
        if (prog_fd == -1 || prog_fd == -3)
            return RAW_BADFILT;

        obj->st = ST_RAWERROR;
        return RAW_UNKFAULT;
    }

    // previously loaded filter auto detached
    rst = setsockopt(
        obj->rfd, SOL_SOCKET, SO_ATTACH_BPF,
        &prog_fd, sizeof(prog_fd));

    close(prog_fd);

    if (rst < 0) {
        if (errno == EINVAL)
            return RAW_BADFILT;

        debug("setsockopt(SOL_SOCKET, SO_ATTACH_BPF) failed with %s",
            strerror(errno));

        obj->st = ST_RAWERROR;
        return RAW_UNKFAULT;
    }

    return RAW_SUCCESS;
}

// libpcap style draining socket recv buf content
int bare_clean_recvbuf(struct raw_socket_obj * obj) {
    static struct sock_filter ret0instr = BPF_STMT(BPF_RET | BPF_K, 0);
    static const struct sock_fprog ret0prog = { 1, &ret0instr };
    uint8_t buf[MAX_PKTSZ];
    ssize_t rst;
    enum raw_error reterr;

    assert(obj != NULL && obj->st == ST_RAWIDLE);

    trace("%s(...)", __func__);

    if (setsockopt(
            obj->rfd, SOL_SOCKET, SO_ATTACH_FILTER,
            (const void *) &ret0prog, (socklen_t) sizeof(ret0prog)) < 0) {
        debug("setsockopt(SOL_SOCKET,SO_ATTACH_FILTER) failed with %s",
            strerror(errno));
        return -1;
    }

    while (1) {
        rst = bare_raw_socket_recv(buf, sizeof(buf), 0, obj, &reterr);

        // RAW_TRYAGAIN can appear due to buffer drain or interrupt
        // Use EINTR to detect interrupt
        if (rst >= 0 || errno == EINTR)
            continue;

        assert(reterr != RAW_SUCCESS && reterr != RAW_INVAL);
        if (reterr == RAW_TRYAGAIN)
            break; // cleanup done

        debug("bare_raw_socket_recv failed with err %d", reterr);
        return -2; // raw_recv failed expectedly
    }

    return 0;
}

void bare_raw_socket_close(struct raw_socket_obj * obj) {
    struct raw_socket_obj_aux * aux;
    struct ifindex_ll * pre;
    struct ifindex_ll * node;

    trace("%s(...)", __func__);

    if (obj == NULL || obj->st <= ST_RAWINIT) {
        free(obj);
        return;
    }

    assert(obj->st != ST_RAWUNKNOWN);

    aux = obj->aux;

    pre = NULL;
    for (node = if_ll; node != NULL; node = node->next) {
        if (node->ifindex == aux->ifindex)
            break;

        pre = node;
    }

    if (node == NULL) {
        debug("closed socket ifindex (%d) not recorded before, "
            "skip if node cleanup process", aux->ifindex);
    } else {
        if (pre == NULL) {
            if_ll = node->next;
        } else {
            pre->next = node->next;
        }

        free(node);
    }

    close(obj->rfd);
    close(obj->wfd);
    free(aux);
    free(obj);
}

bool is_for_self (
    const uint8_t * pkt, size_t pktlen,
    struct raw_socket_obj_aux * aux)
{
    assert(pkt != NULL && aux != NULL);

    if (pktlen == 0)
        return false;

    if (((*pkt) >> 4) == 4) { // ipv4
        if (pktlen < MIN_IP4HDR_LEN)
            return false;

        return \
            (memcmp(
                pkt+(MIN_IP4HDR_LEN-IP4_ALEN),
                aux->intfaddr, IP4_ALEN) == 0);
    } else {
        if (pktlen < MIN_IP6HDR_LEN)
            return false;

        return \
            (memcmp(
                pkt+(MIN_IP6HDR_LEN-IP6_ALEN),
                aux->intfaddr, IP6_ALEN) == 0);
    }
}

int load_ebpf_filter (
    const void * filter, uint_fast32_t filterlen)
{
    union bpf_attr attr = {};
    const struct pktlab_ebpf_instr * pktlab_instr;
    static const char license[] = "proprietary";
    struct bpf_insn * os_ebpf_instr;
    struct bpf_insn * os_ebpf_filter;
    size_t totalmem;
    uint32_t i;
    char logbuf[LOG_BUF_SZ] = {};
    int ret;

    trace("%s(filterlen:%u)",
        __func__, (unsigned) filterlen);

    if (filterlen % PKTLAB_EBPF_INSTR_LEN != 0 ||
        filterlen / PKTLAB_EBPF_INSTR_LEN > UINT32_MAX)
        return -1; // bad filter

    // we mark the filters to be non-GPL compliant
    // as currently there isn't a way to pass that info
    // note this prevents programs from using kernel bpf-helpers

    attr.prog_type = BPF_PROG_TYPE_SOCKET_FILTER;
    attr.license = (uintptr_t) license;
    attr.log_buf = (uintptr_t) logbuf;
    attr.log_size = sizeof(logbuf);
    attr.log_level = 1; // enable verifier log

    // perform instr translation if needed
    os_ebpf_filter = NULL;
    attr.insns = (uintptr_t) filter;
    attr.insn_cnt = filterlen / PKTLAB_EBPF_INSTR_LEN;

    // struct bpf_insn size different than pktlab_ebpf_instr size
    // need to translate
    if (PKTLAB_EBPF_INSTR_LEN != sizeof(*os_ebpf_filter)) {
        totalmem = sizeof(*os_ebpf_filter)*attr.insn_cnt;
        if (totalmem / sizeof(*os_ebpf_filter) != attr.insn_cnt)
            return -2; // overflow

        os_ebpf_filter = safe_malloc(totalmem);

        pktlab_instr = filter;
        os_ebpf_instr = os_ebpf_filter;
        for (i = 0; i < attr.insn_cnt; ++i) {
            os_ebpf_instr->code = pktlab_instr->opcode;
            os_ebpf_instr->dst_reg = pktlab_instr->dstreg;
            os_ebpf_instr->src_reg = pktlab_instr->srcreg;
            os_ebpf_instr->off = pktlab_instr->offset;
            os_ebpf_instr->imm = pktlab_instr->imm;

            ++pktlab_instr;
            ++os_ebpf_instr;
        }

        attr.insns = (uintptr_t) os_ebpf_filter;
    }

    ret = bpf(BPF_PROG_LOAD, &attr, sizeof(attr));
    if (ret < 0) {
        debug("bpf(BPF_PROG_LOAD) failed with %s", strerror(errno));

        if (errno == EACCES) {
            debug("Filter rejected, reason: %s", logbuf);
            ret = -3; // bad filter
        } else if (errno == EINVAL) {
            debug("Filter invalid");
            ret = -3; // bad filter
        } else {
            ret = -4; // unknown fault
        }
    } else
        debug("Loaded prog fd: %d", ret);

    free(os_ebpf_filter);
    return ret;
}