// raw_linux.c
//

// POSIX-compatible include headers
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <limits.h>
#include <net/if.h>
#include <netinet/in.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

#include <pktlab.h>

#include "console.h"
#include "net.h"
#include "raw.h"
#include "util.h"

// Linux dependent headers
#include <ifaddrs.h>

#include "raw_linux.h"

//
// EXPORTED FUNCTION DEFINITIONS
//

// for raw.h
void raw_v4_hdrinfo_read (
    const struct pktlab_vmem_region * restrict rgn,
    uint_fast32_t off, uint_fast32_t len, void * buf)
{
    uint8_t infoarr[PKTLAB_VMEMADDR_SYSPARAM_IP4HDRMOD_FL];

    // OFF HEADER
    // --- -----------
    //   0 Version
    //   1 IHL
    //   2 Type of Service (TOS)
    //   3 Total Length
    //   4 Identification
    //   5 Flags
    //   6 Fragment offset
    //   7 Time to Live (TTL)
    //   8 Protocol
    //   9 Header Checksum
    //  10 Source Address
    //  11 Destination Address

    memset(infoarr, PKTLAB_RAWHDR_NOCHG, sizeof(infoarr));
    infoarr[3] = PKTLAB_RAWHDR_CHG;
    infoarr[4] = PKTLAB_RAWHDR_0FILL;
    infoarr[9] = PKTLAB_RAWHDR_CHG;
    infoarr[10] = PKTLAB_RAWHDR_0FILL;
    memcpy(buf, infoarr + off, len);
}

void raw_v6_hdrinfo_read (
    const struct pktlab_vmem_region * restrict rgn,
    uint_fast32_t off, uint_fast32_t len, void * buf)
{
    uint8_t infoarr[PKTLAB_VMEMADDR_SYSPARAM_IP6HDRMOD_FL];

    // OFF HEADER
    // --- -----------
    //   0 Version
    //   1 IHL
    //   2 Type of Service (TOS)
    //   3 Total Length
    //   4 Identification
    //   5 Flags
    //   6 Fragment offset
    //   7 Time to Live (TTL)
    //   8 Protocol
    //   9 Header Checksum
    //  10 Source Address
    //  11 Destination Address

    memset(infoarr, PKTLAB_RAWHDR_NOCHG, sizeof(infoarr));
    memcpy(buf, infoarr + off, len);
}

// for raw_linux.h
int addr2ifindex (
    int af, const uint8_t * addr)
{
    size_t addrlen;
    struct ifaddrs * ifap;
    struct ifaddrs * ptr;
    uint8_t * intfaddr;
    struct sockaddr_in * sin;
    struct sockaddr_in6 * sin6;

    int ret = -128;

    assert(
        (af == AF_INET || af == AF_INET6) &&
        addr != NULL);

    trace("%s(af:%d)", __func__, af);

    addrlen = (af == AF_INET) ? IP4_ALEN:IP6_ALEN;

    if (getifaddrs(&ifap) < 0) {
        debug("getifaddrs failed with %s", strerror(errno));
        return -1;
    }

    ret = -2; // addr not found
    for (ptr = ifap; ptr != NULL; ptr = ptr->ifa_next) {
        if (ptr->ifa_addr == NULL ||
            ptr->ifa_addr->sa_family != af)
            continue;

        if (af == AF_INET) {
            sin = (struct sockaddr_in *) ptr->ifa_addr;
            intfaddr = (uint8_t *) &(sin->sin_addr.s_addr);
        } else {
            sin6 = (struct sockaddr_in6 *) ptr->ifa_addr;
            intfaddr = (uint8_t *) &(sin6->sin6_addr.s6_addr);
        }

        if (memcmp(intfaddr, addr, addrlen) != 0)
            continue;

        // match found, do translation
        ret = if_nametoindex(ptr->ifa_name);
        if (ret == 0)
            ret = -3; // translation failed

        break;
    }

    freeifaddrs(ifap);
    return ret;
}

// v6 untested
int safe_socket (
    int af, int type, int protocol,
    enum raw_error * err)
{
    int sockfd;
    enum raw_error reterr = RAW_SUCCESS;

    assert(err != NULL);

    trace("%s(af:%d,type:%d,protocol:%d)",
        __func__, af, type, protocol);

    sockfd = socket(af, type, protocol);
    if (sockfd < 0) {
        debug("socket failed with %s", strerror(errno));

        switch(errno) {
        case EPERM:         /* permission errors */
        case EACCES:
            reterr = RAW_PERM;
            break;
        case EMFILE:        /* resource errors */
        case ENFILE:
        case ENOBUFS:
        case ENOMEM:
            reterr = RAW_NORES;
            break;
        case EAFNOSUPPORT:
        default:            /* all others should not happen  */
            reterr = RAW_UNKFAULT;
        }

        *err = reterr;
        return -1;
    }

    if (set_nonblock(sockfd) < 0) {
        close(sockfd);
        reterr = RAW_UNKFAULT;
        sockfd = -1;
    }

    *err = reterr;
    return sockfd;
}

// only allow MSG_PEEK & MSG_TRUNC for flags
bool is_allowed_recv_flags(int flags) {
    if ((flags & ~MSG_PEEK & ~MSG_TRUNC) != 0)
        return false;
    return true;
}