// raw_lkm_linux.c
// pktlab raw socket linux platform impl.
//

// TODO: IPv6 support

//
// POSIX HEADERS
//

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <limits.h>
#include <net/if.h>
#include <netinet/in.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

//
// CUSTOM HEADERS
//

#include <pktlab.h>
#include <nfnlbpf_lib.h>

#include "console.h"
#include "net.h"
#include "raw.h"
#include "raw_linux.h"
#include "util.h"

//
// LINUX-ONLY HEADERS
//

#include <asm/types.h>
#include <ifaddrs.h>
#include <linux/bpf.h>
#include <linux/filter.h>
#include <linux/netlink.h>


//
// INTERNAL GLOBAL VARIABLES
//

struct ifindex_ll * if_ll;

//
// INTERNAL FUNCTION DECLARATIONS
//

static int lkm_raw_socket_setup_v4 (
    const uint8_t * intfaddr,
    struct raw_prog * cert_prog_ll,
    struct raw_socket_obj * obj);

static ssize_t lkm_raw_socket_send (
    const void * buf, size_t len, int flags,
    struct raw_socket_obj * obj, enum raw_error * err);
static ssize_t lkm_raw_socket_recv (
    void * buf, size_t len, int flags,
    struct raw_socket_obj * obj, enum raw_error * err);

static int lkm_raw_socket_set_recvfilter (
    struct raw_prog * filter,
    struct raw_socket_obj * obj);

static int clean_recvbuf(struct raw_socket_obj * obj);

static void lkm_raw_socket_close(struct raw_socket_obj * obj);

//
// EXPORTED FUNCTION DEFINITIONS
//

const struct raw_socket_method_t * raw_socket_steal_method(void) {
    static const struct raw_socket_method_t ret = {
        .raw_socket_setup_v4 = lkm_raw_socket_setup_v4,
        .raw_socket_setup_v6 = NULL, // ipv6 not supported yet
        .raw_socket_close = lkm_raw_socket_close,
        .raw_socket_send = lkm_raw_socket_send,
        .raw_socket_recv = lkm_raw_socket_recv,
        .raw_socket_set_recvfilter = lkm_raw_socket_set_recvfilter,
    };

    return &ret;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int lkm_raw_socket_setup_v4 (
    const uint8_t * intfaddr,
    struct raw_prog * cert_prog_ll,
    struct raw_socket_obj * obj)
{
    int sendfd, recvfd, ifindex, rst, proglen, instrsz;
    struct sock_filter ret0instr = BPF_STMT(BPF_RET | BPF_K, 0);
    struct raw_prog * prog;
    struct raw_socket_obj tmpobj = {};
    enum raw_error ret;

    trace("%s(intfaddr:%p,cert_prog_ll:%p)",
        __func__, intfaddr, cert_prog_ll);

    if (intfaddr == NULL ||
        obj == NULL || obj->st >= ST_RAWIDLE)
        return RAW_INVAL;

    debug("intfaddr:%u.%u.%u.%u",
        (unsigned) intfaddr[0], (unsigned) intfaddr[1],
        (unsigned) intfaddr[2], (unsigned) intfaddr[3]);

    // todo: this is temporary
    // as currently nfnlbpf only support single ifindex
    // should check if in ll
    if (if_ll != NULL) {
        debug("More than one raw socket opened");
        return RAW_UNKFAULT;
    }

    // set up send socket (layer 3 raw socket)
    sendfd = safe_socket(AF_INET, SOCK_RAW, IPPROTO_RAW, &ret);
    if (sendfd < 0)
        return ret;

    if (set_nonblock(sendfd) < 0) {
        debug("set_nonblock failed for sendfd (%d)", sendfd);
        ret = RAW_UNKFAULT;
        goto fail_sendfd;
    }

    // set up recv socket (kernel module passed back netlink socket)
    ifindex = addr2ifindex(AF_INET, intfaddr);
    if (ifindex <= 0) {
        debug("addr2ifindex failed with %d", ifindex);
        ret = RAW_INVAL; // bad intfaddr
        goto fail_sendfd;
    }

    recvfd = init(ifindex);
    if (recvfd < 0) {
        debug("rawrecv:init failed with %d", recvfd);
        ret = RAW_UNKFAULT;
        goto fail_sendfd;
    }

    if (set_nonblock(recvfd) < 0) {
        debug("set_nonblock failed for recvfd (%d)", recvfd);
        ret = RAW_UNKFAULT;
        goto fail_recvfd;
    }

    // insert pass all user filter to force lkm netif node creation
    if (install_user_filter(&ret0instr, 1, false) < 0) {
        debug("rawrecv:install_user_filter failed");
        ret = RAW_UNKFAULT;
        goto fail_recvfd;
    }

    // for better efficiency when there are cert filters
    if (mask_interface() < 0) {
        debug("rawrecv:mask_interface failed");
        ret = RAW_UNKFAULT;
        goto fail_recvfd;
    }

    tmpobj.af = AF_INET;
    tmpobj.rfd = recvfd;
    tmpobj.wfd = -1;
    tmpobj.ifindex = ifindex;
    tmpobj.st = ST_RAWIDLE;
    rst = clean_recvbuf(&tmpobj);
    if (rst < 0) {
        debug("clean_recvbuf failed with %d", rst);
        ret = RAW_UNKFAULT;
        goto fail_recvfd;
    }

    for (prog = cert_prog_ll; prog != NULL; prog = prog->next) {
        instrsz = prog->is_ebpf ? RAW_EBPF_INSTR_SZ:RAW_CBPF_INSTR_SZ;

        if (prog->proglen % instrsz != 0) {
            debug("proglen (%" PRIuFAST32 ") not of multiple of instruction size (%d)",
                prog->proglen, instrsz);
            ret = RAW_BADFILT;
            goto fail_recvfd;
        }

        if (prog->proglen / instrsz > INT_MAX) {
            debug("Actual proglen (%" PRIuFAST32 "/%d) overflow INT_MAX (%d)",
                prog->proglen, instrsz, INT_MAX);
            ret = RAW_BADFILT;
            goto fail_recvfd;
        }

        proglen = prog->proglen / instrsz;
        if (install_cert(
                prog->progptr, proglen,
                prog->is_ebpf) < 0) {
            debug("rawrecv:install_cert(len:%d,is_ebpf:%d) failed",
                proglen, prog->is_ebpf);
            ret = RAW_BADFILT;
            goto fail_recvfd;
        }
    }

    obj->af = AF_INET;
    obj->rfd = recvfd;
    obj->wfd = sendfd;
    obj->ifindex = ifindex;
    obj->st = ST_RAWIDLE;

    // prepend netif to if_ll linkedlist
    if (if_ll == NULL) {
		if_ll = safe_malloc(sizeof(*if_ll));
		if_ll->ifindex = ifindex;
		if_ll->next = NULL;
    } else {
		iftra = safe_malloc(sizeof(*if_ll));
		iftra->ifindex = ifindex;
		iftra->next = if_ll;
		if_ll = iftra;
    }

    return RAW_SUCCESS;

fail_recvfd:
    cleanup();
    close(recvfd);
fail_sendfd:
    close(sendfd);
    return ret;
}

ssize_t lkm_raw_socket_send (
    const void * buf, size_t len, int flags,
    struct raw_socket_obj * obj, enum raw_error * err)
{
    const uint8_t * ptr = buf;
    struct sockaddr_in sin = {};
    struct sockaddr * saddr;
    socklen_t socklen;
    ssize_t ret;
    enum raw_error reterr = RAW_SUCCESS;

    trace("%s(len:%zu,flags:%d)",
        __func__, len, flags);

    if (buf == NULL || len == 0 ||
        obj == NULL || obj->st != ST_RAWIDLE) {
        if (err != NULL)
            *err = RAW_INVAL;
        return -1;
    }

    switch (obj->af) {
    case AF_INET:
        sin.sin_family = obj->af;
        memcpy(&sin.sin_addr.s_addr, ptr+16, IP4_ALEN);
        saddr = (struct sockaddr *) &sin;
        socklen = sizeof(sin);
        break;
    case AF_INET6:
        // should not reach here
        debug("raw_send IPv6 support not implemented yet");
    default:
        assert(0); // bug
    }

    ret = sendto(obj->wfd, buf, len, flags, saddr, socklen);
    if (ret < 0) {
        debug("sendto failed with %s", strerror(errno));

        if (try_again(errno)) {
            reterr = RAW_TRYAGAIN;
        } else {
            switch(errno) {
            case EACCES:
            case EINVAL:
            case EMSGSIZE:
                reterr = RAW_BADPKT;
                break;
            case EINTR:
                reterr = RAW_TRYAGAIN;
                break;
            case EFAULT:
            case EOPNOTSUPP:
                reterr = RAW_INVAL;
                break;
            case ENOBUFS:
            case ENOMEM:
                obj->st = ST_RAWERROR;
                reterr = RAW_NORES;
                break;
            default:
                obj->st = ST_RAWERROR;
                reterr = RAW_UNKFAULT;
            }
        }
    }

    debug("sendto returned %d", (int) ret);

    if (err != NULL)
        *err = reterr;
    return ret;
}

ssize_t lkm_raw_socket_recv (
    void * buf, size_t len, int flags,
    struct raw_socket_obj * obj, enum raw_error * err)
{
    struct msghdr msg = {};
    struct iovec iov;
    uint8_t nlmsg[NLMSG_SPACE(MAX_PKTSZ)];
    struct nlmsghdr * nlh;
    ssize_t ret, actuallen;
    enum raw_error reterr = RAW_SUCCESS;

    trace("%s(len:%zu,flags:%d)",
        __func__, len, flags);

    if (buf == NULL || len == 0 ||
        obj == NULL || obj->st != ST_RAWIDLE ||
        !is_allowed_recv_flags(flags)) {
        if (err != NULL)
            *err = RAW_INVAL;
        return -1;
    }

    iov.iov_base = nlmsg;
    iov.iov_len = sizeof(nlmsg);
    msg.msg_name = NULL;
    msg.msg_namelen = 0;
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
    nlh = (struct nlmsghdr *) nlmsg;

    ret = recvmsg(obj->rfd, &msg, flags);
    if (ret < 0) {
        debug("recvmsg failed with %s", strerror(errno));

        if (try_again(errno)) {
            reterr = RAW_TRYAGAIN;
        } else {
            switch(errno) {
            case EINTR:
                reterr = RAW_TRYAGAIN;
                break;
            case EINVAL:
                reterr = RAW_INVAL;
                break;
            case ENOMEM:
                obj->st = ST_RAWERROR;
                reterr = RAW_NORES;
                break;
            default:
                obj->st = ST_RAWERROR;
                reterr = RAW_UNKFAULT;
            }
        }
    } else if (ret == 0) {
        debug("recvfd EOF");
        obj->st = ST_RAWERROR;
        reterr = RAW_UNKFAULT;
        ret = -1;
    } else if (!NLMSG_OK(nlh, ret)) {
        debug("recvmsg returned bad nlh (ret:%d)", (int) ret);
        obj->st = ST_RAWERROR;
        reterr = RAW_UNKFAULT;
        ret = -1;
    }

    if (reterr == RAW_SUCCESS) {
        assert(ret > 0);

        actuallen = MIN(SSIZE_MAX, len);
        actuallen = MIN(actuallen, NLMSG_PAYLOAD(nlh, 0));
        memcpy(buf, NLMSG_DATA(nlh), actuallen);
        ret = actuallen;

        obj->time = pktlab_time_now();
        debug("recvmsg received %d bytes, "
            "requested len %zu, payload len %u, "
            "actual returned bytes %d",
            (int) ret, len, (unsigned) NLMSG_PAYLOAD(nlh, 0),
            (int) actuallen);
    }

    if (err != NULL)
        *err = reterr;

    return ret;
}

int lkm_raw_socket_set_recvfilter (
    struct raw_prog * filter,
    struct raw_socket_obj * obj)
{
    int instrsz, proglen;

    trace("%s(filter:%p,obj:%p)", __func__, filter, obj);

    if (filter == NULL ||
        (filter->progptr == NULL) != (filter->proglen == 0) ||
        obj == NULL || obj->st != ST_RAWIDLE)
        return RAW_INVAL;

    debug("filter->progptr:%p filter->proglen:%u",
        filter->progptr, (unsigned) filter->proglen);

    if (clean_recvbuf(obj) < 0) {
        debug("clean_recvbuf failed");
        obj->st = ST_RAWERROR;
        return RAW_UNKFAULT;
    }

    // only the first filter is considered
    if (filter->progptr != NULL) {
        instrsz = filter->is_ebpf ? RAW_EBPF_INSTR_SZ:RAW_CBPF_INSTR_SZ;

        if (filter->proglen % instrsz != 0) {
            debug("proglen (%" PRIuFAST32 ") not of multiple of instruction size (%d)",
                filter->proglen, instrsz);
            return RAW_BADFILT;
        }

        if (filter->proglen / instrsz > INT_MAX) {
            debug("Actual proglen (%" PRIuFAST32 "/%d) overflow INT_MAX (%d)",
                filter->proglen, instrsz, INT_MAX);
            return RAW_BADFILT;
        }

        proglen = filter->proglen / instrsz;

        if (install_user_filter(
                filter->progptr, proglen,
                filter->is_ebpf) < 0) {
            debug("install_user_filter failed");
            obj->st = ST_RAWERROR;
            return RAW_UNKFAULT;
        }
    }

    return RAW_SUCCESS;
}

int clean_recvbuf(struct raw_socket_obj * obj) {
    uint8_t buf[MAX_PKTSZ];
    ssize_t rst;
    enum raw_error reterr;

    assert(obj != NULL && obj->st == ST_RAWIDLE);

    trace("%s(...)", __func__);

    if (mask_interface() < 0)
        return -1;

    while (1) {
        rst = raw_socket_recv(buf, sizeof(buf), 0, obj, &reterr);

        // RAW_TRYAGAIN can appear due to buffer drain or interrupt
        // Use EINTR to detect interrupt
        if (rst >= 0 || errno == EINTR)
            continue;

        assert(reterr != RAW_SUCCESS && reterr != RAW_INVAL);
        if (reterr == RAW_TRYAGAIN)
            break; // cleanup done

        debug("raw_recv failed with err %d", reterr);
        return -2; // raw_recv failed expectedly
    }

    return 0;
}

void lkm_raw_socket_close(struct raw_socket_obj * obj) {
    struct ifindex_ll * pre;
    struct ifindex_ll * node;

    trace("%s(...)", __func__);

    if (obj == NULL || obj->st <= ST_RAWINIT) {
        free(obj);
        return;
    }

    assert(obj->st != ST_RAWUNKNOWN);

    pre = NULL;
    for (node = if_ll; node != NULL; node = node->next) {
        if (node->ifindex == obj->ifindex)
            break;

        pre = node;
    }

    if (node == NULL) {
        debug("closed socket ifindex (%d) not recorded before, "
            "skip rawrecv cleanup process", obj->ifindex);
    } else {
        cleanup();

        if (pre == NULL) {
            if_ll = node->next;
        } else {
            pre->next = node->next;
        }

        free(node);
    }

    close(obj->rfd);
    close(obj->wfd);
    free(obj);
    return;
}
