// rlim.c - Resource limits
// Kirill Levchenko

#include "rlim.h"

#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <unistd.h>

#include <pktlab.h>
#include <lmdb.h>

#include "console.h"
#include "util.h"

//
// INTERNAL COMPILE-TIME OPTIONS
//

#ifndef DEFAULT_RLIME_STORE_PATH
#define DEFAULT_RLIME_STORE_PATH "/.pktlab/endpt.rlim"
#define FALLBACK_RLIME_STORE_PATH "endpt.rlim"
#define FILEPATH_BUFSZ 4096
#endif

//
// INTERNAL TYPE DEFINITIONS
//

enum rlacc_type {
	RLACC_NOTYPE,
	RLACC_SNDPKTCNT,
	RLACC_RCVPKTCNT,
	RLACC_SNDBYTECNT,
	RLACC_RCVBYTECNT,
	RLACC_RUNTIME,
	RLACC_COUNT
};

struct rlacc_state {
	enum rlacc_type type;
	uint64_t cnt;
};

struct rlim_dbval {
	pktlab_time_t exptime;
	struct rlacc_state accst[];
};

struct rlim_dbrec {
	void * key;
	uint16_t keylen;
	uint16_t astcnt;
	uint32_t dbvalsz;
	struct rlim_dbval * dbval;
};

//
// INTERNAL FUNCTION DECLARATIONS
//

static uint64_t * rlaccptr(struct rlim_accs * accs, int type);

static void rates_append (
	struct rlim_rates * restrict dst,
	struct rlim_rates * restrict src);

static void rates_optimize(struct rlim_rates * rates);

static struct rlim_rate * ratelist_optimize(struct rlim_rate * head);

static void rates_debug(const struct rlim_rates * rates);

static void ratelist_debug(const char * name, struct rlim_rate * head);

static void ratelist_free(struct rlim_rate * head);

static void exit_if_mdb_error(int result, const char * msg);

//
// EXPORTED GLOBAL VARIABLE DEFINITIONS
//

struct rlim rlim;

//
// INTERNAL GLOBAL VARIABLE DECLARATIONS
//

static char * storename;

static MDB_env * mdbenv;

static struct rlim_dbrec * dbrecs;
static uint32_t dbrcnt; // number of dbrecs
//static bool paused;

static pktlab_time_t time_start;
static struct rlim_accs accs_start;
static uint64_t runtime_rem;

static const struct rlim_accs accs_init = {
	UINT64_MAX, UINT64_MAX, UINT64_MAX, UINT64_MAX, UINT64_MAX
};

static const struct rlim_caps caps_init = {
	UINT16_MAX
};

static const struct rlim_rate sndbyteburst_10Mbps_10sec = {
	.cnt = (uint64_t)10*1000*1000 / 8 * 10, .win = 10 * PKTLAB_TICKS_PER_SECOND
};

static const struct rlim_rates rates_init = {
	.sndbytelist = (struct rlim_rate*)&sndbyteburst_10Mbps_10sec
};

// This is a fake default resource limit record. Until we can extract rlimits
// from certificates, we initialize each certificate's rlim_dbrec with this.

static size_t fake_astcnt = 5;

static const struct rlim_dbval fake_dbval = {
	.exptime = PKTLAB_TIME_MAX,
	.accst = {
		{ .type = RLACC_SNDPKTCNT, .cnt = 100000 },
		{ .type = RLACC_RCVPKTCNT, .cnt = 100000 },
		{ .type = RLACC_SNDBYTECNT, .cnt = 1048576 /* 1MB */ },
		{ .type = RLACC_RCVBYTECNT, .cnt = 1048576 /* 1MB */ },
		{ .type = RLACC_RUNTIME, .cnt = 30 * 60 * PKTLAB_TICKS_PER_SECOND /* 30 mins */ }
	}
};

//
// EXPORTED FUNCTION DEFINITIONS
//

void rlim_init(const char * storepath) {
	uid_t uid;
	int result;
	size_t homepathlen;
	const char * homepath;
	static char storepath_buf[FILEPATH_BUFSZ];

	result = mdb_env_create(&mdbenv);
	exit_if_mdb_error(result, NULL);

	if (storepath == NULL) {
		storepath = DEFAULT_RLIME_STORE_PATH;
		uid = getuid();
		if (uid == 0) {
			homepath = get_passwd_home();
		} else {
			homepath = getenv("HOME");
			if (homepath == NULL)
				homepath = get_passwd_home();
		}

		if (homepath == NULL) {
			warn("Cannot get home directory path");
			warn("Looking for %s in current working directory instead",
					FALLBACK_RLIME_STORE_PATH);
			storepath = FALLBACK_RLIME_STORE_PATH;
		} else {
			homepathlen = strlen(homepath);
			if (homepathlen > sizeof(storepath_buf)-sizeof(DEFAULT_RLIME_STORE_PATH))
				fatal("Store path buffer overflow: "
						"home directory path too long (%zu)", homepathlen);

			strncpy(storepath_buf, homepath, sizeof(storepath_buf));
			strncpy(
					storepath_buf+homepathlen, DEFAULT_RLIME_STORE_PATH,
					sizeof(storepath_buf)-homepathlen); // '\0' appended
			storepath = storepath_buf;
		}
	}

	result = mdb_env_open(mdbenv, storepath, MDB_NOSUBDIR, 0660);
	exit_if_mdb_error(result, storepath);

	// Save rlim store name for error reporting

	storename = strrchr(storepath, '/');

	if (storename == NULL)
		storename = safe_strdup(storepath);
	else
		storename = safe_strdup(storename);

	rlim.state = RLIM_READY;
}

void rlim_finalize() {
	// Call reset to free dbrcs and rlim memory

	switch (rlim.state) {
	case RLIM_UNINIT: // unexpeced
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		break;
	case RLIM_READY: // expected case
		break;
	case RLIM_RUNNING:
	case RLIM_PAUSED:
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		rlim_reset(); // free dbrecs and rlim.rates
		break;
	}

	// Close MDB completely

	mdb_env_close(mdbenv);
	storename = NULL;
	mdbenv = NULL;

	rlim.state = RLIM_UNINIT;
}

void rlim_vacuum(pktlab_time_t now) {
	pktlab_time_t * exptime;
	MDB_txn * txn;
	MDB_dbi dbi;
	MDB_cursor * cur;
	MDB_val key, data;
	uint_fast32_t expcnt;
	uint_fast32_t errcnt;
	uint_fast32_t keycnt;
	int result;

	switch (rlim.state) {
	case RLIM_UNINIT: // can't complete operation
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		return;
	case RLIM_READY: // expected case
		break;
	default: // unexpected, but ok
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		break;
	}

	result = mdb_txn_begin(mdbenv, /* parent */ NULL, MDB_RDONLY, &txn);
	exit_if_mdb_error(result, storename);

	result = mdb_dbi_open(txn, /* name */ NULL, /* flags */ 0, &dbi);
	exit_if_mdb_error(result, storename);

	result = mdb_cursor_open(txn, dbi, &cur);
	exit_if_mdb_error(result, storename);

	keycnt = 0;
	expcnt = 0;
	errcnt = 0;

	// Scan through records in rd-only mode to determine if any records need to
	// be deleted (expire time reached, or malformed).

	result = mdb_cursor_get(cur, &key, &data, MDB_FIRST);

	while (result == 0) {
		if (data.mv_size < sizeof(pktlab_time_t))
			errcnt += 1;

		exptime = data.mv_data;

		debug("Record %" PRIuFAST32 " exptime = %f (%s)",
			keycnt, (double) *exptime / PKTLAB_TICKS_PER_SECOND,
			(*exptime < now) ? "EXPIRED" : "valid");

		if (*exptime < now)
			expcnt += 1;

		result = mdb_cursor_get(cur, &key, &data, MDB_NEXT);
		keycnt += 1;
	}

	if (result != MDB_NOTFOUND)
		exit_if_mdb_error(result, storename);

	// Close cursor and transaction. Database handle dbi closed automatically
	// by aborting transaction.

	mdb_cursor_close(cur);
	mdb_txn_abort(txn);

	if (expcnt > 0)
		debug("MDB contains expired records");

	if (errcnt > 0)
		debug("MDB contains malformed records");

	if (expcnt == 0 && errcnt == 0) {
		debug("All %" PRIuFAST32 " MDB records appear valid and current");
		return;
	}

	// If there are entries that we need to remove, re-open database in write
	// mode and delete any stale entries.

	result = mdb_txn_begin(mdbenv, /* parent */ NULL, /* flags */ 0, &txn);
	exit_if_mdb_error(result, storename);

	result = mdb_dbi_open(txn, /* name */ NULL, /* flags */ 0, &dbi);
	exit_if_mdb_error(result, storename);

	result = mdb_cursor_open(txn, dbi, &cur);
	exit_if_mdb_error(result, storename);

	result = mdb_cursor_get(cur, &key, &data, MDB_FIRST);

	keycnt = 0;

	while (result == 0) {
		exptime = data.mv_data;

		if (data.mv_size < sizeof(pktlab_time_t) || *exptime < now) {
			debug("Deleting record %" PRIuFAST32, keycnt);
			result = mdb_cursor_del(cur, /* flags*/ 0);
			exit_if_mdb_error(result, storename);
		}

		result = mdb_cursor_get(cur, &key, &data, MDB_NEXT);
		keycnt += 1;
	}

	if (result != MDB_NOTFOUND)
		exit_if_mdb_error(result, storename);

	// Cursor closed at commit.

	result = mdb_txn_commit(txn);
	exit_if_mdb_error(result, storename);

	// Do not need to close DB handle dbi.
}

void rlim_begin (
	struct pktlab_certificate ** certs,
	uint_fast32_t certcnt,
	pktlab_time_t now)
{
	struct pktlab_cert_info * cinfo;
	MDB_txn * txn;
	MDB_dbi dbi;
	MDB_val mdbkey;
	MDB_val mdbval;
	uint_fast32_t i, k;
	uint_fast16_t j;
	size_t accstsz;
	uint64_t * cntptr;
	int result;

	switch (rlim.state) {
	case RLIM_UNINIT: // can't complete operation
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		return;
	case RLIM_READY: // expected case
		break;
	case RLIM_RUNNING: // should call rlim_reset() first, but ok
	case RLIM_PAUSED:
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		rlim_reset();
		break;
	}

	// Initialize rlim struct with some defaults.

	rlim.endtime = PKTLAB_TIME_MAX;
	memcpy(&rlim.accs, &accs_init, sizeof(accs_init));
	memcpy(&rlim.caps, &caps_init, sizeof(caps_init));
	rates_append(&rlim.rates, (struct rlim_rates*)&rates_init);

	// Create a read-only transaction for reading data.

	result = mdb_txn_begin(mdbenv, /* parent */ NULL, MDB_RDONLY, &txn);
	exit_if_mdb_error(result, storename);

	result = mdb_dbi_open(txn, /* name */ NULL, /* flags */ 0, &dbi);
	exit_if_mdb_error(result, storename);

	if (dbrecs != NULL)
		rlim_reset();

	dbrecs = safe_calloc(certcnt, sizeof(struct rlim_dbrec));

	for (i = 0, k = 0; i < certcnt; i++) {
		result = pktlab_decode_certificate(certs[i], NULL, &cinfo);

		if (result != PKTLAB_DECODE_SUCCESS) {
			error("Certificate decode error (%d)", result);
			exit(EXIT_FAILURE);
		}

		// Only use PKTLAB_CERT_EXPPRIV and PKTLAB_CERT_DELPRIV certificates
		//

		switch (cinfo->cert_type) {
		case PKTLAB_CERT_EXPPRIV:
		case PKTLAB_CERT_DELPRIV:
			break;
		default:
			free(cinfo);
			continue;
		}

		// Lookup resource limit record in store.

		dbrecs[k].keylen = PKTLAB_KEYID_LEN + cinfo->serialno_len;
		dbrecs[k].key = safe_malloc(dbrecs[k].keylen);
		memcpy(dbrecs[k].key, cinfo->authority_key_id, PKTLAB_KEYID_LEN);
		memcpy(dbrecs[k].key + PKTLAB_KEYID_LEN, cinfo->serialno,
			cinfo->serialno_len);

		mdbkey.mv_size = dbrecs[k].keylen;
		mdbkey.mv_data = dbrecs[k].key;

		result = mdb_get(txn, dbi, &mdbkey, &mdbval);

		if (result != MDB_NOTFOUND)
			exit_if_mdb_error(result, storename);

		accstsz = mdbval.mv_size - sizeof(struct rlim_dbval);

		if (result == 0 && sizeof(struct rlim_dbval) <= mdbval.mv_size) {
			accstsz = mdbval.mv_size - sizeof(struct rlim_dbval);
			if (accstsz % sizeof(struct rlacc_state) == 0) {
				dbrecs[k].dbvalsz = mdbval.mv_size;
				dbrecs[k].dbval = safe_malloc(mdbval.mv_size);
				memcpy(dbrecs[k].dbval, mdbval.mv_data, mdbval.mv_size);
				dbrecs[k].astcnt = accstsz / sizeof(struct rlacc_state);
				debug("Loaded stored record for certificate %" PRIdFAST32, k);
			} else {
				warn("Malformed (%zu-byte) rlim record for certificate %" PRIdFAST32,
					(size_t) mdbval.mv_size, i);
				result = MDB_NOTFOUND;
			}
		}

		if (result == MDB_NOTFOUND) {
			// Parse certificate dbval here. For now, use a fixed dbval.
			debug("Fabricating rlim record for certificate %" PRIdFAST32, k);
			dbrecs[k].astcnt = fake_astcnt;
			dbrecs[k].dbvalsz = sizeof(struct rlim_dbval);
			dbrecs[k].dbvalsz += fake_astcnt * sizeof(struct rlacc_state);
			dbrecs[k].dbval = safe_malloc(dbrecs[k].dbvalsz);
			memcpy(dbrecs[k].dbval, &fake_dbval, dbrecs[k].dbvalsz);
		}

		// exptime always comes from certificate

		dbrecs[k].dbval->exptime = cinfo->valid_not_after;

		// Apply counters to rlim struct. We take the minimum value of all
		// counter values across certificates (and initial values accs_init).

		for (j = 0; j < dbrecs[k].astcnt; j++) {
			cntptr = rlaccptr(&rlim.accs, dbrecs[k].dbval->accst[j].type);
			if (cntptr != NULL && dbrecs[k].dbval->accst[j].cnt < *cntptr)
				*cntptr = dbrecs[k].dbval->accst[j].cnt;
		}

		// FIXME Parse certificate caps and rates.

		// ...

		// Experiment must end by certificate expiration time.

		if (cinfo->valid_not_after < rlim.endtime)
			rlim.endtime = cinfo->valid_not_after;

		free(cinfo);
		k += 1;
	}

	dbrcnt = k;

	if (rlim.accs.sndpktcnt < UINT64_MAX)
		debug("rlim.accs.sndpktcnt: %" PRIu64 " pkts", rlim.accs.sndpktcnt);
	else
		debug("rlim.accs.sndpktcnt: no limit");

	if (rlim.accs.rcvpktcnt < UINT64_MAX)
		debug("rlim.accs.rcvpktcnt: %" PRIu64 " pkts", rlim.accs.rcvpktcnt);
	else
		debug("rlim.accs.rcvpktcnt: no limit");

	if (rlim.accs.sndbytecnt < UINT64_MAX)
		debug("rlim.accs.sndbytecnt: %" PRIu64 " pkts", rlim.accs.sndbytecnt);
	else
		debug("rlim.accs.sndbytecnt: no limit");

	if (rlim.accs.rcvbytecnt < UINT64_MAX)
		debug("rlim.accs.rcvbytecnt: %" PRIu64 " pkts", rlim.accs.rcvbytecnt);
	else
		debug("rlim.accs.rcvbytecnt: no limit");

	if (rlim.accs.runtime < UINT64_MAX)
		debug("runtime: %" PRIu64 " secs",
			rlim.accs.runtime / PKTLAB_TICKS_PER_SECOND);
	else
		debug("runtime: no limit");

	if (now < rlim.endtime && rlim.accs.runtime < rlim.endtime - now)
		rlim.endtime = now + rlim.accs.runtime;

	rates_optimize(&rlim.rates);
	rates_debug(&rlim.rates);

	// Abort transaction, which also closes dbi

	mdb_txn_abort(txn);

	time_start = now;
	runtime_rem = rlim.accs.runtime;
	memcpy(&accs_start, &rlim.accs, sizeof(accs_start));

	rlim.state = RLIM_RUNNING;
}

void rlim_pause(pktlab_time_t now) {
	switch (rlim.state) {
	case RLIM_UNINIT: // can't complete operation
	case RLIM_READY:
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		return;
	case RLIM_RUNNING: // expected case
		break;
	case RLIM_PAUSED: // unexpected but okay
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		break;
	}

	// Debit used time from runtime. We keep our own copy just in case silly
	// programmer decides to modify rlim.accs.runtime.

	if (now < rlim.endtime)
		runtime_rem -= (now - time_start);
	else
		runtime_rem = 0;

	rlim.accs.runtime = runtime_rem;
	time_start = now;

	rlim.state = RLIM_PAUSED;
}

void rlim_resume(pktlab_time_t now) {
	uint_fast32_t i;

	switch (rlim.state) {
	case RLIM_UNINIT: // can't complete operation
	case RLIM_READY:
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		return;
	case RLIM_PAUSED: // expected case
		break;
	case RLIM_RUNNING: // unexpected but okay
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
	}

	// Recalculate new end time.

	rlim.endtime = PKTLAB_TIME_MAX;
	rlim.accs.runtime = runtime_rem;

	for (i = 0; i < dbrcnt; i++) {
		if (dbrecs[i].dbval->exptime < rlim.endtime)
			rlim.endtime = dbrecs[i].dbval->exptime;
	}

	if (now < rlim.endtime && rlim.accs.runtime < rlim.endtime - now)
		rlim.endtime = now + rlim.accs.runtime;

	time_start = now;

	rlim.state = RLIM_RUNNING;
}

void rlim_commit(pktlab_time_t now) {
	MDB_txn * txn;
	MDB_dbi dbi;
	MDB_val mdbkey, mdbval;
	uint64_t * cnt0ptr;
	uint64_t * cntptr;
	uint64_t debit;
	size_t accstsz;
	uint_fast32_t i;
	uint_fast16_t j;
	int result;

	trace("%s(%f)", __func__, (double) now / PKTLAB_TICKS_PER_SECOND);

	switch (rlim.state) {
	case RLIM_UNINIT: // can't complete operation
	case RLIM_READY:
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		return;
	case RLIM_RUNNING:
	case RLIM_PAUSED:
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		break;
	}

	// Update runtime. Equivalent to:
	//
	//   rlim_pause(now)
	//   rlim_resume(now)

	if (now < rlim.endtime)
		runtime_rem -= (now - time_start);
	else
		runtime_rem = 0;

	debug("runtime_rem = %f", (double) runtime_rem / PKTLAB_TICKS_PER_SECOND);

	rlim.accs.runtime = runtime_rem;
	time_start = now;

	// Start txn

	result = mdb_txn_begin(mdbenv, /* parent */ NULL, /* flags */ 0, &txn);
	exit_if_mdb_error(result, storename);

	result = mdb_dbi_open(txn, /* name */ NULL, /* flags */ 0, &dbi);
	exit_if_mdb_error(result, storename);

	// Process dbrecs

	for (i = 0; i < dbrcnt; i++) {
		// Create mdbkey
		mdbkey.mv_data = dbrecs[i].key;
		mdbkey.mv_size = dbrecs[i].keylen;

		// Reload (possibly externally updated) dbrec value from database

		result = mdb_get(txn, dbi, &mdbkey, &mdbval);

		if (result != MDB_NOTFOUND)
			exit_if_mdb_error(result, storename);

		accstsz = mdbval.mv_size - sizeof(struct rlim_dbval);

		if (result == 0 && sizeof(struct rlim_dbval) <= mdbval.mv_size) {
			accstsz = mdbval.mv_size - sizeof(struct rlim_dbval);
			if (accstsz % sizeof(struct rlacc_state) == 0) {
				dbrecs[i].dbvalsz = mdbval.mv_size;
				dbrecs[i].dbval = safe_malloc(mdbval.mv_size);
				memcpy(dbrecs[i].dbval, mdbval.mv_data, mdbval.mv_size);
				dbrecs[i].astcnt = accstsz / sizeof(struct rlacc_state);
				debug("Reloaded stored record for certificate %" PRIdFAST32, i);
			} else {
				warn("Malformed (%zu-byte) rlim record for certificate %" PRIdFAST32,
						(size_t) mdbval.mv_size, i);
				result = MDB_NOTFOUND;
			}
		}

		// Create updated dbrec to store

		mdbval.mv_size = dbrecs[i].dbvalsz;
		mdbval.mv_data = dbrecs[i].dbval;

		for (j = 0; j < dbrecs[i].astcnt; j++) {
			cnt0ptr = rlaccptr(&accs_start, dbrecs[i].dbval->accst[j].type);
			cntptr = rlaccptr(&rlim.accs, dbrecs[i].dbval->accst[j].type);

			if (cntptr == NULL)
				continue;

			debit = *cnt0ptr - *cntptr;

			if (debit < dbrecs[i].dbval->accst[j].cnt)
				dbrecs[i].dbval->accst[j].cnt -= debit;
			else
				dbrecs[i].dbval->accst[j].cnt = 0;
		}

		result = mdb_put(txn, dbi, &mdbkey, &mdbval, /* flags */ 0);
		exit_if_mdb_error(result, storename);
	}

	// commit txn

	result = mdb_txn_commit(txn);
	exit_if_mdb_error(result, storename);

	// Update accs_start struct with new values.

	memcpy(&accs_start, &rlim.accs, sizeof(accs_start));
}

void rlim_reset() {
	struct rlim_rate ** rates;
	uint_fast32_t i;

	switch (rlim.state) {
	case RLIM_UNINIT: // can't complete operation
		debug("%s called in %s state", __func__, rlim_stname(rlim.state));
		return;
	case RLIM_READY:
	case RLIM_RUNNING:
	case RLIM_PAUSED:
		break;
	}

	if (dbrecs != NULL) {
		for (i = 0; i < dbrcnt; i++) {
			free(dbrecs[i].key);
			free(dbrecs[i].dbval);
		}

		free(dbrecs);
		dbrecs = NULL;
		dbrcnt = 0;
	}

	rates = (struct rlim_rate **)&rlim.rates;
	for (i = 0; i < sizeof(rlim.rates) / sizeof(void*); i++)
		ratelist_free(rates[i]);

	memset(&rlim, 0, sizeof(rlim));

	rlim.state = RLIM_READY;
}

void rlim_print(FILE * out) {
	struct rlim_rate * node;

	fprintf(out, "rlim.state = %s\n", rlim_stname(rlim.state));
	fprintf(out, "rlim.endtime = %f\n", (double) rlim.endtime / PKTLAB_TICKS_PER_SECOND);
	fprintf(out, "rlim.accs.sndpktcnt = %" PRIu64 "\n", rlim.accs.sndpktcnt);
	fprintf(out, "rlim.accs.rcvpktcnt = %" PRIu64 "\n", rlim.accs.rcvpktcnt);
	fprintf(out, "rlim.accs.sndbytecnt = %" PRIu64 "\n", rlim.accs.sndbytecnt);
	fprintf(out, "rlim.accs.rcvbytecnt = %" PRIu64 "\n", rlim.accs.rcvbytecnt);
	fprintf(out, "rlim.accs.runtime = %f\n",
		(double) rlim.accs.runtime / PKTLAB_TICKS_PER_SECOND);
	fprintf(out, "rlim.caps.sktcnt = %" PRIu16 "\n", rlim.caps.sktcnt);

	for (node = rlim.rates.sndpktlist; node != NULL; node = node->next)
		fprintf(out, "rlim.rates.sndpktlist += %" PRIu64 " per %f sec\n",
			node->cnt, (double) node->win / PKTLAB_TICKS_PER_SECOND);

	for (node = rlim.rates.sndbytelist; node != NULL; node = node->next)
		fprintf(out, "rlim.rates.sndbytelist += %" PRIu64 " per %f sec\n",
			node->cnt, (double) node->win / PKTLAB_TICKS_PER_SECOND);
}

const char * rlim_stname(enum rlim_state state) {
	static const char * const names[] = {
		[RLIM_UNINIT] = "UNINIT",
		[RLIM_READY] = "READY",
		[RLIM_RUNNING] = "RUNNING",
		[RLIM_PAUSED] = "PAUSED"
	};

	if (0 <= state && state < 4)
		return names[state];
	else
		return NULL;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

uint64_t * rlaccptr(struct rlim_accs * accs, int type) {
	switch (type) {
	case RLACC_SNDPKTCNT:
		return &accs->sndpktcnt;
	case RLACC_RCVPKTCNT:
		return &accs->rcvpktcnt;
	case RLACC_SNDBYTECNT:
		return &accs->sndbytecnt;
	case RLACC_RCVBYTECNT:
		return &accs->rcvbytecnt;
	case RLACC_RUNTIME:
		return &accs->runtime;
	default:
		return NULL;
	}
}

void rates_append (
	struct rlim_rates * restrict dst,
	struct rlim_rates * restrict src)
{
	const unsigned int n = sizeof(struct rlim_rates) / sizeof(void*);
	struct rlim_rate ** srcs = (struct rlim_rate **) src;
	struct rlim_rate ** dsts = (struct rlim_rate **) dst;
	struct rlim_rate * sn;
	struct rlim_rate * dn;
	unsigned int i;

	for (i = 0; i < n; i++) {
		for (sn = srcs[i]; sn != NULL; sn = sn->next) {
			dn = malloc(sizeof(struct rlim_rate));
			dn->win = sn->win;
			dn->cnt = sn->cnt;
			dn->next = dsts[i];
			dsts[i] = dn;
		}
	}
}

void rates_optimize(struct rlim_rates * rates) {
	const unsigned int n = sizeof(struct rlim_rates) / sizeof(void*);
	struct rlim_rate ** ratelists = (struct rlim_rate **)rates;
	unsigned int i;

	for (i = 0; i < n; i++)
		ratelists[i] = ratelist_optimize(ratelists[i]);
}

struct rlim_rate * ratelist_optimize(struct rlim_rate * head) {
	// FIXME Remove duplicates and dominated rates
	return head;
}

void rates_debug(const struct rlim_rates * rates) {
	ratelist_debug("sndpkt", rates->sndpktlist);
	ratelist_debug("sndbyte", rates->sndbytelist);
}

void ratelist_debug(const char * name, struct rlim_rate * head) {
	struct rlim_rate * node;

	for (node = head; node != NULL; node = node->next) {
		debug("%s max rate: %" PRIu64 " per %lu sec", name, node->cnt,
			(unsigned long) node->win / PKTLAB_TICKS_PER_SECOND);
	}
}

void ratelist_free(struct rlim_rate * head) {
	struct rlim_rate * next;

	while (head != NULL) {
		next = head->next;
		free(head);
		head = next;
	}
}

void exit_if_mdb_error(int result, const char * msg) {
	if (result != 0) {
		if (msg != NULL)
			fprintf(stderr, "%s: %s\n", msg, mdb_strerror(result));
		else
			fprintf(stderr, "%s\n", mdb_strerror(result));

		exit(EXIT_FAILURE);
	}
}

