// util.c
//

#include "util.h"

#include <stdarg.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <ctype.h>
#include <unistd.h>

#include <termios.h>
#include <sys/stat.h>
#include <pwd.h>

#include "console.h"

// 
// EXPORTED FUNCTION DEFINITIONS
// 

void * pktlabme_safe_malloc_(size_t size) {
	void * ptr;
	
	ptr = malloc(size);
	
	if (ptr == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
		return NULL;
	} else
		return ptr;
}

void * pktlabme_safe_calloc_(size_t nelts, size_t eltsz) {
	void * ptr;
	
	ptr = calloc(nelts, eltsz);
	
	if (ptr == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
		return NULL;
	} else
		return ptr;
}

void * pktlabme_safe_realloc_(void * ptr, size_t size) {
	
	ptr = realloc(ptr, size);
	
	if (ptr == NULL && size > 0) {
		perror(NULL);
		exit(EXIT_FAILURE);
		return NULL;
	} else
		return ptr;
}

char * pktlabme_safe_strdup_(const char * str) {
	char * newstr;

	newstr = strdup(str);

	if (newstr == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
		return NULL;
	} else
		return newstr;
}

const char * strchr_ext(const char * s, int c, size_t * l) {
    const char * ptr;

    if (s == NULL)
        return NULL;

    if (l != NULL)
        *l = 0;

    for (ptr = s; *ptr != '\0'; ++ptr) {
        if (*ptr == c)
            return ptr;

        if (l != NULL) {
            ++*l;

            if (*l == SIZE_MAX)
                break; // *l will overflow
        }
    }

    return NULL;
}

int copy_str(const char * s, size_t l, char ** copied_str) {
    assert(s != NULL && copied_str != NULL);

    if (l >= SIZE_MAX)
        return -1;

    *copied_str = safe_malloc(l+1);
    memcpy(*copied_str, s, l);
    (*copied_str)[l] = '\0';

    return 0;
}

bool is_valid_filepath(const char * path) {
    struct stat st;

    assert(path != NULL);

    if (stat(path, &st) < 0) {
        debug("stat on %s failed with error %s",
            path, strerror(errno));
        return false;
    }

    // accept if is regular file or symbolic link
    return (S_ISREG(st.st_mode) || S_ISLNK(st.st_mode));
}

bool is_valid_hexstr(const char * hexstr) {
    size_t len, i;

    assert(hexstr != NULL);

    len = strlen(hexstr);
    if (len == 0 || len % 2 != 0)
        return false;

    // 0x at start not allowed
    for (i = 0; i < len; ++i) {
        if (!isxdigit(hexstr[i]))
            return false;
    }

    return true;
}

int read_whole_file (
    const char * path,
    uint8_t ** content, size_t * contentlen)
{
    FILE * fp;
    uint8_t * ptr;
    size_t total, incre, rst;

    assert(
        path != NULL && content != NULL &&
        contentlen != NULL);

    trace("%s(path:%s)", __func__, path);

    fp = fopen(path, "r");
    if (fp == NULL)
        return -1;

    total = incre = 4096;
    *contentlen = 0;
    *content = ptr = safe_malloc(total);
    while (true) {
        assert(total >= *contentlen);
        rst = fread(ptr, 1, total-*contentlen, fp);
        if (ferror(fp)) {
            debug("fread failed with: %s", strerror(errno));
            goto fail;
        }

        *contentlen += rst;
        ptr += rst;

        if (feof(fp))
            break;

        if (total-*contentlen < incre) {
            if (total > SIZE_MAX-incre) {
                debug("Overflow");
                goto fail;
            }

            total += incre;
            safe_realloc(ptr, total);
        }
    }

    fclose(fp);
    return 0;

fail:
    free(ptr);
    fclose(fp);
    return -1;
}

int posix_getpass (
    const char * prompt, char ** passphrase)
{
    FILE * fp;
    int ret = -1;
    struct termios oflags, nflags;
    const size_t max_pplen = 512; // smaller then INT_MAX

    assert(passphrase != NULL);

    fp = fopen("/dev/tty", "r+");
    if (fp == NULL) {
        debug("fopen failed with %s", strerror(errno));
        return -1;
    }

    if (tcgetattr(fileno(fp), &oflags) < 0){
        debug("tcgetattr failed with %s", strerror(errno));
        goto fail;
    }

    nflags = oflags;
    nflags.c_lflag &= ~ECHO;
    nflags.c_lflag |= ECHONL;

    if (tcsetattr(fileno(fp), TCSANOW, &nflags) < 0) {
        debug("tcsetattr failed with %s", strerror(errno));
        goto fail;
    }

    if (prompt != NULL && !fputs(prompt, fp)) {
        debug("fputs failed with %s", strerror(errno));
        goto fail;
    }

    *passphrase = safe_malloc(max_pplen);
    if (fgets(*passphrase, max_pplen, fp) == NULL) {
        debug("fgets failed with %s", strerror(errno));
        free(*passphrase);
        goto fail;
    }

    (*passphrase)[strcspn(*passphrase, "\n")] = '\0';

    if (tcsetattr(fileno(fp), TCSANOW, &oflags) < 0) {
        debug("tcsetattr failed with %s", strerror(errno));
        free(*passphrase);
        goto fail;
    }

    ret = 0;

fail:
    fclose(fp);
    return ret;
}

int overflow_safe_add_uf32 (
    uint_fast32_t u1, uint_fast32_t u2,
    uint_fast32_t * rst)
{
    if (u1 > UINT_FAST32_MAX-u2)
        return -1;

    *rst = u1+u2;
    return 0;
}

int get_hex_val(char c) {
    if (!isxdigit(c))
        return -1;

    if (isdigit(c))
        return c-'0';

    return tolower(c)-'a'+10;
}

int hexstr2bytes (
    const char * hexstr, size_t len,
    uint8_t ** bytes, size_t * blen)
{
    size_t i;
    int u, l;

    assert(
        hexstr != NULL && len != 0 &&
        len % 2 == 0 && bytes != NULL);

    *blen = len/2;
    *bytes = safe_malloc(*blen);

    for (i = 0; i < len; i += 2) {
        u = get_hex_val(hexstr[i]);
        l = get_hex_val(hexstr[i+1]);
        assert(u >= 0 && l >= 0 && u < 16 && l < 16);

        (*bytes)[i/2] = 16*u+l;
    }

    return 0;
}


const char * get_passwd_home(void) {
    const char * loginname;
    struct passwd * pwd;

    errno = 0;
    loginname = getlogin();
    if (loginname == NULL) {
        if (errno != 0) {
            debug("getlogin failed with: %s", strerror(errno));
        } else
            debug("getlogin cannot find user login name");

        return NULL;
    }

    errno = 0;
    pwd = getpwnam(loginname);
    if (pwd == NULL) {
        if (errno != 0) {
            debug("getpwnam failed with: %s", strerror(errno));
        } else
            debug("%s not in user database", loginname);

        return NULL;
    }

    return pwd->pw_dir;
}

