// worker.c
//

#include "worker.h"

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/select.h>

#include "vmem.h"
#include "network.h"
#include "control.h"
#include "console.h"
#include "util.h"
#include "comm.h"

#include <assert.h>

//
// INTERNAL MACROS
//

#define RETRY_GIVE_UP_TIME 5 // sec

//
// EXPORTED GLOBAL VARIABLES
//

pktlab_time_t _time_now;

//
// INTERNAL FUNCTION DECLARATIONS
//

#if DEBUG
static char *strfdset(char *buf, size_t bufsz, const fd_set *fdset);
static char *rfdsetstr(const fd_set *fdset);
static char *wfdsetstr(const fd_set *fdset);
#endif

static int handle_ctrlmsg(int mgrfd, fd_set *set);
static int report_finish(int mgrfd);
static void terminate(int mgrfd);

void worker_initialize(void) {
    trace("%s()", __func__);

    vmem_initialize();
    network_initialize(); // vmem module must be init before network module
    control_initialize();
}

void worker_main(int mgrfd, const char *spec, const char * snikey,
                 struct pktctrl_keycert_store *store,
                 struct pktlab_privatekey *privk, uint32_t exp_priority) {
    pktlab_time_t timeout;
    struct timeval timeout_timeval;
    fd_set rset, wset;
    int nfds, nsel;
    int result;

    trace("%s(mgrfd:%d,spec:\"%s\")", __func__, mgrfd, spec);

    _time_now = pktlab_time_now();

    network_start();
    control_start(spec, snikey, mgrfd,
                  0, RETRY_GIVE_UP_TIME,
                  store, privk, exp_priority);

    while (true) {
        timeout = _time_now + pktlab_time_sec(60);
        memset(&rset, 0, sizeof(rset));
        memset(&wset, 0, sizeof(wset));
        FD_SET(mgrfd, &rset);
        nfds = mgrfd + 1;

        result = control_prepare_select(&rset, &wset, &timeout);
        nfds = MAX(nfds, result);

        result = network_prepare_select(&rset, &wset, &timeout);
        nfds = MAX(nfds, result);

        pktlab_time_to_timeval((timeout > _time_now) ? (timeout - _time_now)
                                                     : 0,
                               &timeout_timeval);

        debug("Calling select(%d, %s, %s, %u sec) ...", nfds, rfdsetstr(&rset),
              wfdsetstr(&wset), (unsigned int)timeout_timeval.tv_sec);

        nsel = select(nfds, &rset, &wset, NULL, &timeout_timeval);

        _time_now = pktlab_time_now();

        debug("=> nsel:%d, rset:%s, wset:%s", nsel, rfdsetstr(&rset),
              wfdsetstr(&wset));

        nsel -= handle_ctrlmsg(mgrfd, &rset);
        if (!control_running())
            break;

        nsel -= control_process_select(nsel, &rset, &wset);
        if (!control_running())
            break;

        nsel -= network_process_select(nsel, &rset, &wset);
    }
    report_finish(mgrfd);
    exit(EXIT_SUCCESS);
    return;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int handle_ctrlmsg(int mgrfd, fd_set *set) {
    struct ctrlmsg msg;
    int result;

    trace("%s", __func__);

    if (FD_ISSET(mgrfd, set)) {
        result = recv_ctrlmsg(mgrfd, &msg);

        if (result != 0) {
            // Something wrong with manager socket.
            // May be that either there is a error or there is an EOF.
            // We just treat that as manager died.
            terminate(mgrfd); // exit here as we don't need to report to manager
        } else {
            trace("%s (type:%d)", __func__, msg.type);

            if (msg.type == CTRLMSG_TYPE_SUSPEND) {
                control_suspend();
                return 1;
            }
        }
    }
    return 0;
}

int report_finish(int mgrfd) {
    int result = send_ctrlmsg(mgrfd, CTRLMSG_TYPE_FINISH, 0);
    assert(result == 0);
    return 0;
}

void terminate(int mgrfd) {
    // Close everything down. For network sockets, probably better not to close
    // them, so TCP sockets don't generate FIN, which suggests an orderly
    // shutdown. We should probably add a network_terminate() function to do
    // this, but for now, just yank the plug.

    control_reset();
    close(mgrfd);

    exit(EXIT_SUCCESS);
}

#ifdef DEBUG

char *strfdset(char *buf, size_t bufsz, const fd_set *fdset) {
    char *start = buf;
    size_t full_len = 0;
    int result;
    int i;

    for (i = 0; i < FD_SETSIZE; i++) {
        if (!FD_ISSET(i, fdset))
            continue;
        result = snprintf(buf, bufsz, ",%d", i);
        full_len += result;
        buf += result;
        bufsz -= result;
    }

    *start = '{';
    if (full_len == 0) {
        full_len += 1;
        buf += 1;
        bufsz -= 1;
    }

    *buf = '}';
    full_len += 1;
    buf += 1;
    bufsz -= 1;

    *buf = '\0';
    buf += 1;
    bufsz -= 1;

    return start;
}

char *rfdsetstr(const fd_set *fdset) {
    static char buf[64];
    return strfdset(buf, sizeof(buf), fdset);
}

char *wfdsetstr(const fd_set *fdset) {
    static char buf[64];
    return strfdset(buf, sizeof(buf), fdset);
}

#endif
