// xdescrpool.c
//

#include <assert.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <pktlab.h>

#include "xdescrpool.h"
#include "console.h"
#include "util.h"

//
// INTERNAL MACROS
//

#define CONTROLLER_ADDR_NAMESTR "controller_addr"
#define CONTROLLER_ADDR_LEN (sizeof(CONTROLLER_ADDR_NAMESTR) - 1)
#define PUB_END_NAMESTR "pub_end"
#define PUB_END_LEN (sizeof(PUB_END_NAMESTR) - 1)
#define EXP_PRIORITY_NAMESTR "exp_priority"
#define EXP_PRIORITY_LEN (sizeof(EXP_PRIORITY_NAMESTR) - 1)
#define SNI_KEY_NAMESTR "sni_key"
#define SNI_KEY_LEN (sizeof(SNI_KEY_NAMESTR) - 1)
#define ENDPOINT_COUNTRY_NAMESTR "endpoint_country"
#define ENDPOINT_COUNTRY_LEN (sizeof(ENDPOINT_COUNTRY_NAMESTR) - 1)
#define ENDPOINT_LOCATION_NAMESTR "endpoint_location"
#define ENDPOINT_LOCATION_LEN (sizeof(ENDPOINT_LOCATION_NAMESTR) - 1)

//
// INTERNAL GLOBAL VARIABLES
//

static struct xdescr_node *xdescr_head = NULL;

//
// EXPORTED FUNCTION DECLARATIONS
//

static bool is_same_xdescr(struct xdescr_node *node1,
                           struct xdescr_node *node2);

//
// EXPORTED FUNCTION DEFINITIONS
//

struct xdescr_node *xdescrnode_construct(const char *xdescrbuf,
                                         uint_fast32_t xdescrlen) {
    struct pktlab_xdescr_dattr_pair *pairs, *p;
    uint_fast32_t paircnt;
    int rst;
    struct xdescr_node *node = NULL;
    char *endptr;
    bool missing_addr = true;

    trace("%s(xdescrlen:%" PRIuFAST32 ")", __func__, xdescrlen);

    rst = pktlab_decode_xdescr(xdescrbuf, xdescrlen, &pairs, &paircnt);
    if (rst != 0) {
        warn("malformed xdescr");
        return NULL;
    }

    node = safe_malloc(sizeof(*node));
    memset(node, 0, sizeof(*node));

    for (int pidx = 0; pidx < paircnt; ++pidx) {
        p = pairs + pidx;
        if (p->namelen == CONTROLLER_ADDR_LEN &&
            strncmp(p->name, CONTROLLER_ADDR_NAMESTR, CONTROLLER_ADDR_LEN) ==
                0) {
            missing_addr = false;
            if (p->valuelen > sizeof(node->controller_addr)-1)
                warn("controller addr larger than node buffer");

            memcpy(node->controller_addr, p->value,
                   MIN(sizeof(node->controller_addr)-1, p->valuelen));
        } else if (p->namelen == PUB_END_LEN &&
                   strncmp(p->name, PUB_END_NAMESTR, PUB_END_LEN) == 0) {

            node->expire = strtoull(p->value, &endptr, 10);
            if (endptr - p->value != p->valuelen) {
                warn("failed to convert xdescr expiration time");
                node->expire = 0; // expired
            }
        } else if (p->namelen == EXP_PRIORITY_LEN &&
                   strncmp(p->name, EXP_PRIORITY_NAMESTR, EXP_PRIORITY_LEN) ==
                       0) {
            node->exp_priority = strtoul(p->value, &endptr, 10);
            if (endptr - p->value != p->valuelen) {
                warn("failed to convert xdescr priority");
                node->exp_priority = 0; // lowest priority
            }
        } else if (p->namelen == SNI_KEY_LEN &&
                   strncmp(p->name, SNI_KEY_NAMESTR, SNI_KEY_LEN) == 0) {
            if (p->valuelen < 1 || p->valuelen > 63) {
                warn("invalid snikey len (%d)", p->valuelen);
            }
            strncpy(node->snikey, p->value, p->valuelen);
        }
    }

    if (missing_addr) {
        warn("xdescr missing controller_addr");
        free(node);
        return NULL;
    }

    node->xdescr = safe_malloc(xdescrlen + 1);
    memset(node->xdescr, 0, xdescrlen + 1);
    node->xdescrlen = xdescrlen;
    memcpy(node->xdescr, xdescrbuf, xdescrlen);

    return node;
}

int xdescrpool_push(struct xdescr_node *node) {
    if (node == NULL)
        return -1;
    trace("%s controller_addr: %s", __func__, node->xdescr);
    node->next = xdescr_head;
    xdescr_head = node;
    return 0;
}

int xdescrpool_remove(struct xdescr_node *node) {
    struct xdescr_node **prev_next = &xdescr_head;
    if (node == NULL)
        return -1;
    trace("%s controller_addr: %s", __func__, node->xdescr);
    for (struct xdescr_node *n = xdescr_head; n != NULL; n = n->next) {
        if (is_same_xdescr(node, n)) {
            *prev_next = n->next;
            return 0;
        }
        prev_next = &n->next;
    }
    return -1;
}

struct xdescr_node *xdescrpool_get(pktlab_time_t *yield) {
    uint_fast32_t max_pri;
    struct xdescr_node *max_pri_xdescr;
    struct xdescr_node *n;
    pktlab_time_t now, first_yield;

    // goal: get the experiment with highest priority that hasn't expired yet
    // also return the earliest yield-hint of higher priority exp

    assert(yield != NULL);

    trace("%s(...)", __func__);

    now = pktlab_time_now();

    // get non-yielding exp having max priority
    max_pri = 0;
    max_pri_xdescr = NULL;

    for (n = xdescr_head; n != NULL; n = n->next) {
        if (n->yield_hint < now &&
            (max_pri_xdescr == NULL || n->exp_priority > max_pri)) {
            max_pri = n->exp_priority;
            max_pri_xdescr = n;
        }
    }

    // get earliest yield end time of yielded exp having higher priority
    first_yield = UINT64_MAX;

    for (n = xdescr_head; n != NULL; n = n->next) {
        if (n->yield_hint >= now &&
            n->exp_priority > max_pri &&
            n->yield_hint < first_yield)
            first_yield = n->yield_hint;
    }

    debug("Got max_pri: %" PRIuFAST32
        ", first_yield: %" PRIu64
        ", max_pri_xdescr->controller_addr: %s",
        max_pri, first_yield,
        (max_pri_xdescr == NULL) ? \
            "n/a": safe_str(max_pri_xdescr->controller_addr));

    *yield = first_yield;
    return max_pri_xdescr;
}

void xdescrnode_destroy(struct xdescr_node *node) {
    if (node == NULL)
        return;
    trace("%s controller_addr: %s", __func__, node->controller_addr);

    if (node->xdescr)
        free(node->xdescr);

    free(node);
}

int get_xdescr_spec(const char * spec, char ** xdescr) {
    size_t speclen, totalmem;
    struct pktlab_xdescr_dattr_pair p[2];
    uint_fast32_t used;

    assert(spec != NULL && xdescr != NULL);

    speclen = strlen(spec);
    if (speclen > UINT_FAST32_MAX ||
        speclen > SIZE_MAX-CONTROLLER_ADDR_LEN-3 ||
        speclen > UINT_FAST32_MAX-CONTROLLER_ADDR_LEN-3)
        return -1;

    p[0].name = CONTROLLER_ADDR_NAMESTR;
    p[0].namelen = CONTROLLER_ADDR_LEN;
    p[0].value = spec;
    p[0].valuelen = speclen;
    p[1].name = SNI_KEY_NAMESTR;
    p[1].namelen = SNI_KEY_LEN;
    p[1].value = "test";
    p[1].valuelen = 4;

    totalmem = speclen+CONTROLLER_ADDR_LEN+2+4+SNI_KEY_LEN+3;
    *xdescr = safe_malloc(totalmem);
    if (pktlab_encode_xdescr(
            p, 2, *xdescr, totalmem, &used) < 0) {
        free(*xdescr);
        return -2;
    }

    assert(used == totalmem-1);
    (*xdescr)[used] = '\0';
    return 0;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

bool is_same_xdescr(struct xdescr_node *node1,
                    struct xdescr_node *node2) {

    const char *xdescr1, *xdescr2;
    uint_fast32_t xdescrlen1, xdescrlen2;

    xdescr1 = node1->xdescr;
    xdescr2 = node2->xdescr;
    xdescrlen1 = node1->xdescrlen;
    xdescrlen2 = node2->xdescrlen;

    if (xdescrlen1 != xdescrlen2)
        return false;
    if (memcmp(xdescr1, xdescr2, xdescrlen1) == 0)
        return true;
    return false;
}