#!/usr/bin/env python3

import os
import sys
import socket
import getpass
import pexpect
import subprocess
import traceback
import netifaces
from shutil import copytree, rmtree

'''
NOTE:
    0: non-raw
    1: write raw to configurations
    2: prompt user
'''
ENABLE_RAW_CONF = RAW_MODE_CONF_FILE_OPTION

PKTLAB_DIR_PATH = os.path.join(os.path.expanduser("~"), ".pktlab")
PPKS_CONF_PATH = os.path.join(PKTLAB_DIR_PATH, "ppks_conf")
PPKS_CERT_PATH = os.path.join(PKTLAB_DIR_PATH, "certs")
PPKS_CSR_PATH = os.path.join(PKTLAB_DIR_PATH, "csrs")
PPKS_KEY_PATH = os.path.join(PKTLAB_DIR_PATH, "keys")

KEY_LIST = ['k_exper', 'k_xc', 'k_eop', 'k_ep']
CERT_LIST = [
        ('k_xc', 'k_xc', 'agent'),
        ('k_ep', 'k_ep', 'agent'),
        ('k_eop', 'k_exper', 'exppriv'),
        ('k_exper', 'k_xc', 'pubcmd'),
        ('k_eop', 'k_ep', 'subcmd')
]
CSR_LIST = [
        ('k_exper', 'exppriv', "", ""),
        ('k_exper', 'delpriv', "--del_exppriv", "_exp"),
        ('k_eop', 'delpriv', "--del_reppriv", "_rep"),
]

KEY_DESCRIPTION = {
        'k_exper': 'Experimenter Key',
        'k_xc': 'Experiment Controller Key',
        'k_eop': 'Endpoint Operator Key',
        'k_ep': 'Measurement Endpoint Key'
}

SETUP_AUX_PATH = os.path.dirname(os.path.realpath(__file__))
PREFIX_PATH = os.path.dirname(SETUP_AUX_PATH)
CONFIG_PATH = os.path.join(PREFIX_PATH, 'config')
CRED_PATH = os.path.join(PREFIX_PATH, 'cred')
PROG_PATH = os.path.join(PREFIX_PATH, 'program')

BOP_KEY_PATH = os.path.join(CRED_PATH, 'caida_k_bop.pub')
BOP_KEY_PATH_COPY = os.path.join(PPKS_KEY_PATH, 'caida_k_bop.pub')
XM_EXAMPLE_CONF_PATH = os.path.join(CONFIG_PATH, 'example.xpmgr.conf')
XM_PROXY_EXAMPLE_CONF_PATH = os.path.join(CONFIG_PATH, 'example.xpmgr.proxy.conf')
EP_EXAMPLE_CONF_PATH = os.path.join(CONFIG_PATH, 'example.endpt.conf')
XM_CONF_PATH = os.path.join(PKTLAB_DIR_PATH, "xpmgr.conf")
XM_PROXY_CONF_PATH = os.path.join(PKTLAB_DIR_PATH, "xpmgr.proxy.conf")
EP_CONF_PATH = os.path.join(PKTLAB_DIR_PATH, "endpt.conf")
PROG_PATH_COPY = os.path.join(PKTLAB_DIR_PATH, "progs")

ASCII_LOGO = \
"""
     ____  ___________
   //   || ||       //   ##########  #      #  ######### #            ##      ########
  //    || ||      //    #        #  #    #        #     #           #  #     #       #
 //     || ||     //     ##########  #  #          #     #          #    #    ########
 \\\\     || ||    //      #           ##  #         #     #         ########   #       #
  \\\\    || ||   //       #           #     #       #     #        #        #  #       #
   \\\\___|| ||__//        #           #       #     #     ####### #          # ########
"""

key_passphrase = {}

def print_welcome_msg():
    print(ASCII_LOGO)
    print('> Welcome to PacketLab!')
    print('> Now running the PacketLab software package user setup initialization script.')
    print('> This script sets up your user home directory for future PacketLab software execution (namely the ~/.pktlab directory).')

def ppks_init():
    print('> Initializing ppksman')

    p = pexpect.spawn(f'ppksman -g {PPKS_CONF_PATH} init')

    # Choose local key manager
    p.expect('0. LocalFSPPKSKeyManager')
    p.sendline('0')

    # Set key manager directory
    p.expect('Enter path to LocalFSPPKSKeyManager directory')
    p.sendline()

    # Choose local state manager
    p.expect('0. LocalFSPPKSStateManager')
    p.sendline('0')

    # Set state manager directory
    p.expect('Enter path to LocalFSPPKSStateManager directory')
    p.sendline()

    p.wait()

    # as directory not created by ppksman
    os.makedirs(PPKS_CSR_PATH)

def do_ppks_keygen(keyname):
    global key_passphrase

    # get and record key passphrase
    print(f'> Generating {KEY_DESCRIPTION[keyname]}')
    key_passphrase[keyname] = getpass.getpass(
        f'Please enter {KEY_DESCRIPTION[keyname]} ({keyname}) new passphrase: ')

    if len(key_passphrase[keyname]) != 0:
        confirm_passphrase = getpass.getpass(
            f'Please re-enter passphrase: ')
        if confirm_passphrase != key_passphrase[keyname]:
            raise ValueError("Passphrase mismatch")

    # generate key based on recorded passphrase
    p = pexpect.spawn(f'ppksman -g {PPKS_CONF_PATH} '
                      f'generate key {keyname}')
    p.expect('\(press enter for no passphrase\): ')
    p.sendline(key_passphrase[keyname])

    if len(key_passphrase[keyname]) != 0:
        p.expect('Please re-enter passphrase: ')
        p.sendline(key_passphrase[keyname])

    p.expect('Key generation completed')
    p.wait()

def ppks_keygen():
    '''
    Generate the following keys using command:
    ppksman -g ~/.pktlab/ppks_conf generate key TEST_KEY

    1. exper
    2. xc
    3. eop
    4. ep
    '''
    for k in KEY_LIST:
        do_ppks_keygen(k)

    # copy pre-shipped keys
    os.popen(f"cp {BOP_KEY_PATH} {BOP_KEY_PATH_COPY}").read()

def do_ppks_sign_cert(signee, signee_pubkey, cert_type,
                      signer, signer_privkey, start_time, end_time, cert_file):
    global key_passphrase

    print(f'> Generating {cert_type} certificate signed by '
          f'{KEY_DESCRIPTION[signer]} ({signer}) '
          f'on {KEY_DESCRIPTION[signee]} ({signee})')

    p = pexpect.spawn(f'ppksman -g {PPKS_CONF_PATH} generate '
                      f'cert -k {signee_pubkey} {cert_type} '
                      f'{signer_privkey} {start_time} {end_time} {cert_file} -y')

    p.expect('Loading signer privkey')

    # send recorded passphrase in
    if len(key_passphrase[signer]) != 0:
        p.expect(f'{signer_privkey}\): ')
        p.sendline(key_passphrase[signer])

    p.wait()

    print(f'> Certificate stored in {cert_file}')

def do_ppks_sign_csr(signee, signee_privkey, cert_type, more_opt, csr_file):
    global key_passphrase

    print(f'> Generating {cert_type} CSR for '
          f'{KEY_DESCRIPTION[signee]} ({signee})')

    p = pexpect.spawn(f'ppksman -g {PPKS_CONF_PATH} generate '
                      f'csr {more_opt} {cert_type} {signee_privkey} {csr_file} -y')

    p.expect('Loading signee privkey')

    if len(key_passphrase[signee]) != 0:
        p.expect(f'{signee_privkey}\): ')
        p.sendline(key_passphrase[signee])

    p.wait()

    print(f'> CSR stored in {csr_file}')

def ppks_sign():
    '''
    TODO: Change expire time
    '''
    start_time = 0
    end_time = 2147483647

    for signer, signee, cert_type in CERT_LIST:
        signer_privkey = os.path.join(PPKS_KEY_PATH, signer)
        signee_pubkey = os.path.join(PPKS_KEY_PATH, f'{signee}.pub')
        cert_name = f"{cert_type}_{signer.split('_')[1]}_{signee.split('_')[1]}.cert"
        cert_file = os.path.join(PPKS_CERT_PATH, cert_name)
        do_ppks_sign_cert(signee, signee_pubkey, cert_type,
                          signer, signer_privkey, start_time, end_time, cert_file)

    for signee, cert_type, more_opt, extra_name in CSR_LIST:
        signee_privkey = os.path.join(PPKS_KEY_PATH, signee)
        csr_name = f"{cert_type}{extra_name}_{signee.split('_')[1]}.csr"
        csr_file = os.path.join(PPKS_CSR_PATH, csr_name)
        do_ppks_sign_csr(signee, signee_privkey, cert_type, more_opt, csr_file)

def generate_credentials():
    ppks_init()
    ppks_keygen()
    ppks_sign()

def rewrite_conf_file(file_path):
    with open(file_path, 'r+') as f:
        data = f.read()
        data = data.replace('INSTALL_KEY_CRED_PATH', PPKS_KEY_PATH)
        data = data.replace('INSTALL_CERT_CRED_PATH', PPKS_CERT_PATH)
        data = data.replace('INSTALL_PROG_PATH', PROG_PATH_COPY)
        f.seek(0)
        f.truncate(0)
        f.write(data)

def change_conf_fields(file_path, fields):
    with open(file_path, 'r+') as f:
        data = f.readlines()
        for idx, line in enumerate(data):
            if not line or ' ' not in line:
                continue
            k = line.split(' ')[0]
            if k in fields:
                data[idx] = f'{k} {fields[k]}\n'
        f.seek(0)
        f.truncate(0)
        f.writelines(data)

def get_local_ipaddr_info():
    try:
        ss = socket.socket()
        ss.settimeout(1) # timeout after 1 sec
        ss.connect(('8.8.8.8', 53))
        ipaddr = ss.getsockname()[0]
        ss.close()
        netif_info = [netifaces.ifaddresses(k) for k in netifaces.interfaces()]
        info_list = [j for v in netif_info for k in v.values() for j in k if 'addr' in j and j['addr'] == ipaddr]
        if not len(info_list):
            raise Exception('IP addr not found')
        netmask = info_list[0]['netmask']
        nm = sum(bin(int(x)).count('1') for x in netmask.split('.'))
        info = f'{ipaddr}/{nm}'
        return info
    except:
        print(
            'Warning: Failed to infer WAN interface address for endpoint config, '+
            'using 127.0.0.1/8 as fallback. '+
            'Please update the VMemLocalAddrStrIPv4 setting in ~/.pktlab/endpt.conf to fix this issue.',
            file=sys.stderr)
        return ''

def raw_mode_prompt():
    valid = {"yes": True, "y": True, "no": False, "n": False}
    while True:
        print("\nEnable Raw Mode? " +
              "(* Reminder: Raw mode is only supported on linux machines.)\n" +
              "Please respond with 'y' for enable or 'n' for disable. [y/n] ", end="")
        choice = input().lower()
        if choice in valid:
            return valid[choice]
        else:
            print("\nPlease respond with 'y' for enable or 'n' for disable. [y/n] ",
                  end="")

def generate_configs():
    print('> Generating configuration files')

    os.popen(f"cp {XM_EXAMPLE_CONF_PATH} {XM_CONF_PATH}").read()
    os.popen(f"cp {XM_PROXY_EXAMPLE_CONF_PATH} {XM_PROXY_CONF_PATH}").read()
    os.popen(f"cp {EP_EXAMPLE_CONF_PATH} {EP_CONF_PATH}").read()

    rewrite_conf_file(XM_CONF_PATH)
    rewrite_conf_file(XM_PROXY_CONF_PATH)
    rewrite_conf_file(EP_CONF_PATH)

    ipaddr_info = get_local_ipaddr_info()
    if ipaddr_info:
        change_conf_fields(EP_CONF_PATH, {'VMemLocalAddrStrIPv4' : ipaddr_info})

    if ENABLE_RAW_CONF == 0:
        raw_mode = False
    elif ENABLE_RAW_CONF == 1:
        raw_mode = True
    else:
        raw_mode = raw_mode_prompt()

    if raw_mode:
        change_conf_fields(EP_CONF_PATH, {'TransprotoSup' : 'raw,tcp,udp'})

def copy_programs():
    print('> Copying preshipped programs')
    copytree(src=PROG_PATH, dst=PROG_PATH_COPY)

def check_ppks_installation():
    p = subprocess.Popen('ppksman', stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=True)
    (output, error) = p.communicate()
    test_str = 'usage: PPKSMan [-h]'
    if test_str in output.decode() or test_str in error.decode():
        return True
    return False

def cleanup():
    # clean up by removing ~/.pktlab
    try:
        rmtree(PKTLAB_DIR_PATH)
    except Exception:
        traceback.print_exc()
        print(f'Warning: Failed to remove {PKTLAB_DIR_PATH} during cleanup')
        print('Please remove it manually before rerunning init script')

if __name__ == "__main__":
    print_welcome_msg()

    if os.path.exists(PKTLAB_DIR_PATH):
        sys.exit(
            'Error: Local ~/.pktlab directory already exists. '+
            'The pktlab_init script can only be run once.')
    if not check_ppks_installation():
        sys.exit(
            'Error: Failed to run PPKSMan command. '+
            'Please install packages in requirements.txt and export $HOME/.local/bin to PATH.')

    try:
        generate_credentials()
        generate_configs()
        copy_programs()

        print("> Setup completed")
    except:
        traceback.print_exc()
        print(f'Failed to set up user home directory for package')
        cleanup()